import lex
from lex import TOKEN

import sys
import re
import os.path
import vvl_objects
import uuid

from vvl_objects import get_unique_prefix

output = None

def write(s):
    output.write(s)

vvl_objects.write = write

def print_line(lexer):
    global filename
    write('\n#line %d "%s"\n' % (lexer.lineno, filename))

def create_definition(typ):
    cls = getattr(vvl_objects, typ, None)
    if cls and isinstance(cls, type) and issubclass(cls, vvl_objects.Definition):
        return cls()
    else:
        raise ValueError("Unknown define: %s" % typ)

states = (
        ('define', 'exclusive'),
        ('definition', 'exclusive'),
        ('ccode', 'inclusive'), # curly brackets
        ('pcode', 'inclusive'), # parenthesis
        ('bcode', 'inclusive'), # square brackets
        ('dowhile', 'inclusive'),
        ('dowhilebody', 'inclusive'),
        ('forall', 'inclusive'),
        ('Cons', 'inclusive'),
        ('pclass', 'inclusive'),
        ('modelviewer', 'exclusive'),
        ('global', 'inclusive'),
        )

tokens = (
        'COMMENT',
        'INT_NUMBER',
        'FLOAT_NUMBER',
        'FIELDS',
        'NAME',
        'LONG_OPERATOR',
        'STRING',
        'CHAR',
        'newline',
        'SPACES',
        'LBRACE',
        'RBRACE',
        'LPAREN',
        'RPAREN',
        'LBRACKET',
        'RBRACKET',
        'LABEL',
        'EQUIV',
        'STRUCT',
        'ENDDEF',
        'DO',
        'WHILE',
        'LINESPEC',
        'DEFINE',
        )

t_FLOAT_NUMBER = r'(\d*\.?|\.\d)\d*([eE][+-]?\d+)'
t_INT_NUMBER = r'(0x[0-9a-fA-F]+|0[0-7]+|\d+)'

def t_ANY_COMMENT(t):
    r'(/\*(.|\n)*?\*/)|(//.*)'
    t.lexer.lineno += t.value.count('\n')
    return t

t_NAME = r'[a-zA-Z_]\w*'
t_LONG_OPERATOR = r'(\.\*|->\*|\|\||==|&&|->|::|\+=|-=|\*=|/=|\^=|\|=|&=|%=|\+\+|--|!=|\#\#|>>=|<<=|>>|<<|>=|<=|<:|:>|<%|%>|%:%:|%:|\.\.\.)'
t_STRING = r'"(\\.|[^"])*"'
t_CHAR = r"'(\\[0-8]{1,3}|\\x[0-9a-fA-F]{1,2}|.|\\.)'"
t_ANY_SPACES = r"[ \t]+"
t_DEFINE = r'\#[ \t]*define'

literals = r'%^&*\-+=?.|/~!#<>;,:'

# Line specification for preprocessing

extract_line_and_file = re.compile(r'\s*#\s*(line)?\s+(\d+)\s+"(.*?)"')

def t_ANY_LINESPEC(t):
    r'[ \t]*\#[ \t]*(line)?[ \t]+\d+[ \t]+".*"([ \t]*\d+)*[ \t]*\n*'
    m = extract_line_and_file.match(t.value)
    if not m:
        raise ValueError("Invalid line specification")
    global filename
    filename = m.group(3)
    t.lexer.lineno = int(m.group(2)) + t.value.count('\n')-1
    return t

# Bracket matching

def t_RPAREN(t):
    r'\)'
    raise ValueError("Unmatched ')'")

def t_RBRACKET(t):
    r'\]'
    raise ValueError("Unmatched ']'")

def t_begin_pcode(t):
    r'\('
    t.lexer.push_state('pcode')
    t.type = 'LPAREN'
    return t

def t_pcode_RPAREN(t):
    r'\)'
    t.lexer.pop_state()
    t.type = 'RPAREN'
    return t

def t_begin_bcode(t):
    r'\['
    t.lexer.push_state('bcode')
    t.type = 'LBRACKET'
    return t

def t_bcode_RBRACKET(t):
    r'\]'
    t.lexer.pop_state()
    t.type = 'RBRACKET'
    return t

def t_begin_ccode(t):
    r'\{'
    t.lexer.push_state("ccode")
    t.type = 'LBRACE'
    return t

# End of a structure ...
def _end_current_0x(t):
    while True:
        state = t.lexer.current_state()
        if state in ("forall", "dowhilebody","Cons"):
            t.lexer.pop_state()
        else:
            return t

def _end_current_std(t):
    while True:
        state = t.lexer.current_state()
        if state == "forall":
            t.value += '}'
            t.lexer.pop_state()
        elif state in ("dowhilebody","Cons"):
            t.lexer.pop_state()
        else:
            return t

_end_current_gcc = _end_current_std

end_current = _end_current_gcc

def t_structure_end(t):
    r'\}'
    state = t.lexer.current_state()
    if state in ['pcode', 'bcode']:
        raise ValueError("Unmatched '}'")
    elif state in ['definition', 'pclass', 'modelviewer']:
        if t.lexer.current.end():
            t.lexer.current = None
        print_line(t.lexer)
        t.lexer.pop_state()
    elif state != 'ccode':
        raise ValueError("Unmatched '}'")
    else:
        t.type = 'RBRACE'
        t.lexer.pop_state()
        return end_current(t)

# Managing do ... while loop

def t_DO(t):
    r'\bdo\b'
    t.lexer.push_state('dowhile')
    t.lexer.push_state('dowhilebody')
    return t

def t_dowhile_WHILE(t):
    r'\bwhile\b'
    t.lexer.pop_state()
    return t

def t_dowhilebody_ENDDEF(t):
    r';'
    t.lexer.pop_state()
    t.type = 'ENDDEF'
    return t

# Handling adding comas in initializer list

@TOKEN(t_NAME)
def t_Cons_NAME(t):
    if t.lexer.newline:
        t.lexer.newline = False
        if t.lexer.first_cons:
            t.lexer.first_cons = False
            if t.lexer.has_implicit_cons:
                t.value = ", %s" % t.value
            else:
                t.value = ": %s" % t.value
    return t

def t_Cons_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)
    t.lexer.newline = True
    return t

# Tokens specialized for defines

@TOKEN(t_NAME)
def t_define_NAME(t):
    t.lexer.current.parseToken(t.value);

@TOKEN(t_NAME)
def t_definition_NAME(t):
    t.lexer.current.parseName(t.value)

def t_definition_EQUIV(t):
    r'=.+;'
    t.lexer.current.startEquiv()
    m = re.match(r'=\s*(.+)\s*;', t.value)
    t.lexer.current.parseName(m.group(1))
    t.lexer.current.endEquiv()

def t_definition_ENDDEF(t):
    r';'
    t.lexer.current.endEquiv()

def t_define_STRUCT(t):
    r':'
    t.lexer.current.validate()

def t_definition_STRUCT(t):
    r':'
    t.lexer.current.validateStruct()

def t_define_LBRACE(t):
    r'\{'
    t.lexer.current.start()
    print_line(t.lexer)
    t.lexer.pop_state()
    t.lexer.push_state("definition")

def t_definition_LBRACE(t):
    r'\{'
    t.lexer.current.startClass()
    print_line(t.lexer)
    t.lexer.push_state("pclass")
    t.type = 'LBRACE'

def t_definition_RBRACE(t):
    r'\}'
    t.lexer.current.end()
    print_line(t.lexer)
    t.lexer.pop_state()

definition_type = re.compile(r"define\s+(in_template\s+)?(\w*\b)")

def t_begin_define(t):
    r'define\s+(in_template\s+)?[a-zA-Z_]\w*\b'
    t.lexer.lineno += t.value.count('\n')
    m = definition_type.match(t.value)
    d = create_definition(m.group(2))
    if m.group(1):
        d.in_template()
    t.lexer.current = d
    t.lexer.push_state('define')

# Tokens specialized for constructor / destructor

#def t_pclass_DESTRUCTOR(t):
#    r'finalize\s*:\s*'
#    t.lexer.lineno += t.value.count('\n')
#    t.lexer.current.destructor()
#    print_line(t.lexer)

#def t_pclass_CONSTRUCTOR(t):
#    r'init\s*:\s*'
#    t.lexer.lineno += t.value.count('\n')
#    t.lexer.first_cons = t.lexer.current.initialize()
#    print_line(t.lexer)
#    t.lexer.newline = True
#    t.lexer.push_state("Cons")

read_label = re.compile(r'([a-zA-Z_]\w*)\s*:')

def t_pclass_LABEL(t):
    r'[a-zA-Z_]\w*\s*:(?!:)\s*'
    t.lexer.lineno += t.value.count('\n')
    lab = read_label.match(t.value).group(1)
    if lab == 'init':
        t.lexer.first_cons = True
        t.lexer.has_implicit_cons = not t.lexer.current.initialize()
        t.lexer.push_state("Cons")
        t.lexer.newline = True
    elif lab == 'finalize':
        t.lexer.current.destructor()
    else:
        return t
    print_line(t.lexer)

# Model and viewer construction

@TOKEN(t_NAME)
def t_modelviewer_NAME(t):
    t.lexer.current.parseName(t.value)

def t_modelviewer_ENDDEF(t):
    r':'
    t.lexer.current.validate()

def t_modelviewer_LBRACE(t):
    r'\{'
    t.lexer.current.startClass()
    print_line(t.lexer)
    t.lexer.begin("pclass")
    t.type = 'LBRACE'

def t_modelviewer_RBRACE(t):
    r'\}'
    t.lexer.current.end()
    print_line(t.lexer)
    t.lexer.pop_state()

def t_global_begin_model(t):
    r'Model(?=([a-zA-Z]\w*|/\*.*?\*/|//.*$|\s*)*:)'
    t.lexer.lineno += t.value.count('\n')
    t.lexer.model = vvl_objects.Model()
    t.lexer.current = t.lexer.model
    print_line(t.lexer)
    t.lexer.push_state("modelviewer")

def t_global_begin_viewer(t):
    r'Viewer(?=([a-zA-Z]\w*|/\*.*?\*/|//.*$|\s*)*:)'
    t.lexer.lineno += t.value.count('\n')
    t.lexer.viewer = vvl_objects.Viewer()
    t.lexer.current = t.lexer.viewer
    print_line(t.lexer)
    t.lexer.push_state("modelviewer")

# Code for managing forall structure

forall_re = re.compile(r'forall\s+([a-zA-Z_:].*\s*)?([*&]*[a-zA-Z_]\w*)\s+in\s+(.*\S)\s*:.*')

def _t_begin_forall_gcc(t):
    r'forall\s+[a-zA-Z_:].*\s+in\s+.+:\s*'
    t.lexer.lineno += t.value.count('\n')
    m = forall_re.match(t.value)
    if not m:
        raise ValueError("Error, syntax not recognised:\n%s" % t.value)
    var_type = m.group(1)
    var_name = m.group(2)
    container = m.group(3)
    if not var_type:
        var_type = ""
    uid = str(uuid.uuid1()).replace('-', '_')
    range_name = "range%s" % (uid,)
    values = {'range': range_name, 'type': var_type, 'var': var_name, 'container': container}
    write("""for(typeof(util::ForAll::range(%(container)s)) %(range)s = util::ForAll::range(%(container)s) ; !%(range)s.atEnd() ; ++%(range)s) {
            %(type)s%(var)s = *%(range)s.current();""" % values)
    print_line(t.lexer)
    t.lexer.push_state('forall')

def _t_begin_forall_0x(t):
    r'forall\s+[a-zA-Z_:].*\s+in\s+.+:\s*'
    t.lexer.lineno += t.value.count('\n')
    m = forall_re.match(t.value)
    if not m:
        raise ValueError("Error, syntax not recognised:\n%s" % t.value)
    var_type = m.group(1)
    var_name = m.group(2)
    container = m.group(3)
    if not var_type:
        var_type = ""
    #uid = str(uuid.uuid1()).replace('-', '_')
    #range_name = "range%s" % (uid,)
    #values = {'range': range_name, 'type': var_type, 'var': var_name, 'container': container}
    #write("""for(auto %(range)s = util::ForAll::range(%(container)s) ; !%(range)s.atEnd() ; ++%(range)s) {
            #%(type)s%(var)s = *%(range)s.current();""" % values)
    values = {'type': var_type, 'var': var_name, 'container': container}
    write("""for(%(type)s%(var)s : %(container)s)""" % values)
    print_line(t.lexer)
    t.lexer.push_state('forall')

def _t_begin_forall_std(t):
    r'forall\s+[a-zA-Z_:].*\s+in\s+.+:\s*'
    t.lexer.lineno += t.value.count('\n')
    m = forall_re.match(t.value)
    if not m:
        raise ValueError("Error, syntax not recognised:\n%s" % t.value)
    var_type = m.group(1)
    var_name = m.group(2)
    container = m.group(3)
    if not var_type:
        var_type = ""
    uid = str(uuid.uuid1()).replace('-', '_')
    iter_name = "it%s" % (uid,)
    range_name = "util::ForAll::make_range(%s)" % (container,)
    values = {'iter': iter_name, 'range': range_name, 'type': var_type, 'var': var_name, 'container': container}
    write("""for(const util::ForAll::BaseForwardIter& %(iter)s = util::ForAll::forwardIter(%(range)s, forall_pointer((%(range)s).first)) ;
            !util::ForAll::castForwardIter( &%(iter)s, forall_pointer( (%(range)s).first ) )->is_simple_end() ;
            ++( util::ForAll::castForwardIter( &%(iter)s, forall_pointer( (%(range)s).first ) )->it ) ) {
                %(type)s%(var)s =util::ForAll::castForwardIter( &%(iter)s, forall_pointer( (%(range)s).first ) )->value();""" % values)
    print_line(t.lexer)
    t.lexer.push_state('forall')

t_begin_forall = _t_begin_forall_gcc

def t_forall_ENDDEF(t):
    r';'
    return end_current(t)

# Define a rule so we can track line numbers
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)
    return t

t_define_definition_newline = t_newline
t_modelviewer_definition_newline = t_newline

def t_ANY_error(t):
    raise ValueError("Illegal token '%s'" % t.value)

def run(out, inp = None):
    lexer = lex.lex()
    lexer.newline = False
    global filename, output
    output = out
    if inp:
        filename = inp
        with open(filename, "rt") as f:
            data = f.read()
    else:
        filename = "<stdin>"
        data = sys.stdin.read()
    # Normalize new lines
    data = data.replace("\r\n", "\n")
    lexer.input(data)
    lexer.lineno = 1
    print_line(lexer)
    lexer.begin('global')
    while True:
        try:
            tok = lexer.token()
        except ValueError, err:
            print >> sys.stderr, "Error in file '%s' on line '%s':" % (filename, lexer.lineno)
            print >> sys.stderr, str(err)
            return False
        if not tok: break
        write(tok.value)
        #print >> sys.stderr, "%s - state: %s" % (tok, lexer.current_state())

    has_model = hasattr(lexer, 'model')
    has_viewer = hasattr(lexer, 'viewer')

    if has_model or has_viewer:
        bn = os.path.splitext(os.path.basename(filename))
        write('\n#include "%s.moc"\n' % bn[0])

    if has_model:
        write("DEFINE_MODEL(%s);\n" % lexer.model.name)

    if has_viewer:
        write("DEFINE_VIEWER(%s);\n" % lexer.viewer.name)

    return True


