import sys
import uuid

def write(s):
    sys.stdout.write(s)

def get_unique_prefix():
    return "p%s_%%s" % str(uuid.uuid1()).replace('-','_')

class Structure(object):
    def __init__(self, type_name, typedef, const_typedef):
        self.structure = None
        self.type_name = "%s_content" % (type_name,)
        self._name = type_name
        self._typedef = typedef
        self._const_typedef = const_typedef
        self._fwd = ""

    def __str__(self):
        return "Structure(%s,%s)" % (self.type_name, self.structure)

    def getName(self):
        return self._name

    def setName(self, name):
        self._name = name

    name = property(getName, setName)

    def typedef(self, template_name, prefix, suffix, typename=""):
        ts = []
        if self._typedef:
            ts.append("typedef %s%s::%s %s%s%s;\n" % (typename, template_name, self._typedef, prefix, self._name, suffix))
        if self._const_typedef:
            ts.append("typedef %s%s::%s const_%s%s%s;\n" % (typename, template_name, self._const_typedef, prefix, self._name, suffix))
        return "".join(ts)

    def setFwd(self, name):
        if name not in ('', 'class', 'struct'):
            raise ValueError("Error, forward declaration must be one of '', 'class' or 'struct'.")
        self._fwd = name

    def getFwd(self):
        return self._fwd

    fwd = property(getFwd, setFwd)

class Definition(object):
    def __init__(self):
        self.name = None
        self.prefix = ""
        self.suffix = ""
        self.next_item = None
        self.struct_prefix = get_unique_prefix()
        self.current_part = None
        self.equiv = False
        self.structures = {}
        self.validated = False
        self.arcs_content = {}
        self.current_arc = None
        self.current_type = "struct"
        self.current_name = None
        self.extra_typedefs = {}
        self.typename = ""
        for name in self.arguments_order:
            self.structures[name] = Structure(name, *self.typedefs.get(name, ("", "")))
        for name in self.arcs:
            self.arcs_content[name] = None
            self.structures[name] = Structure(name, *self.typedefs.get(name, ("", "")))
        for name in self.typedefs:
            if name not in self.structures:
                self.extra_typedefs[name] = Structure(name, *self.typedefs[name])

    # By default, there is nothing ...
    typedefs = {}
    arcs = {}
    default_name = None
    default_args = {}

    def type_name(self):
        return self.__class__.__name__

    def parseToken(self, token):
        if self.validated:
            raise ValueError("No token is valid between the ':' and the '{'")
        if self.next_item == 'prefix':
            self.prefix = token
            self.next_item = None
        elif self.next_item == 'suffix':
            self.suffix = token
            self.next_item = None
        elif token == 'prefix':
            if self.prefix:
                raise ValueError("Definition cannot have more than one prefix")
            self.next_item = 'prefix'
        elif token == 'suffix':
            if self.suffix:
                raise ValueError("Definition cannot have more than one suffix")
            self.next_item = 'suffix'
        elif self.name:
            raise ValueError("After the name only prefix and suffix are valid")
        else:
            self.name = token

    def in_template(self):
        self.typename = "typename "

    def add_definition(self, name):
        if name not in self.structures:
            raise ValueError("You can't define a %s in a definition of %s" % (name, self.type_name()))
        field = self.structures[name]
        if field.structure:
            raise ValueError("A %s has already been defined" % (name,))
        field.structure = self.struct_prefix % field.type_name

    def parseSection(self, name):
        write("%s:\n" % (name,))

    def parseName(self, name):
        if self.equiv:
            if self.current_part.structure:
                raise ValueError("Name of type must be one word")
            names = name.split()
            if len(names) > 2:
                raise ValueError("Name of type must be one word")
            if len(names) == 2:
                if names[0] not in ('class', 'struct', 'typename'):
                    raise ValueError("Only 'class', 'struct' or 'typename' are valid qualifiers for a type")
                if names[0] == 'typename':
                    self.current_part.structure = ' '.join(names)
                else:
                    self.current_part.fwd = names[0];
                    self.current_part.structure = names[1];
            else:
                if names[0] in ('class', 'struct', 'typename'):
                    raise ValueError("'%s' is not a valid class names" % (names[0],))
                self.current_part.structure = names[0]
        elif self.current_arc is not None:
            self.current_arc.append(name)
        elif self.current_part:
            if self.next_item == 'name':
                self.current_part.name = name
                self.next_item = None
            elif name == "named":
                self.next_item = 'name'
            else:
                raise ValueError("Part name is one single word. Just received '%s'" % (name,))
        else:
            if name in self.structures:
                self.current_part = self.structures[name]
                if name in self.arcs_content:
                    self.current_type = "arc"
                else:
                    self.current_type = "struct"
            elif name in self.extra_typedefs:
                self.current_part = self.extra_typedefs[name]
                self.current_type = "typedef"
            else:
                raise ValueError("Type '%s' unknown for define '%s'. The valid types are %s" % (name, self.type_name(), ", ".join(self.structures.keys() + self.extra_typedefs.keys())))
            self.current_name = name
            self.next_item = None
            self.equiv = False

    def initialize(self):
        self.initialized = True
        write("%s()" % (self.current_part.structure,))
        return True

    def destructor(self):
        write("~%s()" % (self.current_part.structure,))

    def startClass(self):
        if self.current_part is None:
            raise ValueError("Start of a structure without naming it")
        self.current_part.structure = self.struct_prefix % self.current_part.type_name
        write("struct %s {\n" % self.current_part.structure)

    def startEquiv(self):
        if self.current_part is None:
            raise ValueError("Start of type equivalence without the type name")
        if self.current_type == "arc":
            raise ValueError("Type equivalence not valid for an arc definition")
        if self.current_type == "typedef":
            raise ValueError("This is a typedef, it doesn't have a content.")
        self.equiv = True

    def endEquiv(self):
        self.current_part = None
        self.current_arc = None
        self.equiv = False

    def validateStruct(self):
        if self.current_part is None and self.current_arc is None:
            raise ValueError("Start of a structure without naming it")
        if self.equiv:
            raise ValueError("Choose between '=' and ':'")
        if self.current_type == 'arc':
            self.current_arc = []
            self.arcs_content[self.current_name] = self.current_arc
        if self.current_type == 'typedef':
            raise ValueError("This is a typedef, it doesn't have a content.")

    def endStruct(self):
        if self.current_type == 'arc':
            self.current_arc = None
        if self.current_part is None:
            raise ValueError("Block without name")
        write("};\n")
        self.current_part = None

    def start(self):
        if not self.validated:
            raise ValueError("Error, you forgot the ':'")

    def init_write(self):
        pass

    def validate(self):
        self.validated = True
        if self.name is None:
            if self.default_name is None:
                raise ValueError("Error, you have to specify a name for this type")
            self.name = self.default_name

    def end(self):
        if self.current_part is not None:
            self.endStruct()
            return False
        args = {}
        for name in self.arguments_order:
            st = self.structures[name]
            if st.structure is None:
                try:
                    val = self.default_args[name][0]
                    if val is None:
                        st.structure = val
                    else:
                        st.structure = val % args
                    st.fwd = self.default_args[name][1]
                except KeyError:
                    raise ValueError("Error, definition of %s requires a structure of type %s" % (self.type_name(), name))
            args[name] = st.structure
#        for name in self.structures:
#            st = self.structures[name]
#            if st.structure is None and name in self.arguments_order:
#                try:
#                    st.structure = self.default_args[name][0]
#                    st.fwd = self.default_args[name][1]
#                except KeyError:
#                    raise ValueError("Error, definition of %s requires a structure of type %s" % (self.type_name(), name))
        for name in self.structures:
            st = self.structures[name]
            if st.fwd:
                write("%s %s;\n" % (st.fwd, st.structure))
        for name in self.arcs_content:
            if self.arcs_content[name] is not None:
                write("void copy_symmetric(%(type)s& out, const %(type)s& in) {" % {'type': self.structures[self.arcs[name]].structure})
                for f in self.arcs_content[name]:
                    write("\n  out.%s = in.%s;" % (f,f))
                write("}\n")
        template_args = [ self.structures[name].structure for name in self.arguments_order ]
        self.init_write()
        write( "typedef %s%s<%s> %s;\n" % (self.typename, self.template_name, ", ".join(ta for ta in template_args if ta is not None), self.name) )
        for s in self.structures:
            write(self.structures[s].typedef(self.name, self.prefix, self.suffix, self.typename))
        for s in self.extra_typedefs:
            write(self.extra_typedefs[s].typedef(self.name, self.prefix, self.suffix, self.typename))
        return True

class VVGraph(Definition):
    template_name = "graph::VVGraph"
    default_name = "Graph"
    arguments_order = ("vertex", "edge", "compact")
    default_args = {"edge": ("graph::_EmptyEdgeContent", ""), "compact": ("false", "")}
    arcs = {"arc": "edge"}
    typedefs = {
                 "vertex": ("vertex_t",""),
                 "edge": ("edge_t", "const_edge_t"),
                 "arc": ("arc_t",""),
               }

class VVBiGraph(Definition):
    template_name = "graph::VVBiGraph"
    default_name = "BiGraph"
    arguments_order = ("vertex1", "vertex2", "edge1", "edge2", "compact")
    default_args = {"edge1": ("graph::_EmptyEdgeContent", ""), "edge2": ("graph::_EmptyEdgeContent", ""), "compact": ("false", "")}
    arcs = {}
    typedefs = {
                 "vertex1": ("vertex1_t",""),
                 "vertex2": ("vertex2_t",""),
                 "edge1": ("edge1_t", "const_edge1_t"),
                 "edge2": ("edge2_t", "const_edge2_t"),
               }

class VVComplexGraph(Definition):
    template_name = "vvcomplex::VVComplexGraph"
    default_name = "VVComplexGraph"
    arguments_order = ("cell", "junction", "cell_junction_edge", "junction_cell_edge")
    default_args = {
            "cell_junction_edge": ("graph::_EmptyEdgeContent", ""),
            "junction_cell_edge": ("graph::_EmptyEdgeContent", ""),
            "compact": ("false", ""),
            }
    arcs = {}
    typedefs = {
                 "cell": ("cell",""),
                 "junction": ("junction",""),
                 "cell_junction_edge": ("cell_junction_edge", "const_cell_junction_edge"),
                 "junction_cell_edge": ("junction_cell_edge", "const_junction_cell_edge"),
               }

class VVComplex(Definition):
    template_name = "vvcomplex::VVComplex"
    default_name = "Complex";
    arguments_order = ("model", "cell", "junction", "wall", "cell_edge", "cell_junction_edge", "junction_cell_edge", "compact")
    arcs = {"wall_arc": "wall", "cell_arc": "cell_edge"}
    default_args = {
            "model": ("VVModel", "class"),
            "wall": ("graph::_EmptyEdgeContent", ""),
            "cell_edge": ("graph::_EmptyEdgeContent", ""),
            "cell_junction_edge": ("graph::_EmptyEdgeContent", ""),
            "junction_cell_edge": ("graph::_EmptyEdgeContent", ""),
            "compact": ("false", ""),
            }
    typedefs = {
                 "cell": ("cell",""),
                 "junction": ("junction",""),
                 "wall": ("wall", "const_wall"),
                 "cell_edge": ("cell_edge", "const_cell_edge"),
                 "cell_junction_edge": ("cell_junction_edge", "const_cell_junction_edge"),
                 "junction_cell_edge": ("junction_cell_edge", "const_junction_cell_edge"),
                 "wall_arc": ("wall_arc",""),
                 "cell_arc": ("cell_arc",""),
                 "cell_graph": ("cell_graph", ""),
                 "wall_graph": ("wall_graph", ""),
                 "complex_graph": ("complex_graph", ""),
               }

class Tissue(VVComplex):
    template_name = "tissue::Tissue"
    default_name = "Tissue"

class List(Definition):
    template_name = "std::list"
    arguments_order = ("content","alloc")
    default_args = {
            "alloc": ("std::allocator<%(content)s> ", "")
            }

class Vector(Definition):
    template_name = "std::vector"
    arguments_order = ("content","alloc")
    default_args = {
            "alloc": ("std::allocator<%(content)s> ", "")
            }

class Queue(Definition):
    template_name = "std::queue"
    arguments_order = ("content","container")
    default_args = {
            "container": ("std::deque<%(content)s> ", "")
            }

class PriorityQueue(Definition):
    template_name = "std::priority_queue"
    arguments_order = ("content","container","compare")
    default_args = {
            "container": ("std::deque<%(content)s> ", ""),
            "compare": (None,"")
            }

class Deque(Definition):
    template_name = "std::deque"
    arguments_order = ("content","alloc")
    default_args = {
            "alloc": ("std::allocator<%(content)s> ", "")
            }

class Set(Definition):
    template_name = "std::set"
    arguments_order = ("key", "compare", "alloc")
    default_args = {
            "compare": ("std::less<%(key)s> ", ""),
            "alloc": ("std::allocator<%(key)s> ", "")
            }

class MultiSet(Set):
    template_name = "std::multiset"

class Map(Definition):
    template_name = "std::map"
    arguments_order = ("key", "value", "compare", "alloc")
    default_args = {
            "compare": ("std::less<%(key)s> ", ""),
            "alloc": ("std::allocator<std::pair<const %(key)s,%(value)s> > ", "")
            }

class MultiMap(Map):
    template_name = "std::multimap"

class UnorderedSet(Definition):
    template_name = "std::unordered_set"
    arguments_order = ("key", "hash", "pred", "alloc")
    default_args = {
            "hash": ("std::hash<%(key)s> ", ""),
            "pred": ("std::equal_to<%(key)s> ", ""),
            "alloc": ("std::allocator<%(key)s> ", "")
            }

class UnorderedMultiSet(UnorderedSet):
    template_name = "std::unordered_multiset"

class UnorderedMap(Definition):
    template_name = "std::unordered_map"
    arguments_order = ("key", "value", "hash", "pred", "alloc")
    default_args = {
            "hash": ("std::hash<%(key)s> ", ""),
            "pred": ("std::equal_to<%(key)s> ", ""),
            "alloc": ("std::allocator<std::pair<const %(key)s,%(value)s> > ", "")
            }

class UnorderedMultiMap(UnorderedMap):
    template_name = "std::unordered_multimap"

class Model(object):
    def __init__(self):
        self.name = None # "VVModel"
        self.initialized = False
        self.finalized = False
        self.valid = False

    def validate(self):
        self.valid = True

    def parseName(self, name):
        self.name = name

    def parseSection(self, name):
        write("%s:\n" % (name,))

    def startClass(self):
        if not self.valid:
            raise ValueError("You forgot the ':' to finish the model declaration")
        if self.name is None:
            self.name = "VVModel"
        write( """class %(name)s : public Model {
                    Q_OBJECT
                  public: """ % {'name': self.name})

    def initialize(self):
        self.initialized = True
        write("""%s(QObject *parent)
    : Model(parent)""" % (self.name,))
        return False

    def destructor(self):
        write("~%s()" % (self.name,))

    def end(self):
        if not self.initialized:
            self.initialize()
            write("{ }")
        if self.finalized:
            raise ValueError("Model was already finalized!")
        self.finalized = True
        write("};\n")
        return True

class Viewer(object):
    def __init__(self):
        self.name = None # "MyViewer"
        self.initialized = False
        self.finalized = False
        self.valid = False

    def validate(self):
        self.valid = True

    def parseName(self, name):
        self.name = name

    def parseSection(self, name):
        write("%s:\n" % (name,))

    def startClass(self):
        if not self.valid:
            raise ValueError("You forgot the ':' to finish the viewer declaration")
        if self.name is None:
            self.name = "MyViewer"
        write( """class %(name)s : public Viewer {
                    Q_OBJECT
                    public: """ % {'name': self.name})

    def initialize(self, full=False):
        self.initialized = True
        write( """%s(QWidget *parent, const QGLWidget *shareWidget = 0, int idx = -1)
                    : Viewer(parent, shareWidget, idx)""" % (self.name,))
        return False

    def destructor(self):
        write("~%s()" % (self.name,))

    def end(self):
        if not self.initialized:
            self.initialize()
            write("{ }")
        if self.finalized:
            raise ValueError("Model was already finalized!")
        self.finalized = True
        write( "};\n")
        return True

import sys
import os
import os.path
import inspect

def checkDefinition(sym, obj, filename):
    def reportError(s):
        raise ImportError("Error importing definition of '%s' from file '%s': %s" % (sym, filename, s))
    if not hasattr(obj, "template_name"):
        reportError("no template name was specified")
    if not hasattr(obj, "arguments_order"):
        reportError("no argument was specified")

# Look for extra definitions
def load_file(filename):
    if os.path.exists(filename):
        defs = {}
        previous = globals()
        execfile(filename, previous, defs)
        for sym in defs:
            obj = defs[sym]
            if inspect.isclass(obj) and issubclass(obj, Definition):
                checkDefinition(sym, obj, filename)
                if sym in previous:
                    print >> sys.stderr, "Warning, definition of '%s' overriden in file '%s'" % (sym, filename)
                previous[sym] = obj

def load_extra():
    try:
        home = os.environ.get("HOME")
        if home is not None and os.path.exists(home) and os.path.isdir(home):
            load_file(os.path.join(home, "vvl_types.py"))
            load_file(os.path.join(home, ".vlab", "vvl_types.py"))
        load_file("./vvl_types.py")
    except ImportError, err:
        print >> sys.stderr, str(err)
        sys.exit(1)

load_extra()

