/**
 * \file algorithms/cellsystem.h
 *
 * Include the definitions necessary for the cell system division algorithm and 
 * model
 */

#ifndef VVELIB_ALGORITHMS_CELLSYSTEM_H
#define VVELIB_ALGORITHMS_CELLSYSTEM_H

#include <config.h>
#include <algorithms/complex.h>
#include <geometry/intersection.h>
#include <geometry/area.h>
#include <util/matrix.h>
#include <model.h>
#include <util/parms.h>
#include <sstream>
#include <algorithms/tissue.h>
#include <storage/storage.h>
#include <storage/complex.h>

#include <iostream>

/**
 * Namespace defining the cell system division algorithm and model.
 */
namespace cell_system
{
#define EPSILON 0.00001

  using namespace vvcomplex;

  using geometry::Point3d;

  /**
   * Convenience function that removes white spaces before and after the string
   */
  std::string removeWhitespace(std::string s);

  /**
   * \class CellSystemDivisionParams cellsystem.h <algorithms/cellsystem.h>
   *
   * Parameters for cell-system division.
   */
  struct CellSystemDivisionParams
  {
    /**
     * Constructor. Only the direction has to be specified.
     */
    CellSystemDivisionParams(const Point3d& dir, double a = 0, double r = 0.5, double eps=0.000001,
                             bool pinch=false, tissue::CellPinchingParams pp = tissue::CellPinchingParams())
      : angle(a)
      , ratio(r)
      , direction(dir)
      , epsilon(eps)
      , minCellWall(0)
      , pinching(pinch)
      , pinchingParam(pp)
    { }

    /**
     * Copy constructor
     */
    CellSystemDivisionParams(const CellSystemDivisionParams& copy)
      : angle(copy.angle)
      , ratio(copy.ratio)
      , direction(copy.direction)
      , epsilon(copy.epsilon)
      , minCellWall(0)
      , pinching(copy.pinching)
      , pinchingParam(copy.pinchingParam)
    { }

    /**
     * Angle between the direction and the normal to the wall
     */
    double angle;

    /**
     * Ratio of the area from the daughter cell opposite from the normal and 
     * the area of the mother cell.
     *
     * If ratio is 0, then just keep the new wall going through the center.
     */
    double ratio;

    /**
     * Reference direction to find the normal of the new wall.
     */
    Point3d direction;

    /**
     * Relative epsilon for floating point comparison.
     */
    double epsilon;

    /**
     * Minimum size of a cell wall
     */
    double minCellWall;

    /**
     * Ask for pinching after the cell wall was found
     */
    bool pinching;

    /**
     * Parameters for cell pinching
     */
    tissue::CellPinchingParams pinchingParam;
  };

  /**
   * Find the division points for a wall going through a given point.
   *
   * \param[in] c Cell to be divided
   * \param[out] center Point defining the line of the division wall. At the 
   * end, \c center is the center of the division wall.
   * \param[in] T Complex containing the cell to divide
   * \param[in] direction Normal to the division wall
   *
   * The division wall is a segment on the line going through \c center and 
   * with \c direction as normal. */
  template <class Complex>
  DivisionData<typename Complex::junction_content>
  findDivisionPoints(const typename Complex::cell& c, Point3d& center, Complex& T,
                     const Point3d& direction)
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_MODEL(Complex, T);
    DivisionData<typename Complex::junction_content> result;
//    cout << "Finding points for center: " << center << endl;
    forall(const junction& v, T.S.neighbors(c))
    {
      const junction& n = T.S.nextTo(c, v);
      const Point3d& p1 = model.position(v);
      const Point3d& p2 = model.position(n);
      Point3d u;
      double s;
      if(geometry::planeLineIntersection(u, s, center, direction, p1, p2) && s >= 0 && s <= 1)
      {
//        cout << "Intersection between " << p1 << " and " << p2 << " at " << u << " with s = " << s << endl;
        if((p1-center)*direction > 0)
        {
//          cout << "Found v1 at " << model.vertexPosition(v) << " with pv = " << u << endl;
//          if(result.v1)
//          {
//            cout << "Warning, finding v1 again" << endl;
//          }
          result.v1 = v;
          result.pv = u;
        }
        else
        {
//          cout << "Found u1 at " << model.vertexPosition(v) << " with pu = " << u << endl;
//          if(result.u1)
//          {
//            cout << "Warning, finding v1 again" << endl;
//          }
          result.u1 = v;
          result.pu = u;
        }
        if(result.v1 && result.u1)
          break;
      }
    }
    if(result.u1 && result.v1)
    {
      center = (result.pu+result.pv)/2;
    }
    return result;
  }

  /**
   * Compute the volume of the left cell of a division
   *
   * Do so with just the division data.
   *
   * \param c Cell being divided
   * \param T Complex containing the cell c
   * \param center A point on the division wall
   * \param result Division data being used
   * \param model Model is use
   */
  template <class Complex>
  double volumeLeftCell(const typename Complex::cell& c, Complex& T, const Point3d& center,
                        const DivisionData<typename Complex::junction_content>& result)
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_MODEL(Complex, T);
    double area = 0;
    junction v = result.v1;
    Point3d prev = result.pv;
    do
    {
      v = T.S.nextTo(c, v);
      const Point3d& cur = model.position(v);
      area += geometry::triangleArea(center, prev, cur);
      prev = cur;
    } while(v != result.u1);
    area += geometry::triangleArea(center, prev, result.pu);
    return area;
  }

  /**
   * Division algorithm for cell-system.
   *
   * \see CellSystemDivisionParams
   */
  template <class Complex>
  DivisionData<typename Complex::junction_content>
  findDivisionPoints(const typename Complex::cell& c, Complex& T,
                     const CellSystemDivisionParams& params)
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_MODEL(Complex, T);
    const Point3d& cell_normal = model.normal(c);
    Point3d direction = util::Matrix<3,3,double>::rotation(cell_normal, params.angle) * params.direction;
//    cout << "Z = " << params.direction << " - angle = " << params.angle << " => direction = " << direction << endl;
//    cout << "Polygon: ";
//    forall(const vertex& j, S.neighbors(cell))
//    {
//      cout << model.vertexPosition(j) << "; ";
//    }
//    cout << endl;
    Point3d center = model.position(c);
    // First, find the division points going through the center
    DivisionData<typename Complex::junction_content> result =
      findDivisionPoints(c, center, T, direction);
    if(!result)
    {
      return DivisionData<typename Complex::junction_content>();
    }

    // Then, if ratio != 0 then adjust volume by dichotomy
    if(params.ratio)
    {
      // Volume of the cell
      double total_volume = model.area(c);
      double target_volume = total_volume*params.ratio;
      double vol_epsilon = params.epsilon * target_volume;
      double proj_min = HUGE_VAL, proj_max = -HUGE_VAL;
      Point3d pos_max, pos_min;
      forall(const junction& j, T.S.neighbors(c))
      {
        const Point3d& jpos = model.position(j);
        double proj = (jpos - center) * direction;
        if(proj > proj_max)
        {
          proj_max = proj;
          pos_max = jpos;
        }
        if(proj < proj_min)
        {
          proj_min = proj;
          pos_min = jpos;
        }
      }
      double current_volume = volumeLeftCell(c, T, center, result);
      Point3d delta;
      if(current_volume < target_volume)
      {
        delta = proj_max*direction;
      }
      else
      {
        delta = -proj_min*direction;
      }
//      cout << endl << endl;
//      cout << "Total volume = " << total_volume << " => target_volume = " << target_volume << endl;
//      cout << "Pos max = " << pos_max << " - Pos min = " << pos_min << endl;
//      cout << "delta = " << delta << endl;
      while(std::abs(current_volume - target_volume) > vol_epsilon)
      {
        delta /= 2;
        if(current_volume < target_volume)
        {
          center += delta;
        }
        else
        {
          center -= delta;
        }
//        cout << "New delta = " << delta << " - new center = " << center << 
//        endl;
        result = findDivisionPoints(c, center, T, direction);
        current_volume = volumeLeftCell(c, T, center, result);
//        cout << "pu = " << result.pu << " - pv = " << result.pv << endl;
      }
//      cout << "Volume = " << current_volume << endl;
    }
    if(params.pinching)
    {
      tissue::cellPinching(c, T, result, params.pinchingParam);
    }
    testDivisionOnVertices(c, result, T, std::max(params.epsilon,params.minCellWall));
    return result;
  }

  /**
   * Type of the label of a cell (for full cell-system only)
   */
  typedef int label_t;

  /**
   * \class CellSystemCell cellsystem.h <algorithms/cellsystem.h>
   *
   * Content of a vertex for the 2D cell system
   */
  struct CellSystemCell
  {
    /**
     * Area of the cell
     */
    double area;
    /**
     * Position of the vertex
     */
    Point3d pos;
    /**
     * Label of the cell
     */
    label_t label;

    /**
     * Serialization method
     */
    bool serialize(storage::VVEStorage&);
  };

  /**
   * \class CellSystemJunction cellsystem.h <algorithms/cellsystem.h>
   *
   * Content of a vertex for the 2D cell system
   */
  struct CellSystemJunction
  {
    /**
     * Position of the vertex
     */
    Point3d pos;
    /**
     * Force of the junction
     */
    Point3d force;
    /**
     * Velocity of the junction
     */
    Point3d velocity;

    /**
     * Serialization method
     */
    bool serialize(storage::VVEStorage&);
  };

  /**
   * \class DivisionParams cellsystem.h <algorithms/cellsystem.h>
   *
   * Structure storing the right hand side of a division rule
   */
  struct DivisionParams
  {
    DivisionParams()
      : left_label(0)
      , right_label(0)
      , angle(0)
      , ratio(0)
    { }

    DivisionParams(label_t ll, label_t rl, double a, double r = 0.5)
      : left_label(ll)
      , right_label(rl)
      , angle(a)
      , ratio(r)
    { }

    DivisionParams(const DivisionParams& copy)
      : left_label(copy.left_label)
      , right_label(copy.right_label)
      , angle(copy.angle)
      , ratio(copy.ratio)
    { }

    /**
     * Label of the "left" daughter cell
     */
    label_t left_label;
    /**
     * Label of the "right" daughter cell
     */
    label_t right_label;
    /**
     * Angle between the reference vector and the normal to the new wall
     */
    double angle;
    /**
     * Volume ratio of the left cell
     *
     * Volume of the left cell divided by the volume of the original cell
     */
    double ratio;
  };

  /**
   * \class CellSystem cellsystem.h <algorithms/cellsystem.h>
   * Full cell-system model
   *
   * Inherit this model to use cell-system with rules defined in the parameter 
   * file.
   *
   * \param Complex Cell complex class to be used (allows for sub-type of 
   * vvcomplex::VVComplex as tissue::Tissue)
   * \param Model Subclass used. Useful to initialise the cell complex class
   */
  template <class Complex, class MyModel>
  struct CellSystem : public Model
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_EDGES(Complex);
    IMPORT_COMPLEX_GRAPHS(Complex);

    /**
     * Cell complex
     */
    Complex T;
    /**
     * Time step for mechanical simulation
     */
    double dt;
    /**
     * Time step between two images
     */
    double showInterval;
    /**
     * Time since the last image
     */
    double showTime;
    /**
     * Stiffness of an external spring
     */
    double ke;
    /**
     * Stiffness of an internal spring
     */
    double ki;
    /**
     * Pressure coefficient in a cell
     */
    double pressure;
    /**
     * Damping parameter
     */
    double damping;
    /**
     * Restlength of the springs
     */
    double restLength;
    /**
     * Stability criteria
     */
    double stability;
    /**
     * Mass of the junctions
     */
    double mass;
    /**
     * Current time
     */
    double time;

    /**
     * Type of rules of type A --> B
     */
    typedef std::map<label_t, label_t> label_change_rules_t;
    /**
     * Type of rules of type A --> B |(a,r) C
     */
    typedef std::map<label_t, DivisionParams> division_rules_t;
    /**
     * Rules of type A --> B
     */
    label_change_rules_t label_change_rules;
    /**
     * Rules of type A --> B |(a,r) C
     */
    division_rules_t division_rules;
    /**
     * Mapping from the label names to the label numbers
     */
    std::map<QString, label_t> label_numbers;
    /**
     * Mapping from the label numbers to the label strings
     */
    std::vector<QString> label_names;

    /**
     * Current division rule
     *
     * Used to setup the cell labels in updateFromOld
     */
    DivisionParams current_div;

    /**
     * Set to true to display the rules when analyzed
     */
    bool viewRulesDefinitions;
    /**
     * Set to true to display the rules when applied
     */
    bool viewRulesApplication;

    /**
     * Update the area and the center of all the cells
     */
    void updateCellsArea()
    {
      forall_named( const cell& c, T.S, cells)
      {
        Point3d center;
        double np = 0;
        c->area = 0.0;
        forall(const junction& n, T.S.neighbors(c))
        {
          // Find area
          const junction& m = T.S.nextTo(c, n);
          c->area += geometry::triangleArea(c->pos, m->pos, n->pos);
          center += n->pos;
          np++;
        }
        center /= np;
        c->pos = center;
        if(c->area <= EPSILON ) // Avoid inf
        {
          cout << "Bad area " << c->area << endl;
          c->area = 0.1;
        }
      }
    }

    /**
     * Read the mechanical parameters
     *
     * The mechanical parameters include:
     *  * dt: Time step for mechanical simulation
     *  * ki: Stiffness of internal springs
     *  * ke: Stability of external springs
     *  * ShowInterval: Time step between two images
     *  * Pressure: Pressure coefficient
     *  * Damping: Damping coefficient
     *  * RestLength: Rest length of the springs
     *  * Stability: Max force allowed on a junction to consider the system at 
     *  equilibrium
     *  * Mass: Mass of a junction
     *
     * \param parms Parameter object to read from
     * \param section Section containing the mechanical parameters
     */
    void readMechanicParam(util::Parms& parms, const QString& section)
    {
      parms(section, "dt", dt);
      parms(section, "ki", ki);
      parms(section, "ke", ke);
      parms(section, "ShowInterval", showInterval);
      if(showInterval < dt)
        showInterval = dt;
      parms(section, "Pressure", pressure);
      parms(section, "Damping", damping);
      parms(section, "RestLength", restLength);
      parms(section, "Stability", stability);
      parms(section, "Mass", mass);
    }

    /**
     * Returns the label number of a symbol
     *
     * \param name Name of the symbol
     *
     * If the symbol do not exist, it is created
     */
    label_t label(QString name)
    {
      std::map<QString,label_t>::const_iterator found = label_numbers.find(name);
      if(found != label_numbers.end())
      {
        return found->second;
      }
      cout << "Error, symbol '" << name.toStdString() << "' used but not defined" << endl;
      return -1;
    }


    /**
     * Read the list of symbols.
     *
     * The symbols are the keys and may have values (but it is not necessary)
     */
    void readSymbols(util::Parms& parms, const QString& section)
    {
      typedef std::map<QString, std::vector<QString> > rmap;
      rmap symbols;
      parms.all(section, symbols);
      if(viewRulesDefinitions)
        util::out << "Reading " << symbols.size() << " symbols from section [" << section << "]" << endl;
      for(rmap::iterator it = symbols.begin() ; it != symbols.end() ; ++it)
      {
        QString name = it->first;
        registerSymbol(name);
      }
    }

    /**
     * Register a new symbol to be used in the cell system.
     *
     * If the symbol already exists, return the id of the existing one.
     *
     * \params name Name of the new symbol
     *
     * \returns The id of the symbol
     */
    label_t registerSymbol(QString name)
    {
      std::map<QString, label_t>::iterator found = label_numbers.find(name);
      if(found == label_numbers.end())
      {
        label_t new_label = label_names.size();
        label_names.push_back(name);
        label_numbers[name] = new_label;
        if(viewRulesDefinitions)
          cout << "Defining symbol '" << name.toStdString() << "' as value " << new_label << endl;
        return new_label;
      }
      return found->second;
    }

    /**
     * Read a prepare the cell-system rules
     *
     * \param parms Parameter object to read from
     * \param section Section containing the rules (and only them)
     *
     * \warning readSymbols() has to be called \b before readRules
     *
     * The rules can take three form:
     *  # A: B, correspond to a relabelling A --> B
     *  # A: B |(a,r) C, correspond to a cell division. \c a is the angle 
     *  between the reference vector and the normal to the new wall, \c r is 
     *  the volume ration VB/VA. B and C and the labels of the new cells. If \c 
     *  r is 0 then the wall goes through the center of the cell.
     *  # A: B |(a) C, is the same as the above rule but with \c r = 0.5.
     *
     * There can be only one rule per symbol (i.e. no two rule can share the 
     * same left hand side symbol).
     */
    void readRules(util::Parms& parms, const QString& section)
    {
      typedef std::map<QString, std::vector<QString> > rmap;
      rmap rules;
      parms.all(section, rules);

      for(rmap::iterator it = rules.begin() ; it != rules.end() ; ++it)
      {
        QString name = it->first;
        label(name);
      }

      label_change_rules.clear();
      division_rules.clear();

      for(rmap::iterator it = rules.begin() ; it != rules.end() ; ++it)
      {
        label_t lhs = label_numbers[it->first];
        if(it->second.size() > 1)
        {
          cout << "Error, more than one rule for '" << lhs << "' ... ignoring" << endl;
          continue;
        }
        QString rhs = it->second[0];
        int pos = rhs.indexOf('|');
        if(pos == -1)
        {
          label_t new_label = label(rhs);
          label_change_rules[lhs] = new_label;
          if(viewRulesDefinitions)
            cout << "Rule: " << label_names[lhs].toStdString() << " --> " << label_names[new_label].toStdString() << endl;
        }
        else
        {
          if(pos == -1)
          {
           cout << "Error, division rule without '|': " << lhs << " --> " << rhs.toStdString() << endl;
           continue;
          }
          QString left_name = rhs.mid(0, pos).trimmed();
          label_t left_label = label(left_name);
          int pos_left_paren, pos_right_paren, pos_comma;
          pos_left_paren = rhs.indexOf('(', pos);
          pos_comma = rhs.indexOf(',', pos_left_paren);
          pos_right_paren = rhs.indexOf(')', pos_left_paren);
          if(pos_left_paren == -1)
          {
           cout << "Error, division rule without '(': " << lhs << " --> " << rhs.toStdString() << endl;
           continue;
          }
          if(pos_right_paren == -1)
          {
           cout << "Error, division rule without ')': " << lhs << " --> " << rhs.toStdString() << endl;
           continue;
          }
          QString right_name = rhs.mid(pos_right_paren+1).trimmed();
          label_t right_label = label(right_name);
          if(pos_comma != -1)
          {
            QString angle_str = rhs.mid(pos_left_paren+1, pos_comma-pos_left_paren-1);
            QString ratio_str = rhs.mid(pos_comma+1, pos_right_paren-pos_comma-1);
            bool ok;
            double angle = angle_str.toDouble(&ok);
            if(!ok)
            {
              cout << "Error, cannot read angle (" << angle_str.toStdString() << ") in " << lhs << " --> " << rhs.toStdString() << endl;
              continue;
            }
            angle *= M_PI/180;
            double ratio = ratio_str.toDouble(&ok);
            if(!ok)
            {
              cout << "Error, cannot read ratio (" << ratio_str.toStdString() << ") in " << lhs << " --> " << rhs.toStdString() << endl;
              continue;
            }
            division_rules[lhs] = DivisionParams(left_label, right_label, angle, ratio);
          }
          else
          {
            QString angle_str = rhs.mid(pos_left_paren+1, pos_right_paren-pos_left_paren-1);
            bool ok;
            double angle = angle_str.toDouble(&ok);
            if(!ok)
            {
              cout << "Error, cannot read angle (" << angle_str.toStdString() << ") in " << lhs << " --> " << rhs.toStdString() << endl;
              continue;
            }
            angle *= M_PI/180;
            division_rules[lhs] = DivisionParams(left_label, right_label, angle);
          }
          const DivisionParams& dp = division_rules[lhs];
          if(viewRulesDefinitions)
            cout << "Rule: " << label_names[lhs].toStdString() << " --> " << label_names[dp.left_label].toStdString() << " |("
              << (dp.angle*180/M_PI) << ","
              << dp.ratio << ") " << label_names[dp.right_label].toStdString() << endl;
        }
      }
      cout << endl;
    }

    /**
     * Read the parameters.
     *
     * Usefull if no other parameters are needed (i.e. very unlikely)
     */
    virtual void readParam()
    {
      util::Parms parms("view.v");
      readMechanicParam(parms, "Mechanic");
      readSymbols(parms, "Symbols");
      readRules(parms, "Rules");
    }

    void modifiedFiles( const std::set<QString>& filenames )
    {
      if(filenames.count("view.v"))
        readParam();
    }

    /**
     * Constructor ...
     */
    CellSystem( QObject* parent )
      : Model( parent )
      , T((MyModel*)this)
      , showTime(0)
      , time(0)
      , viewRulesDefinitions(false)
      , viewRulesApplication(false)
    {
    }

    /**
     * Default step method, just applies the rules and do the mechanics
     */
    void step()
    {
      step_cellsystem();
    }

    /**
     * Compute the mecanic stability of the system for a time showInterval
     *
     * Return true if the system is stable, false otherwise.
     */
    bool step_cellsystem_meca()
    {
      time += dt;
      bool stable;
      do
      {
        showTime += dt;
        // First, mechanical model
        updateCellsArea();
        forall(const junction& v, T.W)
        {
          v->force = 0;
        }
        forall_named(const cell& c, T.S, cells)
        {
          double area= c->area;
          // First, compute the forces due to the pressure
          forall(const junction& j, T.S.neighbors(c))
          {
            const junction& k = T.S.nextTo(c, j);
            Point3d u = k->pos - j->pos;
            double l = util::norm(u);
            u /= l;
            Point3d n(u.y(), -u.x(), 0);
            n *= pressure*l/area/2;
            j->force += n;
            k->force += n;
          }
        }
        // Second, compute the forces due to the springs
        forall(const junction& j, T.W)
        {
          const Point3d& jpos = j->pos;
          forall(const junction& k, T.W.neighbors(j))
          {
            double ks = ki;
            if(T.border(j, k))
              ks = ke;
            Point3d u = k->pos - jpos;
            double l = util::norm(u);
            u *= ks*(1 - restLength/l);
            j->force += u;
          }
        }

        // Now, integrate and check for stability
        stable = true;
        forall(const junction& j, T.W)
        {
          if(util::norm(j->force) > stability) stable = false;
          j->velocity += (j->force - damping*j->velocity)/mass*dt;
          j->pos += j->velocity*dt;
        }
      } while(showTime < showInterval);
      showTime -= showInterval;

      return stable;
    }

    /**
     * Divide the cells according to their current geometry and the rules in 
     * the parameter file.
     */
    void step_cellsystem_division()
    {
      updateCellsArea();

      // Second, division
      std::vector<cell> cells(T.C.begin(), T.C.end());
      forall(const cell& c, cells)
      {
        label_t label = c->label;
        label_change_rules_t::const_iterator label_found = label_change_rules.find(label);
        division_rules_t::const_iterator division_found = division_rules.find(label);
        if(label_found != label_change_rules.end())
        {
          if(viewRulesApplication)
            cout << "Applying rule: " << c->label << " --> " << label_found->second << endl;
          c->label = label_found->second;
        }
        else if(division_found != division_rules.end())
        {
          const DivisionParams& div = division_found->second;
          if(viewRulesApplication)
            cout << "Applying rule: " << c->label << " --> "
              << div.left_label << " |(" << (div.angle*180/M_PI) << "," << div.ratio << ") "
              << div.right_label << endl;
          current_div = div;
          Point3d ref = ((MyModel*)this)->referenceVector(c);
          ref.normalize();
          CellSystemDivisionParams parms(ref, div.angle, div.ratio, EPSILON);
          T.divideCell(c, parms, c);
        }
      }
    }

    /**
     * Cell system step
     *
     * Returns true if the rules were applied, false if just the mechanic was 
     * done (i.e. the system is not stabilised)
     */
    bool step_cellsystem()
    {
      if(!this->step_cellsystem_meca())
        return false;

      step_cellsystem_division();
      return true;
   }

    // Methods required by cell system library

    double area(const cell& c) const
    {
      return c->area;
    }

    Point3d position( const cell& v) const
    {
      return v->pos;
    }

    Point3d position( const junction& v) const
    {
      return v->pos;
    }

    void setPosition( const cell& v, const Point3d& pos)
    {
      v->pos = pos;
    }

    void setPosition( const junction& v, const Point3d& pos)
    {
      v->pos = pos;
    }

    void setPositionHint( const junction& , const junction& , const junction& , double )
    {
    }

    Point3d normal( const junction& ) const
    {
      return Point3d(0,0,1);
    }

    Point3d normal( const cell& ) const
    {
      return Point3d(0,0,1);
    }

    void updateFromOld( const cell& cl, const cell& cr, const cell& c,
                        const typename Complex::division_data&, Complex& T)
    {
      cl->area = 0;
      cl->label = current_div.left_label;
      forall( const junction& n, T.S.neighbors(cl))
      {
        // Find area
        const junction& m = T.S.nextTo(cl, n);
        cl->area += geometry::triangleArea(cl->pos, m->pos, n->pos);
      }
      cr->area = 0;
      cr->label = current_div.right_label;
      forall( const junction& n, T.S.neighbors(cr))
      {
        // Find area
        const junction& m = T.S.nextTo(cr, n);
        cr->area += geometry::triangleArea(cr->pos, m->pos, n->pos);
      }
      if(current_div.ratio > 0)
      {
        double target_volume = current_div.ratio*c->area;
        double vol_epsilon = target_volume*EPSILON;
        if(std::abs(cl->area - target_volume) > vol_epsilon)
        {
          cout << "Error, bad division volume : " << cl->area << " instead of " << target_volume << endl;
        }
        if(std::abs(cl->area+cr->area-c->area)/c->area > 0.001)
        {
          cout << "Error, sum of cl and cr is not c: " << cl->area << " + " << cr->area << " = " << (cl->area +cr->area) << " != " << c->area << endl;
        }
      }
    }

    int versionNumber(const QString& version)
    {
      return storage::versionNumber(version);
    }

    QString version() const
    {
      return "CellSystemModel 1.0";
    }

    bool serialize(storage::VVEStorage& store)
    {
      if(!store.field("T", T))
      {
        return false;
      }
      if(!store.field("Time", time))
      {
        std::cerr << "Warning, cell system saved without time. Setting time to 0" << endl;
        time = 0;
      }
      return true;
    }

  };

#undef EPSILON
}
#endif // VVELIB_ALGORITHMS_CELLSYSTEM_H
