#ifndef VVELIB_ALGORITHMS_COMPLEX_H
#define VVELIB_ALGORITHMS_COMPLEX_H

/**
 * \file algorithms/complex.h
 *
 * Definition of a 2D cell complex.
 */

#include <config.h>
#include <graph/vvgraph.h>
#include <graph/vvbigraph.h>
#include <algorithms/insert.h>
#include <util/parms.h>
#include <util/palette.h>
#include <util/color.h>
#include <util/leaf_class.h>
#include <util/assert.h>
#include <geometry/intersection.h>
#include <geometry/area.h>
#include <storage/fwd.h>

#include <iostream>

/**
 * \namespace vvcomplex
 *
 * \brief Definition of classes and functions for 2d cell complexes.
 *
 * \section cplx_GenInfo General Information
 *
 * This namespace mainly provides a VVComplex class template handling the 
 * creation and evolution of cell complexes.
 *
 * A 2D cell %complex is defined by a set of regions of the space such that two 
 * regions neighbor regions share a continuous set of 1D cells (i.e. line 
 * segments). We represent here the regions as a bipartite %graph where the two 
 * group of vertices are the cells and the junctions. The junctions neighbors 
 * of a cell are ordered such that two junctions next to each other are linked 
 * by a wall. The cells neighbors of a junction are ordered such that two cells 
 * sharing a wall will be next to each other. In addition, two useful graphs 
 * are maintained: the cell graph, linking cells sharing a wall; and the wall 
 * graph, linking junctions parts of the same wall.
 *
 * \image html vve_tissue1.png "Graphs and ordering in 2D cell complexes."
 * \image latex vve_tissue1.png "Graphs and ordering in 2D cell complexes."
 *
 * By default, all graphs are ordered counter-clockwise, but any consistent 
 * ordering is possible.
 *
 * \note The %complex %graph is the only one defining the cell %complex. The 
 * cell %graph can be defined as the %graph linking two cells sharing two 
 * vertices next to each other in their neighborhood. The wall %graph can be 
 * defined as the graph linking two vertices if there exist a cell in which the 
 * two vertices are next to each other.
 *
 * \section cplx_Usage 2D Cell Complex Usage
 *
 * \subsection cplx_Def Definition of a 2D Cell Complex
 *
 * A 2D cell %complex needs 6 types and the type of the model. The 6 types are:
 *  -# The content of a cell
 *  -# The content of a junction
 *  -# The content of a wall (i.e. an edge in the wall %graph)
 *  -# The content of an edge between two cells
 *  -# The content of an edge from a cell to a junction
 *  -# The content of an edge from a junction to a cell
 *
 * For the rest of this page, these types will be called:
 *  -# \c CellContent
 *  -# \c JunctionContent
 *  -# \c WallContent
 *  -# \c CellEdgeContent
 *  -# \c CellJunctionEdgeContent
 *  -# \c JunctionCellEdgeContent
 *
 * Then, if the type of the model is \c MyModel, the cell %complex is defined 
 * as:
 *
 * \code
 * typedef vvcomplex::VVComplex<MyModel,CellContent,JunctionContent,WallContent,CellEdgeContent,CellJunctionEdgeContent,JunctionCellEdgeContent> VVComplex;
 * \endcode
 *
 * All edge content types are optional and defaults to an empty class. Once 
 * created, a macro is provided to help exporting all the types:
 * \code
 * EXPORT_COMPLEX_TYPES(VVComplex);
 * \endcode
 * is equivalent to:
 * \code
 * typedef VVComplex::cell cell;
 * typedef VVComplex::junction junction;
 *
 * typedef VVComplex::wall wall;
 * typedef VVComplex::cell_edge cell_edge;
 * typedef VVComplex::cell_junction_edge cell_junction_edge;
 * typedef VVComplex::junction_cell_edge junction_cell_edge;
 *
 * typedef VVComplex::const_wall const_wall;
 * typedef VVComplex::const_cell_edge const_cell_edge;
 * typedef VVComplex::const_cell_junction_edge const_cell_junction_edge;
 * typedef VVComplex::const_junction_cell_edge const_junction_cell_edge;
 *
 * typedef VVComplex::complex_graph complex_graph;
 * typedef VVComplex::cell_graph cell_graph;
 * typedef VVComplex::wall_graph wall_graph;
 * \endcode
 *
 * Other macro, for partial export or for import in template function or class, 
 * are defined in complex.h
 *
 * \subsection cplx_Edit Creation and Edition of a 2D Cell Complex
 *
 * 2D cell complexes are created by instantiating the type defined in the 
 * previous section, specifying at construction time the pointer on the current 
 * model. In a function, the declaration will look like:
 * \code
 * VVComplex T(&model);
 * \endcode
 *
 * As in the model class, the declaration will look like:
 * \code
 * VVComplex T;
 * MyModel(QObject* parent)
 *   : Model(parent)
 *   , T(this)
 * \endcode
 *
 * Once created, the first cell can be created by creating first junctions and 
 * a cell, then using the method VVComplex::addCell. For example to create 
 * a square cell (supposing the junctions have a \c pos attributes holding a 3D 
 * vector):
 * \code
 * junction j1, j2, j3, j4;
 * cell c;
 * j1->pos = Point3d(0,0,0);
 * j2->pos = Point3d(1,0,0);
 * j3->pos = Point3d(1,1,0);
 * j4->pos = Point3d(0,1,0);
 * T.addCell(c, j1, j2, j3, j4);
 * \endcode
 *
 * Another way is to use one of the factory provided in the complex_factory 
 * namespace.
 *
 * You can then continue to edit the cell %complex using the VVComplex::addCell, 
 * VVComplex::divideCell or VVComplex::deleteCell methods. For finer edition, you 
 * can use VVComplex::splitWall or VVComplex::deleteJunction.
 *
 * For example, given a division algorithm using parameters \c DivisionParams, 
 * you can divide a cell with:
 * \code
 * T.divideCell(c, DivisionParams(), c);
 * \endcode
 * During the division, the \c updateFromOld method of the model object will be 
 * called. During that called, the members VVComplex::OldS, VVComplex::OldC and 
 * VVComplex::OldW will be available with the part of the graph saved (in the 
 * example, only the cell \c c and its junctions are saved). For example, if 
 * you want to increase a generation counter, both for the junctions and the 
 * cells:
 * \code
 * void updateFromOld(const cell& cl, const cell& cr, const cell& c,
 *                    const VVComplex::division_data& ddata, VVComplex& T)
 * {
 *   cl->generation = cr->generation = c->generation+1;
 *   forall( const junction& j , T.S.neighbors(cl))
 *   {
 *     if(!T.OldW.contains(j))
 *       j->generation = cl->generation;
 *   }
 * }
 * \endcode
 * Here, the generation of a junction is defined as the generation of the cells 
 * for which it was created.
 *
 * At last, it can be useful to know the two ways to iterate over cells and 
 * junctions. These notations are pretty much equivalent:
 * \code
 * forall(const cell& c,T.C) { ... }
 * forall_named(const cell& c, T.S, cells) { ... }
 *
 * forall(const junction& j,T.W) { ... }
 * forall_named(const junction& j, T.S, junctions) { ... }
 * \endcode
 * The difference is in performance. It is always better to iterate over the 
 * graph used in the loop itself.
 *
 * \subsection cplx_Model Requirements on the Model using 2D Cell Complexes
 * The following methods must exist in the model class to be able to use the 2d 
 * %cell complexes.
 * \anchor position
 * \code
 * Point3d position(const cell& c);
 * Point3d position(const junction& j);
 * \endcode
 * Returns the position of a cell or a junction.
 *
 * \anchor setPosition
 * \code
 * void setPosition(const cell& c, const Point3d& pos);
 * void setPosition(const junction& j, const Point3d& pos);
 * \endcode
 * Set the position of a cell or a junction.
 *
 * \anchor setPositionHint
 * \code
 * void setPositionHint(const junction& j, const junction& j1, const junction& j2, double ratio);
 * \endcode
 * Used to preset the position of a newly formed junction during cell division.
 *
 * \anchor normal
 * \code
 * Point3d normal(const cell& c);
 * \endcode
 * Returns the normal to a cell.
 *
 * \anchor updateFromOld
 * \code
 * void updateFromOld(const cell& cl, const cell& cr, const cell& c, const VVComplex::division_data& ddata, VVComplex& T);
 * \endcode
 * Method called after the division occured. \c cl and \c cr are the two newly 
 * created cells. \c c is the previous cell. \c ddata is the data used for the 
 * division. The %complex has the old graphs setup with whatever was required to 
 * be saved during division.
 *
 * \subsection cplx_Div Implementing a Cell Division Algorithm
 * Implementing a cell division algorithm consists in implementing the function:
 * \code
 * template <class VVComplex>
 * vvcomplex::DivisionData<typename VVComplex::junction_content>
 *    findDivisionPoints(const typename VVComplex::cell& c,
 *                       VVComplex& T,
 *                       const MyDivisionParam& param);
 * \endcode
 * with \c MyDivisionParams being a type containing the parameters needed for the division and, most 
 * importantly, unique for the division as it will identify which division algorithm to use.
 *
 * This function must not change the graph! Within the function, it might be 
 * convenient to use the macros IMPORT_COMPLEX_TYPES() and 
 * IMPORT_COMPLEX_MODEL() to access easily the structure of the cell complex. 
 * A good example is the very simple the division algorithm using 
 * vvcomplex::InModelDivisionParam as parameters: vvcomplex::findDivisionPoints().
 *
 * \section cplx_Extend Extending the Cell Complex Class
 * Extending the cell %complex can simply consist in adding methods (like the 
 * tissue::Tissue class does) or it can also consist in maintaining extra 
 * structures, like extra graphs. To allow extra graphs to be maintained without 
 * user input, virtual functions can be redefined:
 *  - VVComplex::addCellExtra(const cell&)
 *  - VVComplex::reconnectGraphs()
 *  - VVComplex::saveSubgraphs()
 *  - VVComplex::clearOldGraphs()
 *  - VVComplex::deleteCellInGraphs()
 *  - VVComplex::splitWallExtra()
 *  - VVComplex::deleteJunctionExtra()
 *
 * Each of these method \b must call its parent's method to ensure all graphs 
 * are properly maintained.
 *
 * Also, it is very important for the complex to know the instantiated class. 
 * For that purpose, the last argument of the VVComplex class is the leaf class 
 * instantiated. If you want to provide a class someone else may inherit from, 
 * it is recommended to look at the Tissue class declaration and achieve 
 * something similar.
 *
 * \section Example
 *
 * Here is a commented example of a model using the cell complex.
 *
 * First, we start by defining the data structure for the cell and the 
 * vertices. As we don't label the edges, we ignore them for now.
 *
 * \code
 * #include <vve.h>
 *
 * #include <geometry/geometry.h>
 * #include <algorithms/cellsystem.h>
 * #include <geometry/area.h>
 *
 * using geometry::Point3d;
 * struct CellContent
 * {
 *   Point3d position;
 *   int id;
 * };
 *
 * struct JunctionContent
 * {
 *   Point3d position;
 * };
 * \endcode
 *
 * Now, we declare the complex data structure. As it needs to know about the 
 * model class used, we need the place a forward declaration of the model 
 * class. At last, we export the types defined in the complex class (i.e. the 
 * cell, junction, cell_edge, ... types).
 *
 * \code
 * class MyModel;
 * typedef vvcomplex::VVComplex<MyModel, CellContent, JunctionContent> MyComplex;
 * EXPORT_COMPLEX_TYPES(MyComplex);
 * \endcode
 *
 * Now, the model should contain an instance of MyComplex and the methods 
 * required by the complex class.
 *
 * \code
 * struct MyModel : public Model
 * {
 *   MyComplex T;
 *   Point3d position(const cell& c) { return c->position;}
 *   Point3d position(const junction& j) { return j->position;}
 *   Point3d setPosition(const cell& c, const Point3d& pos) { c->position = pos; }
 *   Point3d setPosition(const junction& j, const Point3d& pos) { j->position = pos; }
 *   void setPositionHint(const junction&, const junction&, const junction&, double) {}
 *   Point3d normal(const cell& ) { return Point3d(0,0,1); }
 *   Point3d normal(const junction& ) { return Point3d(0,0,1); }
 *
 *   void updateFromOld(const cell& cl, const cell& cr, const cell& c, const MyComplex::division_data&, MyComplex& )
 *   {
 *     cr->id = 2*c->id;
 *     cl->id = 2*c->id+1;
 *   }
 * \endcode
 *
 * Now, to use the structure, we need to initialise it in the constructor, and 
 * to add its first cell.
 *
 * \code
 * MyModel(QObject *parent)
 *   : Model(parent)
 *   , T(this)
 * {
 *   // Add a square cell
 *   junction j1, j2, j3, j4;
 *   j1->position = Point3d(0,0,0);
 *   j2->position = Point3d(1,0,0);
 *   j3->position = Point3d(1,1,0);
 *   j4->position = Point3d(0,1,0);
 *
 *   cell c;
 *   c->id = 1;
 *   c->position = Point3d(.5, .5, 0);
 *   T.addCell(c, j1, j2, j3, j4);
 * }
 * \endcode
 *
 * For the step, we'll use the cell system division mechanism alternating the 
 * normal from horizontal to vertical.
 *
 * \code
 *   void step()
 *   {
 *     static bool horizontal = false;
 *     horizontal = !horizontal;
 *     cell_system::CellSystemDivisionParams params(Point3d(1,0,0));
 *     if(horizontal)
 *       params.angle = M_PI/2;
 *     std::vector<cell> cells(T.C.begin(), T.C.end()); // Copy the cells
 *     forall(const cell& c,cells)
 *     {
 *        T.divideCell(c, params);
 *     }
 *   }
 *
 *   // This method is needed for the cell system division
 *   double area(const cell& c)
 *   {
 *     std::vector<Point3d> poss;
 *     forall( const junction& j , T.S.neighbors(c))
 *     {
 *       poss.push_back(j->position);
 *     }
 *     return geometry::polygonArea(poss);
 *   }
 * \endcode
 *
 * At last, we draw the result simply with lines for the walls.
 *
 * \code
 *   void draw(Viewer* viewer)
 *   {
 *     Point3d pmin(HUGE_VAL,HUGE_VAL,HUGE_VAL);
 *     Point3d pmax = -pmin;
 *     glBegin(GL_LINES);
 *     glColor3f(1,1,1);
 *     forall(const junction& j,T.W)
 *     {
 *       const Point3d& p = j->position;
 *       if(p.x() < pmin.x())
 *         pmin.x() = p.x();
 *       if(p.y() < pmin.y())
 *         pmin.y() = p.y();
 *       if(p.z() < pmin.z())
 *         pmin.z() = p.z();
 *       if(p.x() > pmax.x())
 *         pmax.x() = p.x();
 *       if(p.y() > pmax.y())
 *         pmax.y() = p.y();
 *       if(p.z() > pmax.z())
 *         pmax.z() = p.z();
 *       forall( const junction& n , T.W.neighbors(j))
 *       {
 *         glVertex3dv(p.c_data());
 *         glVertex3dv(n->position.c_data());
 *       }
 *     }
 *     glEnd();
 *     viewer->setSceneBoundingBox(Vec(pmin), Vec(pmax));
 *   }
 * };
 *
 * DEFINE_MODEL(MyModel);
 * \endcode
 *
 */
namespace vvcomplex
{
  using util::norm;
  using util::normsq;
  using util::normalized;

  using std::cout;
  using std::cerr;
  using std::endl;

  using geometry::Point3d;

  /**
   * \def EXPORT_COMPLEX_VERTICES(ComplexClass)
   *
   * Export the types of the vertices of the complex, i.e. the cell and 
   * junction. Equivalent to:
   * \code
   * typedef ComplexClass::cell cell;
   * typedef ComplexClass::junction junction
   * \endcode
   *
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_VERTICES(ComplexClass) \
  typedef ComplexClass::cell cell; \
  typedef ComplexClass::junction junction

  /**
   * \def EXPORT_COMPLEX_EDGES(ComplexClass)
   *
   * Export the types of the edges of the complex, i.e. cell_edge, wall, 
   * cell_junction_edge and junction_cell_edge. Equivalent to:
   * \code
   * typedef ComplexClass::cell_edge cell_edge;
   * typedef ComplexClass::wall wall;
   * typedef ComplexClass::cell_junction_edge cell_junction_edge;
   * typedef ComplexClass::junction_cell_edge junction_cell_edge;
   * typedef ComplexClass::const_cell_edge const_cell_edge;
   * typedef ComplexClass::const_wall const_wall;
   * typedef ComplexClass::const_cell_junction_edge const_cell_junction_edge;
   * typedef ComplexClass::const_junction_cell_edge const_junction_cell_edge
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_EDGES(ComplexClass) \
  typedef ComplexClass::cell_edge cell_edge; \
  typedef ComplexClass::wall wall; \
  typedef ComplexClass::cell_junction_edge cell_junction_edge; \
  typedef ComplexClass::junction_cell_edge junction_cell_edge; \
  typedef ComplexClass::const_cell_edge const_cell_edge; \
  typedef ComplexClass::const_wall const_wall; \
  typedef ComplexClass::const_cell_junction_edge const_cell_junction_edge; \
  typedef ComplexClass::const_junction_cell_edge const_junction_cell_edge

  /**
   * \def EXPORT_COMPLEX_GRAPHS(ComplexClass)
   *
   * Export the types of the graphs of the complex, i.e. complex_graph, 
   * cell_graph, wall_graph. Equivalent to:
   * \code
   * typedef ComplexClass::complex_graph complex_graph;
   * typedef ComplexClass::cell_graph cell_graph;
   * typedef ComplexClass::wall_graph wall_graph
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_GRAPHS(ComplexClass) \
  typedef ComplexClass::complex_graph complex_graph; \
  typedef ComplexClass::cell_graph cell_graph; \
  typedef ComplexClass::wall_graph wall_graph

  /**
   * \def EXPORT_COMPLEX_TYPES(ComplexClass)
   *
   * Export all the types of the graph. Equivalent to 
   * \code
   * EXPORT_COMPLEX_VERTICES(ComplexClass);
   * EXPORT_COMPLEX_EDGES(ComplexClass);
   * EXPORT_COMPLEX_GRAPHS(ComplexClass);
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_TYPES(ComplexClass) \
  EXPORT_COMPLEX_VERTICES(ComplexClass); \
  EXPORT_COMPLEX_EDGES(ComplexClass); \
  EXPORT_COMPLEX_GRAPHS(ComplexClass)

  /**
   * \def EXPORT_COMPLEX_VERTICES_PREFIX(ComplexClass, prefix)
   *
   * Export the types of the vertices of the complex, i.e. the cell and 
   * junction, but prefix the defined type. Equivalent to:
   * \code
   * typedef ComplexClass::cell prefix_cell;
   * typedef ComplexClass::junction prefix_junction
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_VERTICES_PREFIX(ComplexClass, prefix) \
  typedef ComplexClass::cell prefix##_cell; \
  typedef ComplexClass::junction prefix##_junction

  /**
   * \def EXPORT_COMPLEX_EDGES_PREFIX(ComplexClass, prefix)
   *
   * Export the types of the edges of the complex, i.e. cell_edge, wall, 
   * cell_junction_edge and junction_cell_edge. Equivalent to:
   * \code
   * typedef ComplexClass::cell_edge prefix_cell_edge;
   * typedef ComplexClass::wall prefix_wall;
   * typedef ComplexClass::cell_junction_edge prefix_cell_junction_edge;
   * typedef ComplexClass::junction_cell_edge prefix_junction_cell_edge;
   * typedef ComplexClass::const_cell_edge prefix_const_cell_edge;
   * typedef ComplexClass::const_wall prefix_const_wall;
   * typedef ComplexClass::const_cell_junction_edge prefix_const_cell_junction_edge;
   * typedef ComplexClass::const_junction_cell_edge prefix_const_junction_cell_edge
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_EDGES_PREFIX(ComplexClass, prefix) \
  typedef ComplexClass::cell_edge prefix##_cell_edge; \
  typedef ComplexClass::wall prefix##_wall; \
  typedef ComplexClass::cell_junction_edge prefix##_cell_junction_edge; \
  typedef ComplexClass::junction_cell_edge prefix##_junction_cell_edge; \
  typedef ComplexClass::const_cell_edge prefix##_const_cell_edge; \
  typedef ComplexClass::const_wall prefix##_const_wall; \
  typedef ComplexClass::const_cell_junction_edge prefix##_const_cell_junction_edge; \
  typedef ComplexClass::const_junction_cell_edge prefix##_const_junction_cell_edge

  /**
   * \def EXPORT_COMPLEX_GRAPHS_PREFIX(ComplexClass, prefix)
   *
   * Export the types of the graphs of the complex, i.e. complex_graph, 
   * cell_graph, wall_graph.
   * \code
   * typedef ComplexClass::complex_graph prefix_complex_graph;
   * typedef ComplexClass::cell_graph prefix_cell_graph;
   * typedef ComplexClass::wall_graph prefix_wall_graph
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_GRAPHS_PREFIX(ComplexClass, prefix) \
  typedef ComplexClass::complex_graph prefix##_complex_graph; \
  typedef ComplexClass::cell_graph prefix##_cell_graph; \
  typedef ComplexClass::wall_graph prefix##_wall_graph

  /**
   * \def EXPORT_COMPLEX_TYPES_PREFIX(ComplexClass, prefix)
   *
   * Export all the types of the graph. Equivalent to 
   * \code
   * EXPORT_COMPLEX_VERTICES_PREFIX(ComplexClass, prefix);
   * EXPORT_COMPLEX_EDGES_PREFIX(ComplexClass, prefix);
   * EXPORT_COMPLEX_GRAPHS_PREFIX(ComplexClass, prefix);
   * \endcode
   *
   * \relates vvcomplex
   */
#define EXPORT_COMPLEX_TYPES_PREFIX(ComplexClass, prefix) \
  EXPORT_COMPLEX_VERTICES_PREFIX(ComplexClass, prefix); \
  EXPORT_COMPLEX_EDGES_PREFIX(ComplexClass, prefix); \
  EXPORT_COMPLEX_GRAPHS_PREFIX(ComplexClass, prefix)

  /**
   * \def IMPORT_COMPLEX_VERTICES(ComplexClass)
   *
   * Import the vertices types for another template where the complex class is 
   * not completely instantiated.
   *
   * \relates vvcomplex
   */
#define IMPORT_COMPLEX_VERTICES(ComplexClass) \
  typedef typename ComplexClass::cell cell; \
  typedef typename ComplexClass::junction junction

  /**
   * \def IMPORT_COMPLEX_EDGES(ComplexClass)
   *
   * Import the edges types for another template where the complex class is not 
   * completely instantiated.
   *
   * \relates vvcomplex
   */
#define IMPORT_COMPLEX_EDGES(ComplexClass) \
  typedef typename ComplexClass::cell_edge cell_edge; \
  typedef typename ComplexClass::wall wall; \
  typedef typename ComplexClass::cell_junction_edge cell_junction_edge; \
  typedef typename ComplexClass::junction_cell_edge junction_cell_edge; \
  typedef typename ComplexClass::const_cell_edge const_cell_edge; \
  typedef typename ComplexClass::const_wall const_wall; \
  typedef typename ComplexClass::const_cell_junction_edge const_cell_junction_edge; \
  typedef typename ComplexClass::const_junction_cell_edge const_junction_cell_edge

  /**
   * \def IMPORT_COMPLEX_GRAPHS(ComplexClass)
   *
   * Import the graphs types for another template where the complex class is not 
   * completely instantiated.
   *
   * \relates vvcomplex
   */
#define IMPORT_COMPLEX_GRAPHS(ComplexClass) \
  typedef typename ComplexClass::complex_graph complex_graph; \
  typedef typename ComplexClass::cell_graph cell_graph; \
  typedef typename ComplexClass::wall_graph wall_graph

  /**
   * \def IMPORT_COMPLEX_MODEL(ComplexClass, cplx)
   *
   * Import the model from the instance \c cplx of the complex of type \c 
   * ComplexClass. As a result, a variable named \c model is defined as 
   * a reference to the model.
   *
   * \relates vvcomplex
   */
#define IMPORT_COMPLEX_MODEL(ComplexClass, cplx) \
  typename ComplexClass::model_t &model = *(cplx).model;

  /**
   * \def IMPORT_COMPLEX_TYPES(ComplexClass)
   *
   * Import all the types of the complex for another template where the complex 
   * class is not completely instantiated.
   *
   * \relates vvcomplex
   */
#define IMPORT_COMPLEX_TYPES(ComplexClass) \
  IMPORT_COMPLEX_VERTICES(ComplexClass); \
  IMPORT_COMPLEX_EDGES(ComplexClass); \
  IMPORT_COMPLEX_GRAPHS(ComplexClass)

  /*
   * Documentation for the types imported
   *
   * \typedef typename VVComplex::cell cell
   * \brief Type of a cell
   *
   * \typedef typename VVComplex::junction junction
   * \brief Type of a junction
   *
   * \typedef typename VVComplex::cell_edge cell_edge
   * \brief Type of an edge in the cell graph
   *
   * \typedef typename VVComplex::wall wall
   * \brief Type of a wall (i.e. an edge in the wall graph)
   *
   * \typedef typename VVComplex::cell_junction_edge cell_junction_edge
   * \brief Type of an edge from a cell to a junction
   *
   * \typedef typename VVComplex::junction_cell_edge junction_cell_edge
   * \brief Type of an edge from a junction to a cell
   *
   * \typedef typename VVComplex::complex_graph complex_graph;
   * \brief Type of the complex graph (i.e. linking cells to junctions and 
   * vice-versa)
   *
   * \typedef typename VVComplex::cell_graph cell_graph;
   * \brief Type of the cell graph
   *
   * \typedef typename VVComplex::wall_graph wall_graph
   * \brief Type of the wall graph
   */

#ifdef JUST_FOR_DOXYGEN
  //@{
  /// \name Types defined by EXPORT_COMPLEX_TYPES(VVComplex)

  /// Type of a cell
  typedef VVComplex::cell cell;

  /// Type of a junction
  typedef VVComplex::junction junction;

  /// Type of an edge in the cell graph
  typedef VVComplex::cell_edge cell_edge;

  /// Type of an edge in the cell graph
  typedef VVComplex::const_cell_edge const_cell_edge;

  /// Type of a wall (i.e. an edge in the wall graph)
  typedef VVComplex::wall wall;

  /// Type of a wall (i.e. an edge in the wall graph)
  typedef VVComplex::const_wall const_wall;

  /// Type of an edge from a cell to a junction
  typedef VVComplex::cell_junction_edge cell_junction_edge;

  /// Type of an edge from a cell to a junction
  typedef VVComplex::const_cell_junction_edge const_cell_junction_edge;

  /// Type of an edge from a junction to a cell
  typedef VVComplex::junction_cell_edge junction_cell_edge;

  /// Type of an edge from a junction to a cell
  typedef VVComplex::const_junction_cell_edge const_junction_cell_edge;

  /// Type of the complex graph (i.e. linking cells to junctions and vice-versa)
  typedef VVComplex::complex_graph complex_graph;

  /// Type of the cell graph
  typedef VVComplex::cell_graph cell_graph;

  /// Type of the wall graph
  typedef VVComplex::wall_graph wall_graph;
  //@}
#endif

  /**
   * \class DivisionData complex.h <algorithms/complex.h>
   *
   * \brief Class holding the data needed to actually divide a cell
   *
   * This is the return type of the different division algorithm. It provides 
   * all the informations needed to modify the graphs to account for the cell 
   * division.
   *
   * The division will cut the walls (u1,u2) and (v1,v2) and will place the new 
   * vertexes at positions pu on the (u1,u2) wall and pv on the (v1,v2) wall.
   */
  template <typename JunctionContent>
    struct DivisionData
  {
    /**
     * Type of graph vertex
     */
    typedef graph::Vertex<JunctionContent> junction;

    /**
     * Default constructor do not require pinching.
     */
    DivisionData()
      : u1(0)
      , v1(0)
      , divide_at_u1(false)
      , divide_at_v1(false)
    { }

    /**
     * Complete constructor
     */
    DivisionData(Point3d _pu, Point3d _pv,
                 const junction& _u1,
                 const junction& _v1,
                 bool _du1 = false,
                 bool _dv1 = false)
      : pu(_pu)
      , pv(_pv)
      , u1(_u1)
      , v1(_v1)
      , divide_at_u1(_du1)
      , divide_at_v1(_dv1)
    { }

    /**
     * Copy constructor
     */
    DivisionData(const DivisionData& copy)
      : pu(copy.pu)
      , pv(copy.pv)
      , u1(copy.u1)
      , v1(copy.v1)
      , divide_at_u1(copy.divide_at_u1)
      , divide_at_v1(copy.divide_at_v1)
    { }

    /**
     * Convert the result to true if everything is ok.
     *
     * If the algorithm fails for any reason, the returned structure should 
     * have u1 or v1 set to null vertexes. That way, the boolean operator 
     * will convert it to false.
     */
    operator bool()
    {
      return u1 and v1;
    }

    /**
     * Convert the result to false if everything is ok
     */
    bool operator!() const
    {
      return !u1 or !v1;
    }

    /**
     * \var vertex u1
     * First junction of the 'U' wall, u2 is the vertex after u1 in the 
     * neighborhood of c
     */
    /**
     * \var vertex v1
     * First junction of the 'V' wall, v2 is the vertex after v1 in the 
     * neighborhood of c
     */
    /**
     * \var Point3d pu
     * Position of the junction inserted in the U wall
     */
    /**
     * \var Point3d pv
     * Position of the junction inserted in the V wall
     */
    /**
     * \var divide_at_u1
     * If true, pu is ignored and the division wall goes through \c u1
     */
    /**
     * \var divide_at_v1
     * If true, pv is ignored and the division wall goes through \c v1
     */
    Point3d pu, pv;
    junction u1, v1;
    bool divide_at_u1, divide_at_v1;
  };


#ifdef USE_ALLOCATOR
#   define COMPLEX_ARGS CellContent,JunctionContent,CellJunctionContent,JunctionCellContent,compact,Alloc
#else // USE_ALLOCATOR
#   define COMPLEX_ARGS CellContent,JunctionContent,CellJunctionContent,JunctionCellContent,compact
#endif // USE_ALLOCATOR

  /**
   * \class VVComplexGraph complex.h <algorithms/complex.h>
   *
   * Definition of the bipartite graph with specialized methods for the complex 
   * graph.
   */
  template <typename CellContent, typename JunctionContent, typename CellJunctionContent = graph::_EmptyEdgeContent, typename JunctionCellContent = graph::_EmptyEdgeContent, bool compact = false
#ifdef USE_ALLOCATOR
    , typename Alloc = DEFAULT_ALLOC(CellContent)
#endif
    >
  struct VVComplexGraph : public graph::VVBiGraph<COMPLEX_ARGS>
  {
    typedef graph::VVBiGraph<COMPLEX_ARGS> ParentClass;
    typedef typename ParentClass::iterator1 iterator1;
    typedef typename ParentClass::iterator2 iterator2;
    typedef typename ParentClass::const_iterator1 const_iterator1;
    typedef typename ParentClass::const_iterator2 const_iterator2;
    typedef typename ParentClass::vertex1_t cell;
    typedef typename ParentClass::vertex2_t junction;
    typedef typename ParentClass::edge1_t cell_junction_edge;
    typedef typename ParentClass::edge2_t junction_cell_edge;
    typedef typename ParentClass::const_edge1_t const_cell_junction_edge;
    typedef typename ParentClass::const_edge2_t const_junction_cell_edge;
    typedef typename ParentClass::size_type size_type;
    typedef typename ParentClass::range_vertex1 range_cell;
    typedef typename ParentClass::range_vertex2 range_junction;
    typedef typename ParentClass::const_range_vertex1 const_range_cell;
    typedef typename ParentClass::const_range_vertex2 const_range_junction;

    VVComplexGraph()
      : ParentClass()
    {}

    VVComplexGraph(const ParentClass& other)
      : ParentClass(other)
    {}

    /**
     * Returns a cell from the complex graph
     */
    const cell& any_cell() const { return this->any_vertex1(); }
    /**
     * Retrns a junction from the complex graph
     */
    const junction& any_junction() const { return this->any_vertex2(); }

    ///\{
    ///\name STL iterators
    iterator1 begin_cells() { return this->begin_vertex1(); }
    iterator1 end_cells() { return this->end_vertex1(); }

    iterator2 begin_junctions() { return this->begin_vertex2(); }
    iterator2 end_junctions() { return this->end_vertex2(); }

    const_iterator1 begin_cells() const { return this->begin_vertex1(); }
    const_iterator1 end_cells() const { return this->end_vertex1(); }

    const_iterator2 begin_junctions() const { return this->begin_vertex2(); }
    const_iterator2 end_junctions() const { return this->end_vertex2(); }

    range_cell cells() { return this->vertices1(); }
    range_junction junctions() { return this->vertices2(); }

    const_range_cell cells() const { return this->vertices1(); }
    const_range_junction junctions() const { return this->vertices2(); }
    ///\}

    /**
     * Returns the number of cells in the graph
     */
    size_t nb_cells() const { return this->nb_vertices1(); }
    /**
     * Returns the number of junctions in the graph
     */
    size_t nb_junctions() const { return this->nb_vertices2(); }

    /**
     * True if there are no cells
     */
    bool no_cell() const { return this->no_vertex1(); }
    /**
     * True if there are no junctions
     */
    bool no_junction() const { return this->no_vertex2(); }

    /**
     * Returns the cell number idx
     */
    const cell& get_cell(size_type idx) const { return this->get_vertex1(idx); }
    /**
     * Returns the junction number idx
     */
    const junction& get_junction(size_type idx) const { return this->get_vertex2(idx); }

    /**
     * Add a new cell with three vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3)
    {
      std::vector<junction> junctions(3, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with four vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4)
    {
      std::vector<junction> junctions(4, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with five vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5)
    {
      std::vector<junction> junctions(5, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with six vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6)
    {
      std::vector<junction> junctions(6, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with seven vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6, const junction& j7)
    {
      std::vector<junction> junctions(7, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      junctions[6] = j7;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with height vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6, const junction& j7,
                 const junction& j8)
    {
      std::vector<junction> junctions(8, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      junctions[6] = j7;
      junctions[7] = j8;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with nine vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6, const junction& j7,
                 const junction& j8, const junction& j9)
    {
      std::vector<junction> junctions(9, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      junctions[6] = j7;
      junctions[7] = j8;
      junctions[8] = j9;
      return this->addCell(c, junctions);
    }

    template <typename JunctionContainer>
    bool addCell(const cell& c, const JunctionContainer& junctions)
    {
      std::list<junction> inserted;
      insert(c);
      junction pj(0), fj(0);
      // First, create the outgoing edges and insert cell and junctions
      forall(const junction& j,junctions)
      {
        if(!contains(j))
        {
          inserted.push_back(j);
          insert(j);
        }
        if(pj.isNull())
        {
          fj = j;
          insertEdge(c, j);
        }
        else
        {
          spliceAfter(c, pj, j);
        }
        pj = j;
      }
      // Next, connect the junctions to the cells
      forall( const junction& j , neighbors(c))
      {
        if(valence(j) < 2)
          insertEdge(j, c);
        else
        {
          bool done = false;
          // First, try to connect with respect to the previous junction
          const junction& nj = this->nextTo(c, j);
          forall( const cell& nc , neighbors(j))
          {
            if(edge(nc, nj))
            {
              spliceAfter(j, nc, c);
              done = true;
              break;
            }
          }
          if(!done)
          {
            const junction& nj = prevTo(c, j);
            forall( const cell& nc , neighbors(j))
            {
              if(edge(nc, nj))
              {
                spliceBefore(j, nc, c);
                done = true;
                break;
              }
            }
          }
          if(!done)
          {
            erase(c);
            forall(const junction& j, inserted)
            {
              erase(j);
            }
            return false;
          }
        }
      }
      return true;
    }

    /**
     * Test if the wall of the cell c starting at j is at the border of the 
     * complex graph.
     */
    bool border(const cell& c, const junction& j) const
    {
      const cell& c2 = this->prevTo(j, c);
      return this->nextTo(c, j) != this->prevTo(c2, j);
    }

    /**
     * Test if a junction is at the border of the complex graph
     */
    bool border(const junction& j) const
    {
      forall( const cell& c , this->neighbors(j))
      {
        if(border(c, j)) return true;
      }
      return false;
    }

    /**
     * Test if a junction is at the border of the complex graph
     */
    bool border(const cell& c) const
    {
      forall( const junction& j , neighbors(c))
      {
        if(border(c, j)) return true;
      }
      return false;
    }

    bool border(const junction& j1, const junction& j2) const
    {
      forall( const cell& c , neighbors(j1))
      {
        if(nextTo(c, j1) == j2)
          return prevTo(prevTo(j1, c), j1) != j2;
        else if(prevTo(c, j1) == j2)
          return nextTo(nextTo(j1, c), j1) != j2;
      }
      return false;
    }

    /**
     * Returns the cell adjacent the c sharing the wall (j, S.nextTo(c,j))
     */
    const cell& adjacentCell(const cell& c, const junction& j) const
    {
      const junction& nj = this->nextTo(c, j);
      const cell& adj = this->prevTo(j, c);
      if(this->prevTo(adj, j) == nj)
        return adj;
      return cell::null;
    }

    /**
     * Returns the cell(s) sharing the wall (j1,j2).
     *
     * \returns a pair of cells (c1,c2) such that in c1, j1 is before j2 and in 
     * c2, j2 is before j1.
     */
    std::pair<cell,cell> adjacentCells(const junction& j1, const junction& j2) const
    {
      vvassert(j1 != j2);
      forall(const cell& c , neighbors(j1))
      {
        if(nextTo(c, j1) == j2)
        {
          const cell& c2 = prevTo(j1, c);
          if(prevTo(c2, j1) == j2)
            return std::make_pair(c,c2);
          return std::make_pair(c, cell::null);
        }
        else if(prevTo(c, j1) == j2)
        {
          const cell& c1 = nextTo(j1, c);
          if(nextTo(c1, j1) == j2)
            return std::make_pair(c1, c);
          return std::make_pair(cell::null, c);
        }
      }
      return std::make_pair(cell::null, cell::null);
    }

  private:
    const junction& startInterfaceWall(const junction& j, const cell& c1, const cell& c2, junction start = junction(0)) const
    {
      if(not start) start = j;
      const junction& pj = prevTo(c1, j);
      if(pj == nextTo(c2, j)) // We are not at the start of the interface
      {
        if(pj == start) // The whole interface is with c2, so there is no beginning
          return pj;
        return startInterfaceWall(pj, c1, c2, start);
      }
      return j;
    }

    /**
     * Returns the end of the interface wall between c1 and c2 that starts on j
     */
    const junction& endInterfaceWall(const junction& j, const cell& c1, const cell& c2, junction start = junction(0)) const
    {
      if(not start) start = j;
      const junction& nj = nextTo(c1, j);
      if(nj == prevTo(c2, j)) // We are not at the end of the interface
      {
        if(nj == start) // The interface is with c2 so there is no end
          return nj;
        return endInterfaceWall(nj, c1, c2, start);
      }
      return j;
    }

  public:

    /**
     * \returns the junction in \c c1 starting the wall common between \c c1 
     * and \c c2.
     */
    const junction& interfaceWall(const cell& c1, const cell& c2) const
    {
      forall( const junction& j, neighbors(c1))
      {
        if(prevTo(j, c1) == c2)
        {
          if(nextTo(c1, j) == prevTo(c2, j))
          {
            return startInterfaceWall(j, c1, c2);
          }
        }
      }
      return junction::null;
    }

    /**
     * \returns the pair of junctions in \c c1 such that the interface between 
     * \c c1 and \c c2 starts at the first junction and end at the second. In 
     * case the \c c1 is entirely contained in \c c2 (or the other way around), 
     * both junctions will be the same.
     */
    std::pair<junction, junction> interfaceWallSpan(const cell& c1, const cell& c2) const
    {
      forall(const junction& j, neighbors(c1))
      {
        if(prevTo(j, c1) == c2)
        {
          if(nextTo(c1, j) == prevTo(c2, j))
          {
            return std::make_pair(startInterfaceWall(j, c1, c2),
                                  endInterfaceWall(j, c1, c2));
          }
        }
      }
      return std::make_pair(junction::null, junction::null);
    }

    /**
     * Delete a cell and all the junctions that become orphans
     *
     * \returns the list of junctions deleted
     */
    std::vector<junction> deleteCell(const cell& c)
    {
      typename ParentClass::neighbor_iterator1_range p = neighbors(c);
      std::vector<junction> js(p.begin(), p.end());
      std::vector<junction> deleted;
      erase(c);
      forall(const junction& j,js) if(empty(j))
      {
        erase(j);
        deleted.push_back(j);
      }
      return deleted;
    }

    /**
     * Split a wall in two and return the new junction.
     */
    const junction& splitWall(const junction& j1, const junction& j2, const junction& x = junction(0))
    {
      const junction& j = *insert(x?x:junction());
      forall(const cell& c, neighbors(j1))
      {
        if(nextTo(c, j1) == j2)
        {
          spliceAfter(c, j1, j);
          insertEdge(j, c);
        }
        else if(prevTo(c, j1) == j2)
        {
          spliceBefore(c, j1, j);
          insertEdge(j, c);
        }
      }
      return j;
    }

    /**
     * Split a wall in two and return the new junction.
     */
    const junction& splitWall(const cell& c, const junction& j1, const junction& x = junction(0))
    {
      const cell& c2 = adjacentCell(c, j1);
      const junction& j = *insert(x?x:junction());
      spliceAfter(c, j1, j);
      insertEdge(j, c);
      if(c2)
      {
        spliceBefore(c2, j1, j);
        insertEdge(j, c2);
      }
      return j;
    }

    /**
     * Structure describing a cell division
     */
    typedef DivisionData<JunctionContent> division_data;

    /**
     * \class division_result_t
     *
     * Describe the result of a cell division.
     *
     * Can be converted to boolean to check the division happened, and return 
     * the two new cells.
     */
    struct division_result_t
    {
      division_result_t()
        : cl(0)
        , cr(0)
        , u1(0)
        , u2(0)
        , u(0)
        , v1(0)
        , v2(0)
        , v(0)
      {}

      division_result_t(cell cl, cell cr,
                        junction u1, junction u2, junction u,
                        junction v1, junction v2, junction v)
        : cl(cl)
        , cr(cr)
        , u1(u1)
        , u2(u2)
        , u(u)
        , v1(v1)
        , v2(v2)
        , v(v)
      {}

      division_result_t(const division_result_t& copy)
        : cl(copy.cl)
        , cr(copy.cr)
        , u1(copy.u1)
        , u2(copy.u2)
        , u(copy.u)
        , v1(copy.v1)
        , v2(copy.v2)
        , v(copy.v)
      {}

      cell cl, cr;
      junction u1, u2, u, v1, v2, v;
      operator bool() { return cl and cr; }
    };


    /**
     * Divide a cell
     */
    division_result_t divideCell( const cell& c,
                                  const division_data& ddata )
    {
      //Point3d pu = ddata.pu, pv = ddata.pv;
      junction u1 = ddata.u1, v1 = ddata.v1;
      junction u2 = this->nextTo(c, u1), v2 = this->nextTo(c, v1);

      junction u(0), v(0);

      // First, find the neighbor cells
      const cell& nu = adjacentCell(c, u1);
      const cell& nv = adjacentCell(c, v1);

      // Insert a new vertex on the new wall
      if(ddata.divide_at_u1)
        u = u1;
      else
      {
        u = junction();
        this->insert(u);
        this->spliceAfter(c, u1, u);
        this->insertEdge(u, c);
      }

      // Insert a new vertex on the other side
      if(ddata.divide_at_v1)
        v = v1;
      else
      {
        v = junction();
        this->insert(v);
        this->spliceAfter(c, v1, v);
        this->insertEdge(v, c);
      }

      // If necessary, adjust the connectivity of the adjacent cells
      // Then compute the position of the new vertices and adjust the one of 
      // the cells around
      if(!ddata.divide_at_u1)
      {
        if(nu)
        {
          this->spliceBefore(nu, u1, u);
          this->insertEdge(u, nu);
        }
      }

      if(!ddata.divide_at_v1)
      {
        if(nv)
        {
          this->spliceBefore(nv, v1, v);
          this->insertEdge(v, nv);
        }
      }

      // Create the two new cell centers
      cell cl, cr;
      this->insert(cl);
      this->insert(cr);

      // Connect new junctions
      this->insertEdge(cl, u);
      this->insertEdge(cl, v);
      this->insertEdge(cr, u);
      this->insertEdge(cr, v);

      this->spliceAfter(u, c, cl);
      this->replace(u, c, cr);
      this->spliceBefore(v, c, cl);
      this->replace(v, c, cr);

      // Connect the new centers
      forall( const junction& x , this->neighbors(c, this->nextTo(c, u)))
      {
        if(x == v) break;
        this->spliceBefore(cr, v, x);
        this->replace(x, c, cr);
      }

      forall( const junction& y , this->neighbors(c, this->nextTo(c, v)))
      {
        if(y == u) break;
        this->spliceBefore(cl, u, y);
        this->replace(y, c, cl);
      }

      // Remove c
      this->erase(c);

      return division_result_t(cl, cr, u1, u2, u, v1, v2, v);
    }

    /**
     * Merge two adjacent cells.
     *
     * Only c1 is kept, c2 is being deleted.
     *
     * \returns the junctions at the start and end of the wall that has been 
     * removed
     */
    std::pair<junction, junction> mergeCells(const cell& c1, const cell& c2, std::vector<junction>* deleted_junctions = 0)
    {
      std::pair<junction, junction> result = interfaceWallSpan(c1, c2);
      if(!result.first)
        return result;
      junction j1 = result.first;
      junction j2 = result.second;
      std::vector<junction> local_deleted_junctions;
      if(deleted_junctions == 0) deleted_junctions = &local_deleted_junctions;
      forall(const junction& j, neighbors(c1, nextTo(c1, j1)))
      {
        if(j == j2)
          break;
        deleted_junctions->push_back(j);
      }

      // Remove junctions from c1 and W
      forall(const junction& j, *deleted_junctions)
      {
        eraseEdge(c1, j);
      }

      // Update S
      forall(const junction& j, neighbors(c2, nextTo(c2, j1)))
      {
        if(j == j2) break;
        spliceBefore(c1, j2, j);
        replace(j, c2, c1);
      }

      // Erase c2 from C and S
      erase(c2);

      return result;
    }

    std::vector<junction> nextContourVertex(const junction& first, const cell& c,
                                            const junction& j,
                                            std::vector<junction> acc) const
    {
      const junction& n = nextTo(c, j);
      if(n == first)
        return acc;
      else if(acc.size() >= nb_junctions())
        return std::vector<junction>();
      int nb_common_cells = 0;
      forall( const cell& cn , neighbors(j))
      {
        if(edge(cn, n))
          nb_common_cells++;
      }
      if(nb_common_cells == 1)
      {
        acc.push_back(n);
        return nextContourVertex(first, c, n, acc);
      }
      const cell& cn = nextTo(j, c);
      const junction& jn = nextTo(cn, j);
      if(jn == first)
        return acc;
      acc.push_back(jn);
      return nextContourVertex(first, cn, jn, acc);
    }

    std::vector<junction> contour() const
    {
      std::vector<junction> result;
      switch(nb_cells())
      {
        case 0:
          return result; // No contour !
        case 1:
          {
            const cell& c = any_cell();
            const junction& j1 = anyIn(c);
            const junction& j2 = nextTo(c, j1);
            const junction& j3 = nextTo(c, j2);
            result.push_back(j1);
            result.push_back(j2);
            result.push_back(j3);
          }
          break;
        case 2:
          {
            const cell& c1 = reference(get_cell(0));
            const cell& c2 = reference(get_cell(1));
            forall( const junction& j , neighbors(c1))
            {
              if(valence(j) == 1)
              {
                result.push_back(prevTo(c1, j));
                result.push_back(j);
                result.push_back(nextTo(c1, j));
                break;
              }
            }
            forall( const junction& j , neighbors(c2))
            {
              if(valence(j) == 1)
              {
                result.push_back(j);
                break;
              }
            }
          }
          break;
        default:
          // First, find one point of the contour
          forall(const junction& j, junctions())
          {
            if(border(j))
            {
              result.push_back(j);
              forall( const cell& c , neighbors(j))
              {
                const junction& jn = nextTo(c, j);
                int nb_common_cells = 0;
                forall( const cell& cn , neighbors(jn))
                {
                  if(edge(j, cn))
                    nb_common_cells++;
                }
                if(nb_common_cells == 1)
                {
                  result.push_back(jn);
                  return nextContourVertex(j, c, jn, result);
                }
              }
              return result;
            }
          }
      }
      return result; // Should not be reached!
    }


  };

  template <typename CplxGraph>
    void create_cplxgraph_methods(CplxGraph& G)
    {
      if(G.size())
      {
        if(G.valence(G.any_vertex1()) > G.size())
        {
          graph::create_bigraph_methods(G);
          typename CplxGraph::cell c = G.any_cell();
          typename CplxGraph::junction j = G.any_junction();
          G.nb_cells();
          G.nb_junctions();
          G.no_cell();
          G.no_junction();
          G.get_cell(10);
          G.get_junction(10);
          G.border(j);
          G.border(c);
          G.border(c,j);
          G.border(j,j);
          G.adjacentCell(c, j);
          G.adjacentCells(j, j);
          G.deleteCell(c);
          G.begin_cells();
          G.end_cells();
          G.begin_junctions();
          G.end_junctions();
          G.interfaceWall(c,c);
          G.interfaceWallSpan(c,c);
        }
      }
    }

#ifndef DOXYGEN_SHOULD_SKIP_THIS

#define VVBIGRAPH graph::VVBiGraph<CellContent,JunctionContent,CellJunctionContent,JunctionContent>
#define CPLXGRAPH complex_graph<CellContent,JunctionContent,CellJunctionContent,JunctionCellContent>
#undef VVBIGRAPH
#undef CPLXGRAPH

#endif // DOXYGEN_SHOULD_SKIP_THIS

  /**
   * \class InModelDivisionParam complex.h <algorithms/complex.h>
   *
   * Class used to define the division parameters from within the model class.
   *
   * Use this class parameters when dividing a cell and the following method 
   * will be used in your model:
   *
   * \code
   * DivisionData<VertexContent> divisionParameters(const vertex& c, vvgraph& 
   * S, const vvcomplex::InModelDivisionParam& param);
   * \endcode
   *
   * A typical use will be to inherit this class to put parameters:
   * \code
   *  struct MyDivisionParams : public vvcomplex::InModelDivisionParam
   *  {
   *    int divisionType;
   *  }
   * \endcode
   *
   * And use this to divide according to the parameters:
   * \code
   * vvcomplex::DivisionData<VertexContent> divisionParameters(const vertex& c, 
   * vvgraph& S, const vvcomplex::InModelDivisionParam& p)
   * {
   *   DivisionData<VertexContent> result;
   *   MyDivisionParams& param = dynamic_cast<const InModelDivisionParam&>(p);
   *   if(param.divisionType == 0)
   *   {
   *     // First type of division
   *   }
   *   else
   *   {
   *     // Other type of division
   *   }
   *   return result;
   * }
   * \endcode
   */
  struct InModelDivisionParam
  { };

  /**
   * Implementation of the in model division scheme.
   *
   * Implementation:
   * \code
   template <class VVComplex>
   DivisionData<typename VVComplex::junction_content> findDivisionPoints(const typename VVComplex::cell& c,
   VVComplex& T,
   const InModelDivisionParam& param)
   {
   IMPORT_COMPLEX_MODEL(VVComplex, T);
   return model.divisionParameters(c, T, param);
   }
   * \endcode
   */
  template <class VVComplex>
    DivisionData<typename VVComplex::junction_content> findDivisionPoints(const typename VVComplex::cell& c,
                                                                          VVComplex& T,
                                                                          const InModelDivisionParam& param)
    {
      IMPORT_COMPLEX_MODEL(VVComplex, T);
      return model.divisionParameters(c, T, param);
    }

  /**
   * Move point \p v to avoid segment borders.
   *
   * Find out how to move the point \p v on the segment <tt>[v1,v2]</tt> to be 
   * at least at a distance \p mw from the \p v1 and \p v2.
   *
   * \param[in,out] v Point to displace if needed
   * \param[in] v1 Extremity of the segment
   * \param[in] v2 Extremity of the segment
   * \param[in] mw Minimum distance from the extremities
   * \param[out] displacement If non-null, it is set to be the displacement of 
   * the point.
   *
   * \returns True if it was possible, false otherwise. In case it wasn't 
   * possible, it displaces the point exactly between v1 and v2.
   */
  bool FindWallMin(Point3d &v, const Point3d& v1, const Point3d& v2, double mw, double *displacement = 0);

  /**
   * Find the wall opposite to the point already selected.
   *
   * This function suppose the first division point was found. The other wall 
   * is the one intersecting the line starting from the point \p result.pu and 
   * going through the cell center. Then, the point is displaced to endure the 
   * minimum size of a wall.
   *
   * \param[in] c Cell to divide
   * \param[in,out] result Structure to update with the opposite wall
   * \param[in] S %VV graph
   * \param[in] cellWallMin minimum size of a wall
   * \param[in] model Model using the algorithm
   */
  template <typename VVComplex>
    void FindOppositeWall(const typename VVComplex::cell& c,
                          DivisionData<typename VVComplex::junction_content>& result,
                          VVComplex& T,
                          double cellWallMin, bool strictCellWallMin)
    {
      IMPORT_COMPLEX_VERTICES(VVComplex);
      IMPORT_COMPLEX_MODEL(VVComplex, T);
      double mindis = HUGE_VAL;
      const Point3d& cnml = model.normal(c);
      Point3d minu = result.pu;
      const Point3d& cpos = model.position(c);
      junction p_u1 = result.u1;
      junction p_u2 = T.S.nextTo(c, p_u1);
      forall( const junction& tv1 , T.S.neighbors(c))
      {
        const junction& tv2 = T.S.nextTo(c, tv1);
        const Point3d& v1 = model.position(tv1);
        const Point3d& v2 = model.position(tv2);
        if(tv2 == p_u2)
        {
          continue;
        }
        // Keep uc close to same plane as v1v2
        Point3d uc = cpos - minu;
        uc -= (uc * cnml) * cnml;
        Point3d n = cnml^uc;
        uc += minu;
        double s;
        Point3d v;
        bool found = geometry::planeLineIntersection(v, s, uc, n, v1, v2);
        if(found)
        {
          double dis;
          if(FindWallMin(v, v1, v2, cellWallMin, &dis) or not strictCellWallMin)
          {
            if(dis < mindis)
            {
              mindis = dis;
              result.pv = v;
              result.v1 = tv1;
            }
            if(dis == 0) break;
          }
        }
      }
    }

  /**
   * Test if the division point in result is close enough to one of the 
   * extremum. If so, setup the division to go through the vertex.
   */
  template <typename VVComplex>
    void testDivisionOnVertices(const typename VVComplex::cell& c,
                                DivisionData<typename VVComplex::junction_content>& result,
                                VVComplex& T,
                                double epsilon)
    {
      IMPORT_COMPLEX_MODEL(VVComplex, T);
      double eps_sq = epsilon*epsilon;
      if(util::normsq(result.pu-model.position(result.u1)) < eps_sq)
      {
        result.divide_at_u1 = true;
      }
      else if(util::normsq(result.pu-model.position(T.S.nextTo(c,result.u1))) < eps_sq)
      {
        result.u1 = T.S.nextTo(c,result.u1);
        result.divide_at_u1 = true;
      }
      if(util::normsq(result.pv-model.position(result.v1)) < eps_sq)
      {
        result.divide_at_v1 = true;
      }
      else if(util::normsq(result.pv-model.position(T.S.nextTo(c,result.v1))) < eps_sq)
      {
        result.v1 = T.S.nextTo(c,result.v1);
        result.divide_at_v1 = true;
      }
    }

  /**
   * Return the position of the center of a cell.
   *
   * This function compute the barycenter of a cell using the position of its 
   * junctions.
   */
  template <class VVComplex>
    Point3d findCenter(const typename VVComplex::cell& c, VVComplex& T)
    {
      IMPORT_COMPLEX_VERTICES(VVComplex);
      IMPORT_COMPLEX_MODEL(VVComplex, T);
      // First get average of points as estimate to center
      Point3d cest;
      forall( const junction& n , T.S.neighbors(c))
      {
        cest += model.position(n);
      }
      cest /= T.S.valence(c);
      // Average Area * midpoint of each triangle
      Point3d cpos;
      double sum = 0;
      forall( const junction& m , T.S.neighbors(c))
      {
        const junction& n = T.S.prevTo(c, m);
        const Point3d& mpos = model.position(m);
        const Point3d& npos = model.position(n);
        double ta = geometry::triangleArea(mpos, npos, cest);
        cpos += ta * (cest + ((mpos - cest) + (npos - cest))/3.0);
        sum += ta;
      }
      cpos /= sum;

      return cpos;
    }


  /**
   * Compute the position of the center of a cell.
   *
   * This function compute the barycenter of a cell using the position of its 
   * junctions.
   *
   * \note The position of the center is only updated, thus there is no call to 
   * \ref setPositionHint
   *
   * \deprecated You should use vvcomplex::findCenter instead and update yourself 
   * the position of the cell.
   */
  template <class VVComplex>
    void FindCenter(const typename VVComplex::cell& c, VVComplex& T)
    {
      IMPORT_COMPLEX_MODEL(VVComplex, T);
      const Point3d& cpos = findCenter(c, T);
      model.setPosition(c, cpos);
    }

#ifdef USE_ALLOCATOR
#  define ALL_COMPLEX_TEMPLATE_ARGS Model,CellContent,JunctionContent,WallContent,CellEdgeContent,CellJunctionContent,JunctionCellContent,compact,Alloc,LeafClass
#  define MANDATORY_COMPLEX_TEMPLATE_ARGS Model,CellContent,JunctionContent,WallContent,CellEdgeContent,CellJunctionContent,JunctionCellContent,compact,Alloc
#else // USE_ALLOCATOR
#  define ALL_COMPLEX_TEMPLATE_ARGS Model,CellContent,JunctionContent,WallContent,CellEdgeContent,CellJunctionContent,JunctionCellContent,compact,LeafClass
#  define MANDATORY_COMPLEX_TEMPLATE_ARGS Model,CellContent,JunctionContent,WallContent,CellEdgeContent,CellJunctionContent,JunctionCellContent,compact
#endif // USE_ALLOCATOR

  /**
   * \class VVComplex complex.h <algorithms/complex.h>
   *
   * Class handling the representation and development of 2D cell %complex.
   */
  template <typename Model,
           typename CellContent, typename JunctionContent,
           typename WallContent = graph::_EmptyEdgeContent, typename CellEdgeContent = graph::_EmptyEdgeContent,
           typename CellJunctionContent = graph::_EmptyEdgeContent, typename JunctionCellContent = graph::_EmptyEdgeContent,
           bool compact = false,
#ifdef USE_ALLOCATOR
           typename Alloc = DEFAULT_ALLOC(CellContent),
#endif
           typename LeafClass = template_utils::this_class
             >
  class VVComplex
  {
    /**
     * Retrieve the leaf class in case it derived from VVComplex
     */
    typedef RESOLVE_LEAF_CLASS(LeafClass,VVComplex) leaf_class;

  public:
    typedef CellContent cell_content;
    typedef JunctionContent junction_content;
    typedef WallContent wall_content;
    typedef CellJunctionContent cell_junction_content;
    typedef JunctionCellContent junction_cell_content;
    typedef Model model_t;

    /**
     * Type of a cell vertex
     */
    typedef graph::Vertex<CellContent> cell;

    /**
     * Type of a junction vertex
     */
    typedef graph::Vertex<JunctionContent> junction;

    /**
     * Type of the cell graph
     */
    typedef graph::VVGraph<CellContent,CellEdgeContent,compact> cell_graph;

    /**
     * Type of the wall graph
     */
    typedef graph::VVGraph<JunctionContent,WallContent,compact> wall_graph;

    /**
     * Type of the cell complex graph
     */
    typedef VVComplexGraph<CellContent,JunctionContent,CellJunctionContent,JunctionCellContent,compact> complex_graph;

    /**
     * Type of an edge in the cell graph
     */
    typedef typename cell_graph::edge_t cell_edge;

    /**
     * Type of an arc in the cell graph
     */
    typedef typename cell_graph::arc_t cell_arc;

    /**
     * Type of an edge in the cell graph
     */
    typedef typename cell_graph::const_edge_t const_cell_edge;

    /**
     * Type of an edge in the wall graph;
     */
    typedef typename wall_graph::edge_t wall;

    /**
     * Type of an arc in the wall graph;
     */
    typedef typename wall_graph::arc_t wall_arc;

    /**
     * Type of an edge in the wall graph;
     */
    typedef typename wall_graph::const_edge_t const_wall;

    /**
     * Type of the edges from cell to junction in the cell complex graph
     */
    typedef typename complex_graph::edge1_t cell_junction_edge;

    /**
     * Type of the edges from cell to junction in the cell complex graph
     */
    typedef typename complex_graph::const_edge1_t const_cell_junction_edge;

    /**
     * Type of the edges from junction to cell in the cell complex graph
     */
    typedef typename complex_graph::edge2_t junction_cell_edge;

    /**
     * Type of the edges from junction to cell in the cell complex graph
     */
    typedef typename complex_graph::const_edge2_t const_junction_cell_edge;

    /**
     * Type of the division data used for the complex
     */
    typedef typename complex_graph::division_data division_data;

    /**
     * Model linked to this complex
     */
    Model* model;

    /**
     * Cell graph
     */
    cell_graph C;
    /**
     * Saved cell graph. Only during cell division!
     */
    cell_graph OldC;

    /**
     * Wall graph
     */
    wall_graph W;
    /**
     * Saved wall graph. Only during cell division!
     */
    wall_graph OldW;

    /**
     * Cell complex graph
     */
    complex_graph S;
    /**
     * Saved complex graph. Only during cell division!
     */
    complex_graph OldS;

    /**
     * If true, the parameters are also serialized.
     *
     * Not that this parameter will never be serialized ...
     */
    bool save_parameters;

    /**
     * Constructor
     */
    VVComplex(Model *mod)
      : model(mod)
      , save_parameters(false)
      { }

    /**
     * Copy constructor
     */
    VVComplex(const VVComplex& other)
      : model(other.model)
      , C(other.C)
      , W(other.W)
      , S(other.S)
      , save_parameters(other.save_parameters)
      { }

    /**
     * Virtual destructor allow for inheritance
     */
    virtual ~VVComplex() {}

    /// \cond InternalDoc
    struct AngledJunction
    {
      junction v;
      double angle;

      AngledJunction(const junction& v, double a)
        : v(v)
        , angle(a)
        {}

      AngledJunction(const AngledJunction& copy)
        : v(copy.v)
        , angle(copy.angle)
        {}

      bool operator<(const AngledJunction& other) const
      {
        if(angle == other.angle)
          return v < other.v;
        return angle < other.angle;
      }

      bool operator==(const AngledJunction& other) const
      {
        return (angle == other.angle) && (v == other.v);
      }
    };
    /// \endcond

    /**
     * Erase the graphs of the complex
     */
    virtual void clear()
    {
      S.clear();
      C.clear();
      W.clear();
      OldS.clear();
      OldC.clear();
      OldW.clear();
    }

    /**
     * Add a new cell with three vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3)
    {
      std::vector<junction> junctions(3, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with four vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4)
    {
      std::vector<junction> junctions(4, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with five vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5)
    {
      std::vector<junction> junctions(5, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with six vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6)
    {
      std::vector<junction> junctions(6, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with seven vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6, const junction& j7)
    {
      std::vector<junction> junctions(7, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      junctions[6] = j7;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with height vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6, const junction& j7,
                 const junction& j8)
    {
      std::vector<junction> junctions(8, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      junctions[6] = j7;
      junctions[7] = j8;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell with nine vertices to the %complex
     */
    bool addCell(const cell& c, const junction& j1, const junction& j2, const junction& j3,
                 const junction& j4, const junction& j5, const junction& j6, const junction& j7,
                 const junction& j8, const junction& j9)
    {
      std::vector<junction> junctions(9, junction(0));
      junctions[0] = j1;
      junctions[1] = j2;
      junctions[2] = j3;
      junctions[3] = j4;
      junctions[4] = j5;
      junctions[5] = j6;
      junctions[6] = j7;
      junctions[7] = j8;
      junctions[8] = j9;
      return this->addCell(c, junctions);
    }

    /**
     * Add a new cell to the %complex
     *
     * \param c Vertex holding the cell center
     * \param j Container holding the counter-clockwise ordered cycle 
     * defining the cell.
     *
     * \note \p c must not be in the complex already or result is undefined. 
     * However, all or part of the junctions may be in the vv graph 
     * already.
     *
     * \returns false if the addition fails
     */
    template <typename JunctionContainer>
      bool addCell(const cell& c, const JunctionContainer& junctions)
      {
        if(!S.addCell(c, junctions) ) return false;
        return this->addCellExtra(c);
      }

    /**
     * Method to override to add a cell in extra graphs.
     *
     * \note Do not forget to call this method again if you extend the complex.
     */
    virtual bool addCellExtra(const cell& c)
    {
      bool result = this->addCell_cell_graph(c) && this->addCell_wall_graph(c);
      return result;
    }

    /// \cond InternalDoc
    bool addCell_wall_graph(const cell& c)
    {
      // Next, connect from the walls
      forall(const junction& j, S.neighbors(c)) if(!W.contains(j))
        W.insert(j);
      forall(const junction& j, S.neighbors(c))
      {
        const junction& pj = S.prevTo(c, j);
        const junction& nj = S.nextTo(c, j);
        if(!W.edge(j, pj)) // Then connect j -> pj
        {
          if(W.valence(j) < 2)
          {
            W.insertEdge(j, pj);
          }
          else if(W.edge(j, nj))
          {
            W.spliceAfter(j, nj, pj);
          }
          else
          {
            const cell& nc = S.nextTo(j, c);
            if(nc == c)
            {
              cerr << "There is something wrong: the junction already has more than two neighbors in W but only one cell in c" << endl;
              return false;
            }
            W.spliceBefore(j, S.nextTo(nc, j), pj);
          }
        }
        if(!W.edge(j, nj)) // Then connect j -> nj
        {
          if(W.valence(j) < 2)
          {
            W.insertEdge(j, nj);
          }
          else if(W.edge(j, pj))
          {
            W.spliceBefore(j, pj, nj);
          }
          else
          {
            const cell& pc = S.prevTo(j, c);
            if(pc == c)
            {
              cerr << "There is something wrong: the junction already has more than two neighbors in W but only one cell in c" << endl;
              return false;
            }
            W.spliceAfter(j, S.prevTo(pc, j), nj);
          }
        }
      }
      return true;
    }

  private:
    // Function used in addCell_cell_graph
    const cell& acg_find_next_cell(const junction& j, const cell& nc, const cell& c)
    {
      const junction& nj = S.nextTo(nc, j);
      forall( const cell& nnc , S.neighbors(j))
      {
        if(nnc != nc and nnc != c and S.edge(nj, nnc))
          return nnc;
      }
      return acg_find_next_cell(nj, nc, c);
    }

  public:
    bool addCell_cell_graph(const cell& c)
    {
      C.insert(c);
      // At last, connect the cells
      cell pc(0);
      forall(const junction& j, S.neighbors(c))
      {
        const cell& nc = S.adjacentCell(c, j);
        if(nc and !C.edge(c,nc))
        {
          if(pc)
          {
            if(!C.spliceAfter(c, pc, nc))
            {
              cerr << "error in "  << __FILE__ << " line " << __LINE__ << endl;
              return false;
            }
          }
          else
          {
            if(!C.insertEdge(c, nc))
            {
              cerr << "error in "  << __FILE__ << " line " << __LINE__ << endl;
              return false;
            }
          }
          if(C.valence(nc) < 2)
          {
            if(!C.insertEdge(nc, c))
            {
              cerr << "error in "  << __FILE__ << " line " << __LINE__ << endl;
              return false;
            }
          }
          else
          {
            const cell& nnc = acg_find_next_cell(j, nc, c);
            if(!C.spliceBefore(nc, nnc, c))
            {
              cerr << "error in "  << __FILE__ << " line " << __LINE__ << endl;
              return false;
            }
          }
          pc = nc;
        }
      }
      return true;
    }
    ///\endcond

    /**
     * Add a new cell to the complex
     *
     * This function add one cell to the %complex. The set of junctions is 
     * first sorted in counterclockwise order around the position of the cell 
     * \p c and then the addCell() method is called. For the sorting to work, 
     * the cell center need to be correctly positioned with its normal and 
     * the junctions just need to be positioned.
     *
     * \param c First cell of the %complex
     * \param j Set of junctions of the cell
     */
    template <typename JunctionContainer>
      bool addCellUnsorted(const cell& c, const JunctionContainer& j)
      {
        Point3d normal = model->normal(c);
        Point3d cpos, ref, u1, u2;
        cpos = model->position(c);
        normal.normalize();
        bool ref_found = false;
        std::set<AngledJunction> sorted;
        forall(const junction& v,j)
        {
          if(ref_found)
          {
            const Point3d& pos = model->position(v);
            Point3d u = pos - cpos;
            Point3d proj = normalized(u - (u*normal)*normal);
            double cosa = proj*u1;
            double sina = (u1^proj)*normal;
            double angle = std::atan2(sina, cosa);
            sorted.insert(AngledJunction(v, angle));
          }
          else
          {
            ref = model->position(v);
            u1 = normalized(ref - cpos);
            sorted.insert(AngledJunction(v, 0));
            ref_found = true;
          }
        }
        std::vector<junction> sorted_list;
        forall(const AngledJunction& aj,sorted)
        {
          sorted_list.push_back(aj.v);
        }
        return this->addCell(c, sorted_list);
      }

    typedef typename complex_graph::division_result_t division_result_t;

    /**
     * Divide a cell of the %complex.
     *
     * In the \c updateFromOld call, cl will be the cell having the wall (u-v) 
     * and cr the cell having the wall (v-u). In the same way, cl has the 
     * vertices of c from v to u and cr from u to v.
     *
     * \param c Cell to divide
     * \param ddata Specification of where the division points should appear
     * \param to_keep Set of cells to keep to update the data in the new cells
     *
     * \see \ref updateFromOld
     */
    division_result_t divideCell( const cell& c,
                                  const division_data& ddata,
                                  const std::vector<cell>&to_keep)
    {
      std::set<junction> junctions_to_keep;

      forall(const cell& k,to_keep)
      {
        forall( const junction& j , S.neighbors(k))
        {
          junctions_to_keep.insert(j);
        }
      }

      OldS = S.subgraph(to_keep, junctions_to_keep);
      this->saveSubgraphs(to_keep, junctions_to_keep);

      division_result_t res = S.divideCell(c, ddata);
      if(!res)
        return res;

      // Set the position of u
      if(!ddata.divide_at_u1)
      {
        const Point3d& pu1 = model->position(res.u1);
        const Point3d& pu2 = model->position(res.u2);
        double ru = ((ddata.pu-pu1)*(pu2-pu1)) / normsq(pu2-pu1);
        model->setPositionHint(res.u, res.u1, res.u2, ru);
        model->setPosition(res.u, ddata.pu);
      }

      // Set the position of v
      if(!ddata.divide_at_v1)
      {
        const Point3d& pv1 = model->position(res.v1);
        const Point3d& pv2 = model->position(res.v2);
        double rv = ((ddata.pv-pv1)*(pv2-pv1)) / normsq(pv2-pv1);
        model->setPositionHint(res.v, res.v1, res.v2, rv);
        model->setPosition(res.v, ddata.pv);
      }

      // Set new centers, clear, and update concentrations from parent
      //FindCenter(cl, static_cast<leaf_class&>(*this));
      //FindCenter(cr, static_cast<leaf_class&>(*this));

      // Connect the other graphs
      this->reconnectGraphs(c, res.cl, res.cr, res.u1, res.u2, res.v1, res.v2, res.u, res.v);

      model->updateFromOld(res.cl, res.cr, c, ddata, static_cast<leaf_class&>(*this));
      OldS.clear();
      this->clearOldGraphs();
      return res;
    }

    division_result_t divideCell( const cell& c,
                                  const division_data& ddata )
    {
      return divideCell(c, ddata, std::vector<cell>());
    }

    /**
     * Divide a cell with the cells to keep defined by any kind of container.
     */
    template <typename CellContainer>
      division_result_t divideCell( const cell& c,
                                    const division_data& ddata,
                                    const CellContainer& to_keep)
      {
        std::vector<cell> tk(to_keep.begin(), to_keep.end());
        return this->divideCell(c, ddata, tk);
      }

    /**
     * Divide a cell of the %complex
     *
     * \param c Cell to divide
     * \param params Parameter of the algorithm to use. The type of the 
     * params will decide which algorithm to use.
     * \param to_keep Set of cells to keep to update the data in the new cells
     *
     * \see \ref updateFromOld
     */
    template <typename AlgoParameter>
      division_result_t divideCell( const cell& c, // cell to divide
                                    const AlgoParameter& params,
                                    const std::vector<cell>& to_keep)
      {
        // Exit if division fails
        const division_data& result = findDivisionPoints(c, static_cast<leaf_class&>(*this), params);
        if(!result)
        {
          OldS.clear();
          this->clearOldGraphs();
          cerr << "Error, division failed" << endl;
          return division_result_t();
        }

        return divideCell(c, result, to_keep);
      }

    void reconnectWallGraph(const junction& u1, const junction& u2,
                            const junction& v1, const junction& v2,
                            const junction& nu, const junction nv)
    {
      // First, update W
      if(nu != u1)
        algorithms::insert(u1, u2, W, nu);
      if(nv != v1)
        algorithms::insert(v1, v2, W, nv);

      W.spliceAfter(nu, u2, nv);
      W.spliceAfter(nv, v2, nu);
    }

    void reconnectCellGraph(const cell& cl, const cell& cr, const cell& c)
    {
      // Then C
      C.insert(cl);
      C.insert(cr);

      connectFromJunctions(c, cl, cr);
      connectFromJunctions(c, cr, cl);

      C.erase(c);
    }

    /**
     * Virtual method to reconnect the graphs after cell division.
     * This method input all the parameters for the cell division. The division 
     * itself already occurred only for the complex graph at that point. This 
     * function maintain the consistency with other graphs.
     *
     * \note Don't forget to call this method from your overloaded version.
     */
    virtual void reconnectGraphs(const cell& c, const cell& cl, const cell& cr,
                                 const junction& u1, const junction& u2,
                                 const junction& v1, const junction& v2,
                                 const junction& nu, const junction& nv)
    {
      reconnectWallGraph(u1, u2, v1, v2, nu, nv);
      reconnectCellGraph(cl, cr, c);
    }

    /**
     * Divide a cell with the cells to keep defined by any kind of container.
     */
    template <typename AlgoParameter, typename CellContainer>
      division_result_t divideCell( const cell& c, // cell to divide
                                    const AlgoParameter& params,
                                    const CellContainer& to_keep )
      {
        std::vector<cell> tk(to_keep.begin(), to_keep.end());
        return this->divideCell(c, params, tk);
      }

    /**
     * Virtual method to redefine to save the status of the graphs maintained by 
     * an extension.
     *
     * \note Don't forget to call this method.
     */
    virtual void saveSubgraphs(const std::vector<cell>& cells_to_keep,
                               const std::set<junction>& junctions_to_keep)
    {
      OldC = C.subgraph(cells_to_keep);
      OldW = W.subgraph(junctions_to_keep);
    }

    /**
     * Called after the division, erased the old graphs maintained during the 
     * division itself.
     */
    virtual void clearOldGraphs()
    {
      OldC.clear();
      OldW.clear();
    }

    /**
     * Divide a cell
     *
     * Convenience function where a single cell is to be kept
     */
    division_result_t divideCell( const cell& to_divide, // cell to divide
                                  const division_data& params,
                                  const cell& cell_kept )
    {
      std::vector<cell> k;
      k.push_back(cell_kept);
      return this->divideCell(to_divide, params, k);
    }

    /**
     * Divide a cell
     *
     * Convenience function where a single cell is to be kept
     */
    template <typename AlgoParameter>
      division_result_t divideCell( const cell& to_divide, // cell to divide
                                    const AlgoParameter& params,
                                    const cell& cell_kept )
      {
        std::vector<cell> k;
        k.push_back(cell_kept);
        return this->divideCell(to_divide, params, k);
      }

    /**
     * Delete a cell from the graph
     */
    void deleteCell(const cell& c)
    {
      const std::vector<junction>& deleted =  S.deleteCell(c);
      this->deleteCellInGraphs(c, deleted);
    }

    /**
     * Function to override \b and \b call to handle cell deletion if other 
     * graphs are maintained.
     */
    virtual void deleteCellInGraphs(const cell& c, const std::vector<junction>& deleted)
    {
      C.erase(c);
      forall(const junction& j,deleted)
        W.erase(j);
    }

    /**
     * Merge two adjacent cells.
     *
     * Only c1 is kept, c2 is being deleted.
     *
     * \returns the junctions at the start and end of the wall that has been 
     * removed
     */
    std::pair<junction, junction> mergeCells(const cell& c1, const cell& c2)
    {
      std::vector<junction> deleted_junctions;
      std::pair<junction,junction> result = S.mergeCells(c1, c2, &deleted_junctions);

      const junction& j1 = result.first;
      const junction& j2 = result.second;

      // Remove junctions from W
      forall(const junction& j, deleted_junctions)
      {
        W.erase(j);
      }
      W.spliceAfter(j1, S.prevTo(c1, j1), j2);
      W.spliceBefore(j2, S.nextTo(c1, j2), j1);

      // Update cell graph
      const cell& after = C.nextTo(c1, c2);
      forall(const cell& c, C.neighbors(c2, C.nextTo(c2, c1)))
      {
        if(c == c1) break;
        C.spliceBefore(c1, after, c);
        C.replace(c, c2, c1);
      }

      // Erase c2 from C and S
      C.erase(c2);

      return result;
    }


    /**
     * Returns the cell adjacent the c sharing the wall (j, S.nextTo(c,j))
     */
    const cell& adjacentCell(const cell& c, const junction& j) const
    {
      return S.adjacentCell(c, j);
    }

    /**
     * Returns the cell(s) sharing the wall (j1,j2).
     *
     * \returns a pair of cells (c1,c2) such that in c1, j1 is before j2 and in 
     * c2, j2 is before j1.
     */
    std::pair<cell,cell> adjacentCells(const junction& j1, const junction& j2) const
    {
      return S.adjacentCells(j1, j2);
    }

    /**
     * \returns the junction in \c c1 starting the wall common between \c c1 
     * and \c c2.
     */
    const junction& interfaceWall(const cell& c1, const cell& c2) const
    {
      return S.interfaceWall(c1, c2);
    }

    /**
     * \returns the pair of junctions in \c c1 such that the interface between 
     * \c c1 and \c c2 starts at the first junction and end at the second. In 
     * case the \c c1 is entirely contained in \c c2 (or the other way around), 
     * both junctions will be the same.
     */
    std::pair<junction, junction> interfaceWallSpan(const cell& c1, const cell& c2) const
    {
      return S.interfaceWallSpan(c1, c2);
    }

    /**
     * Split a wall in two and return the new junction.
     */
    const junction& splitWall(const junction& j1, const junction& j2, const junction& x = junction(0))
    {
      const junction& j = algorithms::insert(j1, j2, W, x);
      S.splitWall(j1, j2, j);
      this->splitWallExtra(j1, j2, j);
      return j;
    }

    /**
     * Split a wall in two and return the new junction.
     */
    const junction& splitWall(const cell& c, const junction& j1, const junction& x = junction(0))
    {
      const junction& j2 = S.nextTo(c, j1);
      const junction& j = algorithms::insert(j1, j2, W, x);
      S.splitWall(c, j1, j);
      this->splitWallExtra(j1, j2, j);
      return j;
    }

    /**
     * Called after a wall was split was done.
     *
     * The argument is the junction inserted. Getting the original wall is 
     * easily done using W.anyIn(j) and W.nextTo(W.anyIn(j)).
     */
    virtual void splitWallExtra(const junction& , const junction& , const junction& ) { }

    /**
     * Erase a junction.
     *
     * This is possible only if the junction do not change the structure of the 
     * cell graph.
     */
    bool deleteJunction(const junction& j)
    {
      size_t nb_cells = S.valence(j);
      size_t nb_junctions = W.valence(j);
      if(nb_cells > 2)
        return false;
      if(nb_cells == 2 and nb_junctions != 2)
        return false;
      if(nb_cells > 0)
      {
        const cell& c = S.anyIn(j);
        const junction& pj = S.prevTo(c, j);
        const junction& nj = S.nextTo(c, j);
        W.replace(pj, j, nj);
        W.replace(nj, j, pj);
      }
      W.erase(j);
      S.erase(j);
      this->deleteJunctionExtra(j);
      return true;
    }

    /**
     * Called after a junction have been erased.
     */
    virtual void deleteJunctionExtra(const junction& /*j*/) {}

    /**
     * True if a cell has an border face.
     */
    bool border(const cell& c) const
    {
      size_t nb_walls = 0;
      forall(const junction& j, S.neighbors(c))
      {
        if(W.valence(j) > 2)
          nb_walls++;
      }
      return C.empty(c) or nb_walls != C.valence(c);
    }

    /**
     * True if the junction is on the border
     */
    bool border(const junction& j) const
    {
      return S.valence(j) != W.valence(j);
    }

    /**
     * True if the wall on the cell c starting at junction j is on the border
     */
    bool border(const cell& c, const junction &j ) const
    {
      return S.border(c, j);
    }

    /**
     * True if the wall from j1 to j2 is at the border of the cell complex.
     */
    bool border(const junction& j1, const junction& j2) const
    {
      return S.border(j1, j2);
    }

    /**
     * Serialization method
     *
     * By default, only the three main graphs are serialized: S, T and W. No 
     * parameter and no old graph are saved.
     *
     * \warning You need to include <storage/complex.h> to use this 
     * serialization method
     */
    bool serialize(storage::VVEStorage&);

  protected:
    /**
     * Find the connections of cell c in the cell graph, when it's correctly 
     * connected in the vv graph.
     *
     * \warning For internal use only
     */
    void connectFromJunctions( const cell& old_c, const cell& c, const cell& cc)
    {
      C.insertEdge(c, cc);
      // Connect c to its neighbors
      cell pn(0);
      forall( const junction& jn , S.neighbors(c))
      {
        const junction& jp = S.prevTo(c, jn);
        const cell& n = S.nextTo(jn, c);
        // If we are not at a border ...
        if(n != c and S.edge(jp, n))
        {
          if(n != cc)
          {
            // Two cases:
            // 1 - old_c is in the neighborhood of n -> we replace it by the 
            // current
            // 2 - cc is in the neighborhood of n -> we figure if we should 
            // place it before or after
            cell_edge e = C.replace(n, old_c, c);
            if(e.isNull()) // c was not in the neighborhood of n
            {
              const cell& cc1 = S.nextTo(jp, c);
              if(cc1 == cc)
              {
                C.spliceBefore(n, cc, c);
              }
              else
              {
                C.spliceAfter(n, cc, c);
              }
            }
            if(pn.isNull())
            {
              C.insertEdge(c, n);
            }
            else
            {
              C.spliceAfter(c, pn, n);
            }
          }
          pn = n;
        }
      }
    }

  };

  template <typename Complex>
    std::vector<typename Complex::junction> contour(const Complex& T)
    {
      return T.S.contour();
    }

}

#endif // VVELIB_ALGORITHMS_COMPLEX_H

