#ifndef VVELIB_ALGORITHMS_DRAW_CONNECTIONS_H
#define VVELIB_ALGORITHMS_DRAW_CONNECTIONS_H

#include <config.h>
#include <util/palette.h>
#include <util/parms.h>
#include <viewer.h>
#include <geometry/geometry.h>

namespace algorithms
{

  using geometry::Point3d;

  template <typename src_t, typename tgt_t, typename Model, typename Graph, typename iterator>
  void _drawGraphConnections(Model* model, Viewer* viewer, const Graph& G, iterator (Graph::*begin)() const, iterator (Graph::*end)() const, bool order_neighbors)
  {
    bool has_cull_face = glIsEnabled(GL_CULL_FACE);
    glDisable(GL_CULL_FACE);
    forall_range(const src_t& c, std::make_pair((G.*begin)(), (G.*end)()))
    {
      double nc = G.valence(c);
      double min_l = HUGE_VAL;
      const Point3d& cpos = model->position(c);
      if(order_neighbors)
      {
        forall(const tgt_t& n, G.neighbors(c))
        {
          const Point3d& npos = model->position(n);
          double l = norm(npos - cpos);
          if(l < min_l)
            min_l = l;
        }
        min_l /= 2*nc;
      }

      int i = 1;
      forall(const tgt_t& n, G.neighbors(c))
      {
        const Point3d& npos = model->position(n);
        Point3d endpos = cpos;
        if(order_neighbors)
        {
          Point3d u = util::normalized(npos - cpos);
          endpos += u*min_l*i;
          i++;
        }
        else
        {
          endpos += (npos-cpos)/2;
        }
        viewer->drawArrow(Vec(cpos), Vec(endpos));
      }
    }
    if(has_cull_face)
      glEnable(GL_CULL_FACE);
  }

  /**
   * Draw the connections of a graph using the drawArrow method of the viewer.
   *
   * This method is required in the model:
   * \code
   * Point3d position(const vertex& v);
   * \endcode
   */
  template <typename Model, typename GraphType>
  void drawVVGraphConnections(Model* model, Viewer* viewer, const GraphType& G, bool order_neighbors)
  {
    typedef typename GraphType::vertex_t vertex;
    typedef typename GraphType::const_iterator (GraphType::*range_fct)() const;
    _drawGraphConnections<vertex,vertex>(model, viewer, G, (range_fct)&GraphType::begin, (range_fct)&GraphType::end, order_neighbors);
  }

  /**
   * Draw the connections of a bigraph using the drawArrow method of the viewer.
   *
   * This function draws the connections from the vertex1 to the vertex2.
   *
   * This method is required in the model:
   * \code
   * Point3d position(const vertex& v);
   * \endcode
   */
  template <typename Model, typename GraphType>
  void drawVVBiGraphConnections1(Model* model, Viewer* viewer, const GraphType& G, bool order_neighbors)
  {
    typedef typename GraphType::vertex1_t vertex1;
    typedef typename GraphType::vertex2_t vertex2;
    typedef typename GraphType::const_iterator1 (GraphType::*range_fct)() const;
    _drawGraphConnections<vertex1,vertex2>(model, viewer, G, (range_fct)&GraphType::begin_vertex1, (range_fct)&GraphType::end_vertex1, order_neighbors);
  }

  /**
   * Draw the connections of a bigraph using the drawArrow method of the viewer.
   *
   * This function draws the connections from the vertex2 to the vertex1.
   *
   * This method is required in the model:
   * \code
   * Point3d position(const vertex& v);
   * \endcode
   */
  template <typename Model, typename GraphType>
  void drawVVBiGraphConnections2(Model* model, Viewer* viewer, const GraphType& G, bool order_neighbors)
  {
    typedef typename GraphType::vertex1_t vertex1;
    typedef typename GraphType::vertex2_t vertex2;
    typedef typename GraphType::const_iterator2 (GraphType::*range_fct)() const;
    _drawGraphConnections<vertex2,vertex1>(model, viewer, G, (range_fct)&GraphType::begin_vertex2, (range_fct)&GraphType::end_vertex2, order_neighbors);
  }

}

#endif // VVELIB_ALGORITHMS_DRAW_CONNECTIONS_H

