/**
 * \file algorithms/graph.h
 *
 * Implement common algorithms on graphs
 */

#ifndef VVELIB_ALGORITHMS_GRAPH_H
#define VVELIB_ALGORITHMS_GRAPH_H

#include <config.h>
#include <graph/vvgraph.h>
#include <set>
#include <util/unorderedset.h>
#include <utility>

namespace algorithms
{

  /// \cond NoDoc

  namespace shortest_path
  {
    template <typename VertexContent>
    struct DistanceStructure
    {
      DistanceStructure(const graph::Vertex<VertexContent>& tgt, double d, const graph::Vertex<VertexContent>& src)
        : distance(d)
        , source(src)
        , target(tgt)
      { }
      double distance;
      graph::Vertex<VertexContent> source, target;
    };

    template <typename VertexContent>
    DistanceStructure<VertexContent> make_distance(const graph::Vertex<VertexContent>& tgt, double distance, const graph::Vertex<VertexContent>& src)
    {
      return DistanceStructure<VertexContent>(tgt, distance, src);
    }

    struct no_sp_tag {};

    template <typename VertexContent, typename Model, typename tag_t>
    double distance( const graph::Vertex<VertexContent>& tgt, const Model& m, const tag_t& t)
    {
      return m.graphDistance(tgt, t);
    }

    template <typename VertexContent, typename Model>
    double distance( const graph::Vertex<VertexContent>& tgt, const Model& m, const no_sp_tag&)
    {
      return m.graphDistance(tgt);
    }

    template <typename VertexContent, typename Model, typename tag_t>
    double distance( const graph::Vertex<VertexContent>& src, const graph::Vertex<VertexContent>& tgt, const Model& m, const tag_t& t)
    {
      if(src == tgt)
        return 0;
      return m.graphDistance(src, tgt, t);
    }

    template <typename VertexContent, typename Model>
    double distance( const graph::Vertex<VertexContent>& src, const graph::Vertex<VertexContent>& tgt, const Model& m, const no_sp_tag&)
    {
      if(src == tgt)
        return 0;
      return m.graphDistance(src, tgt);
    }

    template <typename VertexContent, typename Model, typename tag_t>
    void setDistance( const graph::Vertex<VertexContent>& src, const graph::Vertex<VertexContent>& tgt, double dist, Model& m, const tag_t& t)
    {
      m.setGraphDistance(src, tgt, dist, t);
    }

    template <typename VertexContent, typename Model>
    void setDistance( const graph::Vertex<VertexContent>& src, const graph::Vertex<VertexContent>& tgt, double dist, Model& m, const no_sp_tag&)
    {
      m.setGraphDistance(src, tgt, dist);
    }

    template <typename VertexContent, typename Model, typename tag_t>
    double edgeWeight( const Vertex<VertexContent>& src, const Vertex<VertexContent>& tgt, const Model& m, const tag_t& t)
    {
      return m.edgeWeight(src, tgt, t);
    }

    template <typename VertexContent, typename Model>
    double edgeWeight( const Vertex<VertexContent>& src, const Vertex<VertexContent>& tgt, const Model& m, const no_sp_tag&)
    {
      return m.edgeWeight(src, tgt);
    }

    template <typename VertexContent>
    struct CompareWeightedVertexes
    {
      typedef graph::Vertex<VertexContent> vertex;
      typedef std::pair<vertex,double> value_type;
      bool operator()(const value_type& v1, const value_type& v2) const
      {
        if(v1.second == v2.second)
          return v1.first < v2.first;
        return v1.second < v2.second;
      }
    };

    template <typename VertexContent>
    struct CompareDistances
    {
      typedef DistanceStructure<VertexContent> value_type;
      bool operator()(const value_type& v1, const value_type& v2) const
      {
        double d1 = v1.distance;
        double d2 = v2.distance;
        if(d1 == d2)
          return v1.target < v2.target;
        return d1 < d2;
      }
    };

  }

  /// \endcond

  /**
   * \fn void shortest_paths_FloydWarshall( const graph::VVGraph<VertexContent,EdgeContent>& m, Model& model, const tag_t& tag)
   * Compute the shortest path for all pair of vertices in the %graph.
   *
   * This function uses the Floyd-Marshall method (cf. http://en.wikipedia.org/wiki/Floyd-Warshall_algorithm).
   * \param m VVGraph to compute the distances on
   * \param model %Model used to gather/save data
   * \param tag This tag is used to allow to call this function on different graphs from within the same model. It will be sent back to the model in each method call.
   * The model is expected to provide a few methods:
   *
   * \code
   * void setGraphDistance( const vertex& src, const vertex& tgt, double distance, const tag_t&);
   * \endcode
   * Ask the model to store the distance from \c src to \c tgt.
   *
   * \code
   * double graphDistance( const vertex& src, const vertex& tgt, const tag_t&) const;
   * \endcode
   * Returns the latest stored distance from \c src to \c tgt. This function will never be called before the corresponding \c setGraphDistance has been called.
   *
   * \code
   * double edgeWeight( const vertex& src, const vertex& tgt, const tag_t&) const;
   * \endcode
   * Return the weigth of the edge from \c src to \c tgt. You can assume that 
   * the edge from \c src to \c tgt exists. In case the edge exists but you want 
   * it to be not considered, return \c HUGE_VAL.
   */
  template <typename VertexContent, typename EdgeContent, bool compact, typename Model, typename tag_t>
  void shortest_paths_FloydWarshall( const graph::VVGraph<VertexContent,EdgeContent,compact>& m, Model& model, const tag_t& t )
  {
    typedef graph::VVGraph<VertexContent,EdgeContent,compact> vvgraph;
    typedef typename vvgraph::vertex_t vertex;
    // Initialization
    forall(const vertex& v1, m)
    {
      forall(const vertex& v2, m)
      {
        if(v1 != v2)
        {
          shortest_path::setDistance(v1, v2, m.edge(v1, v2)?shortest_path::edgeWeight(v1,v2,model,t):HUGE_VAL,model,t);
          shortest_path::setDistance(v2, v1, m.edge(v2, v1)?shortest_path::edgeWeight(v2,v1,model,t):HUGE_VAL,model,t);
        }
      }
    }
    // Main loop
    forall(const vertex& k, m)
    {
      forall(const vertex& i, m)
      {
        forall(const vertex& j, m)
        {
          if( i != j )
          {
            double dist = shortest_path::distance(i,k,model,t) + shortest_path::distance(k,j,model,t);
            if(dist < shortest_path::distance(i,j,model,t))
            {
              shortest_path::setDistance(i,j,dist,model,t);
            }
          }
        }
      }
    }
  }

  /**
   * Compute the shortest path for all pair of vertices in the %graph.
   *
   * This version do not have a tag and will call all the required methods in the model without a tag (to use either as a default or if there is no other call to this function)
   * \see shortest_paths_FloydWarshall(const graph::VVGraph<VertexContent,EdgeContent>&, Model&, const tag_t&)
   */
  template <typename VVGraph, typename Model>
  void shortest_paths_FloydWarshall( const VVGraph& m, const Model& model)
  {
    shortest_paths_FloydWarshall(m, model, shortest_path::no_sp_tag());
  }

  /**
   * \fn void shortest_paths_Dijkstra( const Vertex<VertexContent>& src, const graph::VVGraph<VertexContent,EdgeContent>& m, Model& model, const tag_t& tag)
   * Compute the shortest path from a source to all other vertex of the %graph.
   *
   * This function uses the Dijkstra algorithm (http://en.wikipedia.org/wiki/Dijkstra%27s_algorithm).
   *
   * \param src Source from which the distance are computed
   * \param m VVGraph to work on
   * \param model %Model defining the required methods
   * \param tag This tag is used to allow to call this function on different graphs from within the same model. It will be sent back to the model in each method call.
   *
   * The method required in the model are the same as for shortest_paths_FloydWarshall(const graph::VVGraph<VertexContent,EdgeContent>&, Model&, const tag_t&)
   */
  template <typename VertexContent, typename EdgeContent, bool compact, typename Model, typename tag_t>
  void shortest_paths_Dijkstra( const Vertex<VertexContent>& src, const graph::VVGraph<VertexContent,EdgeContent,compact>& m, Model& model, const tag_t& t)
  {
    typedef graph::VVGraph<VertexContent,EdgeContent,compact> vvgraph;
    typedef typename vvgraph::vertex_t vertex;
    std::set<std::pair<vertex,double>,shortest_path::CompareWeightedVertexes<VertexContent> > active_vertexes;
    active_vertexes.insert(std::make_pair(src,0));
    forall(const vertex& v, m)
    {
      if(v != src)
      {
        shortest_path::setDistance(src, v, HUGE_VAL, model, t);
      }
    }
    while(!active_vertexes.empty())
    {
      std::pair<vertex,double> av = *active_vertexes.begin();
      active_vertexes.erase(active_vertexes.begin());
      double dist = av.second;
      forall(const vertex& v, m.neighbors(av.first))
      {
        double w = shortest_path::edgeWeight(av.first,v,model,t);
        double d1 = dist+w;
        if(d1 < shortest_path::distance(src,v,model,t))
        {
          shortest_path::setDistance(src,v,d1,model,t);
          active_vertexes.insert(std::make_pair(v,d1));
        }
      }
    }
  }

  template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
  void shortest_paths_Dijkstra( const Vertex<VertexContent>& src, const graph::VVGraph<VertexContent,EdgeContent,compact>& m, const Model& model)
  {
    shortest_paths_Dijkstra(src, m, model, shortest_path::no_sp_tag());
  }

  /**
   *\fn shortest_paths_Dijkstra( const Vertex<VertexContent>& src, const graph::VVGraph<VertexContent,EdgeContent>& m, Model& model, const tag_t&) 
   */
  template <typename VertexContent, typename EdgeContent, bool compact, typename Model, typename tag_t>
  void shortest_paths_Dijkstra( const std::unordered_set<Vertex<VertexContent> >& srcs, const graph::VVGraph<VertexContent,EdgeContent,compact>& m, Model& model, const tag_t& t)
  {
    typedef graph::VVGraph<VertexContent,EdgeContent,compact> vvgraph;
    typedef typename vvgraph::vertex_t vertex;
    std::set<shortest_path::DistanceStructure<VertexContent>,shortest_path::CompareDistances<VertexContent> > active_vertexes;
    forall(const vertex& v, srcs)
    {
      active_vertexes.insert(shortest_path::make_distance(v,0,v));
    }
    while(!active_vertexes.empty())
    {
      shortest_path::DistanceStructure<VertexContent> av = *active_vertexes.begin();
      active_vertexes.erase(active_vertexes.begin());
      double dist = av.distance;
      forall(const vertex& v, m.neighbors(av.target))
      {
        double w = shortest_path::edgeWeight(av.target,v,model,t);
        double d1 = dist+w;
        if(d1 < shortest_path::distance(v,model,t))
        {
          shortest_path::setDistance(av.source,v,d1,model,t);
          active_vertexes.insert(shortest_path::make_distance(v,d1,av.source));
        }
      }
    }
  }


};

#endif // VVELIB_ALGORITHMS_GRAPH_H
