#ifndef VVELIB_ALGORITHMS_INSERT_H
#define VVELIB_ALGORITHMS_INSERT_H
/**
 * \file insert.h
 *
 * Defines the algorithms::Insert class template
 */

#include <config.h>
#include <utility>
#include <util/static_assert.h>
#include <graph/vvgraph.h>

/**
 * Namespace containing various useful algorithms
 */
namespace algorithms
{

  /**
   * \class Insert insert.h <algorithms/insert.h>
   * \brief Insert a new vertex on an edge.
   * \param a an existing vertex
   * \param b an exsisting vertex
   * \param S the vv graph to edit
   *
   * This function creates a new vertex and inserts it between two existing
   * vertices by replacing it into the existing neighbourhoods.  If the vertices
   * have no relation or an assymmetric relation, a warning is printed to stderr
   * and an empty vertex is returned.
   *
   * Example:
   * \code
   * vvgraph S;
   * vertex v1, v2;
   * S.insert(v1);
   * S.insert(v2);
   * S.insertEdge(v1,v2);
   * S.insertEdge(v2,v1);
   * Insert<vvgraph> insert;
   * vertex v = insert(v1, v2, S);
   * \endcode
   *
   * The pre-condition is that edges (v1,v2) and (v2,v1) exist.
   *
   * The post-condition is that v replaced v2 in the neighborhood of v1 and v1 
   * in the neighborhood of v2, keeping the double connections.
   */
  template <class vvgraph, bool do_checks = true>
    class Insert
#ifdef JUST_FOR_DOXYGEN
  { };
#else
    ;
#endif

  template <class vvgraph>
    class Insert<vvgraph, true>
    {
    public:
      typedef typename vvgraph::vertex_t vertex;

      vertex operator()(const vertex& a, const vertex& b, vvgraph& S, const vertex& x = vertex(0)) const
      {
        if (a.isNull() || b.isNull())
        {
          return vertex(0);
        }

        unsigned int check = 0;
        if (S.edge(a,b)) check++;
        if (S.edge(b,a)) check++;

        switch (check)
        {
          case 0:
            std::cerr << "Warning: Attempt to insert a vertex between vertices that have no relation." << std::endl;
            return vertex(0);
            break;
          case 1:
            std::cerr << "Warning: Attempt to insert a vertex between vertices that have an assymetric relation." << std::endl;
            return vertex(0);
            break;
          default:
            break;
        }

        const vertex& r = *S.insert(x?x:vertex());

        S.insert(r);

        S.replace(a, b, r);
        S.replace(b, a, r);

        S.insertEdge(r, a);
        S.insertEdge(r, b);

        return r;
      }
    };

  template <class vvgraph>
    class Insert<vvgraph, false>
    {
    public:
      typedef typename vvgraph::vertex_t vertex;

      const vertex& operator()(const vertex& a, const vertex& b, vvgraph& S, const vertex& x = vertex(0)) const
      {
        const vertex& r = *S.insert(x?x:vertex());

        S.replace(a, b, r);
        S.replace(b, a, r);

        S.insertEdge(r, a);
        S.insertEdge(r, b);

        return r;
      }
    };

  template <class VertexContent, class EdgeContent, bool compact>
  const graph::Vertex<VertexContent>& insert( const graph::Vertex<VertexContent>& a,
                                              const graph::Vertex<VertexContent>& b,
                                              graph::VVGraph<VertexContent,EdgeContent,compact>& S,
                                              const graph::Vertex<VertexContent>& x = graph::Vertex<VertexContent>(0) )
  {
    typedef Insert<graph::VVGraph<VertexContent,EdgeContent,compact>,false> Insert;
    static const Insert fct = Insert();
    return fct(a,b,S,x);
  }

  /**
   * Splice \c nv after \c ref in \c v if ref is not null. Otherwise just 
   * insert the edge (v,nv) in S.
   */
  template <class Graph>
  typename Graph::edge_t insertAfter(const typename Graph::vertex_t& v,
                                     const typename Graph::vertex_t& ref,
                                     const typename Graph::vertex_t& nv,
                                     Graph& S)
  {
    if(ref)
      return S.spliceAfter(v, ref, nv);
    return S.insertEdge(v, nv);
  }

  /**
   * Splice \c nv before \c ref in \c v if ref is not null. Otherwise just 
   * insert the edge (v,nv) in S.
   */
  template <class Graph>
  typename Graph::edge_t insertBefore(const typename Graph::vertex_t& v,
                                      const typename Graph::vertex_t& ref,
                                      const typename Graph::vertex_t& nv,
                                      Graph& S)
  {
    if(ref)
      return S.spliceBefore(v, ref, nv);
    return S.insertEdge(v, nv);
  }

}

#endif // VVELIB_ALGORITHMS_INSERT_H

