#ifndef VVELIB_ALGORITHMS_PARALLEL_H
#define VVELIB_ALGORITHMS_PARALLEL_H

/**
 * \file parallel.h
 *
 * Defines the help class for parallel execution
 */

#include <config.h>
#include <cstdlib>

/**
 * \namespace parallel
 *
 * Define a thread pool for multi-threading computation
 */
namespace parallel
{

  /**
   * \class ThreadEval parallel.h <algorithms/parallel.h>
   * Base class for function evaluated in a thread
   */
  struct ThreadEval
  {
    virtual ~ThreadEval() {}
    virtual void operator()() const = 0;
  };

  class ThreadPoolImpl;

  /**
   * \class ParallelControler parallel.h <algorithms/parallel.h>
   *
   * This class is used to control how many threads are used for parallel 
   * processing.
   *
   * It also detect automatically how many processors are available on the 
   * machine.
   */
  struct ThreadPool
  {
    ThreadPool();
    ~ThreadPool();

    int nbProcessors();
    int nbThreads();
    void setNbThreads(int n);
    void setNbThreadsToDefault();

    /**
     * Evaluate \a fct on each available thread
     */
    void join();

    void clear();

    static ThreadPool& threadPool()
    {
      static ThreadPool* pool = 0;
      if(!pool)
      {
        pool = new ThreadPool();
        atexit(&ThreadPool::clearPool);
      }
      return *pool;
    }

    static void clearPool()
    {
      threadPool().clear();
    }

    template <typename Function>
    static void eval(const Function* fct)
    {
      ThreadPool& pool = threadPool();
      if(pool.nbThreads() == 0)
      {
        (*fct)();
        delete fct;
      }
      else
      {
        pool.thread_eval(fct);
      }
    }

  protected:
    void thread_eval( const ThreadEval* fct);

  private:
    ThreadPoolImpl *impl;
    int nbComputations;
  };

  inline void threadJoin()
  {
    ThreadPool& pool = ThreadPool::threadPool();
    pool.join();
  }

#define ONLY_ARG(F) typename F::argument_type
#define FIRST_ARG(F) typename F::first_argument_type
#define SECOND_ARG(F) typename F::second_argument_type
#define THIRD_ARG(F) typename F::third_argument_type
#define FOURTH_ARG(F) typename F::fourth_argument_type

  /**
   * Evaluate a function without arguments using the thread pool
   */
  template <typename Function>
  void threadEval(const Function& fct);

  template <typename Function>
  void threadEval(const Function& fct, ONLY_ARG(Function) a);

  template <typename Function>
  void threadEval(const Function& fct, FIRST_ARG(Function) a1, SECOND_ARG(Function) a2);

  template <typename Function>
  void threadEval(const Function& fct, FIRST_ARG(Function) a1, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3);

  template <typename Function>
  void threadEval(const Function& fct, FIRST_ARG(Function) a1, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3, FOURTH_ARG(Function) a4);

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, Container& container);

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, SECOND_ARG(Function) a2, Container& container);

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3, Container& container);

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3, FOURTH_ARG(Function) a4, Container& container);

  template <class Class>
  struct MemberFunction0
  {
    MemberFunction0( void (Class::*m)() )
      : mem(m)
    { }
    void operator()( Class& c ) const
    {
      (c.*mem)();
    }
    void (Class::*mem)();
  };

  template <class Class>
  MemberFunction0<Class> memberFunction(void (Class::*m)())
  { return MemberFunction0<Class>(m); }

  template <class Class, class Arg>
  struct MemberFunction1
  {
    typedef Class& first_argument_type;
    typedef Arg second_argument_type;
    typedef void result_type;
    MemberFunction1( void (Class::*m)(Arg) )
      : mem(m)
    { }
    void operator()( Class& c, Arg a ) const
    {
      (c.*mem)(a);
    }
    void (Class::*mem)(Arg);
  };

  template <class Class, typename Arg>
  MemberFunction1<Class,Arg> memberFunction(void (Class::*m)(Arg))
  { return MemberFunction1<Class,Arg>(m); }

  template <class Class, typename Arg1, typename Arg2>
  struct MemberFunction2
  {
    typedef Class& first_argument_type;
    typedef Arg1 second_argument_type;
    typedef Arg2 third_argument_type;
    typedef void result_type;
    MemberFunction2( void (Class::*m)(Arg1,Arg2) )
      : mem(m)
    { }
    void operator()( Class& c, Arg1 a1, Arg2 a2 ) const
    {
      (c.*mem)(a1,a2);
    }
    void (Class::*mem)(Arg1,Arg2);
  };

  template <class Class, typename Arg1, typename Arg2>
  MemberFunction2<Class,Arg1,Arg2> memberFunction(void (Class::*m)(Arg1,Arg2))
  { return MemberFunction2<Class,Arg1,Arg2>(m); }

  template <class Class, typename Arg1, typename Arg2, typename Arg3>
  struct MemberFunction3
  {
    typedef Class& first_argument_type;
    typedef Arg1 second_argument_type;
    typedef Arg2 third_argument_type;
    typedef Arg3 fourth_argument_type;
    typedef void result_type;
    MemberFunction3( void (Class::*m)(Arg1,Arg2,Arg3) )
      : mem(m)
    { }
    void operator()( Class& c, Arg1 a1, Arg2 a2, Arg3 a3 ) const
    {
      (c.*mem)(a1,a2,a3);
    }
    void (Class::*mem)(Arg1,Arg2,Arg3);
  };

  template <class Class, typename Arg1, typename Arg2, typename Arg3>
  MemberFunction3<Class,Arg1,Arg2,Arg3> memberFunction(void (Class::*m)(Arg1,Arg2,Arg3))
  { return MemberFunction3<Class,Arg1,Arg2,Arg3>(m); }

  template <class Arg>
  struct FreeFunction1
  {
    typedef Arg argument_type;
    typedef void result_type;
    FreeFunction1(void (*f)(Arg))
      : fct(f)
    { }
    void operator()( Arg a ) const
    {
      fct(a);
    }
    void (*fct)(Arg);
  };

  template <class Arg1>
  FreeFunction1<Arg1> freeFunction(void (*f)(Arg1))
  { return FreeFunction1<Arg1>(f); }

  template <class Arg1, class Arg2>
  struct FreeFunction2
  {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef void result_type;
    FreeFunction2(void (*f)(Arg1,Arg2))
      : fct(f)
    { }
    void operator()( Arg1 a1, Arg2 a2 ) const
    {
      fct(a1, a2);
    }
    void (*fct)(Arg1,Arg2);
  };

  template <class Arg1, class Arg2>
  FreeFunction2<Arg1,Arg2> freeFunction(void (*f)(Arg1,Arg2))
  { return FreeFunction2<Arg1,Arg2>(f); }

  template <class Arg1, class Arg2, class Arg3>
  struct FreeFunction3
  {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Arg3 third_argument_type;
    typedef void result_type;
    FreeFunction3(void (*f)(Arg1,Arg2,Arg3))
      : fct(f)
    { }
    void operator()( Arg1 a1, Arg2 a2, Arg3 a3 ) const
    {
      fct(a1, a2, a3);
    }
    void (*fct)(Arg1,Arg2,Arg3);
  };

  template <class Arg1, class Arg2, class Arg3>
  FreeFunction3<Arg1,Arg2,Arg3> freeFunction(void (*f)(Arg1,Arg2,Arg3))
  { return FreeFunction3<Arg1,Arg2,Arg3>(f); }

  template <class Arg1, class Arg2, class Arg3, class Arg4>
  struct FreeFunction4
  {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Arg3 third_argument_type;
    typedef Arg4 fourth_argument_type;
    typedef void result_type;
    FreeFunction4(void (*f)(Arg1,Arg2,Arg3,Arg4))
      : fct(f)
    { }
    void operator()( Arg1 a1, Arg2 a2, Arg3 a3, Arg4 a4 ) const
    {
      fct(a1, a2, a3, a4);
    }
    void (*fct)(Arg1,Arg2,Arg3,Arg4);
  };

  template <class Arg1, class Arg2, class Arg3, class Arg4>
  FreeFunction4<Arg1,Arg2,Arg3,Arg4> freeFunction(void (*f)(Arg1,Arg2,Arg3,Arg4))
  { return FreeFunction4<Arg1,Arg2,Arg3,Arg4>(f); }

}

#include <algorithms/parallel_impl.h>

#undef FIRST_ARG
#undef SECOND_ARG
#undef THIRD_ARG
#undef FOURTH_ARG

#endif // VVELIB_ALGORITHMS_PARALLEL_H

