#ifndef VVELIB_ALGORITHMS_PARALLEL_IMPL_H
#define VVELIB_ALGORITHMS_PARALLEL_IMPL_H

#include <config.h>
#include <iterator>

/// \cond NoDoc

#include <vector>

namespace parallel
{

  template <typename Function>
  struct FunctionReduce0 : public ThreadEval
  {
    FunctionReduce0(const Function& f)
      : fct(f)
    { }

    void operator()() const
    {
      return fct();
    }

    const Function& fct;
  };

  template <typename Function>
  FunctionReduce0<Function>* functionReduce(const Function& f)
  {
    return new FunctionReduce0<Function>(f);
  }

  template <typename Function>
  struct FunctionReduce1 : public ThreadEval
  {
    typedef ONLY_ARG(Function) Arg;
    typedef void result_type;
    FunctionReduce1(const Function& f, Arg a)
      : fct(f)
      , arg(a)
    { }

    void operator()() const
    {
      return fct(arg);
    }

    const Function& fct;
    Arg arg;
  };

  template <typename Function>
  FunctionReduce1<Function>* functionReduce(const Function& f, ONLY_ARG(Function) arg)
  {
    return new FunctionReduce1<Function>(f, arg);
  }

  template <typename Function>
  struct FunctionReduce2 : public ThreadEval
  {
    typedef FIRST_ARG(Function) Arg1;
    typedef SECOND_ARG(Function) Arg2;
    typedef void result_type;
    FunctionReduce2(const Function& f, Arg1 a1, Arg2 a2)
      : fct(f)
      , arg1(a1)
      , arg2(a2)
    { }

    void operator()() const
    {
      return fct(arg1, arg2);
    }

    const Function& fct;
    Arg1 arg1;
    Arg2 arg2;
  };

  template <typename Function>
  FunctionReduce2<Function>* functionReduce(const Function& f, FIRST_ARG(Function) arg1, SECOND_ARG(Function) arg2)
  {
    return new FunctionReduce2<Function>(f, arg1, arg2);
  }

  template <typename Function>
  struct FunctionReduce3 : public ThreadEval
  {
    typedef FIRST_ARG(Function) Arg1;
    typedef SECOND_ARG(Function) Arg2;
    typedef THIRD_ARG(Function) Arg3;
    typedef void result_type;
    FunctionReduce3(const Function& f, Arg1 a1, Arg2 a2, Arg3 a3)
      : fct(f)
      , arg1(a1)
      , arg2(a2)
      , arg3(a3)
    { }

    void operator()() const
    {
      return fct(arg1, arg2, arg3);
    }

    const Function& fct;
    Arg1 arg1;
    Arg2 arg2;
    Arg3 arg3;
  };

  template <typename Function>
  FunctionReduce3<Function>* functionReduce(const Function& f, FIRST_ARG(Function) arg1, SECOND_ARG(Function) arg2, THIRD_ARG(Function) arg3)
  {
    return new FunctionReduce3<Function>(f, arg1, arg2, arg3);
  }

  template <typename Function>
  struct FunctionReduce4 : public ThreadEval
  {
    typedef FIRST_ARG(Function) Arg1;
    typedef SECOND_ARG(Function) Arg2;
    typedef THIRD_ARG(Function) Arg3;
    typedef FOURTH_ARG(Function) Arg4;
    typedef void result_type;
    FunctionReduce4(const Function& f, Arg1 a1, Arg2 a2, Arg3 a3, Arg4 a4)
      : fct(f)
      , arg1(a1)
      , arg2(a2)
      , arg3(a3)
      , arg4(a4)
    { }

    void operator()() const
    {
      return fct(arg1, arg2, arg3, arg4);
    }

    const Function& fct;
    Arg1 arg1;
    Arg2 arg2;
    Arg3 arg3;
    Arg4 arg4;
  };

  template <typename Function>
  FunctionReduce4<Function>* functionReduce(const Function& f, FIRST_ARG(Function) arg1, SECOND_ARG(Function) arg2, THIRD_ARG(Function) arg3, FOURTH_ARG(Function) arg4)
  {
    return new FunctionReduce4<Function>(f, arg1, arg2, arg3, arg4);
  }

  template <typename Function>
  void threadEval(const Function& fct)
  {
    ThreadPool::eval(functionReduce(fct));
  }

  template <typename Function>
  void threadEval(const Function& fct, ONLY_ARG(Function) a)
  {
    ThreadPool::eval(functionReduce(fct, a));
  }

  template <typename Function>
  void threadEval(const Function& fct, FIRST_ARG(Function) a1, SECOND_ARG(Function) a2)
  {
    ThreadPool::eval(functionReduce(fct, a1, a2));
  }

  template <typename Function>
  void threadEval(const Function& fct, FIRST_ARG(Function) a1, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3)
  {
    ThreadPool::eval(functionReduce(fct, a1, a2, a3));
  }

  template <typename Function>
  void threadEval(const Function& fct, FIRST_ARG(Function) a1, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3, FOURTH_ARG(Function) a4)
  {
    ThreadPool::eval(functionReduce(fct, a1, a2, a3, a4));
  }

//  template <typename Function, typename Container>
//  struct ThreadTransform : public ThreadEval
//  {
//    ThreadTransform(const Function& f)
//      : fct(f)
//    { }

//    void operator()()
//    {
//      for(typename Container::iterator it = container.begin() ;
//          it != container.end() ; ++it)
//      {
//        f(*it);
//      }
//    }

//    const Function& f;
//    Container& container;
//  };

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, Container& container)
//  {
//    return new ThreadTransform(fct, container);
//  }

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, SECOND_ARG(Function) a2, Container& container)
//  {
//  }

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3, Container& container)
//  {
//  }

//  template <typename Function, typename Container>
//  void threadTransform(const Function& fct, SECOND_ARG(Function) a2, THIRD_ARG(Function) a3, FOURTH_ARG(Function) a4, Container& container)
//  {
//  }

}

/// \endcond

#endif // VVELIB_ALGORITHMS_PARALLEL_IMPL_H

