#ifndef VVELIB_ALGORITHMS_SOLVER_H
#define VVELIB_ALGORITHMS_SOLVER_H

/**
 * \file solver.h
 * Defines the solver namespace
 */

#include <config.h>

#ifdef WIN32
#  include <float.h>
#  define isnan _isnan
#  define finite _finite
#else
#  include <cmath>
#  ifndef GXX_USE_GNUXX0X
using std::isnan;
#  endif
#endif

#include <iostream>
#include <string>
#include <util/vector.h>
#include <util/matrix.h>
#include <algorithms/parallel.h>

/**
 * Implement a generic solver of ODE on a graph
 *
 * \section GenInfo General information
 *
 * This module implements a set of ODE solvers on a graph, from forward Euler 
 * to Cranck-Nicholson. It assumes a symmetric graph (i.e. for each vertex \c v1 
 * and \c v2, if there is an edge from \c v1 to \c v2, there must be an edge 
 * from \c v2 to \c v1).
 *
 * \section Usage Module Usage
 *
 * This module is used by constructing a solver::Solver object which holds the 
 * parameters for solving and the time stepping.
 *
 * Most solvers requires an internal structure to be placed in the content of 
 * your vertexes and edge.
 *
 * For example, to solve a 3-components system (i.e. 3 component on each vertex 
 * of the graph):
 * \code
 * typedef solver::Solver<3> RDSolver;
 *
 * struct VertexContent
 * {
 * // ...
 * RDSolver::VertexInternals interns;
 * };
 *
 * struct EdgeContent
 * {
 * // ...
 * RDSolver::EdgeInternals interns;
 * };
 *
 * class MyModel : public Model
 * {
 * public:
 *   RDSolver solve;
 *   typedef RDSolver::tag_t tag_t;
 *   vvgraph S;
 *
 *   void step()
 *   {
 *   solve(S, *this);
 *   }
 * };
 * \endcode
 *
 * To use this module, the user needs to define some methods in its model 
 * class. Each class accept as argument a tag to differentiate between 
 * different solvers used in the same model. If you have different solvers with 
 * the same template arguments, then you can add an optional argument to 
 * differentiate them:
 * \code
 * struct reaction_diffusion_solver {};
 * struct mass_spring_solver {};
 * typedef solver::Solver<2,reaction_diffusion_solver> RDSolver;
 * typedef solver::Solver<2,mass_spring_solver> MSSolver;
 *
 * struct VertexContent
 * {
 * // ...
 * RDSolver::VertexInternals rd_interns;
 * MSSolver::VertexInternals ms_interns;
 * };
 *
 * struct EdgeContent
 * {
 * // ...
 * RDSolver::EdgeInternals rd_interns;
 * MSSolver::EdgeInternals ms_interns;
 * };
 *
 * class MyModel : public Model
 * {
 *   RDSolver rd_solve;
 *   typedef RDSolver::tag_t rd_tag_t;
 *
 *   MSSolver ms_solve;
 *   typeder MSSolver::tag_t ms_tag_t;
 * };
 * \endcode
 *
 * Here are the methods to define in your model:
 * \code
 * util::Vector<nb_chemicals,double>& values(const vertex& v, const tag_t& tag);
 * \endcode
 * Returns a reference on the vector containing the values for the different 
 * components of the ODE on the vertex \c v.
 *
 * \code
 * util::Vector<nb_vars,double>& derivatives(const vertex& v, const tag_t& tag);
 * \endcode
 * Returns a reference on the vector containing the derivatives for the 
 * different components of the ODE on the vertex \c v.
 *
 * \code
 * void updateDerivatives(const vertex& v, const tag_t& tag);
 * \endcode
 * Update the time derivatives of the components on the vertex \c v. You must 
 * not change any data outside the vertex \c v. Plus, you should not assume any 
 * order in the evaluation of this function. You might even assume all the 
 * updateDerivatives may be evaluated in parallel (I hope it will eventually by 
 * the case).
 *
 * \code
 * Solver::VertexInternals& vertexInternals(const vertex& v, const tag_t& tag);
 * \endcode
 * This method is not used for the forward Euler method. For all other you have 
 * to implement it. It means also that your vertex has to contain an instance \c 
 * Solver::VertexInternals. This method should return a reference on that 
 * structure.
 *
 * \code
 * Solver::EdgeInternals& edgeInternals(const vertex& src, const vertex& tgt, graph& S, const tag_t& tag);
 * \endcode
 * This method is only used in the Cranck-Nicholson solver, so if you do not 
 * intent on using it, you don't need edge internals. Note however that if you 
 * let the user choose, you have to implement ALL the methods, even if you never 
 * use them. It should return a reference on the edgeInternals structure for the 
 * solver.
 */
namespace solver
{
#define VERTEX graph::Vertex<VertexContent>

  using std::endl;
  using std::cout;
  using std::cerr;

  using parallel::memberFunction;
  using parallel::freeFunction;
  using parallel::threadEval;
  using parallel::threadJoin;

#ifdef __GNUC__
  double max(double a, double b) { return (a>b)?a:b; }
  double min(double a, double b) { return (a<b)?a:b; }
#endif

  struct default_id_t {};

  template <size_t nb_vars, typename identifier>
    struct tag_holder {};

  template <size_t nb_vars>
    struct _VertexInternals
    {
      typedef util::Vector<nb_vars, double> Vec;
      typedef util::Matrix<nb_vars,nb_vars,double> Mat;

      Vec pDerivs;
      Vec ppDerivs;
      Vec pVars; // previous values
      Vec mVars; // previous values
      Vec fVars; // first step values
      Vec k1, k2, k3, k4;

      // Temps for matrix solvers
      Mat AA;
      Mat AT;
      Vec b;

      // Temps for bi-conjugate gradient solver
      Vec x;                                    // Correction vector
      Vec x0;                                   // Back up of data vars
      Vec t0;                                   // Vars at t0
      Vec r1;                                   // Residual vectors
      Vec r2;
      Vec p1;                                   // Basis vectors
      Vec p2;
      Vec AAp1;                                 // Temp from matrix mult
      Vec ATp2;                                 // Temp from matrix mult
    };

  template <size_t nb_vars>
  struct _EdgeInternals
  {
    typedef util::Matrix<nb_vars,nb_vars,double> Mat;

    Mat AA;
    Mat AT;
  };

  /**
   * Available solving methods.
   *
   * There is stream operator defined to read and write the methods with the exact same name
   */
  enum SolvingMethod
  {
    /// Forward euler method
    Euler,
    /// Adaptive forward euler method
    AdaptiveEuler,
    /// Fixed point method (i.e. find the solution for the derivatives = 0)
    Fixedpoint,
    /// Mid-point method (i.e. order 2 Runge-Kutta)
    Midpoint,
    /// Order 4 Runge-Kutta
    RungeKutta,
    /// Adaptative order 4 Runge-Kutta
    AdaptiveRungeKutta,
    /// Adaptative Crank-Nicholson
    AdaptiveCrankNicholson
  };

  inline QTextStream& operator<<(QTextStream& s, const SolvingMethod& m)
  {
    switch(m)
    {
      case Euler:
        s << "Euler";
        break;
      case AdaptiveEuler:
        s << "AdaptiveEuler";
        break;
      case Fixedpoint:
        s << "Fixedpoint";
        break;
      case Midpoint:
        s << "Midpoint";
        break;
      case RungeKutta:
        s << "RungeKutta";
        break;
      case AdaptiveRungeKutta:
        s << "AdaptiveRungeKutta";
        break;
      case AdaptiveCrankNicholson:
        s << "AdaptiveCrankNicholson";
        break;
    }
    return s;
  }

 inline  QTextStream& operator>>(QTextStream& s, SolvingMethod& m)
  {
    QString name;
    s >> name;
    if( name == "Euler" )
      m = Euler;
    else if( name == "AdaptiveEuler" )
      m = AdaptiveEuler;
    else if( name == "Fixedpoint" )
      m = Fixedpoint;
    else if( name == "Midpoint" )
      m = Midpoint;
    else if( name == "RungeKutta" )
      m = RungeKutta;
    else if( name == "AdaptiveRungeKutta" )
      m = AdaptiveRungeKutta;
    else if( name == "AdaptiveCrankNicholson" )
      m = AdaptiveCrankNicholson;
    else
      m = Euler;
    return s;
  }


 inline  std::ostream& operator<<(std::ostream& s, const SolvingMethod& m)
  {
    switch(m)
    {
      case Euler:
        s << "Euler";
        break;
      case AdaptiveEuler:
        s << "AdaptiveEuler";
        break;
      case Fixedpoint:
        s << "Fixedpoint";
        break;
      case Midpoint:
        s << "Midpoint";
        break;
      case RungeKutta:
        s << "RungeKutta";
        break;
      case AdaptiveRungeKutta:
        s << "AdaptiveRungeKutta";
        break;
      case AdaptiveCrankNicholson:
        s << "AdaptiveCrankNicholson";
        break;
    }
    return s;
  }

  inline std::istream& operator>>(std::istream& s, SolvingMethod& m)
  {
    std::string name;
    s >> name;
    if( name == "Euler" )
      m = Euler;
    else if( name == "AdaptiveEuler" )
      m = AdaptiveEuler;
    else if( name == "Fixedpoint" )
      m = Fixedpoint;
    else if( name == "Midpoint" )
      m = Midpoint;
    else if( name == "RungeKutta" )
      m = RungeKutta;
    else if( name == "AdaptiveRungeKutta" )
      m = AdaptiveRungeKutta;
    else if( name == "AdaptiveCrankNicholson" )
      m = AdaptiveCrankNicholson;
    else
      m = Euler;
    return s;
  }

  /**
   * Type of the tolerances
   *
   * A stream operator define the read/write methods
   */
  enum ToleranceType
  {
    /// Extract the maximum value to compute the precision (name: MaxComponent)
    MAX_COMPONENT,
    /// Extract the mean of the values to compute the precision (name: MeanComponent)
    MEAN_COMPONENT
  };

  inline std::ostream& operator<<(std::ostream& s, const ToleranceType& m)
  {
    switch(m)
    {
      case MAX_COMPONENT:
        s << "MaxComponent";
        break;
      case MEAN_COMPONENT:
        s << "MeanComponent";
        break;
    }
    return s;
  }

  inline std::istream& operator>>(std::istream& s, ToleranceType& m)
  {
    std::string name;
    s >> name;
    if( name == "MaxComponent" )
      m = MAX_COMPONENT;
    else if( name == "MeanComponent" )
      m = MEAN_COMPONENT;
    else
      m = MAX_COMPONENT;
    return s;
  }

  /**
   * \class Solver solver.h <algorithms/solver.h>
   *
   * Implement a set of solvers for ODE on a %graph.
   *
   * The %solver class implements 7 ODE %algorithms:
   *  -# Forward Euler
   *  -# Adaptative forward Euler
   *  -# Fixed point
   *  -# Mid point
   *  -# Runge-Kutta order 4
   *  -# Adaptative Runge-Kutta order 4
   *  -# Adaptative Crank-Nicholson
   */
  template <size_t nb_vars, typename identifier = default_id_t>
  struct Solver
  {
    /// Type of a vector used to hold the data
    typedef util::Vector<nb_vars, double> Vec;
    /// Type of a matrix used to store interaction on the edges
    typedef util::Matrix<nb_vars,nb_vars,double> Mat;
    /// Type of the internals stored on the vertices for some algorithms
    typedef _VertexInternals<nb_vars> VertexInternals;
    /// Type of the internals stored on the edges for some algorithms
    typedef _EdgeInternals<nb_vars> EdgeInternals;

    /// Tag identifying this specific solver type
    typedef tag_holder<nb_vars,identifier> tag_t;

    //@{
    ///name Parameters

    /// Method read in the parameter file
    SolvingMethod method;
    /// Method currently used (i.e. it is set only after initialization)
    SolvingMethod current_method;
    /// Current dt
    double dt;
    /// Current computation step
    int step;

    /// Level of debuging output
    int PrintStats;

    //@{
    ///\name Different Dt

    /// Timestep for Forward Euler solver
    double EulerDt;
    /// Timestep for FixedPoint iteration solver
    double FixedPointDt;
    /// Timestep for MultiPoint iteration solver
    double MidPointDt;
    /// Timestep for Runge-Kutta solver
    double RungeKuttaDt;
    /// Timestep for adaptive solvers
    double InitialDt;
    /// Minimum Dt for adapative solvers
    double MinDt;
    /// Maximum Dt for adapative solvers
    double MaxDt;
    //@}

    //@{
    ///\name Parameters for Adaptative Euler

    /// Type of the tolerance
    ToleranceType AEulerTolType;
    /// Dt increment
    double AEulerIncDt;
    /// Restart Dt decrement
    double AEulerResDt;
    /// restart tolerance
    double AEulerResTol;
    /// low water mark
    double AEulerLowTol;
    /// high water mark
    double AEulerHighTol;
    //@}

    //@{
    ///\name Fixed Point iteration parms

    /// Max steps for fixed point iteration
    double FixedPointMaxSteps;
    /// Tolerance for fixed point iteration
    double FixedPointTol;
    /// Type of tolerance
    ToleranceType FixedPointTolType;
    //@}

    //@{
    ///\name Adaptive Runge-Kutta parms

    /// Dt increment/decrement
    double ARungeIncDt;
    /// restart Dt decrement
    double ARungeResDt;
    /// Tolerence type
    ToleranceType ARungeTolType;
    /// restart tolerance
    double ARungeResTol;
    /// low water mark
    double ARungeLowTol;
    /// high water mark
    double ARungeHighTol;
    //@}

    //@{
    ///\name Crank Nicholson params

    /// Increment for timestep based on efficiency (mulitple of current size)
    double CRIncDt;
    /// Decrement for timestep if unsucc
    double CRResDt;
    /// Weight of current dt in eff average
    double CRAvgCPU;
    /// Minimun CPU, below this alway increases timestep
    double CRMinCPU;
    //@}

    //@{
    ///\name Newton Tolerance (used by Crank Nicholson)

    /// Tolerence for Newton's method
    double NewtTol;
    /// Tolerance type
    ToleranceType NewtTolType;
    /// Max steps for Newton's method
    int NewtMaxSteps;
    //@}

    //@{
    ///\name Conjugate gradient params (used by Crank Nicholson)

    /// Tolerence for conjugate gradient
    double ConjGradTol;
    /// Tolerance type
    ToleranceType ConjGradTolType;
    /// Max steps for conjugate gradient (multiple of N)
    double ConjGradMaxSteps;
    /// Set if partials to neighbors are constant (diffusion only)
    bool ConstNbPartials;
    //@}

    //@{
    ///\name Misc general parms

    /// Delta for numerical diff
    double Dx;
    /// Print Matrix (Conj-Grad)
    bool PrintMatrix;
    //@}

    //@{
    ///\name Previous efficiency and dt for Cranck-Nicholson

    /// Previous efficiency
    double peff;
    /// Previous dt
    double pdt;
    //@}
    //@}

    /// True if the solver is initialized
    bool initialized;

    /// Default constructor
    Solver()
      : initialized(false)
      {}

    /// Constructor reading the parameters instantly
    Solver( util::Parms& parms, const QString& section, bool selectAlgorithm = true )
      : initialized(false)
      {
        readParms(parms, section, selectAlgorithm);
        initialize();
      }

    /**
     * Read the parameters needed for the different algorithms,
     * \param parms Parameter object pointing to the parameter file
     * \param section Section including the parameters for the solver
     * \param selectAlgorithm If true, the algorithm will be read from the 
     * parameter file.
     */
    void readParms( util::Parms& parms, const QString& section, bool selectAlgorithm = true )
    {
      if(selectAlgorithm)
      {
        parms(section, "Solver", method);
        cout << "Solver selected: " << method << endl;
      }

      // Level of debuging output
      parms(section, "PrintStats", PrintStats);

      // Timesteps for the different algorithms
      parms(section, "EulerDt", EulerDt);
      parms(section, "FixedPointDt", FixedPointDt);
      parms(section, "MidPointDt", MidPointDt);
      parms(section, "RungeKuttaDt", RungeKuttaDt);
      parms(section, "InitialDt", InitialDt);
      parms(section, "MaxDt", MaxDt);

      // Parameters for Adaptative Euler
      parms(section, "AEulerTolType", AEulerTolType);
      parms(section, "AEulerIncDt", AEulerIncDt);
      parms(section, "AEulerResDt", AEulerResDt);
      parms(section, "AEulerResTol", AEulerResTol);
      parms(section, "AEulerLowTol", AEulerLowTol);
      parms(section, "AEulerHighTol", AEulerHighTol);

      // Fixed Point iteration parms
      parms(section, "FixedPointMaxSteps", FixedPointMaxSteps);
      parms(section, "FixedPointTol", FixedPointTol);
      parms(section, "FixedPointTolType", FixedPointTolType);

      // Adaptive Runge-Kutta parms
      parms(section, "ARungeIncDt", ARungeIncDt);
      parms(section, "ARungeResDt", ARungeResDt);
      parms(section, "ARungeTolType", ARungeTolType);
      parms(section, "ARungeResTol", ARungeResTol);
      parms(section, "ARungeLowTol", ARungeLowTol);
      parms(section, "ARungeHighTol", ARungeHighTol);

      // Crank Nicholson params
      parms(section, "CRIncDt", CRIncDt);
      parms(section, "CRResDt", CRResDt);
      parms(section, "CRAvgCPU", CRAvgCPU);
      parms(section, "CRMinCPU", CRMinCPU);

      // Newton Tolerance (used by Crank Nicholson)
      parms(section, "NewtTol", NewtTol);
      parms(section, "NewtTolType", NewtTolType);
      parms(section, "NewtMaxSteps", NewtMaxSteps);

      // Conjugate gradient params (used by Crank Nicholson)
      parms(section, "ConjGradTol", ConjGradTol);
      parms(section, "ConjGradTolType", ConjGradTolType);
      parms(section, "ConjGradMaxSteps", ConjGradMaxSteps);
      parms(section, "ConstNbPartials", ConstNbPartials);

      // Misc general parms
      parms(section, "Dx", Dx);
      parms(section, "PrintMatrix", PrintMatrix);

      initialize(false);
    }

    /**
     * Initialize the solver.
     *
     * Call this method after changing the solver type.
     */
    void initialize(bool update_dt = true)
    {
      if(update_dt or !initialized)
      {
        peff = 0;
        pdt = 0;
        step = 0;
        dt = 0;
      }

      current_method = method;
      switch(method)
      {
        case Euler:
          dt = EulerDt;
          break;
        case AdaptiveEuler:
          if(update_dt)
            dt = EulerDt;
          break;
        case Fixedpoint:
          dt = FixedPointDt;
          break;
        case Midpoint:
          dt = MidPointDt;
          break;
        case RungeKutta:
          dt = RungeKuttaDt;
          break;
        case AdaptiveRungeKutta:
        case AdaptiveCrankNicholson:
          if(update_dt or !initialized)
            dt = InitialDt;
          break;
      }
      initialized = true;
    }

    /**
     * Invoke the solver, forcing the use of the forward euler method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveEuler(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      typedef graph::VVGraph<VertexContent,EdgeContent,compact> vvgraph;
      forall(const vertex& v, S)
      {
        model.updateDerivatives(v, tag_t());
      }
      forall(const vertex& v, S)
      {
        Vec& vars = model.values(v, tag_t());
        const Vec& derivs = model.derivatives(v, tag_t());
        vars += dt * derivs;
      }
      step++;
    }

    /**
     * Invoke the solver, forcing the use of the adaptative forward euler 
     * method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveAdaptiveEuler(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      double err = 0;
      forall(const vertex& v, S)
      {
        model.updateDerivatives(v, tag_t());
      }
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec& vars = model.values(v, tag_t());
        const Vec& derivs = model.derivatives(v, tag_t());
        double derr = norm(derivs - interns.pDerivs);
        switch(AEulerTolType)
        {
          case MEAN_COMPONENT:
            err += derr;
            break;
          case MAX_COMPONENT:
          default:
            if(err < derr)
              err = derr;
        }
        vars += dt * derivs;
        interns.ppDerivs = interns.pDerivs;
        interns.pDerivs = derivs;
      }
      if(AEulerTolType == MEAN_COMPONENT)
        err /= S.size();

      if(step)
      {
        if(err > AEulerResTol && dt > MinDt)
        { // Backup a step and restart
          forall( vertex v, S )
          {
            VertexInternals& interns = model.vertexInternals(v, tag_t());
            Vec& vars = model.values(v, tag_t());
            const Vec& derivs = model.derivatives(v, tag_t());
            vars -= dt * (1.0 - AEulerResDt) * derivs;
            interns.pDerivs = interns.ppDerivs;
          }
          dt *= AEulerResDt;
          if(PrintStats >= 1)
            cout << "Euler restart Err:" << err << " Timestep:" << dt << endl;
        }
        else if(err > AEulerHighTol) // Over high water, reduce step
          dt -= dt * AEulerIncDt;
        else if(err < AEulerLowTol)    // Under high water, increase step
          dt += dt * AEulerIncDt;

        // Clip to max/min
        dt = min(MaxDt, max(MinDt, dt));
      }
      step++;
    }

    /**
     * Invoke the solver, forcing the use of the fixed point method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveFixedpoint(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      // Find initial deltas
      forall(const vertex& v, S)
      {
        model.updateDerivatives(v, tag_t());
      }
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        // save original vars
        interns.pDerivs = model.derivatives(v, tag_t());
        Vec& vars = model.values(v, tag_t());
        interns.pVars = vars;
        // Find first estimate
        vars += dt * interns.pDerivs;
      }

      // Loop to improve approximation
      double fxerr = 0;
      int fxsteps = 0;
      while(true)
      {
        fxsteps++;
        fxerr = 0;
        forall(const vertex& v, S)
        {
          model.updateDerivatives(v, tag_t());
        }

        forall(const vertex& v, S)
        {
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          Vec derivs = model.derivatives(v, tag_t());
          Vec& vars = model.values(v, tag_t());
          vars = interns.pVars + dt * (derivs + interns.pDerivs)/2.0;
          double err = norm(derivs - interns.pDerivs);
          switch(FixedPointTolType)
          {
            case MEAN_COMPONENT:
              fxerr += err;
              break;
            case MAX_COMPONENT:
            default:
              if(err > fxerr)
                fxerr = err;
          }
          interns.pDerivs = derivs;
        }
        if(FixedPointTolType == MEAN_COMPONENT)
          fxerr /= S.size();

        if(fxerr < FixedPointTol || fxsteps >= FixedPointMaxSteps)
          break;
      }
      if(PrintStats >= 2)
        cout << "Fixed point steps:" << fxsteps << ", Err:" << fxerr << endl;
      step++;
    }

    /**
     * Invoke the solver, forcing the use of the mid point method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveMidpoint(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      // Save original data values and deltas at start of step - k1
      forall( vertex v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        interns.pVars = model.values(v, tag_t());
        model.updateDerivatives(v, tag_t());
        interns.k1 = model.derivatives(v, tag_t());
      }

      // Find deltas at trial midpoint - k2
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec &vars = model.values(v, tag_t());
        vars += dt * 0.5 * interns.k1;
      }

      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        model.updateDerivatives(v, tag_t());
        interns.k2 = model.derivatives(v, tag_t());
      }

      // Update concentrations based on midpoint deltas
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec &vars = model.values(v, tag_t());
        vars = interns.pVars + dt * interns.k2;
      }
      step++;
    }

    /**
     * Invoke the solver, forcing the use of the order 4 runge-kutta method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveRungeKutta(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      // Save original data values and deltas at start of step - k1
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        interns.pVars = model.values(v, tag_t());
        model.updateDerivatives(v, tag_t());
        interns.k1 = model.derivatives(v, tag_t());
      }

      // Find deltas at first trial midpoint - k2
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec &vars = model.values(v, tag_t());
        vars = interns.pVars + dt * .5 * interns.k1;
      }
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        model.updateDerivatives(v, tag_t());
        interns.k2 = model.derivatives(v, tag_t());
      }

      // Find deltas at second trial midpoint - k3
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec &vars = model.values(v, tag_t());
        vars = interns.pVars + dt * .5 * interns.k2;
      }
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        model.updateDerivatives(v, tag_t());
        interns.k3 = model.derivatives(v, tag_t());
      }

      // Find deltas at trial endpoint (based on second midpoint estimate) - k3
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec &vars = model.values(v, tag_t());
        vars = interns.pVars + dt * interns.k3;
      }
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        model.updateDerivatives(v, tag_t());
        interns.k4 = model.derivatives(v, tag_t());
      }

      // Update concentrations based on Runge-Kutta weighted average
      forall(const vertex& v, S)
      {
        VertexInternals& interns = model.vertexInternals(v, tag_t());
        Vec &vars = model.values(v, tag_t());
        vars = interns.pVars;
        vars += dt * ((interns.k1+interns.k4)/6.0 + (interns.k2 + interns.k3)/3.0);
      }
      step++;
    }

    /**
     * Invoke the solver, forcing the use of the adaptative order 4 runge-kutta 
     * method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveAdaptiveRungeKutta(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      double err = 0.0;
      for(int rstep = 1; rstep <= 3; rstep++)
      {
        double ldt = dt;
        if(rstep != 1) // 1st step at full dt
          ldt *= .5;

        // Save original data values and deltas at start of step - k1
        if(rstep != 2) // only compute k1 on 1st and 3rd steps
        {
          forall(const vertex& v, S)
          {
            VertexInternals& interns = model.vertexInternals(v, tag_t());
            if(rstep == 1) // 1st time through save original vars
              interns.pVars = model.values(v, tag_t());
            model.updateDerivatives(v, tag_t());
            interns.k1 = model.derivatives(v, tag_t());
          }
        }

        // Find deltas at first trial midpoint - k2
        forall(const vertex& v, S)
        {
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          Vec &vars = model.values(v, tag_t());
          if(rstep != 3) // Steps 1 & 2 start from beginning of timestep
            vars = interns.pVars;
          else
            vars = interns.mVars;
          vars += ldt * .5 * interns.k1;
        }
        forall(const vertex& v, S)
        {
          model.updateDerivatives(v, tag_t());
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          interns.k2 = model.derivatives(v, tag_t());
        }

        // Find deltas at second trial midpoint - k3
        forall(const vertex& v, S)
        {
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          Vec &vars = model.values(v, tag_t());
          if(rstep != 3) // Steps 1 & 2 start from beginning of timestep
            vars = interns.pVars;
          else
            vars = interns.mVars;
          vars += ldt * .5 * interns.k2;
        }
        forall(const vertex& v, S)
        {
          model.updateDerivatives(v, tag_t());
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          interns.k3 = model.derivatives(v, tag_t());
        }

        // Find deltas at trial endpoint (based on second midpoint estimate) - k3
        forall(const vertex& v, S)
        {
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          Vec &vars = model.values(v, tag_t());
          if(rstep != 3) // Steps 1 & 2 start from beginning of timestep
            vars = interns.pVars;
          else
            vars = interns.mVars;
          vars += ldt * interns.k3;
        }
        forall(const vertex& v, S)
        {
          model.updateDerivatives(v, tag_t());
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          interns.k4 = model.derivatives(v, tag_t());
        }

        // Update concentrations based on Runge-Kutta weighted average
        forall(const vertex& v, S)
        {
          VertexInternals& interns = model.vertexInternals(v, tag_t());
          Vec &vars = model.values(v, tag_t());
          if(rstep != 3) // Steps 1 & 2 update from beginning of timestep
            vars = interns.pVars;
          else
            vars = interns.mVars;

          vars += ldt * ((interns.k1 + interns.k4)/6.0 + (interns.k2 + interns.k3)/3.0);
          if(rstep == 1)
          {
            interns.fVars = vars;
          }
          else if(rstep == 2)
          {
            interns.mVars = vars;
          }
          else // Calc truncation err
          {
            double derr = norm(vars - interns.fVars)/norm(interns.fVars);
            switch(ARungeTolType)
            {
              case MEAN_COMPONENT:
                err += derr;
                break;
              case MAX_COMPONENT:
              default:
                if(derr > err)
                  err = derr;
            }
          }
        }
      }
      if(ARungeTolType == MEAN_COMPONENT)
        err /= S.size();

      // Check truncation error (diif of midpoints)
      if(step != 0)
      {
        if(err > ARungeResTol && dt > MinDt) // Backup a step and restart
        {
          forall(const vertex& v, S)
          {
            VertexInternals &interns = model.vertexInternals(v, tag_t());
            Vec &vars = model.values(v, tag_t());
            vars = interns.pVars;
          }
          dt *= ARungeResDt;
          if(PrintStats >= 1)
            cout << "Runge-Kutta restart Err:" << err << " Timestep:" << dt << endl;
        }
        else if(err > ARungeHighTol) // Over high water, reduce step
          dt -= dt * ARungeIncDt;
        else if(err < ARungeLowTol)    // Under high water, increase step
          dt += dt * ARungeIncDt;

        // Clip to max/min
        dt = min(MaxDt, max(MinDt, dt));
      }
      step++;
    }

    /**
     * Find the partial derivative by successive approximation on each 
     * neighbors
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void FindPartials(const VERTEX& v, graph::VVGraph<VertexContent,EdgeContent,compact>& S, double dt,
                      int newtstep, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      // Save old vars and deltas
      VertexInternals& interns = model.vertexInternals(v, tag_t());
      model.updateDerivatives(v, tag_t());
      Vec ovars = model.values(v, tag_t());
      Vec odels = model.derivatives(v, tag_t());

      // Find partials for current node
      for(size_t i = 0; i < nb_vars; i++)
      {
        Vec& vars = model.values(v, tag_t());
        vars = ovars;
        vars[i] += Dx;
        model.updateDerivatives(v, tag_t());
        Vec dels = model.derivatives(v, tag_t());
        Vec partials = (dels - odels)/Dx * dt/2.0;
        partials[i] -= 1.0;
        interns.AT[i] = partials;
      }
      interns.AA = transpose(interns.AT);
      model.values(v, tag_t()) = ovars;

      // If only a diffusive interaction with neighbors, do not re-calc partials (constant)
      if(newtstep == 0 || !ConstNbPartials)
      {
        // Find partials for neighbors
        forall(const vertex& n, S.neighbors(v))
        {
          // Save velocity and postition and deltas
          Vec ovars = model.values(n, tag_t());
          EdgeInternals &einterns = model.edgeInternals(v,n,S, tag_t());

          // Find partials for current node
          for(size_t i = 0; i < nb_vars; i++)
          {
            Vec &vars = model.values(n, tag_t());
            vars = ovars;
            vars[i] += Dx;
            model.updateDerivatives(v, tag_t());
            Vec dels = model.derivatives(v, tag_t());
            Vec partials = (dels - odels)/Dx * dt/2.0;
            einterns.AT[i] = partials;
          }
          model.edgeInternals(n,v,S,tag_t()).AA = transpose(einterns.AT);
          model.values(n, tag_t()) = ovars;
        }
      }
      model.derivatives(v, tag_t()) = odels;
    }


    /**
     * Invoke the solver, forcing the use of the adaptative Crank-Nicholson 
     * method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void solveAdaptiveCrankNicholson(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      if(!initialized)
      {
        cerr << "Error, trying to use uninitialized solver" << endl;
        return;
      }
      typedef graph::Vertex<VertexContent> vertex;
      int conjGradMaxSteps = int(ConjGradMaxSteps*S.size());
      int newthalf = (int)(NewtMaxSteps/2.0);
      int cghalf = (int)(conjGradMaxSteps/2.0);

      int cgsteps = 0;
      int newtsteps = 0;
      long cpu = 0;

      double rr = 0;
      newtsteps = 0;

      // Back up data values in case of restart
      forall(const vertex& v, S)
      {
        model.vertexInternals(v, tag_t()).x0 = model.values(v, tag_t());
      }

      // Newton iteration
      bool restart = false;
      while(true)
      {
        if(restart)
        {
          cgsteps = newtsteps = 0;
          cpu = 1;
          dt *= CRResDt;
          restart = false;
          // Restore data
          forall(const vertex& v, S)
          {
            model.values(v, tag_t()) = model.vertexInternals(v, tag_t()).x0;
          }
        }
        if(newtsteps == 0)
        {
          // Find t0 values
          forall(const vertex& v, S)
          {
            model.updateDerivatives(v, tag_t());
            VertexInternals &interns = model.vertexInternals(v, tag_t());
            interns.t0 = model.derivatives(v, tag_t()) * dt/2.0;
            interns.t0 += interns.x0;
          }
        }

        // Initialize matrix A and vector b in Ax=b
        forall(const vertex& v, S)
        {
          FindPartials(v, S, dt, newtsteps, model);
          VertexInternals &interns = model.vertexInternals(v, tag_t());
          interns.b = (model.values(v, tag_t())
                       - model.derivatives(v, tag_t())*dt/2.0
                       - interns.t0);
        }

        newtsteps++;

        // Calculate cpu, add penalty for too many steps to avoid restart
        long newtcpu = 1;
        int newtover = newtsteps - newthalf;
        if(newtover > 1)
        {
          newtcpu += newtover;
          if(PrintStats >= 3)
            cout << "Newton step:" << newtsteps << " penalized by:" << newtover << endl;
        }
        if(!ConstNbPartials)
          newtcpu *= nb_vars;
        cpu += newtcpu;

        // Print matrix if required
        if(PrintMatrix)
        {
          cout << "Bi-conjugate Gradient Matrix" << endl;
          //PrintMatrix();
        }

        // Find first residual: Ax - b (note x = 0)
        rr = 0;
        double orr = 0;
        forall(const vertex& v, S)
        {
          VertexInternals &interns = model.vertexInternals(v, tag_t());
          interns.x = 0.0;
          interns.r1 = interns.r2 = -interns.b;
          interns.p1 = -interns.r1;
          interns.p2 = -interns.r2;
          rr += interns.r2 * interns.r1;
        }

        // Print out initial residual
        if(PrintStats >= 3)
          cout << "Bi-conjugate gradient initial residual:" << fabs(rr)/S.size() << " step size:" << dt << endl;

        // Main conj-grad iteration loop
        cgsteps = 0;
        double cgerr = 0;
        while(true)
        {
          cgsteps++;

          // Calculate cpu, add penalty for too many steps to avoid restart
          int cgcpu = 1;
          int cgover = cgsteps - cghalf;
          if(cgover > 1)
          {
            cgcpu += cgover;
            if(PrintStats >= 3)
              cout << "Bi-conjugate gradient step:" << cgsteps << " penalized by:" << cgover << endl;
          }
          cpu += cgcpu;

          // Find AAp1, ATp2, and p2AAp1
          double p2AAp1 = 0;
          forall(const vertex& v, S)
          {
            VertexInternals &v_interns = model.vertexInternals(v, tag_t());
            v_interns.AAp1 = v_interns.AA * v_interns.p1;
            v_interns.ATp2 = v_interns.AT * v_interns.p2;
            forall(const vertex& n, S.neighbors(v))
            {
              VertexInternals &n_interns = model.vertexInternals(n, tag_t());
              EdgeInternals &e_interns = model.edgeInternals(v, n, S, tag_t());
              v_interns.AAp1 += e_interns.AA * n_interns.p1;
              v_interns.ATp2 += e_interns.AT * n_interns.p2;
            }
            p2AAp1 += v_interns.p2 * v_interns.AAp1;
          }

          // Find coefficients for basis vectors (alpha)
          double alpha = rr / p2AAp1;

          // Find next solution, residual, error, and new rr
          orr = rr;
          rr = 0;
          cgerr = 0;
          forall(const vertex& v, S)
          {
            VertexInternals &interns = model.vertexInternals(v, tag_t());
            // Improve solution vector
            interns.x += alpha * interns.p1;

            // Next residual
            interns.r1 += alpha * interns.AAp1;
            interns.r2 += alpha * interns.ATp2;
            double res = interns.r2 * interns.r1;
            rr += res;
            res = fabs(res);
            switch(ConjGradTolType)
            {
              case MEAN_COMPONENT:
                cgerr += res;
                break;
              case MAX_COMPONENT:
              default:
                if(res > cgerr)
                  cgerr = res;
            }
          }
          if(ConjGradTolType == MEAN_COMPONENT)
            cgerr /= S.size();

          // Calculate next vector
          double beta = rr/orr;
          forall(const vertex& v, S)
          {
            VertexInternals &interns = model.vertexInternals(v, tag_t());
            interns.p1 = beta * interns.p1 - interns.r1;
            interns.p2 = beta * interns.p2 - interns.r2;
          }

          // Print Stats
          if(PrintStats >= 3)
            cout << "Bi-conj grad step:" << cgsteps << " Residual:" << cgerr << endl;

          // Check if we are done
          if(cgerr < ConjGradTol)
            break;

          // Check if we need to restart
          if(cgsteps >= conjGradMaxSteps || isnan(cgerr) || !finite(cgerr))
          {
            if(PrintStats >= 1)
            {
              if(cgsteps >= conjGradMaxSteps)
                cout << "Restarting step:" << step << ", Conj-Grad too many iterations at stepsize:" << dt << endl;
              else
                cout << "Restarting step:" << step << ", Conj-Grad residual out of bounds" << endl;
            }
            restart = true;
            break;
          }
        }

        double newterr = 0.0;
        if(cgsteps > 0 && !restart)
        {
          // Update model values  
          forall(const vertex& v, S)
          {
            VertexInternals &interns = model.vertexInternals(v, tag_t());
            Vec &vars = model.values(v, tag_t());
            vars += interns.x;
            double corr = norm(interns.x);
            switch(NewtTolType)
            {
              case MEAN_COMPONENT:
                newterr += corr;
                break;
              case MAX_COMPONENT:
              default:
                if(corr > newterr)
                  newterr = corr;
            }
          }
          if(NewtTolType == MEAN_COMPONENT)
            newterr /= S.size();

          if(PrintStats >= 2)
          {
            cout << "Newton step:" << newtsteps << " Corr Size:" << newterr << " ";
            cout << "Bi-Conj steps:" << cgsteps << " Residual:" << cgerr << endl;
          }
          if(newterr < NewtTol && !restart)
            break;
        }

        if(newtsteps >= NewtMaxSteps || isnan(newterr) || !finite(newterr))
        {
          if(PrintStats >= 1)
          {
            if(newtsteps >= NewtMaxSteps)
              cout << "Restarting:" << step << ", Newton iterations exceeded at stepsize:" << dt << endl;
            else
              cout << "Restarting:" << step << ", Newton correction out of bounds" << endl;
          }
          restart = true;
        }
      }

      // Adaptive stepsize based on cpu use
      double eff = dt/double(cpu);
      double inc = dt * CRIncDt;
      if(cpu > CRMinCPU)
      {
        if(peff > eff)
        {
          if(dt >= pdt)
            inc *= -1;
        }
        else
        {
          if(dt < pdt)
            inc *= -1;
        }
      }

      peff = (peff * (1.0 - CRAvgCPU) + CRAvgCPU * eff);
      pdt = dt;

      if(PrintStats >= 2)
        cout << "Stepsize:" << dt << ", current increment:" << inc << endl;

      dt += inc;

      if(dt > MaxDt)
        dt = MaxDt;
    }

    /**
     * Invoke the solver, using the currently selected method
     */
    template <typename VertexContent, typename EdgeContent, bool compact, typename Model>
    void operator()(graph::VVGraph<VertexContent,EdgeContent,compact>& S, Model& model)
    {
      switch(current_method)
      {
        case Euler:
          solveEuler(S, model);
          break;
        case AdaptiveEuler:
          solveAdaptiveEuler(S, model);
          break;
        case Fixedpoint:
          solveFixedpoint(S, model);
          break;
        case Midpoint:
          solveMidpoint(S, model);
          break;
        case RungeKutta:
          solveRungeKutta(S, model);
          break;
        case AdaptiveRungeKutta:
          solveAdaptiveRungeKutta(S, model);
          break;
        case AdaptiveCrankNicholson:
          solveAdaptiveCrankNicholson(S, model);
          break;
      }
    }
  };

#undef VERTEX
};

#endif // VVELIB_ALGORITHMS_SOLVER_H

