/**
 * \file tissue.h
 *
 * Definition of the tissue algorithm
 */

#ifndef VVELIB_ALGORITHMS_TISSUE_H
#define VVELIB_ALGORITHMS_TISSUE_H

#include <config.h>
#include <algorithms/complex.h>
#include <geometry/geometry.h>

/**
 * \namespace tissue
 *
 * The tissue module handle cell division in a cell %tissue.
 *
 * Compared to the 2d cell complex module, this module provides:
 *  - Drawing function for the tissue
 *  - Division algorithms
 */

namespace tissue
{
  using util::norm;
  using util::normsq;
  using util::normalized;

  using vvcomplex::DivisionData;
  using vvcomplex::InModelDivisionParam;
  using vvcomplex::FindWallMin;
  using vvcomplex::FindOppositeWall;
  using vvcomplex::FindCenter;

#ifndef max
  using std::max;
#endif

  /**
   * Type of the 3D positions
   */
  using geometry::Point3d;

  /**
   * Enumeration of the provided division algorithms
   */
  enum CELL_DIVISION_ALGORITHM
  {
    CLOSEST_MID,
    SHORT_WALL,
    CLOSEST_WALL
  };

#ifdef JUST_FOR_DOXYGEN
  /**
   * Documentation alias for the type of a graph vertex
   */
  typedef Tissue::vertex vertex;
  /**
   * Documentation alias for the type of the cell graph
   */
  typedef Tissue::cell_graph cell_graph;
  /**
   * Documentation alias for the type of the %tissue graph
   */
  typedef Tissue::tissue_graph tissue_graph;
#endif

  /**
   * Parameters of the cell pinching algorithm
   */
  struct CellPinchingParams
  {
    /**
     * Default and initialization constructor
     */
    CellPinchingParams(double cp = 0, double cmp = 0)
      : cellPinch(cp)
      , cellMaxPinch(cmp)
    { }

    /**
     * Copy constructor
     */
    CellPinchingParams(const CellPinchingParams& copy)
      : cellPinch(copy.cellPinch)
      , cellMaxPinch(copy.cellMaxPinch)
    { }

    /**
     * Pinching ratio.
     */
    double cellPinch;
    /**
     * Maximum displacement due to the pinching.
     */
    double cellMaxPinch;
  };

  /**
   * Pinching algorithm.
   *
   * This function displace the position of the newly inserted vertexes to 
   * account for the pinching of the cell. The pinching is always in the 
   * direction of the newly formed wall. A vertex on the side of the %tissue 
   * won't be moved. Otherwise, the pinching is proportional to the distance to 
   * the closest extremity of the cut segment.
   *
   * \param[in] c Cell to be divided
   * \param[in] S %Tissue graph
   * \param[in,out] data Parameters for the division to be updated
   * \param[in] params Parameters for the pinching
   * \param[in] model Model using the algorithm
   */
  template <class Complex>
    void cellPinching(const typename Complex::cell& c,
                      Complex& T,
                      DivisionData<typename Complex::junction_content>& data,
                      const CellPinchingParams& params)
  {
    typedef typename Complex::cell cell;
    typedef typename Complex::junction junction;
    typename Complex::model_t& model = *T.model;
    // Calculate cell pinch
    double mind = 0.0;
    double pinch = 0.0;

    junction p_u1 = data.u1;
    junction p_u2 = T.S.nextTo(c, p_u1);

    junction p_v1 = data.v1;
    junction p_v2 = T.S.nextTo(c, p_v1);

    Point3d u1 = model.position(p_u1);
    Point3d u2 = model.position(p_u2);

    // If on edge, don't pinch
    if(!T.border(p_u1, p_u2))
    {
      Point3d uc = (data.pv - data.pu);
      double ucl = norm(uc);
      uc /= ucl;
      double uu1 = norm(data.pu-u1);
      double uu2 = norm(data.pu-u2);
      if(uu1 > uu2)
        mind = uu2;
      else
        mind = uu1;
      pinch = mind * ucl * params.cellPinch;
      if(pinch > params.cellMaxPinch)
        pinch = params.cellMaxPinch;
      data.pu += uc * pinch;
    }

    Point3d v1 = model.position(p_v1);
    Point3d v2 = model.position(p_v2);

    if(!T.border(p_v1, p_v2))
    {
      Point3d vc = (data.pu - data.pv);
      double vcl = norm(vc);
      vc /= vcl;
      double vv1 = norm(data.pv-v1);
      double vv2 = norm(data.pv-v2);
      if(vv1 > vv2)
        mind = vv2;
      else
        mind = vv1;
      pinch = mind * vcl * params.cellPinch;
      if(pinch > params.cellMaxPinch)
        pinch = params.cellMaxPinch;
      data.pv += vc * pinch;
    }
  }

  /**
   * Parameters for the shortest wall algorithm.
   */
  struct ShortWallAlgoParams : public CellPinchingParams
  {
    /**
     * Default and initialization constructor
     */
    ShortWallAlgoParams(double cwm = 0, bool scwm = true, double cws = 0.1, double d = 0.1)
      : CellPinchingParams()
      , cellWallMin(cwm)
      , strictCellWallMin(scwm)
      , cellWallSample(cws)
      , dx(d)
      { }

    /**
     * Copy constructor
     */
    ShortWallAlgoParams(const ShortWallAlgoParams& copy)
      : CellPinchingParams(copy)
      , cellWallMin(copy.cellWallMin)
      , strictCellWallMin(copy.strictCellWallMin)
      , cellWallSample(copy.cellWallSample)
      , dx(copy.dx)
      { }

    /**
     * Minimum size of a wall
     */
    double cellWallMin;
    /**
     * If true, the minimum size of a wall is strictly enforce, otherwise, it 
     * will not imply a change of division wall
     */
    bool strictCellWallMin;
    /**
     * Size of sampling to find the minimum wall size
     */
    double cellWallSample;
    /**
     * Precision required to consider a point in a segment
     */
    double dx;
  };

  /**
   * Parameters for the closest mid
   */
  struct ClosestMidAlgoParams : public CellPinchingParams
  {
    /**
     * Default and initialization constructor
     */
    ClosestMidAlgoParams(double cwm = 0, bool scwm = true)
      : CellPinchingParams()
      , cellWallMin(cwm)
      , strictCellWallMin(scwm)
      { }

    /**
     * Copy constructor
     */
    ClosestMidAlgoParams(const ClosestMidAlgoParams& copy)
      : CellPinchingParams(copy)
      , cellWallMin(copy.cellWallMin)
      , strictCellWallMin(copy.strictCellWallMin)
      { }

    /**
     * Minimum size of a wall
     */
    double cellWallMin;
    /**
     * If true, the minimum size of a wall is strictly enforce, otherwise, it 
     * will not imply a change of division wall
     */
    bool strictCellWallMin;
  };

  /**
   * Parameters for the closest wall algorithm
   */
  struct ClosestWallAlgoParams : public CellPinchingParams
  {
    /**
     * Default and initialization constructor
     */
    ClosestWallAlgoParams(double cwm = 0, bool scwm = true)
      : CellPinchingParams()
      , cellWallMin(cwm)
      , strictCellWallMin(scwm)
      { }

    /**
     * Copy constructor
     */
    ClosestWallAlgoParams(const ClosestWallAlgoParams& copy)
      : CellPinchingParams(copy)
      , cellWallMin(copy.cellWallMin)
      , strictCellWallMin(copy.strictCellWallMin)
      { }

    /**
     * Minimum size of a wall
     */
    double cellWallMin;
    /**
     * If true, the minimum size of a wall is strictly enforce, otherwise, it 
     * will not imply a change of division wall
     */
    bool strictCellWallMin;
  };

  /**
   * Implementation of the shortest wall algorithm
   *
   * This algorithm search for the shortest segment traversing the cell and 
   * going through the cell center.
   *
   * \param c Cell to divide
   * \param S %Tissue graph
   * \param params Parameters
   * \param model Model using the %tissue
   *
   * \returns The parameters for the division
   *
   * \see ShortWallAlgoParams
   */
  template <class Complex>
  DivisionData<typename Complex::junction_content> findDivisionPoints(const typename Complex::cell& c,
                                                                      Complex& T,
                                                                      const ShortWallAlgoParams& params)
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_MODEL(Complex, T);
    DivisionData<typename Complex::junction_content> result;
    // p_u1-u2, first edge, p_v1-v2 opposite edge
    Point3d minu, minv;

    // Find center (in case growth has changed it)
    //FindCenter(c, S, model);

    Point3d cpos = model.position(c);

    // Find shortest line through center
    double mindis = HUGE_VAL;
    Point3d n;
    n = model.normal(c);
    double epsilon = 0;
    forall(const junction& tu1, T.S.neighbors(c))
    {
      const junction& tu2 = T.S.nextTo(c, tu1);
      Point3d u1 = model.position(tu1);
      Point3d u2 = model.position(tu2);
      Point3d u1u2 = u2 - u1;
      double u1u2l = norm(u1u2);
      u1u2 /= u1u2l;
      epsilon = max(epsilon, VVE_REL_EPSILON*u1u2l);
      forall(const junction& tv1, T.S.neighbors(c))
      {
        if(tv1 == tu1)
        {
          continue;
        }
        const junction& tv2 = T.S.nextTo(c, tv1);
        const Point3d& v1 = model.position(tv1);
        const Point3d& v2 = model.position(tv2);

        double ule = u1u2l - params.cellWallMin;
        for(double ul = params.cellWallMin; ul < ule; ul += params.cellWallSample) 
        {
          Point3d u = u1 + ul * u1u2;
          Point3d uc = cpos - u;
          Point3d nu = n^uc;
          Point3d v;
          double s;
          if(geometry::planeLineIntersection(v, s, u, nu, v1, v2))
          {
            if(s >= -params.dx  && s <= (1.0 + params.dx))
            {
              if(FindWallMin(v, v1, v2, params.cellWallMin) or not params.strictCellWallMin)
              {
                double dis = norm(u-v);
                if(dis < mindis and (u-cpos)*(v-cpos) < 0)
                {
                  mindis = dis;
                  minu = u;
                  minv = v;
                  result.u1 = tu1;
                  result.v1 = tv1;
                }
              }
            }
          }
        } // Samples
      } // v1v2
    } // u1u2
    if(mindis == HUGE_VAL)
    {
      util::out << "findDivisionPoints::Error:Failed to find divide line" << endl;
      return DivisionData<typename Complex::junction_content>();
    }
    result.pu = minu;
    result.pv = minv;

    if(params.cellWallMin == 0)
      testDivisionOnVertices(c, result, T, epsilon);

    cellPinching(c, T, result, params);
    return result;
  }

  /**
   * Implementation of the closest mid algorithm
   *
   * This algorithm defining the new wall as the one starting from the mid wall 
   * the closest to the cell center and going through the cell center.
   *
   * \param c Cell to divide
   * \param S %Tissue graph
   * \param params Parameters
   * \param model Model using the %tissue
   *
   * \returns The parameters for the division
   *
   * \see ClosestMidAlgoParams
   */
  template <class Complex>
  DivisionData<typename Complex::junction_content> findDivisionPoints(const typename Complex::cell& c,
                                                                      Complex& T,
                                                                      const ClosestMidAlgoParams& params)
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_MODEL(Complex, T);
    DivisionData<typename Complex::junction_content> result;
    // p_u1-u2, first edge, p_v1-v2 opposite edge
    Point3d cpos = model.position(c);

    // Find center (in case growth has changed it)
    //FindCenter(c, S, model);

    // Find closest midpoint on wall
    double mindis = HUGE_VAL;
    forall(const junction& tu2, T.S.neighbors(c))
    {
      // other end of wall
      const junction& tu1 = T.S.prevTo(c, tu2);
      Point3d u1 = model.position(tu1);
      Point3d u2 = model.position(tu2);
      Point3d u = (u1 + u2)/2.0;
      double dis = norm(u-cpos);
      if(dis < mindis)
      {
        mindis = dis;
        result.pu = u;
        result.u1 = tu1;
      }
    }

    FindOppositeWall(c, result, T, params.cellWallMin, params.strictCellWallMin);

    cellPinching(c, T, result, params);
    return result;
  }

  /**
   * Implementation of the closest wall algorithm
   *
   * This algorithm defining the new wall as the one starting from the point on 
   * the walls the closest to the cell center and going through the cell center.
   *
   * \param c Cell to divide
   * \param S %Tissue graph
   * \param params Parameters
   * \param model Model using the %tissue
   *
   * \returns The parameters for the division
   *
   * \see ClosestWallAlgoParams
   */
  template <class Complex>
  DivisionData<typename Complex::junction_content> findDivisionPoints(const typename Complex::cell& c,
                                                                      Complex& T,
                                                                      const ClosestWallAlgoParams& params)
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    IMPORT_COMPLEX_MODEL(Complex, T);
    DivisionData<typename Complex::junction_content> result;
    // p_u1-u2, first edge, p_v1-v2 opposite edge
    Point3d cpos = model.position(c);

    // Find center (in case growth has changed it)
    //FindCenter(c, S, model);

    // Find closest point on wall
    double mindis = 1000000;
    double epsilon = 0;
    forall( const junction& tu2, T.S.neighbors(c))
    {
      const junction& tu1 = T.S.prevTo(c, tu2);
      Point3d u1 = model.position(tu1);
      Point3d u2 = model.position(tu2);
      Point3d u1u2 = u2 - u1;
      double u1u2l = norm(u1u2);
      u1u2 /= u1u2l;
      epsilon = max(epsilon, VVE_REL_EPSILON*u1u2l);
      Point3d u = u1 + ((cpos - u1) * u1u2) * u1u2;
      if(FindWallMin(u, u1, u2, params.cellWallMin) or not params.strictCellWallMin)
      {
        double dis = norm(u-cpos);
        if(dis < mindis)
        {
          mindis = dis;
          result.pu = u;
          result.u1 = tu1;
        }
      }
    }

    FindOppositeWall(c, result, T, params.cellWallMin, params.strictCellWallMin);

    if(params.cellWallMin == 0)
      testDivisionOnVertices(c, result, T, epsilon);

    cellPinching(c, T, result, params);
    return result;
  }

  /**
   * \class Tissue tissue.h <algorithms/tissue.h>
   *
   * Class handling the development and representation of a cell %tissue.
   *
   * \param VertexContent Data structure for the vertex content
   * \param TissueEdgeContent Data structure for the edge content of the tissue 
   * graph
   * \param CellEdgeContent Data structure for the edge content of the cell graph
   * \param Model Type of your model
   */
  template <typename Model,
           typename CellContent, typename JunctionContent,
           typename WallContent = graph::_EmptyEdgeContent, typename CellEdgeContent = graph::_EmptyEdgeContent,
           typename CellJunctionContent = graph::_EmptyEdgeContent, typename JunctionCellContent = graph::_EmptyEdgeContent,
           bool compact = false,
#ifdef USE_ALLOCATOR
           typename Alloc = DEFAULT_ALLOC(CellContent),
#endif
           typename LeafClass = template_utils::this_class
             >
  struct Tissue : public vvcomplex::VVComplex<MANDATORY_COMPLEX_TEMPLATE_ARGS,
                                              RESOLVE_LEAF_CLASS(LeafClass,Tissue<ALL_COMPLEX_TEMPLATE_ARGS>)>
  {
    typedef RESOLVE_LEAF_CLASS(LeafClass,Tissue) leaf_class;
    typedef vvcomplex::VVComplex<MANDATORY_COMPLEX_TEMPLATE_ARGS,leaf_class> VVComplex;
    IMPORT_COMPLEX_TYPES(VVComplex);
    typedef typename VVComplex::division_result_t division_result_t;
    typedef typename VVComplex::division_data division_data;

    typedef Model model_t;

    // Algorithm parameters
    /**
     * Default algorithm to use
     */
    CELL_DIVISION_ALGORITHM cellDivAlg;
    /**
     * Cell pinching ratio
     */
    double cellPinch;
    /**
     * Maximum cell pinching
     */
    double cellMaxPinch;
    /**
     * Minimum size of a cell wall
     */
    double cellWallMin;
    /**
     * If true, the cellWallMin property is strictly enforced. If false, it's 
     * losely inforce, meaning it will never lead to a change of cell wall.
     */
    bool strictCellWallMin;
    /**
     * Number of samples to find the shortest wall
     */
    double cellWallSample;
    /**
     * precision required to consider a point on a segment.
     */
    double sampleDx;

    // Drawing parameters
    /**
     * Palette to use for the colors
     */
    const util::Palette* palette;
    /**
     * Color of the high values for the cell
     */
    int colorBegin;
    /**
     * Color of the low values for the cell
     */
    int colorEnd;
    /**
     * Increase in color of the center (0 < colorCenter < 1)
     */
    double colorCenter;
    /**
     * Color of the contour of the cell
     */
    int contourColor;
    /**
     * Size of the cell corners
     */
    double cellWallCorner;
    /**
     * The inside of the cell is drawn depending on the value is true, 
     * otherwise the end color is used.
     */
    bool drawInsides;
    /**
     * Draw thick corners if true, otherwise draw \c cellWallWidth pixel 
     * borders
     */
    bool drawBorders;
    /**
     * Blending for the exterior of the cell
     *
     * (i.e. alpha value)
     */
    double blending;
    /**
     * Blending for the center of the cell
     *
     * (i.e. alpha value)
     */
    double center_blending;
    /**
     * Width of the "thin" cell wall
     */
    double cellWallWidth;

    /**
     * Constructor with just the model
     */
    Tissue( Model* mod )
      : VVComplex(mod)
      , cellDivAlg(SHORT_WALL)
      , cellPinch(0)
      , cellMaxPinch(0)
      , cellWallMin(0)
      , strictCellWallMin(true)
      , cellWallSample(100)
      , sampleDx(0.1)
      , palette(NULL)
      , colorBegin(0)
      , colorEnd(1)
      , colorCenter(0.2)
      , contourColor(3)
      , cellWallCorner(0.15)
      , drawInsides(true)
      , drawBorders(true)
      , blending(1)
      , center_blending(1)
      , cellWallWidth(0.2)
      { }

    /**
     * Constructor setting up both the palette and the model
     */
    Tissue( const util::Palette& pal, Model* mod )
      : VVComplex(mod)
      , cellDivAlg(SHORT_WALL)
      , cellPinch(0)
      , cellMaxPinch(0)
      , cellWallMin(0)
      , strictCellWallMin(true)
      , cellWallSample(100)
      , sampleDx(0.1)
      , palette(&pal)
      , colorBegin(0)
      , colorEnd(1)
      , colorCenter(0.2)
      , contourColor(3)
      , cellWallCorner(0.15)
      , drawInsides(true)
      , drawBorders(true)
      , blending(1)
      , center_blending(1)
      , cellWallWidth(0.2)
      { }

    /**
     * Copy constructor
     */
    Tissue( const Tissue& copy )
      : VVComplex(copy)
      , cellDivAlg(copy.cellDivAlg)
      , cellPinch(copy.cellPinch)
      , cellMaxPinch(copy.cellMaxPinch)
      , cellWallMin(copy.cellWallMin)
      , strictCellWallMin(copy.strictCellWallMin)
      , cellWallSample(copy.cellWallSample)
      , sampleDx(copy.sampleDx)
      , palette(copy.palette)
      , colorBegin(copy.colorBegin)
      , colorEnd(copy.colorEnd)
      , colorCenter(copy.colorCenter)
      , contourColor(copy.contourColor)
      , cellWallCorner(copy.cellWallCorner)
      , drawInsides(copy.drawInsides)
      , drawBorders(copy.drawBorders)
      , blending(copy.blending)
      , center_blending(copy.center_blending)
      , cellWallWidth(copy.cellWallWidth)
      { }

    /**
     * Read the parameters for the algorithmic part of the object.
     *
     * \param parms Parameter object to read from
     * \param section Section containing the parameters
     */
    void readParms( util::Parms& parms, const QString& section )
    {
      QString algo;
      parms(section, "DivisionAlgorithm", algo);
      algo = algo.toLower();
      if( algo == "closestmid" )
        cellDivAlg = CLOSEST_MID;
      else if(algo == "shortwall")
        cellDivAlg = SHORT_WALL;
      else if(algo == "closestwall")
        cellDivAlg = CLOSEST_WALL;
      else
      {
        util::err << "Invalid division algorithm, should be ClosestMid, ShortWall or ClosestWall. Choosing ShortWall by default." << endl;
        cellDivAlg = SHORT_WALL;
      }
      parms(section, "CellPinch", cellPinch);
      parms(section, "CellMaxPinch", cellMaxPinch);
      parms(section, "CellWallMin", cellWallMin);
      parms(section, "StrictCellWallMin", strictCellWallMin);
      parms(section, "CellWallSample", cellWallSample);
      parms(section, "CellSampleDx", sampleDx);
    }

    /**
     * Read the parameters for the drawing of a cell
     *
     * \param parms Parameter object to read from
     * \param section Section containing the parameters
     */
    void readViewParms( util::Parms& parms, const QString& section )
    {
      parms(section, "CellColorBegin", colorBegin);
      parms(section, "CellColorEnd", colorEnd);
      parms(section, "CellColorCenter", colorCenter);
      parms(section, "CellContourColor", contourColor);
      parms(section, "CellWallCorner", cellWallCorner);
      parms(section, "DrawInsideCells", drawInsides);
      parms(section, "DrawCellBorders", drawBorders);
      parms(section, "CellBlending", blending);
      parms(section, "CellCenterBlending", center_blending);
      parms(section, "CellWallWidth", cellWallWidth);
    }

    /**
     * Set the pinching parameters to be used.
     *
     * By default, the parameters will be read from the parameter file
     */
    void setCellPinchingParams(CellPinchingParams& params)
    {
      params.cellPinch = cellPinch;
      params.cellMaxPinch = cellMaxPinch;
    }

    /**
     * Returns the cell pinching parameter object
     */
    CellPinchingParams getCellPinchingParams()
    {
      return CellPinchingParams(cellPinch, cellMaxPinch);
    }

    /**
     * Returns the shortest wall algorithm parameter object
     */
    ShortWallAlgoParams getShortWallAlgoParams()
    {
      ShortWallAlgoParams p(cellWallMin, strictCellWallMin, cellWallSample, sampleDx);
      setCellPinchingParams(p);
      return p;
    }

    /**
     * Returns the closest mid algorithm parameter object
     */
    ClosestMidAlgoParams getClosestMidAlgoParams()
    {
      ClosestMidAlgoParams p(cellWallMin, strictCellWallMin);
      setCellPinchingParams(p);
      return p;
    }

    /**
     * Returns the closest wall algorithm parameter object
     */
    ClosestWallAlgoParams getCloseWallAlgoParams()
    {
      ClosestWallAlgoParams p(cellWallMin, strictCellWallMin);
      setCellPinchingParams(p);
      return p;
    }

    division_result_t divideCell( const cell& c,
                                  const division_data& ddata )
    {
      division_result_t res = VVComplex::divideCell(c, ddata, std::vector<cell>());
      if(res)
      {
        FindCenter(res.cl, static_cast<leaf_class&>(*this));
        FindCenter(res.cr, static_cast<leaf_class&>(*this));
      }
      return res;
    }

    /**
     * Divide a cell with the cells to keep defined by any kind of container.
     */
    template <typename CellContainer>
      division_result_t divideCell( const cell& c,
                                    const division_data& ddata,
                                    const CellContainer& to_keep)
      {
        division_result_t res = VVComplex::divideCell(c, ddata, to_keep);
        if(res)
        {
          FindCenter(res.cl, static_cast<leaf_class&>(*this));
          FindCenter(res.cr, static_cast<leaf_class&>(*this));
        }
        return res;
      }

    /**
     * Divide a cell with the cells to keep defined by any kind of container.
     */
    division_result_t divideCell( const cell& c,
                                  const division_data& ddata,
                                  const cell& to_keep)
    {
      division_result_t res = VVComplex::divideCell(c, ddata, to_keep);
      if(res)
      {
        FindCenter(res.cl, static_cast<leaf_class&>(*this));
        FindCenter(res.cr, static_cast<leaf_class&>(*this));
      }
      return res;
    }


    /**
     * Divide a cell of the %tissue
     *
     * \param c Cell to divide
     * \param params Parameter of the algorithm to use. The type of the 
     * params will decide which algorithm to use.
     * \param to_keep Set of cells to keep to update the data in the new cells
     *
     * \see \ref updateFromOld
     */
    template <typename AlgoParameter, typename CellContainer>
      division_result_t divideCell( const cell& c, // cell to divide
                                    const AlgoParameter& params,
                                    const CellContainer& to_keep)
    {
      division_result_t res = VVComplex::divideCell(c, params, to_keep);
      if(res)
      {
        FindCenter(res.cl, static_cast<leaf_class&>(*this));
        FindCenter(res.cr, static_cast<leaf_class&>(*this));
      }
      return res;
    }

    /**
     * Divide a cell
     *
     * Convenience function where a single cell is to be kept
     */
    template <typename AlgoParameter>
      division_result_t divideCell( const cell& to_divide, // cell to divide
                                    const AlgoParameter& params,
                                    const cell& cell_kept )
    {
      division_result_t res = VVComplex::divideCell(to_divide, params, cell_kept);
      if(res)
      {
        FindCenter(res.cl, static_cast<leaf_class&>(*this));
        FindCenter(res.cr, static_cast<leaf_class&>(*this));
      }
      return res;
    }

    /**
     * Divide a cell
     *
     * Convenience function where the algorithm used is the default one.
     */
    template <typename CellContainer>
      division_result_t divideCell( const cell& to_divide, // cell to divide
                                    const CellContainer& kept_cells )
    {
      switch(cellDivAlg)
      {
        case CLOSEST_MID:
          return divideCell(to_divide, getClosestMidAlgoParams(), kept_cells);
        case SHORT_WALL:
          return divideCell(to_divide, getShortWallAlgoParams(), kept_cells);
        case CLOSEST_WALL:
          return divideCell(to_divide, getCloseWallAlgoParams(), kept_cells);
        default:
          util::err << "Error, unknown algorithm specified ... doing nothing." << endl;
      }
      return division_result_t();
    }

    /**
     * Divide a cell
     *
     * Convenience function where the algorithm used is the default one and 
     * no context is required to setup the data of the new cells.
     */
    division_result_t divideCell( const cell& to_divide)
    {
      std::vector<cell> k;
      division_result_t res = this->divideCell(to_divide, k);
      if(res)
      {
        FindCenter(res.cl, static_cast<leaf_class&>(*this));
        FindCenter(res.cr, static_cast<leaf_class&>(*this));
      }
      return res;
    }

    /**
     * Divide a cell
     *
     * Convenience function where the algorithm used is the default one and one 
     * cell is enough to setup the data of the new cells.
     */
    division_result_t divideCell( const cell& to_divide, // cell to divide
                                  const cell& cell_kept )
    {
      std::vector<cell> k;
      k.push_back(cell_kept);
      division_result_t res = this->divideCell(to_divide, k);
      if(res)
      {
        FindCenter(res.cl, static_cast<leaf_class&>(*this));
        FindCenter(res.cr, static_cast<leaf_class&>(*this));
      }
      return res;
    }

    /**
     * Setup the OpenGL context to draw the cells
     */
    void preDraw()
    {
      glEnable(GL_POLYGON_OFFSET_FILL);
      glEnable(GL_POLYGON_OFFSET_LINE);
      glLineWidth(1.0);
    }

    /**
     * Restore the OpenGL context after the drawing
     */
    void postDraw()
    {
      glDisable(GL_POLYGON_OFFSET_FILL);
      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    /**
     * Draw a single cell with differentiated walls.
     *
     * The colors used for the cell are specified by \p cell_color and \p 
     * center_color. The color used by the wall depend on the value on the wall.
     *
     * This method requires two new methods:
     *
     *   util::Palette::Color cellWallColor(const vertex& cell, const vertex& 
     *   wall_junction)
     *
     *   double cellWallOrder(const vertex& cell, const vertex& wall_junction)
     *
     *
     * The wall considered is the one from \c wall_junction to the next vertex 
     * in the neighborhood of cell in the VVGraph. For the order, the value has 
     * to be between 0 and 1. A value of 1 will put the representation of the 
     * wall on top.
     */
    void drawWalledCell(const cell& c,
                        util::Palette::Color cell_color,
                        util::Palette::Color center_color)
    {
      Point3d cpos = this->model->position(c);

      if(!drawInsides)
      {
        cell_color = palette->getColor(colorEnd, blending);
        center_color = palette->getColor(colorEnd, center_blending);
      }

      // Draw cell faces
      glNormal3dv(this->model->normal(c).c_data());

      glPolygonMode(GL_FRONT, GL_FILL);
      glPolygonOffset(3.0, 1.0);
      glBegin(GL_TRIANGLE_FAN);
      glColor4fv(center_color.c_data());
      glVertex3dv(cpos.c_data());
      glColor4fv(cell_color.c_data());
      junction fv(0);
      forall(const junction& k, this->S.neighbors(c))
      {
        if(fv.isNull())
          fv = k;
        glVertex3dv(this->model->position(k).c_data());
      }
      glVertex3dv(this->model->position(fv).c_data());
      glEnd();

      const std::map<junction,double>& quadsWidth = calcQuads(c);

      glPolygonOffset(-1.0, -1.0);
      forall(const junction& k, this->S.neighbors(c))
      {
        const junction& l = this->S.nextTo(c, k);
        const junction& j = this->S.prevTo(c, k);
        // Get vectors towards neighbors and normalize
        Point3d jpos = this->model->position(j);
        Point3d kpos = this->model->position(k);

        if(drawBorders)
        {
          util::Palette::Color col = this->model->cellWallColor(c, j);
          double j_order = 2*this->model->cellWallOrder(c, j);
          double l_order = 2*this->model->cellWallOrder(c, l);
          double order = std::min(j_order, l_order);
          Point3d lpos = this->model->position(l);
          Point3d kj = jpos - kpos;
          Point3d kl = lpos - kpos;
          Point3d kc = cpos - kpos;
          Point3d jc = cpos - jpos;
          Point3d kjn = kj, kln =kl, kcn = kc, jcn = jc;
          kjn /= norm(kjn);
          kln /= norm(kln);
          kcn /= norm(kcn);
          jcn /= norm(jcn);

          // Corner size along cell wall, clip to wall length
          double kjl = norm(kj);
          if(kjl > cellWallCorner)
            kjl = cellWallCorner;
          double kll = norm(kl);
          if(kll > cellWallCorner)
            kll = cellWallCorner;

          double kcl = .75 * cellWallCorner * exp((kjn * kcn + kln * kcn)/2.0);

          // Width of quads for walls
          double kcqw = quadsWidth.find(k)->second * cellWallWidth;
          double jcqw = quadsWidth.find(j)->second * cellWallWidth;

          // Adjust corner centers if required
          bool ccvx; // convex
          ccvx = ((kll*kln - kcl*kcn)^(kjl*kjn - kcl*kcn)) * this->model->normal(k) > 0;
          if(!ccvx || kcl < kcqw)
            kcl = kcqw;

          Point3d p;

          // Draw corners
          glPolygonOffset((1.0-order)+1, -3.0);
          glColor4fv(col.c_data());
          glBegin(GL_TRIANGLE_FAN);
          glVertex3dv(kpos.c_data());
          p = kpos + kll * kln;
          glVertex3dv(p.c_data());
          p = kpos + kcl * kcn;
          glVertex3dv(p.c_data());
          p = kpos + kjl * kjn;
          glVertex3dv(p.c_data());
          glEnd();

          // Draw cell walls
          glPolygonOffset((1.0-j_order)+1, -3.0);
          glBegin(GL_QUADS);
          glVertex3dv(jpos.c_data());
          glVertex3dv(kpos.c_data());
          p = kpos + kcqw * kcn;
          glVertex3dv(p.c_data());
          p = jpos + jcqw * jcn;
          glVertex3dv(p.c_data());
          glEnd();
        }
      }
    }

    /**
     * Draw a cell with walls, specifying the value.
     */
    void drawWalledCell(const cell& c,
                        double value)
    {
      Point3d cpos = this->model->position(c);

      if(!drawInsides)
      {
        value = 0;
      }

      // Draw cell faces
      glNormal3dv(this->model->normal(c).c_data());

      drawWalledCell(c, valueColor(value), valueCenterColor(value));
    }

    /**
     * Draw a single cell.
     *
     * The colors used for the cell are specified by \p cell_color and \p 
     * center_color.
     */
    void drawCell(const cell& c, util::Palette::Color cell_color,
                  util::Palette::Color center_color)
    {
      Point3d cpos = this->model->position(c);

      if(!drawInsides)
      {
        cell_color = palette->getColor(colorEnd, blending);
        center_color = palette->getColor(colorEnd, center_blending);
      }

      // Draw cell faces
      glNormal3dv(this->model->normal(c).c_data());

      glPolygonMode(GL_FRONT, GL_FILL);
      glPolygonOffset(3.0, 1.0);
      glBegin(GL_TRIANGLE_FAN);
      glColor4fv(center_color.c_data());
      glVertex3dv(cpos.c_data());
      glColor4fv(cell_color.c_data());
      junction fv(0);
      forall(const junction& k, this->S.neighbors(c))
      {
        if(fv.isNull())
          fv = k;
        glVertex3dv(this->model->position(k).c_data());
      }
      glVertex3dv(this->model->position(fv).c_data());
      glEnd();

      const std::map<junction,double>& quadsWidth = calcQuads(c);

      if(drawBorders)
      {
        palette->useColor(contourColor);
        glPolygonOffset(-1.0, -1.0);
        forall(const junction& k, this->S.neighbors(c))
        {
          // Get vectors towards neighbors and normalize
          const junction& j = this->S.prevTo(c, k);
          const junction& l = this->S.nextTo(c, k);

          const Point3d& jpos = this->model->position(j);
          const Point3d& kpos = this->model->position(k);
          const Point3d& lpos = this->model->position(l);

          Point3d kj = jpos - kpos;
          Point3d kl = lpos - kpos;
          Point3d kc = cpos - kpos;
          Point3d jc = cpos - jpos;
          Point3d kjn = kj, kln =kl, kcn = kc, jcn = jc;
          kjn /= norm(kjn);
          kln /= norm(kln);
          kcn /= norm(kcn);
          jcn /= norm(jcn);

          // Corner size along cell wall, clip to wall length
          double kjl = norm(kj);
          if(kjl > cellWallCorner)
            kjl = cellWallCorner;
          double kll = norm(kl);
          if(kll > cellWallCorner)
            kll = cellWallCorner;

          double kcl = .75 * cellWallCorner * (kjn * kcn + kln * kcn)/2.0;

          // Width of quads for walls
          double kcqw = quadsWidth.find(k)->second * cellWallWidth;
          double jcqw = quadsWidth.find(j)->second * cellWallWidth;

          // Adjust corner centers if required
          bool ccvx; // convex
          ccvx = ((kll*kln - kcl*kcn)^(kjl*kjn - kcl*kcn)) * this->model->normal(k) > 0;
          if(!ccvx || kcl < kcqw)
            kcl = kcqw;

          Point3d p;

          // Draw corners
          glBegin(GL_TRIANGLE_FAN);
          glVertex3dv(kpos.c_data());
          p = kpos + kll * kln;
          glVertex3dv(p.c_data());
          p = kpos + kcl * kcn;
          glVertex3dv(p.c_data());
          p = kpos + kjl * kjn;
          glVertex3dv(p.c_data());
          glEnd();

          // Draw cell walls
          glBegin(GL_QUADS);
          glVertex3dv(jpos.c_data());
          glVertex3dv(kpos.c_data());
          p = kpos + kcqw * kcn;
          glVertex3dv(p.c_data());
          p = jpos + jcqw * jcn;
          glVertex3dv(p.c_data());
          glEnd();
        }
      }
    }

    /**
     * Draw just the contour of the cells.
     *
     * This function was extracted as, due to a bug in latest NVidia drivers, 
     * alternating between LINE and FILL to draw polygon crashes text 
     * rendering.
     */
    void drawCellContour(const cell& c)
    {
      palette->useColor(contourColor);
      glPolygonMode(GL_FRONT, GL_LINE);
      glPolygonOffset(1.0, 5.0);
      glBegin(GL_POLYGON);
      forall(const junction& n, this->S.neighbors(c))
      {
        glVertex3dv(this->model->position(n).c_data());
      }
      glEnd();
    }

    /**
     * Draw the cell as a simple colored polygon.
     *
     * Useful for quick drawing of the shape, like for selection system.
     */
    void drawSimplifiedCell(const cell& c)
    {
      glBegin(GL_POLYGON);
      forall(const junction& n, this->S.neighbors(c))
      {
        glVertex3dv(this->model->position(n).c_data());
      }
      glEnd();
    }

    /**
     * Draw a single cell. The value is used to setup the color (using 
     * colorBegin, colorEnd) if the interior of the cell is drawn.
     *
     * If \p center_color is specified, then it is used instead of the usual 
     * modification of the color of the cell color.
     */
    void drawCell(const cell& c, double value)
    {
      Point3d cpos = this->model->position(c);

      if(!drawInsides)
      {
        value = 0;
      }

      // Draw cell faces
      glNormal3dv(this->model->normal(c).c_data());

      drawCell(c, valueColor(value), valueCenterColor(value));
    }

    /**
     * Get the color corresponding to a given value
     */
    util::Palette::Color valueColor(double value)
    {
      if(value > 1.0)
        value = 1.0;
      value *= (1.0 - colorCenter);
      return palette->selectColor(colorBegin, colorEnd, value, blending);
    }

    /**
     * Get the center color corresponding to a given value
     */
    util::Palette::Color valueCenterColor(double value)
    {
      if(value > 1.0)
        value = 1.0;
      value *= (1.0 - colorCenter);
      double value_center = value + colorCenter;
      return palette->selectColor(colorBegin, colorEnd, value_center, blending);
    }

  protected:
    /**
     * Compute parameters for the width of the corners.
     *
     * \warning For internal use only
     */
    std::map<junction,double>
    calcQuads( const cell& c )
    {
      std::map<junction,double> quadsWidth;
      forall(const junction& k, this->S.neighbors(c))
      {
        const junction& l = this->S.nextTo(c, k);
        const junction& j = this->S.prevTo(c, k);

        // Get vectors towards neighbors
        Point3d kpos = this->model->position(k);
        Point3d kj = this->model->position(j) - kpos;
        Point3d kl = this->model->position(l) - kpos;
        Point3d kc = this->model->position(c) - kpos;

        // Info for quads
        double lsz = norm(kc ^ util::normalized(kl)); // norm(kc - kl * (kc * kl));
        if(lsz > 0)
          lsz = norm(kc)/lsz;
        double jsz = norm(kc ^ util::normalized(kj)); // norm(kc - kj * (kc * kj));
        if(jsz > 0)
          jsz = norm(kc)/jsz;

        quadsWidth[k] = max(lsz, jsz);
      }
      return quadsWidth;
    }

  };
#undef CELL_MESH
#undef TISSUE_MESH
#undef VERTEX
}
#endif // VVELIB_ALGORITHMS_TISSUE_H
