#ifndef VVELIB_ALGORITHMS_TRIANGLE_GROWTH_H
#define VVELIB_ALGORITHMS_TRIANGLE_GROWTH_H

/**
 * \file algorithms/triangle_growth.h
 *
 * This file contains the classes needed to describe the growth of a tissue 
 * from a triangular mesh growing through time.
 */

#include <config.h>
#include <geometry/geometry.h>
#include <vector>
#include <algorithms/complex.h>
#include <QString>
#include <storage/storage.h>
#include <util/unorderedset.h>
#include <util/unorderedmap.h>
#include <util/assert.h>
#include <QTextStream>

namespace algorithms
{
  using geometry::Point3d;

  /**
   * \class TriangleSurface triangle_growth.h <algorithms/triangle_growth.h>
   *
   * Class representing a triangulated surface.
   */
  class TriangleSurface
  {
  public:
    /**
     * Type of a cell, i.e. a triangle
     */
    struct Cell
    {
      /// Normal to the cell
      Point3d normal;
      /// Id of the cell
      int id;
    };

    /**
     * Type of a vertex, i.e. a junction
     */
    struct Vertex
    {
      /// 3D position
      Point3d pos;
      /// Normal
      Point3d normal;
      /// Id of the junction
      int id;
    };

    /// Type of the complex graph holding the triangle structure
    typedef vvcomplex::VVComplexGraph<Cell, Vertex> CellComplex;
    EXPORT_COMPLEX_VERTICES(CellComplex);

    /// Complex graph holding the triangle structure
    CellComplex S;

    /**
     * Structure describing the position of a point on a triangulated surface,
     * relatively to this surface.
     */
    struct Position
    {
      Point3d barycentric;
      int v1, v2, v3;
      int face;

      Position() : v1(0), v2(0), v3(0), face(0) {}
      Position(const Point3d& bar, int vv1, int vv2, int vv3, int f)
        : barycentric(bar)
        , v1(vv1) , v2(vv2) , v3(vv3) , face(f)
        {}
      Position(const Position& copy)
        : barycentric(copy.barycentric)
        , v1(copy.v1), v2(copy.v2), v3(copy.v3), face(copy.face)
        {}
#ifdef USE_CXX0X
      Position(Position&& copy) = default;
#endif

      bool serialize(storage::VVEStorage&);

      operator bool() const { return face != 0; }
    };

    TriangleSurface();
#ifdef USE_CXX0X
    TriangleSurface(TriangleSurface&& ) = default;
    TriangleSurface(const TriangleSurface& ) = default;
#endif

    /**
     * Get the 3d position from the position relative to the triangular
     * surface
     */
    Point3d point3d(const Position& pos) const;
    /**
     * Get the normal to the surface at position \c pos
     */
    Point3d normal(const Position& pos) const;
    /**
     * Get the "smoothed" normal.
     *
     * To each vertex is associated a normal, mostly used for visualization.
     * This function will interpolate between these specified normals rather
     * that using the actual normal of the triangle.
     */
    Point3d smoothNormal(const Position& pos) const;

    /**
     * Give the position, relatively to the surface, of the point closest to
     * \c pos.
     */
    Position position(const Point3d& pos) const;
    /**
     * Same as position(const Point3d&), but specifies a starting point.
     *
     * The position will be searched, starting from the \c start position and
     * going from triangle to triangle. The first traversed triangle in which
     * \c pos can be projected will be used as reference.
     *
     * \note This method works only for surfaces regular enough, at least from
     * \c start to \c pos.
     */
    Position position(const Position& start, const Point3d& pos) const;

    /**
     * Returns the position of the center of the face.
     *
     * Effectively, this will be the points of barycentric coordinates 
     * (1,1,1)/3 on the face.
     */
    Position centerPosition(int face) const;

    /**
     * Returns the coordinates of the center of the face.
     *
     * Effectively, this will be the points of barycentric coordinates 
     * (1,1,1)/3 on the face.
     */
    Point3d center3d(int face) const;

    /**
     * Return the vector represented by \c pos
     *
     * The vector is such that its orientation and size are deformed with the
     * surface. Also, the vector will always be in the plane of the triangle
     * it is defined on.
     */
    Point3d vector3d(const Position& pos) const;

    /**
     * Creates a vector attached to the surface.
     */
    Position vector(const Point3d& ref, const Point3d& pos) const;
    /**
     * Creates a vector attached to the surface.
     */
    Position vector(const Position& ref, const Point3d& pos) const;

    /**
     * Reset the structure ...
     */
    void clear();

    /**
     * Return a description of the contour of the surface independent of the
     * time
     */
    std::vector<Position> contour() const;

    /// Time of this triangle surface
    double time;
    int time_index;

    /// List of vertices
    std::vector<junction> vertices;
    /// List of faces
    std::vector<cell> faces;

    /// Bounding box of the surface
    Point3d pmin, pmax;
  protected:
    typedef std::unordered_set<cell> cell_set_t;
    Position find_position(const cell& c, const Point3d& pos, cell_set_t& visited) const;
    Position vertexPosition(const junction& j) const;
    std::vector<Position> nextContourVertex(const junction& first, const cell& c, const junction& j, std::vector<Position> acc) const;
  };

  /**
   * \class TriangleGrowth triangle_growth.h <algorithms/triangle_growth.h>
   * Growth description using a set of triangles moving.
   *
   * The number of points and triangles must be constant for this class to 
   * work. Also, the different OBJ files must keep the same indices for the 
   * same points.
   */
  class TriangleGrowth
  {
  public:
    typedef TriangleSurface::Position Position;

    struct Cell
    {
      int id;
    };

    struct Vertex
    {
      std::vector<Point3d> pos;
      std::vector<Point3d> normal;
      int id;
    };

    TriangleGrowth();

    /**
     * Initialize the growth with a description file or a directory
     *
     * \returns True if everything was fine. If not, the error is but in the 
     * error message of the class.
     */
    bool readOBJs(QString desc_filepath);

    /**
     * Initialize the growth with a description file or a directory
     *
     * \returns True if everything was fine. If not, the error is but in the 
     * error message of the class.
     */
    bool readOBJs(const std::string& desc_filepath) { return readOBJs(QString::fromStdString(desc_filepath)); }

    /**
     * Initialize the growth from a data structure.
     *
     * \param cplx Graph complex representing the triangulated mesh
     * \param times List of key frame times, ordered from the earliest to the 
     * latest (i.e. times must be strictely increasing).
     * \param model Object able interrogated for informations about vertices 
     * and cells.
     *
     * The Model class is suppose to include a few methods:
     * \code
     * Point3d position(const junction& v, int keyframe) const;
     * \endcode
     * Returns the position of the junction v at the specified key frame.
     *
     * \code
     * Point3d normal(const junction& v, int keyframe) const;
     * \endcode
     * Returns the normal to the junction v at the specified key frame.
     */
    template <typename Model, typename ComplexGraph>
      bool init(const ComplexGraph& cplx, std::vector<double> times, Model *model);

    /**
     * \returns True if the growth is correctly initialized
     */
    bool valid() const { return _valid; }

    /**
     * If the object is not valid, return the reason
     */
    const QString& errorMsg() const { return _errorMsg; }

    /**
     * Returns the number of time points
     */
    size_t size() const;

    /**
     * Returns the start time
     */
    double startTime() const { return times.front(); }

    /**
     * Returns the end time
     */
    double endTime() const { return times.back(); }

    /**
     * Set the current time.
     *
     * \warning If you try to set a time less than the start time or greater 
     * than the end time, the time will be clamped to the start or end value.
     */
    void setTime(double t);

    /**
     * Return the current time.
     */
    double time() const { return _time; }

    /**
     * Return the current position in the time scale
     */
    int timePos() const { return _time_pos; }

    /**
     * Return the surface at the current time
     */
    const TriangleSurface& surface() const;

    /**
     * Return the surface at the current time
     */
    TriangleSurface surface(double time);

    /**
     * Return a description of the contour of the surface independent of the
     * time
     */
    std::vector<Position> contour() const;

  protected:
    void setSurface(TriangleSurface& s) const;

    typedef vvcomplex::VVComplexGraph<Cell,Vertex> CellComplex;
    EXPORT_COMPLEX_VERTICES(CellComplex);

    Position vertexPosition(const junction& j) const;

    mutable TriangleSurface TS;

    CellComplex S;
    std::vector<double> times;
    double _time;
    int _time_pos;
    bool _valid;
    QString _errorMsg;
    std::vector<cell> faces;
    std::vector<junction> vertices;
  };

  template <typename Model, typename ComplexGraph>
  bool TriangleGrowth::init(const ComplexGraph& cplx, std::vector<double> _times, Model *model)
  {
    typedef typename ComplexGraph::cell ext_cell;
    typedef typename ComplexGraph::junction ext_junction;

    //QTextStream out(stdout);

    _valid = false;

    std::unordered_map<ext_cell, cell> cell_convert;
    std::unordered_map<ext_junction, junction> junction_convert;

    faces.clear();
    vertices.clear();
    S.clear();

    //out << "Init the structure " << flush;

    if(_times.empty())
      return false;
    size_t nb_times = _times.size();
    times = _times;
    _time = times.front();
    _time_pos = 0;
    faces.resize(cplx.nb_cells(), cell(0));
    vertices.resize(cplx.nb_junctions(), junction(0));

    //out << "done" << endl;

    //out << "Create the faces " << flush;

    // Create faces
    int i = 0;
    forall(const ext_cell& ec, cplx.cells())
    {
      //out << '.' << flush;
      cell c;
      c->id = i;
      faces[i] = c;
      S.insert(c);
      cell_convert[ec] = c;
      i++;
    }
    //out << " OK" << endl;

    //out << "Create the vertices " << flush;

    // Create vertices
    i = 0;
    forall(const ext_junction& ej, cplx.junctions())
    {
      //out << '.' << flush;
      junction j;
      j->id = i;
      j->pos.resize(nb_times);
      j->normal.resize(nb_times);
      for(size_t t = 0 ; t < nb_times ; ++t)
      {
        j->pos[t] = model->position(ej, t);
        j->normal[t] = model->normal(ej, t);
      }
      vertices[i] = j;
      S.insert(j);
      junction_convert[ej] = j;
      i++;
    }
    //out << " OK" << endl;

    //out << "Copy faces neighborhood " << flush;

    // Copy faces neighborhood
    forall(const ext_cell& ec, cplx.cells())
    {
      //out << '.' << flush;
      const cell& c = S.reference(cell_convert[ec]);
      forall(const ext_junction& ej, cplx.neighbors(ec))
      {
        junction j = junction_convert[ej];
        if(S.valence(c) < 2)
        {
          S.insertEdge(c, j);
        }
        else
        {
          junction pj = junction_convert[cplx.prevTo(ec, ej)];
          S.spliceAfter(c, pj, j);
        }
      }
      vvassert_msg(S.valence(c) == 3, QString("Face %1 has %2 junctions instead of 3").arg(c->id).arg(S.valence(c)));
    }
    //out << " OK" << endl;

    //out << "Copy faces neighborhood " << flush;
    // Copy vertices neighborhood
    forall(const ext_junction& ej, cplx.junctions())
    {
      //out << '.' << flush;
      const junction& j = S.reference(junction_convert[ej]);
      forall(const ext_cell& ec, cplx.neighbors(ej))
      {
        cell c = cell_convert[ec];
        if(S.valence(j) < 2)
        {
          S.insertEdge(j, c);
        }
        else
        {
          cell pc = cell_convert[cplx.prevTo(ej, ec)];
          S.spliceAfter(j, pc, c);
        }
      }
    }
    //out << " OK" << endl;

    _valid = true;
    return true;
  }
}

#endif // VVELIB_ALGORITHMS_TRIANGLE_GROWTH_H

