#ifndef BSPLINE_TISSUE_MODEL_H
#define BSPLINE_TISSUE_MODEL_H

/**
 * \file bspline_tissue_model.h
 *
 * This files include the bspline ::tissue model helper. This helper define 
 * a tissue whose growth is driven by a set of bspline surfaces defining the 
 * geometry at key-points.
 *
 * The namespace bspline_tissue_model contains the base class for this helper.
 *
 * To use the helper, simply include <bspline_tissue_model.h>.
 * Any extra attributes for vertices or edges should be defined before 
 * including this file.
 *
 * Once included, the model starts with
 * \code
 * StartModel
 * \endcode
 * and ends with
 * \code
 * EndModel
 * \endcode
 *
 * The tissue defined in the model uses cells which contains these attributes:
 * \code
 *  double area;              // Area of the cell
 *  Point3d pos;              // Position of the cell
 *  Point3d normal;           // Normal to the surface
 *  Point2d uv;               // Local coordinate of the cell on the bspline
 *  double value;             // Value used to draw the cell
 * \endcode
 * and junctions which contains these attributes:
 * \code
 *  Point3d pos;              // Position of the junction
 *  Point3d normal;           // Normal to the surface
 *  Point2d uv;               // Local coordinate of the junction on the bspline
 * \endcode
 *
 * In addition, the user can add its own attributes by defining the macro \c VertexAttributes \b before including this file:
 * \code
 * #define CellAttribute \
 *   double attr1; \
 *   int attr2; \
 *   bool attr3;
 * #define JunctionAttribute \
 *   double attr1; \
 *   bool attr2; \
 *   int attr3;
 *
 * #include <bspline_tissue_model.h>
 * \endcode
 * \note As this is a macro, to span over many lines you need to end the inner lines with a backslash.
 *
 * Similarly, the user can add attributes to the cell and tissue edges by 
 * defining the macros \c CellEdgeAttributes, \c WallAttributes, \c 
 * CellJunctionEdgeAttributes and \c JunctionCellEdgeAttributes.
 *
 * Within the model class, named \c ModelClass, all the methods and variable of 
 * the class bspline_tissue_model::TissueModel are available. Reference to the 
 * parent class can be done using the \c ParentClass typedef.
 *
 * For example, if yo want to augment the drawing method, you might write:
 * \code
 * void draw(Viewer* viewer)
 * {
 *   ParentClass::draw(viewer);
 *   // my other commands
 * }
 * \endcode
 *
 * As a small example, here is a code that creates a single cell and change its 
 * color when the user click on it.
 * \code
 * #define CellAttributes \
 *   bool clicked; // Is the current cell clicked?
 *
 * #include <bspline_tissue_model.h>
 *
 * StartModel
 *
 * // Define
 * int clickedColor, unclickedColor;
 * double dt;      // How much mecanic advance at each step
 * double maxArea; // Maximum area of a cell
 *
 * // Read the extra parameter
 * void readParam(util::Parms& parms)
 * {
 *   parms("Main", "dt", dt);
 *   parms("Main", "MaxArea", maxArea);
 *   parms("View", "ClickedColor", clickedColor);
 *   parms("View", "UnclickedColor", unclickedColor);
 * }
 *
 * // Initialize the tissue
 * void initialize()
 * {
 *   // First, create the cell
 *   initTissue();
 *   // Then, initialise the content of the cells
 *   forall(const cell& c, T.C)
 *     c->clicked = false;
 * }
 *
 * // Update the status of the cell after user interaction
 * void postSelection(const QPoint&, Viewer* viewer)
 * {
 *   const cell& c = cellFromId(viewer->selectedName());
 *   // If a cell was selected
 *   if(c)
 *   {
 *     c->clicked ^= true; // Invert its state
 *   }
 * }
 *
 * // Main loop: tissue growth and cell division
 * void step()
 * {
 *    // Change the current time, then update the vertices positions and cell areas
 *    time += dt;
 *    updatePositions();
 *    updateCellsArea();
 *    ordered_cells_t ordered_divide;
 *
 *    // Find the cells to be divided, order them and divide them
 *    forall(const cell& c, T.C)
 *    {
 *      vvcomplex::FindCenter(c, T);
 *      if(c->area > maxArea)
 *      {
 *        ordered_divide.insert(c);
 *      }
 *    }
 *
 *    // How to divide
 *    forall(const cell& c, ordered_divide)
 *    {
 *      T.divideCell(c);
 *    }
 * }
 *
 * // Define the color depending on the clicked state
 * Color getCellColor(const cell& c)
 * {
 *   if(c->clicked)
 *     return palette.getColor(clickedColor);
 *   else
 *     return palette.getColor(unclickedColor);
 * }
 *
 * // Define the color depending on the clicked state
 * Color getCellCenterColor(const cell& c)
 * {
 *   if(c->clicked)
 *     return palette.getColor(clickedColor);
 *   else
 *     return palette.getColor(unclickedColor);
 * }
 *
 * // When a cell is clicked, the property is inherited by the children
 * void updateFromOld(const cell& cl, const cell& cr, const cell& c,
 *                    const Tissue::division_data&, Tissue&)
 * {
 *    cl->clicked = cr->clicked = c->clicked;
 * }
 *
 * EndModel
 *
 * \endcode
 */

#include <vve.h>
#include <util/parms.h>
#include <util/palette.h>

#include <util/vector.h>
#include <util/watchdog.h>

#include <util/bsurface.h>
#include <util/key_framer.h>

#include <algorithms/tissue.h>
#include <iostream>

using util::Point3d;
using util::Point2d;

/**
 * Typedef to easy the use of colors
 */
typedef util::Palette::Color Color;

/**
 * \namespace bspline_tissue_model
 *
 * Namespace containing the base class for bspline tissue model helper
 */
namespace bspline_tissue_model
{
  /**
   * Relative error for geometry
   */
  const double epsilon=0.0001;

  using std::cout;
  using std::endl;

  /**
   * \class TissueModel bspline_tissue_model.h <bspline_tissue_model.h>
   *
   * Base class for the bspline tissue model helper
   */
  template <typename RealModel, typename TissueClass>
  struct TissueModel : public Model
  {
    IMPORT_COMPLEX_TYPES(TissueClass);

    /**
     * Operator class to compare the size of cells
     */
    struct CompareSize : public std::binary_function<cell, cell, bool>
    {
      bool operator()(const cell& c1, const cell& c2) const
      {
        return c1->area < c2->area;
      }
    };

    /**
     * Ordered set of vertices.
     *
     * The vertices must represent cells. They will be ordered by area.
     */
    typedef std::set<cell,CompareSize> ordered_cells_t;

    util::WatchDog rex;        /// Watchdog for file modification
    util::Palette palette;     /// Palette to retrieve colors
    TissueClass T;             /// Main tissue

    util::KeyFramer leaf;      /// Description of the leaf growth
    util::BSplineSurface leafs;/// One surface of the leaf at a time

    int cellInitWalls;         /// Initial number of walls in the cell
    double dt;                 /// Time step
    double growthStartTime;    /// Initial time for the leaf description
    int backColor;             /// Color of the background
    bool drawNeighborhood;     /// Draw the neighborhood as arrows ?

    double time;               /// Current time

    /**
     * Default constructor
     *
     * Creates the tissue, the aliases, the palette and the watchdog
     */
    TissueModel(QObject *parent)
      : Model(parent)
      , rex(this)
      , palette("pal.map")
      , T(palette, (RealModel*)this)
    {
      rex.addObject(&palette);
      rex.addObject(&leaf);
    }

    /**
     * Keep track of the modification of the parameter file ("view.v"), and of 
     * all the files registered to the watchdog.
     */
    void modifiedFiles( const std::set<std::string>& filenames )
    {
      if(filenames.count("view.v"))
        readTissueParam();
      rex.watch(filenames);
    }

    /**
     * Method to redefine to read extra parameters
     */
    virtual void readParam( util::Parms& ) { }

    /**
     * Set the position of the cell \c x with respect to the current time.
     */
    void SetPos( const cell& c )
    {
      c->pos = leafs.EvalN(c->uv[0], c->uv[1]);
      c->normal = leafs.NormalEvalN(c->uv[0], c->uv[1]);
    }

    /**
     * Set the position of the junction \c j with respect to the current time.
     */
    void SetPos( const junction& j )
    {
      j->pos = leafs.EvalN(j->uv[0], j->uv[1]);
      j->normal = leafs.NormalEvalN(j->uv[0], j->uv[1]);
    }

    /**
     * Initialize the tissue as one single cell whose shape is the shape of the 
     * current b-spline surface.
     */
    void initTissue()
    {
      time = growthStartTime;
      leaf.SetTime(time);
      leafs = leaf.GetSurfaceTime();

      // Find center and set cell area and length of cell walls
      cell c;
      c->uv[0] = c->uv[1] = .5;
      c->value = 0.5;
      SetPos(c);

      std::vector<junction> junctions(cellInitWalls, junction(0));
      for(int i = 0 ; i < cellInitWalls ; i++)
      {
        double s = (double)(cellInitWalls-i)/(double)cellInitWalls;
        junction v;
        v->uv = leafs.ContourInverse(s);
        SetPos(v);
        junctions[i] = v;
      }

      T.addCell(c, junctions);

      updateCellsArea();
    }

    /**
     * Update the positions of all the vertices to reflect the new b-spline 
     * surface.
     */
    void updatePositions()
    {
      leaf.SetTime(time);
      leafs = leaf.GetSurfaceTime();
      forall(const cell& c, T.C)
      {
        SetPos(c);
      }
      forall(const junction& j, T.W)
      {
        SetPos(j);
      }
    }

    /**
     * Update the cells area using the current positions of the vertices.
     */
    void updateCellsArea()
    {
      forall_named( const cell& c, T.S, cells )
      {
        c->area = 0.0;
        forall( const junction& n, T.S.neighbors(c) )
        {
          // Find area
          const junction& m = T.S.nextTo(c, n);
          c->area += geometry::triangleArea(c->pos, m->pos, n->pos);
        }
        if(c->area <= epsilon ) // Avoid inf
        {
          cout << "Bad area " << c->area << endl;
          c->area = 0.1;
        }
      }
    }

    /**
     * Register the parameter file.
     */
    void registerFiles()
    {
      registerFile("view.v"); 
    }

    /**
     * Read parameters needed for the tissue and tissue growth
     */
    void readTissueParam()
    {
      util::Parms parms("view.v");
      std::string newKeyFrameFile;

      parms("Main", "dt", dt);
      parms("Main", "KeyFrameFile", newKeyFrameFile);
      parms("Main", "GrowthStartTime", growthStartTime);
      parms("Main", "CellInitWalls", cellInitWalls);

      rex.changeFilename(&leaf, newKeyFrameFile);

      T.readViewParms(parms, "View");
      parms("View", "BackColor", backColor);
      parms("View", "DrawNeighborhood", drawNeighborhood);

      T.readParms(parms, "Tissue");

      readParam(parms);
    }

    void initDraw()
    {
      glEnable(GL_NICEST);
      glDisable(GL_CULL_FACE);
    }

    void preDraw()
    {
      Color bg = palette.getColor(backColor);
      glClearColor(bg.r(), bg.g(), bg.b(), 1.0);
      T.preDraw();
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    }

    void postDraw()
    {
      glDisable(GL_BLEND);
      T.postDraw();
    }

    void draw(Viewer *viewer)
    {
      forall_named(const cell& c, T.S, cells)
      {
        Color color = getCellColor(c);
        Color center_color = getCellCenterColor(c);
        T.drawCell(c, color, center_color);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        if(drawNeighborhood)
        {
          int cnt = 3;
          double ds = 0.1 / (3+T.C.valence(c));
          glColor3f(1.0, 0, 0);
          Vec from(c->pos);
          forall(const cell& n, T.C.neighbors(c))
          {
            glColor3f(1.0, 1.0, 1.0);
            Vec to(n->pos);
            Vec u = to-from;
            u.normalize();
            to = from + cnt*ds*u;
            cnt++;
            viewer->drawArrow(from, to, -1, 6);
          }
        }
      }

      double vmin[3], vmax[3];
      leafs.BoundingBox(vmin[0], vmin[1], vmin[2], vmax[0], vmax[1], vmax[2]);
      Vec pmin(vmin), pmax(vmax);
      viewer->setSceneBoundingBox(pmin, pmax);
    }

    /**
     * Draw for cell selection
     */
    void drawWithNames()
    {
      bool db = T.drawBorders;
      T.drawBorders = false;
      int i = 0;
      forall_named(const cell& c, T.S, cells)
      {
        glPushName(i++);
        T.drawCell(c, c->value);
        glPopName();
      }
      T.drawBorders = db;
    }

    /**
     * Function to override to change the color used for the cell
     */
    virtual Color getCellColor(const cell &c)
    {
      return T.valueColor(c->value);
    }

    /**
     * Function to override to change the color used for the center of the cell
     */
    virtual Color getCellCenterColor(const cell &c)
    {
      return T.valueCenterColor(c->value);
    }

    /**
     * Returns the cell of id \c i
     *
     * This id is the one used for OpenGL selection and is returned by 
     * QGLViewer::selectedName()
     */
    const cell& cellFromId(int i)
    {
      if(i == -1)
        return cell::null;
      return T.S.get_cell(i);
    }

    //@{
    ///\name Methods required by the tissue library
    Point3d position( const cell& v) const
    {
      return v->pos;
    }
    Point3d position( const junction& v) const
    {
      return v->pos;
    }

    void setPosition( const cell& v, const Point3d& pos)
    {
      leafs.PointInversionN(v->uv, v->uv, pos);
      SetPos(v);
    }

    void setPosition( const junction& v, const Point3d& pos)
    {
      leafs.PointInversionN(v->uv, v->uv, pos);
      SetPos(v);
    }

    void setPositionHint( const junction& v, const junction& n1, const junction& n2, double r)
    {
      v->uv = (1-r)*n1->uv + r*n2->uv;
    }

    Point3d normal( const cell& v) const
    {
      return v->normal;
    }

    Point3d normal( const junction& v) const
    {
      return v->normal;
    }

    void updateFromOld( const cell& , const cell& , const cell& ,
                        const typename TissueClass::division_data&, TissueClass& )
    {
    }
    //@}

    /**
     * Method to override to provide the user's model
     */
    void step() { }

    /**
     * Method to override for user's initialization
     */
    void initialize() {}

  };
}

#ifndef CellAttributes
#  define CellAttributes
#endif

#ifndef JunctionAttributes
#  define JunctionAttributes
#endif

#ifndef CellEdgeAttributes
#  define CellEdgeAttributes
#endif

#ifndef WallAttributes
#  define WallAttributes
#endif

#ifndef CellJunctionEdgeAttributes
#  define CellJunctionEdgeAttributes
#endif

#ifndef JunctionCellEdgeAttributes
#  define JunctionCellEdgeAttributes
#endif

#ifndef ModelInit
#  define ModelInit
#endif

/**
 * Macro defining the start of the model
 */
#define StartModel \
  struct CellType \
  { \
    double area; \
    Point3d pos; \
    Point3d normal; \
    Point2d uv; \
    double value; \
    CellAttributes\
  }; \
  \
  struct JunctionType \
  { \
    Point3d pos; \
    Point3d normal; \
    Point2d uv; \
    JunctionAttributes\
  }; \
 \
  struct CellEdgeType \
  { \
    CellEdgeAttributes\
  }; \
 \
  struct WallType \
  { \
    WallAttributes\
  }; \
 \
  struct CellJunctionEdgeType \
  { \
    CellJunctionEdgeAttributes\
  }; \
 \
  struct JunctionCellEdgeType \
  { \
    JunctionCellEdgeAttributes\
  }; \
 \
  class ModelClass; \
  typedef tissue::Tissue<ModelClass, CellType, JunctionType, WallType, CellEdgeType,CellJunctionEdgeType, JunctionCellEdgeType> Tissue; \
 \
  struct ModelClass : public bspline_tissue_model::TissueModel<ModelClass,Tissue> \
  { \
    typedef bspline_tissue_model::TissueModel<ModelClass,Tissue> ParentClass; \
    ModelClass(QObject *parent) \
      : ParentClass(parent)\
      ModelInit \
    { \
      readTissueParam(); \
      registerFiles(); \
      initialize(); \
      if(T.C.empty()) initTissue(); \
    }

/**
 * Macro defining the end of the model
 */
#define EndModel }; DEFINE_MODEL(ModelClass);

#endif // BSPLINE_TISSUE_MODEL_H
