#ifndef CELLSYSTEM_MODEL_H
#define CELLSYSTEM_MODEL_H

#include <vve.h>
#include <util/parms.h>
#include <util/palette.h>
#include <util/watchdog.h>
#include <util/vector.h>

#include <algorithms/tissue.h>
#include <algorithms/cellsystem.h>

#include <iostream>

typedef util::Vector<3,double> Point3d;

namespace cellsystem_model
{
  using std::endl;
  using std::cout;

  template <typename TissueClass, typename RealModel>
  struct CellSystemModel : public cell_system::CellSystem<TissueClass,RealModel>
  {
    typedef cell_system::CellSystem<TissueClass,RealModel> ParentClass;

    IMPORT_COMPLEX_TYPES(TissueClass);

    typedef graph::VVBiGraph<typename cell::content_t, typename junction::content_t> ReferenceGraph;

    util::WatchDog rex;
    util::Palette palette;

    int backColor;
    int wallColor;
    bool division;
    bool showReference;
    bool torusWorld;
    double referenceVectorSize;
    bool stopAfterRules, stopBeforeRules, applyRule;
    std::vector<int> polygon_sizes;
    std::vector<int> symbol_color;

    ReferenceGraph reference;
    cell_system::label_t axiom;
    Point3d init_ref;

    TissueClass& T;

    virtual void readParms()
    {
      util::Parms parms("view.v");
      parms("Main", "ViewRulesApplication", this->viewRulesApplication);
      parms("Main", "ViewRulesDefinitions", this->viewRulesDefinitions);
      parms("Main", "ShowReference", this->showReference);
      parms("Main", "ReferenceVectorSize", this->referenceVectorSize);
      parms("Main", "StopAfterRules", stopAfterRules);
      parms("Main", "StopBeforeRules", stopBeforeRules);
      parms("Main", "TorusWorld", torusWorld);

      parms("View", "BackColor", backColor);
      T.readViewParms(parms, "View");
      this->readMechanicParam(parms, "Mechanic");
      this->readSymbols(parms, "Symbols");
      this->readRules(parms, "Rules");

      readSymbolsParams(parms, "Symbols");

      QString axiom_name;
      parms("Init", "Axiom", axiom_name);
      axiom = this->label_numbers[axiom_name];
      parms("Init", "Reference", init_ref);
      if(this->viewRulesDefinitions)
      {
        cout << "Axiom: " << this->label_names[axiom].toStdString() << endl;
        cout << "Reference vector: " << init_ref << endl;
      }

      readParam(parms);
    }

    virtual void readParam(util::Parms&) {}

    void readSymbolsParams(util::Parms& parms, QString section)
    {
      polygon_sizes.clear();
      symbol_color.clear();
      polygon_sizes.resize(this->label_names.size(), 0);
      symbol_color.resize(this->label_names.size(), 1);
      forall(const QString& symbol, this->label_names)
      {
        int lab = this->label_numbers[symbol];
        QString val;
        //cout << "Parameters of symbol " << this->label_names[lab] << ":";
        if(parms(section, symbol, val))
        {
          QStringList lst = val.split(" ");
          forall(const QString& field, lst)
          {
            QStringList s = field.split("=");
            if(s.size() != 2)
            {
              cout << "Error, field '" << field.toStdString() << ", is not valid" << endl;
              continue;
            }
            if(s[0] == "polygon")
            {
              bool ok;
              int n = s[1].toInt(&ok);
              if(!ok)
              {
                cout << "Cannot read number of sides of polygon ('" << s[1].toStdString() << "')" << endl;
              }
              else
                polygon_sizes[lab] = n;
            }
            else if(s[0] == "color")
            {
              bool ok;
              int n = s[1].toInt(&ok);
              if(!ok)
              {
                cout << "Cannot read color number ('" << s[1].toStdString() << "')" << endl;
              }
              else
                symbol_color[lab] = n;
            }
          }
        }
      }
    }

    void modifiedFiles( const std::set<std::string>& filenames )
    {
      if(filenames.count("view.v"))
        readParms();
      rex.watch(filenames);
    }

    CellSystemModel(QObject *parent)
      : cell_system::CellSystem<TissueClass, RealModel>(parent)
      , rex(this)
      , palette("pal.map")
      , T(ParentClass::T)
    {
      rex.addObject(&palette);
      T.palette = &palette;
      applyRule = false;
    }

    void initDraw()
    {
      glEnable(GL_NICEST);
      glDisable(GL_CULL_FACE);
    }

    void preDraw()
    {
      util::Palette::Color bg = palette.getColor(backColor);
      glClearColor(bg.r(), bg.g(), bg.b(), 1.0);
      T.preDraw();
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    }

    void postDraw()
    {
      glDisable(GL_BLEND);
      T.postDraw();
    }

    void draw(Viewer *viewer)
    {
      qglviewer::Vec pmin(HUGE_VAL, HUGE_VAL, HUGE_VAL), pmax(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
      palette.useColor(wallColor);
      forall_named(const cell& c, T.S, cells)
      {
        util::Palette::Color cell_color;
        if(c->label >= symbol_color.size())
        {
          cell_color = palette.getColor(backColor);
        }
        else
        {
          cell_color = palette.getColor(symbol_color[c->label]);
        }
        T.drawCell(c, cell_color, cell_color);
        forall(const junction& j, T.S.neighbors(c))
        {
          const Point3d& p = j->pos;
          if(p.x() < pmin.x)
            pmin.x = p.x();
          if(p.y() < pmin.y)
            pmin.y = p.y();
          if(p.z() < pmin.z)
            pmin.z = p.z();
          if(p.x() > pmax.x)
            pmax.x = p.x();
          if(p.y() > pmax.y)
            pmax.y = p.y();
          if(p.z() > pmax.z)
            pmax.z = p.z();
        }
        if(showReference)
        {
          glColor3f(1,1,1);
          Point3d s = c->pos;
          Point3d ref = referenceVector(c);
          ref.normalize();
          ref *= referenceVectorSize;
          s -= ref/2;
          Point3d e = s+ref;
          viewer->drawArrow(Vec(s), Vec(e));
        }
      }

      viewer->setSceneBoundingBox(pmin, pmax);
    }

    void updateFromOld( const cell& cl, const cell& cr, const cell& c,
                        const typename TissueClass::division_data& ddata, TissueClass& T)
    {
      ParentClass::updateFromOld(cl, cr, c, ddata, T);

      std::vector<junction> to_clean;
      forall(const junction& j, reference.neighbors(c))
      {
        if(reference.iValence(j) == 1)
          to_clean.push_back(j);
      }
      reference.erase(c);
      forall(const junction& j, to_clean)
      {
        reference.erase(j);
      }
      reference.insert(cl);
      reference.insert(cr);
      forall(const junction& j, T.S.neighbors(cl))
      {
        if(!T.OldS.edge(c, j))
        {
          reference.insert(j);
          reference.insertEdge(cl, j);
          reference.insertEdge(cr, j);
        }
      }
    }

    void insertMissingEdges(const cell& c)
    {
      int size_cell = polygon_sizes[c->label];
      int diff = size_cell - T.S.valence(c);
      if(diff > 0)
      {
        if(this->viewRulesApplication)
        {
          cout << "Cell of label " << this->label_names[c->label].toStdString() << " has only " << T.S.valence(c) << " edges" << endl;
        }
        // Then add the missing number of vertices to an external edge
        // If there are no external edge, then do nothing
        forall(const junction& j, T.S.neighbors(c))
        {
          const junction& jn = T.S.nextTo(c, j);
          if(T.border(j, jn))
          {
            Point3d u = jn->pos - j->pos;
            u /= diff+1;
            junction prev = j;
            for(int i = 0 ; i < diff ; ++i)
            {
              const junction& v = algorithms::insert(prev, jn, T.W);
              v->pos = prev->pos + u;
              T.S.insert(v);
              T.S.spliceBefore(c, jn, v);
              T.S.insertEdge(v, c);
              prev = v;
            }
            break;
          }
        }
      }
    }

    virtual Point3d referenceVector(const cell& c)
    {
      if(reference.valence(c) < 2)
      {
        return init_ref;
      }
      else
      {
        const junction& v = reference.anyIn(c);
        const junction& n = reference.nextTo(c, v);
        Point3d ref = v->pos - n->pos;
        ref = Point3d(ref.y(), -ref.x(), 0);
        if(ref * (v->pos - c->pos) > 0)
        {
          ref *= -1;
        }
        return ref;
      }
    }

    bool apply_division()
    {
      this->step_cellsystem_division();
      if(torusWorld)
      {
        forall_named(const cell& c, T.S, cells)
        {
          insertMissingEdges(c);
        }
      }
      if(stopAfterRules)
      {
        this->stop();
        return true;
      }
      return false;
    }

    void cellsystem_step()
    {
      if(stopBeforeRules and applyRule)
      {
        applyRule = false;
        if(apply_division())
          return;
      }
      if(this->step_cellsystem_meca())
      {
        if(stopBeforeRules)
        {
          cout << "Rule to be applied ..." << endl;
          applyRule = true;
          this->stop();
        }
        else
        {
          apply_division();
        }
      }
    }

    virtual void initialize() {}

    void defineReferenceWall(const cell& c, const junction& v1, const junction& v2)
    {
      reference.insert(c);
      reference.insert(v1);
      reference.insert(v2);
      reference.insertEdge(c, v1);
      reference.insertEdge(c, v2);
    }

    void initReference()
    {
      forall(const cell& c, T.C)
      {
        reference.insert(c);
      }
    }

    void initTissue()
    {
      if(T.C.empty())
      {
        int nb_sides = polygon_sizes[axiom];
        if(nb_sides == 0)
        {
          cout << "You have to specify a number of sides for the axiom" << endl;
          return;
        }
        std::vector<junction> junctions(nb_sides, junction(0));
        for(int i = 0 ; i < nb_sides ; i++)
        {
          junction v;
          v->pos = Point3d(cos(2*M_PI/nb_sides*i), sin(2*M_PI/nb_sides*i), 0);
          junctions[i] = v;
        }

        cell c;
        c->pos = Point3d(0,0,0);

        T.addCell(c, junctions);
        c->label = axiom;

        this->updateCellsArea();
      }
    }

    void step()
    {
      this->cellsystem_step();
    }

  };
}

#ifndef CellAttributes
#  define CellAttributes
#endif

#ifndef JunctionAttributes
#  define JunctionAttributes
#endif

#ifndef CellEdgeAttributes
#  define CellEdgeAttributes
#endif

#ifndef WallAttributes
#  define WallAttributes
#endif

#ifndef CellJunctionEdgeAttributes
#  define CellJunctionEdgeAttributes
#endif

#ifndef JunctionCellEdgeAttributes
#  define JunctionCellEdgeAttributes
#endif

#ifndef ModelInit
#  define ModelInit
#endif

#define StartModel \
  struct CellType : public cell_system::CellSystemCell \
  { \
    CellAttributes\
  }; \
 \
  struct JunctionType : public cell_system::CellSystemJunction \
  { \
    JunctionAttributes\
  }; \
 \
  struct CellEdgeType \
  { \
    bool serialize(storage::VVEStorage&) { return true; } \
    CellEdgeAttributes\
  }; \
 \
  struct WallType \
  { \
    bool serialize(storage::VVEStorage&) { return true; } \
    WallAttributes\
  }; \
 \
  struct CellJunctionEdgeType \
  { \
    bool serialize(storage::VVEStorage&) { return true; } \
    CellJunctionEdgeAttributes\
  }; \
 \
  struct JunctionCellEdgeType \
  { \
    bool serialize(storage::VVEStorage&) { return true; } \
    JunctionCellEdgeAttributes\
  }; \
 \
  class ModelClass; \
  typedef tissue::Tissue<ModelClass, CellType, JunctionType, CellEdgeType, WallType, CellJunctionEdgeType, JunctionCellEdgeType> Tissue; \
 \
  struct ModelClass : public cellsystem_model::CellSystemModel<Tissue,ModelClass> \
  { \
    typedef cellsystem_model::CellSystemModel<Tissue,ModelClass> ParentClass;\
    ModelClass(QObject *parent) \
      : ParentClass(parent) \
      ModelInit \
    { \
      readParms(); \
      registerFile("view.v"); \
      initialize(); \
      initTissue(); \
      initReference(); \
      updateCellsArea(); \
    } \

#define EndModel }; DEFINE_MODEL(ModelClass);

#endif // CELLSYSTEM_MODEL_H
