#ifndef VVELIB_FACTORY_COMPLEX_GRID_H
#define VVELIB_FACTORY_COMPLEX_GRID_H
/**
 * \file complex_grid.h
 *
 * Functions to initialise a complex as a regular grid.
 */

#include <config.h>
#include <util/vector.h>
#include <factory/grid.h>
#include <algorithms/insert.h>
#include <algorithms/complex.h>
#include <cassert>
#include <cmath>

#define TF_CORNER(i,j) corners[(i)*(M+1)+(j)]
#define TF_TOP_CORNER(i,j) top_corner[(((i)>N)?N:(i))*(M+1)+(j)]
#define TF_BOTTOM_CORNER(i,j) bottom_corner[(i)*(M+1)+(j)]

/**
 * \namespace complex_factory
 *
 * Contains functions to initialise a complex.
 */
namespace complex_factory
{
  using algorithms::insertAfter;
  typedef util::Vector<3,double> Point3d;
  typedef util::Vector<2,size_t> Point2u;

  /**
   * For internal use only!
   *
   * Compute the position of each point
   */
  template <typename Complex, class Model>
  struct SquareFiller
  {
    IMPORT_COMPLEX_VERTICES(Complex);
    typedef typename Complex::cell_graph cell_graph;

    SquareFiller(const Point3d& bl,
                 const Point3d& sr,
                 const Point3d& su,
                 Complex& T,
                 Model& model)
      : bottom_left(bl)
      , shift_right(sr)
      , shift_up(su)
      , delta_right(sr/2)
      , delta_up(su/2)
      , T(T)
      , model(model)
    { }

    void initGridVertex(size_t i, size_t j, const cell& v, cell_graph&)
    {
      Point3d pos = (delta_right + (1.0*j)*shift_right) + (delta_up + (1.0*i)*shift_up);
      model.setPosition(v, pos);
      model.saveGridCellPosition(v, Point2u(i,j), T);
    }

    Point3d bottom_left, shift_right, shift_up;
    Point3d delta_right, delta_up;
    Complex& T;
    Model& model;
  };

  /**
   * Create a square filler
   */
  template <typename Complex, class Model>
  SquareFiller<Complex,Model> squareFiller(const Point3d& bottom_left,
                                           const Point3d& shift_right,
                                           const Point3d& shift_up,
                                           Complex& T,
                                           Model& model)
  {
    return SquareFiller<Complex,Model>(bottom_left, shift_right, shift_up, T, model);
  }

  /**
   * Create a square grid with N lines and M columns.
   *
   * \param N Number of lines
   * \param M Number of columns
   * \param G Graph to initialize as a grid
   * \param m Model containing the initialization functions
   * \param bottom_left Position of the bottom-left corner of the bottom_left 
   * cell
   * \param shift_right Vector describing the width of a cell
   * \param shift_up Vector describing the height of a cell
   *
   * \returns True if successful
   *
   * The model should contain this method:
   *  \code
   *   void saveGridCellPosition(const cell&, Point2u pos, Complex&)
   *   \endcode
   *   - Initialize a newly created vertex at line i and column j. This method 
   *   is called only on centers, before the junctions exist.
   *
   *  \code
   *   Point2u gridCellPosition(const cell&, Complex&)
   *   \endcode
   *
   *
   * The neighbors are oriented in that order: bottom -> right -> top -> left
   *
   * With the bottom-left corner being the one of coordinate (0,0).
   *
   * \Note The graph is cleared at the beginning of the process
   */
  template <typename Complex, class Model>
  bool square_grid(size_t N, size_t M, Complex& T, Model& model,
                   const Point3d& bottom_left = Point3d(0,0,0),
                   const Point3d& shift_right = Point3d(1,0,0),
                   const Point3d& shift_up = Point3d(0,1,0))
  {
    IMPORT_COMPLEX_TYPES(Complex);
    SquareFiller<Complex,Model> sf(bottom_left, shift_right, shift_up, T, model);
    if(!factory::square_grid(N, M, T.C, sf))
      return false;
    // Store the bottom-left corner of each cell
    std::vector<junction> corners((N+1)*(M+1), junction(0));

    // Grid of cells
    std::vector<std::vector<cell> > cell_grid(N, std::vector<cell>(M, cell(0)));
    forall(const cell& c, T.C)
    {
      Point2u ij = model.gridCellPosition(c, T);
      cell_grid[ij.x()][ij.y()] = c;
    }

    // Create the corners and connect them to the cells
    size_t k = 0;
    for(size_t i = 0 ; i < N+1 ; ++i)
      for(size_t j = 0 ; j < M+1 ; ++j, ++k)
      {
        Point3d pos = bottom_left + (1.0*i)*shift_up + (1.0*j)*shift_right;
        junction v;
        T.S.insert(v);
        T.W.insert(v);
        model.setPosition(v, pos);
        corners[k] = v;
      }
    // Connect the corners to each other
    const size_t delta_right = 1;
    const size_t delta_up = M+1;
    k = 0;
    for(size_t i = 0 ; i < N+1 ; ++i)
      for(size_t j = 0 ; j < M+1 ; ++j, ++k)
      {
        const junction& v = corners[k];
        junction pv(0);
        // First, connect up
        if(i < N)
        {
          pv = corners[k+delta_up];
          T.W.insertEdge(v, pv);
        }
        // Second, connect left
        if(j > 0)
        {
          pv = corners[k-delta_right];
          T.W.insertEdge(v, pv);
        }
        // Third, connect down
        if(i > 0)
        {
          const junction& nv = corners[k-delta_up];
          insertAfter(v, pv, nv, T.W);
          pv = nv;
        }
        // At last, connect right
        if(j < M)
        {
          const junction& nv = corners[k+delta_right];
          insertAfter(v, pv, nv, T.W);
        }
      }
    // Next, connect the cells to the corners
    for(size_t i = 0 ; i < N ; ++i)
      for(size_t j = 0 ; j < M ; ++j, ++k)
      {
        const cell& c = cell_grid[i][j];
        T.S.insert(c);
        junction c1 = TF_CORNER(i,j);
        junction c2 = TF_CORNER(i,j+1);
        junction c3 = TF_CORNER(i+1,j+1);
        junction c4 = TF_CORNER(i+1,j);
        T.S.insertEdge(c, c1);
        T.S.insertEdge(c, c2);
        T.S.spliceAfter(c, c2, c3);
        T.S.spliceAfter(c, c3, c4);

        if(i == 0)
        {
          T.S.insertEdge(c1, c);
          T.S.insertEdge(c2, c);
        }
        else
        {
          forall(const cell& cc, T.S.neighbors(c1))
          {
            Point2u ij = model.gridCellPosition(cc, T);
            if(ij.x() == i-1 and ij.y() == j)
            {
              T.S.spliceAfter(c1, cc, c);
              T.S.spliceBefore(c2, cc, c);
              break;
            }
          }
        }

        T.S.insertEdge(c3, c);
        T.S.insertEdge(c4, c);

        /*
           T.S.spliceAfter(c1, c2, c);
           T.S.spliceAfter(c2, c3, c);
           T.S.spliceAfter(c3, c4, c);
           T.S.spliceAfter(c4, c1, c);
           */
      }
    return true;
  }

  /**
   * For internal use only!
   *
   * Compute the position of each point
   */
  template <typename Complex, class Model>
  struct HexFiller
  {
    IMPORT_COMPLEX_TYPES(Complex);

    HexFiller(const Point3d& bl,
              const Point3d& u,
              const Point3d& v,
              Complex& T,
              Model& model)
      : bottom_left(bl)
      , shift_right(sqrt(3)*u)
      , shift_up(1.5*v)
      , delta_odd(std::sqrt(3)/2*u)
      , T(T)
      , model(model)
    { }

    void initGridVertex(size_t i, size_t j, const cell& v, cell_graph&)
    {
      Point3d pos = (double)i*shift_up + (double)j*shift_right;
      if(i%2)
        pos += delta_odd;
      model.setPosition(v, pos);
      model.saveGridCellPosition(v, Point2u(i,j), T);
    }

    Point3d bottom_left, shift_right, shift_up;
    Point3d delta_odd;
    Complex& T;
    Model& model;
  };

  /**
   * Connect \c j to \c nj knowing than \c pj is the junction before \c j in 
   * the current cell.
   *
   * \note For internal use only!
   */
  template <class Graph>
  void connectJunction(const typename Graph::vertex_t& j,
                       const typename Graph::vertex_t& nj,
                       const typename Graph::vertex_t& pj,
                       Graph& S,
                       bool direct = true)
  {
    typedef typename Graph::vertex_t vertex;
    if(S.empty(j))
    {
      S.insertEdge(j, nj);
      return;
    }
    if(S.edge(j, pj))
    {
      if(direct)
        S.spliceBefore(j, pj, nj);
      else
        S.spliceAfter(j, pj, nj);
      return;
    }
    const vertex& v = S.anyIn(j);
    S.spliceBefore(j, v, nj);
  }

  /**
   * Create a hexagonal grid with N lines and M columns.
   *
   * \param N Number of lines
   * \param M Number of columns
   * \param G Graph to initialize as a grid
   * \param m Model containing the initialization functions
   * \param bottom_left Position of the center of the bottom-left cell
   * \param u Vector describing the reference system. if u=(1,0,0) and 
   * v=(0,1,0), then the hexagon drawn would be of radius 1 and placed 
   * vertically.
   * \param v Vector describing the reference system.
   *
   * \returns True if successful
   *
   * The model should contain this method:
   *  \code
   *   void saveGridCellPosition(const cell& c, Point2u pos, Complex&)
   *   \endcode
   *   - Initialize a newly created vertex at line i and column j. This method 
   *   is called only on centers, before the junctions exist.
   *
   *  \code
   *  Point2u gridCellPosition(const cell& c, Complex&)
   *  \endcode
   *
   *
   * The neighbors are oriented in that order: bottom -> right -> top -> left
   *
   * With the bottom-left corner being the one of coordinate (0,0).
   *
   * \Note The graph is cleared at the beginning of the process
   */
  template <typename Complex, class Model>
  bool hex_grid(size_t N, size_t M, Complex& T, Model& model,
                const Point3d& bottom_left = Point3d(0,0,0),
                const Point3d& u = Point3d(1,0,0),
                const Point3d& v = Point3d(0,1,0))
  {
    IMPORT_COMPLEX_TYPES(Complex);
    HexFiller<Complex,Model> sf(bottom_left, u, v, T, model);
    if(!factory::hex_grid(N, M, T.C, sf))
      return false;
    size_t nb_corners = (M+1)*(N+1)-1;
    // Bottom-left and top-left corners
    std::vector<junction> bottom_corner(nb_corners,junction(0)), top_corner(nb_corners, junction(0));
    size_t k = 0;
    Point3d shift_top_left = bottom_left + v*0.5 - u*std::sqrt(3)/2;
    Point3d shift_bottom_left = bottom_left - v*0.5 - u*std::sqrt(3)/2;
    Point3d delta_pos_j = std::sqrt(3)*u;
    Point3d delta_pos_i = 1.5*v;
    Point3d odd_shift_pos = std::sqrt(3)/2*u;

    // Grid of cells
    std::vector<std::vector<cell> > cell_grid(N, std::vector<cell>(M, cell(0)));
    forall(const cell& c, T.C)
    {
      Point2u ij = model.gridCellPosition(c, T);
      cell_grid[ij.x()][ij.y()] = c;
    }

    for(size_t i = 0 ; i < N+1 ; ++i)
      for(size_t j = 0 ; j < M+1 ; ++j, ++k)
      {
        if(i == N and j == M)
          continue;
        junction top;
        junction bottom;
        T.W.insert(top); T.W.insert(bottom);
        T.S.insert(top); T.S.insert(bottom);
        Point3d delta_pos;
        delta_pos = (double)j*delta_pos_j + (double)i*delta_pos_i;
        if(i%2)
          delta_pos += odd_shift_pos;
        else if(i == N)
          delta_pos += 2*odd_shift_pos;
        Point3d pos_top;
        if(i == N)
        {
          pos_top = (double)j*delta_pos_j - v;
        }
        else
          pos_top = shift_top_left + delta_pos;
        Point3d pos_bottom = shift_bottom_left + delta_pos;
        model.setPosition(top, pos_top);
        model.setPosition(bottom, pos_bottom);
        bottom_corner[k] = bottom;
        top_corner[k] = top;
      }

    // Connect everything
    for(size_t i = 0 ; i < N ; ++i)
      for(size_t j = 0 ; j < M ; ++j)
      {
        const cell& c = cell_grid[i][j];
        T.S.insert(c);
        const junction& c1 = TF_TOP_CORNER(i,j);
        const junction& c2 = TF_BOTTOM_CORNER(i,j);
        const junction& c3 = (i%2)?TF_TOP_CORNER(i-1,j+1):TF_TOP_CORNER(i-1,j);
        const junction& c4 = TF_BOTTOM_CORNER(i,j+1);
        const junction& c5 = TF_TOP_CORNER(i,j+1);
        const junction& c6 = (i%2)?TF_BOTTOM_CORNER(i+1,(i==N-1)?j:j+1):TF_BOTTOM_CORNER(i+1,j);
        // Connect the junctions to each other
        connectJunction(c1, c2, c6, T.W);
        connectJunction(c2, c3, c1, T.W);
        connectJunction(c3, c4, c2, T.W);
        connectJunction(c4, c5, c3, T.W);
        connectJunction(c5, c6, c4, T.W);
        connectJunction(c6, c1, c5, T.W);

        if(i == 0)
        {
          connectJunction(c3, c2, c4, T.W, false);
          connectJunction(c4, c3, c5, T.W, false);
        }
        if(j == 0)
        {
          if(i%2 == 0)
          {
            if(i != 0)
              connectJunction(c3, c2, c4, T.W, false);
            if(i != N-1)
              connectJunction(c1, c6, c2, T.W, false);
          }
          connectJunction(c2, c1, c3, T.W, false);
        }
        if(i == N-1)
        {
          connectJunction(c1, c6, c2, T.W, false);
          connectJunction(c6, c5, c1, T.W, false);
        }
        if(j == M-1)
        {
          if(i%2)
          {
            if(i != N-1)
              connectJunction(c6, c5, c1, T.W, false);
            if(i != 0)
              connectJunction(c4, c3, c5, T.W, false);
          }
          connectJunction(c5, c4, c6, T.W, false);
        }

        // Connect the center to the junctions
        T.S.insertEdge(c, c1);
        T.S.insertEdge(c, c2);
        T.S.spliceAfter(c, c2, c3);
        T.S.spliceAfter(c, c3, c4);
        T.S.spliceAfter(c, c4, c5);
        T.S.spliceAfter(c, c5, c6);

        // Connect the junctions to the center
        if(i == 0 or j == 0)
        {
          T.S.insertEdge(c1, c);
          T.S.insertEdge(c2, c);
        }
        else
        {
          const cell& nc = T.S.anyIn(c1);
          T.S.spliceAfter(c1, nc, c);
          T.S.spliceBefore(c2, nc, c);
        }
        {
          const cell& nc = T.S.anyIn(c4);
          if(nc)
            T.S.spliceAfter(c3, nc, c);
          else
            T.S.insertEdge(c3, c);
        }
        T.S.insertEdge(c4, c);
        T.S.insertEdge(c5, c);
        T.S.insertEdge(c6, c);
      }
    return true;
  }
}

#undef TF_CORNER
#undef TF_TOP_CORNER
#undef TF_BOTTOM_CORNER
#endif // VVELIB_FACTORY_COMPLEX_GRID_H
