#ifndef VVELIB_FACTORY_GRID_H
#define VVELIB_FACTORY_GRID_H
/**
 * \file grid.h
 *
 * Contains factories to generate grids
 */

#include <config.h>
#include <list>

/**
 * Contains predefined ways to initialize a graph
 */
namespace factory
{
  /**
   * Create a square grid with N lines and M columns.
   *
   * \param N Number of lines
   * \param M Number of columns
   * \param G Graph to initialize as a grid
   * \param m Model containing the initialization functions
   * \param torus Connect the top cells to the bottom ones and the left cells 
   * to the right ones if true.
   * \param connect8 Set to true to get a 8-connected grid, false to get 
   * a 4-connected grid.
   *
   * \returns True if successful
   *
   * The model should contain this method:
   *
   *  \code
   *  void initGridVertex(size_t i, size_t j, const vertex&, Graph&)
   *  \endcode
   *
   *  Initialize a newly created vertex at line i and column j
   *
   * The neighbors are oriented in that order: bottom -> right -> top -> left
   *
   * With the bottom-left corner being the one of coordinate (0,0).
   *
   * \Note The graph is cleared at the beginning of the process
   */
  template <typename Graph, class Model>
  bool square_grid(size_t N, size_t M, Graph& G, Model& model, bool torus = false, bool connect8 = false)
  {
    typedef typename Graph::vertex_t vertex;
    typedef typename Graph::edge_t edge;
    // First, clear the graph
    G.clear();
    // Array of vertices by position
    std::vector<vertex> vertices(M*N, vertex(0));
    // Create the vertices and insert them
    for(size_t k = 0 ; k < M*N ; ++k)
    {
      vertex v;
      G.insert(v);
      vertices[k] = v;
    }
    // Create the edges
    size_t k = 0;
    const size_t delta_border_i = M*(N-1);
    const size_t delta_border_j = M-1;
    const size_t delta_i = M;
    const size_t delta_j = 1;
    for(size_t i = 0 ; i < N ; ++i)
      for(size_t j = 0 ; j < M ; ++j, ++k)
      {
        vertex v = vertices[k];
        // Connect to the right
        if(j == M-1)
        {
          if(torus)
            G.insertEdge(v, vertices[k-delta_border_j]);
        }
        else
          G.insertEdge(v, vertices[k+delta_j]);
        // Connect to the top
        if(i == N-1)
        {
          if(torus)
            G.insertEdge(v, vertices[k-delta_border_i]);
        }
        else
          G.insertEdge(v, vertices[k+delta_i]);
        // Connect to the left
        if(j == 0)
        {
          if(torus)
          {
            if(i==N-1)
              G.spliceAfter(v, vertices[k-delta_border_i], vertices[k+delta_border_j]);
            else
              G.spliceAfter(v, vertices[k+delta_i], vertices[k+delta_border_j]);
          }
        }
        else
        {
          if(i==N-1)
          {
            if(j == M-1)
              G.insertEdge(v, vertices[k-delta_j]);
            else
              G.spliceAfter(v, vertices[k+delta_j], vertices[k-delta_j]);
          }
          else
            G.spliceAfter(v, vertices[k+delta_i], vertices[k-delta_j]);
        }
        // Connect to the bottom
        if(i == 0)
        {
          if(torus)
          {
            if(j==0)
              G.spliceAfter(v, vertices[k+delta_border_j], vertices[k+delta_border_i]);
            else
              G.spliceAfter(v, vertices[k-delta_j], vertices[k+delta_border_i]);
          }
        }
        else
        {
          if(j==0)
            G.spliceBefore(v, vertices[k+delta_j], vertices[k-delta_i]);
          else
            G.spliceAfter(v, vertices[k-delta_j], vertices[k-delta_i]);
        }
      }
    if(connect8)
    {
      const size_t delta_ij = M+1;
      const size_t delta_mij = M-1;
      const size_t bottom_left_corner = 0;
      const size_t bottom_right_corner = M-1;
      const size_t top_left_corner = M*(N-1);
      const size_t top_right_corner = M*N-1;
      const size_t delta_border_bottom_ij = M*(N-1)-1;
      const size_t delta_border_right_ij = 1;
      const size_t delta_border_bottom_mij = M*(N-1)+1;
      const size_t delta_border_left_mij = 2*M-1;
      k = 0;
      for(size_t i = 0 ; i < N ; ++i)
        for(size_t j = 0 ; j < M ; ++j, ++k)
        {
          vertex v = vertices[k];
          // ** Connect to the bottom-left
          if(i == 0 && j == 0)
          {
            // If in the corner
            if(torus)
              G.spliceAfter(v, vertices[bottom_right_corner], vertices[top_right_corner]);
          }
          else if(i == 0)
          {
            // If on the bottom line
            if(torus)
              G.spliceAfter(v, vertices[k-delta_j], vertices[k+delta_border_bottom_ij]);
          }
          else if(j == 0)
          {
            // If on the left border
            if(torus)
              G.spliceBefore(v, vertices[k-delta_i], vertices[k-delta_border_right_ij]);
          }
          else
            G.spliceAfter(v, vertices[k-delta_j], vertices[k-delta_ij]);

          // ** Connect the bottom-right
          if(i == 0 && j == M-1)
          {
            // If in the corner
            if(torus)
              G.spliceAfter(v, vertices[top_right_corner], vertices[top_left_corner]);
          }
          else if(i == 0)
          {
            // If on the bottom line
            if(torus)
              G.spliceBefore(v, vertices[k+delta_j], vertices[k+delta_border_bottom_mij]);
          }
          else if(j == M-1)
          {
            // If on the right line
            if(torus)
              G.spliceAfter(v, vertices[k-delta_i], vertices[k-delta_border_left_mij]);
          }
          else
            G.spliceAfter(v, vertices[k-delta_i], vertices[k-delta_mij]);

          // ** Connect to the top-right
          if(i == N-1 && j == M-1)
          {
            // If on the corner
            if(torus)
              G.spliceAfter(v, vertices[top_left_corner], vertices[bottom_left_corner]);
          }
          else if(i == N-1)
          {
            // If on the top line
            if(torus)
              G.spliceAfter(v, vertices[k+delta_j], vertices[k-delta_border_bottom_ij]);
          }
          else if(j == M-1)
          {
            // If on the right line
            if(torus)
              G.spliceBefore(v, vertices[k+delta_i], vertices[k+delta_border_right_ij]);
          }
          else
            G.spliceAfter(v, vertices[k+delta_j], vertices[k+delta_ij]);

          // ** Connect to the top-left
          if(i == N-1 && j == 0)
          {
            // If on the bottom-left corner
            if(torus)
              G.spliceAfter(v, vertices[bottom_left_corner], vertices[bottom_right_corner]);
          }
          else if(i == N-1)
          {
            // If on the top line
            if(torus)
              G.spliceBefore(v, vertices[k-delta_j], vertices[k-delta_border_bottom_mij]);
          }
          else if(j == 0)
          {
            // If on the left line
            if(torus)
              G.spliceAfter(v, vertices[k+delta_i], vertices[k+delta_border_left_mij]);
          }
          else
            G.spliceBefore(v, vertices[k-delta_j], vertices[k+delta_mij]);
        }
    }
    k = 0;
    for(size_t i = 0 ; i < N ; ++i)
      for(size_t j = 0 ; j < M ; ++j, ++k)
      {
        model.initGridVertex(i, j, vertices[k], G);
      }
    return true;
  }

  /**
   * Build a hexagonal grid with N lines and M columns.
   *
   * \param N Number of lines of the grid
   * \param M Number of columns of the grid
   * \param G Graph to fill in
   * \param model Model containing initialisation method
   * \param torus True to connect the top cells to the bottom ones and the left 
   * cells to the right ones. Requires an even number of lines.
   *
   * \returns True if successful
   */
  template <typename Graph, class Model>
  bool hex_grid(size_t N, size_t M, Graph& G, Model& model, bool torus = false)
  {
    if(torus and N%2)
      return false;
    typedef typename Graph::vertex_t vertex;
    typedef typename Graph::edge_t edge;
    // First, clear the graph
    G.clear();
    // Array of vertices by position
    std::vector<vertex> vertices(M*N, vertex(0));
    // Create the vertices and insert them
    for(size_t k = 0 ; k < M*N ; ++k)
    {
      vertex v;
      G.insert(v);
      vertices[k] = v;
    }
    // Create the edges
    size_t k = 0;
    const size_t delta_right = 1;
    const size_t delta_top_right_even = M;
    const size_t delta_top_right_odd = M+1;
    const size_t delta_top_left_even = M-1;
    const size_t delta_top_left_odd = M;
    const size_t delta_bottom_border_bottom_right = M*(N-1);
    const size_t delta_bottom_border_bottom_left = M*(N-1)-1;
    const size_t delta_right_border_top_right = 1;
    const size_t delta_left_border_left = M-1;
    const size_t delta_left_border_top_left = 2*M-1;
    const size_t top_right_corner = M*N-1;
    const size_t bottom_left_corner = 0;
    for(size_t i = 0 ; i < N ; ++i)
      for(size_t j = 0 ; j < M ; ++j, ++k)
      {
        vertex v = vertices[k];
        // ** Connect to the right
        if(j == M-1)
        {
          if(torus)
            G.insertEdge(v, vertices[k-delta_left_border_left]);
        }
        else
          G.insertEdge(v, vertices[k+delta_right]);
        // ** Connect to the left
        if(j==0)
        {
          if(torus)
            G.insertEdge(v, vertices[k+delta_left_border_left]);
        }
        else
          G.insertEdge(v, vertices[k-delta_right]);
        // ** Connect to the top-right
        if(k == top_right_corner)
        {
          if(torus)
            G.spliceAfter(v, vertices[k-delta_left_border_left], vertices[bottom_left_corner]);
        }
        else if(i == N-1)
        {
          if(torus)
            G.spliceAfter(v, vertices[k+delta_right], vertices[k-delta_bottom_border_bottom_left]);
        }
        else if(i%2)
        {
          if(j == M-1)
          {
            if(torus)
              G.spliceAfter(v, vertices[k-delta_left_border_left], vertices[k+delta_right_border_top_right]);
          }
          else
            G.spliceAfter(v, vertices[k+delta_right], vertices[k+delta_top_right_odd]);
        }
        else if(j == M-1)
          G.spliceBefore(v, vertices[k-delta_right], vertices[k+delta_top_right_even]);
        else
          G.spliceAfter(v, vertices[k+delta_right], vertices[k+delta_top_right_even]);
        // ** Connect to the top-left
        if(k == top_right_corner)
        {
          if(torus)
            G.spliceAfter(v, vertices[bottom_left_corner], vertices[k-delta_bottom_border_bottom_right]);
        }
        else if(i == N-1)
        {
          if(torus)
            G.spliceBefore(v, vertices[k-delta_right], vertices[k-delta_bottom_border_bottom_right]);
        }
        else if(i%2)
        {
          if(j == 0)
            G.spliceAfter(v, vertices[k+delta_top_right_odd], vertices[k+delta_top_left_odd]);
          else
            G.spliceBefore(v, vertices[k-delta_right], vertices[k+delta_top_left_odd]);
        }
        else
        {
          if(j == 0)
          {
            if(torus)
              G.spliceAfter(v, vertices[k+delta_top_right_even], vertices[k+delta_left_border_top_left]);
          }
          else
            G.spliceBefore(v, vertices[k-delta_right], vertices[k+delta_top_left_even]);
        }
        // ** Connect to the bottom-left
        if(k == bottom_left_corner)
        {
          if(torus)
            G.spliceBefore(v, vertices[k+delta_right], vertices[top_right_corner]);
        }
        else if(i == 0)
        {
          if(torus)
            G.spliceAfter(v, vertices[k-delta_right], vertices[k+delta_bottom_border_bottom_left]);
        }
        else if(i%2)
        {
          if(j==0)
            G.spliceBefore(v, vertices[k+delta_right], vertices[k-delta_top_right_even]);
          else
            G.spliceAfter(v, vertices[k-delta_right], vertices[k-delta_top_right_even]);
        }
        else if(j == 0)
        {
          if(torus)
            G.spliceBefore(v, vertices[k+delta_right], vertices[k-delta_right_border_top_right]);
        }
        else
          G.spliceAfter(v, vertices[k-delta_right], vertices[k-delta_top_right_odd]);
        // ** Connect to the bottom-right
        if(k == bottom_left_corner)
        {
          if(torus)
            G.spliceAfter(v, vertices[top_right_corner], vertices[k+delta_bottom_border_bottom_right]);
        }
        else if(i == 0)
        {
          if(torus)
            G.spliceAfter(v, vertices[k+delta_bottom_border_bottom_left], vertices[k+delta_bottom_border_bottom_right]);
        }
        else if(i%2)
        {
          if(j==M-1)
          {
            if(torus)
              G.spliceBefore(v, vertices[k-delta_left_border_left], vertices[k-delta_left_border_top_left]);
          }
          else
            G.spliceBefore(v, vertices[k+delta_right], vertices[k-delta_top_left_even]);
        }
        else if(j == M-1)
          G.spliceAfter(v, vertices[k-delta_top_right_odd], vertices[k-delta_top_left_odd]);
        else
          G.spliceBefore(v, vertices[k+delta_right], vertices[k-delta_top_left_odd]);
      }
    k = 0;
    for(size_t i = 0 ; i < N ; ++i)
      for(size_t j = 0 ; j < M ; ++j, ++k)
      {
        model.initGridVertex(i, j, vertices[k], G);
      }
    return true;
  }
}

#endif // VVELIB_FACTORY_GRID_H
