#ifndef VVELIB_FACTORY_OBJREADER_H
#define VVELIB_FACTORY_OBJREADER_H
/**
 * \file objreader.h
 *
 * Contains factories to generate a cell complex from an obj file
 */

#include <config.h>
#include <string>
#include <util/vector.h>
#include <QIODevice>
#include <QTextStream>
#include <QString>
#include <QFile>
#include <iostream>
#include <algorithms/complex.h>

namespace complex_factory
{
  typedef util::Vector<3,double> Point3d;
  using std::cerr;
  using std::endl;

  /**
   * \class ObjReaderError objreader.h <factory/objreader.h>
   *
   * Error object returned by the complex_factory::objreader functions.
   */
  class ObjReaderError
  {
  public:
    /**
     * Type of the error
     */
      enum Type
      {
        NO_ERRORS, // Due to windows crap, NO_ERROR is not usable (but equivalent?)
        BAD_NUMBER_OF_DIMENSIONS,
        FILE_FORMAT,
        DOUBLE_EXPECTED,
        INTEGER_EXPECTED,
        INVALID_FACE,
        CANNOT_OPEN_FILE
      };

      ObjReaderError() {}
      ObjReaderError(Type t);
      ObjReaderError(Type t, QString msg);
      ObjReaderError(const ObjReaderError& copy)
        : _string(copy._string)
        , _type(copy._type)
      {}

      /**
       * Returns a human-readable string describing the error
       */
      virtual QString string() { return _string; }

      /**
       * Returns the type of the error
       */
      virtual int type() { return _type; }

      /**
       * Convert to true if the object correspond to an error
       */
      operator bool() { return _type != NO_ERRORS; }

  protected:
      void initFromType();

      QString _string;
      Type _type;
  };

  /**
   * Read a Wavefront OBJ file describing a surface and return the complex 
   * representing it.
   *
   * \param content Stream containing the OBJ file
   * \param T Complex to fill with the OBJ file mesh
   * \param model Object containing needed methods
   *
   * This version interpret vertices, faces and normals on vertices.
   *
   * The \c model object should include, in addition to the methods required by 
   * the complex object, the following methods:
   * \code
   * void setVertexId(const junction& v, int id);
   * \endcode
   * Associate the identifier \c id with the vertex \c v. The identifiers for 
   * faces and junctions are independant. Typically, the id will be store in 
   * the vertex itself.
   *
   * \code
   * void setFaceId(const cell& f, int id);
   * \endcode
   * Associate the identifier \c id with the face \c f. The identifiers for 
   * faces and junctions are independant. Typically, the id will be store in 
   * the face itself.
   *
   * \code
   * int vertexId(const junction& v);
   * \endcode
   * Returns the identifier that was previously associated with the vertex \c 
   * v.
   *
   * \code
   * int faceId(const cell& f);
   * \endcode
   * Returns the identifier that was previously associated with the face \c f.
   *
   * \code
   * void setNormal(const cell& c, const Point3d& normal);
   * \endcode
   * Set the normal of the face \c v to \c normal. The normal is just for information, and do not have to be used.
   *
   * \code
   * void setNormal(const junction& v, const Point3d& normal);
   * \endcode
   * Set the normal of the vertex \c v to \c normal. The normal is just for 
   * information, and do not have to be used.
   */
  template <class CellComplex, class Model>
  ObjReaderError objreader(QTextStream& content, CellComplex& T, Model& model)
  {
    IMPORT_COMPLEX_TYPES(CellComplex);

    std::vector<Point3d> vertices_pos;
    std::vector<Point3d> normals;
    std::vector<std::vector<int> > faces_idx;

    QString token;

    while(!content.atEnd())
    {
      content >> token;
      QString buffer = content.readLine();
//      cerr << "Reading token '" << token.toStdString() << "'" << endl;
//      cerr << "Buffer: '" << buffer.toStdString() << "'" << endl;
      if(token == QString("v"))
      {
        QList<QString> pos = buffer.split(' ', QString::SkipEmptyParts);
        if(pos.size() != 3)
        {
          QString msg = "Error, vertex " + QString::number(vertices_pos.size()) +
            " has " + QString::number(pos.size()) +  " dimensions instead of 3";
          return ObjReaderError(ObjReaderError::BAD_NUMBER_OF_DIMENSIONS, msg);
        }
        Point3d p;
        bool ok, ok_tmp;
        p.x() = pos[0].toDouble(&ok);
        p.y() = pos[1].toDouble(&ok_tmp);
        ok &= ok_tmp;
        p.z() = pos[2].toDouble(&ok_tmp);
        ok &= ok_tmp;
        if(!ok)
        {
          QString msg = "Error on vertex " + QString::number(vertices_pos.size()) +
            ": double value expected";
          return ObjReaderError(ObjReaderError::DOUBLE_EXPECTED, msg);
        }
        vertices_pos.push_back(p);
      }
      else if(token == QString("vn"))
      {
        QList<QString> pos = buffer.split(' ');
        if(pos.size() != 3)
        {
          QString msg = "Error, normal " + QString::number(normals.size()) + " has " +
            QString::number(pos.size()) + " dimensions instead of 3";
          return ObjReaderError(ObjReaderError::BAD_NUMBER_OF_DIMENSIONS, msg);
        }
        Point3d p;
        bool ok, ok_tmp;
        p.x() = pos[0].toDouble(&ok);
        p.y() = pos[1].toDouble(&ok_tmp);
        ok &= ok_tmp;
        p.z() = pos[2].toDouble(&ok_tmp);
        ok &= ok_tmp;
        if(!ok)
        {
          QString msg = "Error on normal " + QString::number(vertices_pos.size()) + ": double value expected";
          return ObjReaderError(ObjReaderError::DOUBLE_EXPECTED, msg);
        }
        normals.push_back(p);
      }
      else if(token == QString("f"))
      {
        std::vector<int> f;
        QList<QString> face = buffer.split(' ', QString::SkipEmptyParts);
        if(face.size() < 3)
        {
          QString msg = "Error, face " + QString::number(faces_idx.size()) + " has less than 3 vertices.";
          return ObjReaderError(ObjReaderError::INVALID_FACE, msg);
        }
        bool ok = true;
        forall(QString& value, face)
        {
          int vn = value.toInt(&ok);
          if(!ok)
          {
            QString msg = "Error on face " + QString::number(faces_idx.size()) + ": integer values expected";
            return ObjReaderError(ObjReaderError::INTEGER_EXPECTED, msg);
          }
          f.push_back(vn);
        }
        faces_idx.push_back(f);
      }
    }
    if(normals.size() > vertices_pos.size())
    {
      cerr << "Warning: more normals than vertices" << endl;
    }
    if(!normals.empty() and normals.size() < vertices_pos.size())
    {
      cerr << "Warning: not all vertices have a normal, ignoring the specified normals." << endl;
      normals.clear();
    }
    // First, add the vertices
    std::vector<junction> vertices;
    std::vector<cell> faces;
    int id = 0;
    forall(const Point3d& pos, vertices_pos)
    {
      junction v;
      T.S.insert(v);
      T.W.insert(v);
      model.setPosition(v, pos);
      vertices.push_back(v);
      model.setVertexId(v, ++id);
    }
    // Then add the faces and connect the center to the junctions (in the 
    // correct order)
    id = 0;
    forall(const std::vector<int>& face_idx, faces_idx)
    {
      cell face;
      junction pf(0);
      T.S.insert(face);
      T.C.insert(face);
      Point3d center;
      forall(int i, face_idx)
      {
        center += vertices_pos[i-1];
        junction current = vertices[i-1];
        if(pf)
        {
          T.S.spliceAfter(face, pf, current);
        }
        else
          T.S.insertEdge(face, current);
        pf = current;
      }
      center /= face_idx.size();
      model.setPosition(face, center);
      model.setFaceId(face, ++id);
    }
    // Connect the vertices to the faces
    forall_named(const junction& j, T.S, junctions)
    {
      if(T.S.iValence(j) < 3)
      {
        // There is nothing to sort ...
        forall(const cell& f, T.S.iNeighbors(j))
        {
          T.S.insertEdge(j, f);
        }
      }
      else
      {
        const cell& f_init = T.S.iAnyIn(j);
        cell pf = f_init;
        T.S.insertEdge(j, f_init);
        do
        {
          cell f(0);
          const junction& m = T.S.prevTo(pf, j);
          // Now, find the face that contains also m
          forall(const cell& of, T.S.iNeighbors(j))
          {
            if(of == pf)
              continue;
            if(T.S.edge(of, m))
            {
              f = of;
              break;
            }
          }
          if(!f or f == f_init)
          {
            break;
          }
          // Insert f after pf in j
          T.S.spliceAfter(j, pf, f);
          pf = f;
        } while( T.S.iValence(j) != T.S.valence(j));
        pf = f_init;
        // If we didn't find all the neighbors, it means we have to go the 
        // other way around
        while(T.S.iValence(j) != T.S.valence(j))
        {
          cell f(0);
          const junction& m = T.S.nextTo(pf, j);
          forall(const cell& of, T.S.iNeighbors(j))
          {
            if(of == pf)
              continue;
            if(T.S.edge(of, m))
            {
              f = of;
              break;
            }
          }
          if(!f)
            break;
          T.S.spliceBefore(j, pf, f);
          pf = f;
        }
        if(T.S.valence(j) != T.S.iValence(j))
        {
          cerr << "Error, vertex " << model.vertexId(j) << " was not properly connected to the faces" << endl;
        }
      }
    }
    // Connect the junctions
    forall_named(const junction& j, T.S, junctions)
    {
      junction prev(0);
      forall(const cell& c, T.S.neighbors(j))
      {
        const junction& jn = T.S.nextTo(c, j);
        const junction& jp = T.S.prevTo(c, j);
        if(prev.isNull())
        {
          T.W.insertEdge(j, jn);
          prev = jn;
        }
        else if(prev != jn)
        {
          T.W.spliceAfter(j, prev, jn);
        }
        T.W.spliceAfter(j, jn, jp);
        prev = jp;
      }
    }
    // if the normals aren't specified, then compute them
    if(normals.empty())
    {
      forall_named(const junction& v, T.S, junctions)
      {
        Point3d pos = model.position(v);
        Point3d normal;
        if(T.border(v))
        {
          // If this is a border index, first find the border
          forall(const junction& n, T.W.neighbors(v))
          {
            const junction& m = T.W.nextTo(v,n);
            if(T.border(v, m) and T.border(v, n))
            {
              junction n1 = m;
              do
              {
                const junction& n2 = T.W.nextTo(v, n1);
                Point3d p1 = model.position(n1) - pos;
                Point3d p2 = model.position(n2) - pos;
                normal += p1^p2;
                n1 = n2;
              } while(n1 != n);
              break;
            }
          }
        }
        else
        {
          // If this is an inside vertex
          forall(const junction& n, T.W.neighbors(v))
          {
            Point3d p1 = model.position(n) - pos;
            Point3d p2 = model.position(T.W.nextTo(v, n)) - pos;
            normal += p1^p2;
          }
        }
        normal.normalize();
        model.setNormal(v, normal);
      }
    }
    else
    {
      int k = 0;
      forall(const Point3d& n, normals)
      {
        junction v = vertices[k++];
        model.setNormal(v, n);
      }
    }
    // Compute normals for the faces
    forall_named(const cell& f, T.S, cells)
    {
      Point3d pos = model.position(f);
      Point3d normal;
      forall(const junction& v, T.S.neighbors(f))
      {
        const junction& v1 = T.S.nextTo(f, v);
        Point3d p1 = model.position(v) - pos;
        Point3d p2 = model.position(v1) - pos;
        normal += p1^p2;
      }
      normal.normalize();
      model.setNormal(f, normal);
    }
    // At last, connect the cell graph
    forall_named(const cell& f, T.S, cells)
    {
      cell pf(0);
      forall(const junction& j, T.S.neighbors(f))
      {
        const cell& f2 = T.adjacentCell(f, j);
        if(f2)
        {
          if(pf)
          {
            T.C.spliceAfter(f, pf, f2);
          }
          else
          {
            T.C.insertEdge(f, f2);
          }
          pf = f2;
        }
      }
    }
    return ObjReaderError(ObjReaderError::NO_ERRORS);
  }

  /**
   * Convenience methods that read a Wavefront OBJ file
   *
   * \see objreader(QTextStream&,CellComplex&,Model&)
   */
  template <class CellComplex, class Model>
  ObjReaderError objreader(const QString& filename, CellComplex& T, Model& model)
  {
    QFile f(filename);
    if(!f.open(QIODevice::ReadOnly))
    {
      return ObjReaderError(ObjReaderError::CANNOT_OPEN_FILE, f.errorString());
    }
    QTextStream ts(&f);
    return objreader(ts, T, model);
  }

  /**
   * Convenience methods that read a Wavefront OBJ file
   *
   * \see objreader(QTextStream&,CellComplex&,Model&)
   */
  template <class CellComplex, class Model>
  ObjReaderError objreader(std::string filename, CellComplex& T, Model& model)
  {
    QString fn = QString::fromLocal8Bit(filename.c_str());
    return objreader(fn, T, model);
  }

}
#endif // VVELIB_FACTORY_OBJREADER_H
