#ifndef GEOMETRY_COORDINATES_H
#define GEOMETRY_COORDINATES_H

/**
 * \file coordinates.h
 *
 * Includes functions to convert between different type of coordinate systems
 */

#include <config.h>
#include <geometry/geometry.h>

namespace geometry
{
  /**
   * Convert cartesian coordinates to barycentric coordinates in a triangle
   *
   * The triangle is defined by \c p1, \c p2, \c p3 and the result is such 
   * that: pos = result[0]*p1 + result[1]*p2 + result[2]*p3
   */
  Point3d cartesianToBarycentric(const Point3d& pos,
                                 const Point3d& p1, const Point3d& p2, const Point3d& p3);

  /**
   * Convert cartesian coordinates to barycentric coordinates in a triangle
   *
   * This version simply returns the matrix that convert from cartesian to 
   * barycentric. The resulting matrix has to be used like this:
   * \code
   *   Point3d pos(x,y,z);
   *   int index_one;
   *   Matrix3d mat = cartesianToBarycentric_matrix(p1, p2, p3, index_one);
   *   pos[index_one] = 1;
   *   Point3d bar = mat*pos;
   * \endcode
   *
   * \param index_one Index of the position that has to be set to one.
   */
  Matrix3d cartesianToBarycentric_matrix(const Point3d& p1, const Point3d& p2, const Point3d& p3,
                                         int& index_one);

  /**
   * Convert barycentric coordinates to cartesian
   *
   * The triangle is defined by \c p1, \c p2, \c p3 and the result is such 
   * that: result = barycentric[0]*p1 + barycentric[1]*p2 + barycentric[2]*p3
   */
  Point3d barycentricToCartesian(const Point3d& barycentric,
                                 const Point3d& p1, const Point3d& p2, const Point3d& p3);

  /**
   * Convert cartesian coordinates to barycentric coordinates in a triangle
   *
   * This version simply returns the matrix that convert from cartesian to 
   * barycentric.
   */
  Matrix3d barycentricToCartesian_matrix(const Point3d& p1, const Point3d& p2, const Point3d& p3);
}

#endif // GEOMETRY_COORDINATES_H

