#ifndef VVELIB_GEOMETRY_INTERSECTION_H
#define VVELIB_GEOMETRY_INTERSECTION_H

/**
 * \file intersection.h
 *
 * Algorithms to compute intersections.
 */

#include <config.h>
#include <geometry/geometry.h>
#include <vector>

namespace geometry
{

  /**
   * Plane-Line Intersection.
   *
   * This function compute the intersection of the line <tt>[u1,u2]</tt> and the 
   * plane defined by the point \p p and the normal \p n.
   *
   * \param[out] u Intersection point if any
   * \param[out] s Position of the intersection on the line. Between 0 and 1, 
   * the intersection is on the segment. Greater than 0, the position is on the 
   * ray starting on \p u1 and going through \p u2.
   * \param[in] p Point contained in the plane
   * \param[in] n Normal of the plane
   * \param[in] u1 Starting point of the segment
   * \param[in] u2 Ending point of the segment
   *
   * \returns True if there is an intersection.
   */
  bool planeLineIntersection(Point3d &u, double &s, 
                             const Point3d& p, const Point3d& n, const Point3d& u1, const Point3d& u2);


  /**
   * Point in triangle test (2D)
   *
   * Test if the point \p p is in the triangle (\p p1,\p p2,\p p3)
   *
   * \bug Not working properly at all !
   */
  bool pointInTriangle(const Point2d& p,
                       const Point2d& p1, const Point2d& p2, const Point2d& p3);


  /**
   * Point in triangle test (3D)
   *
   * \param u modified to the projection of the point in the triangle plane and 
   * \param s the square distance from the point to the plane.
   * \param p Point to test
   * \param p1 First point of the triangle
   * \param p2 Second point of the triangle
   * \param p3 Third point of the triangle
   *
   * Test if the point \p p is projected in the triangle (\p p1,\p p2,\p p3).
   */
  bool pointInTriangle(Point3d& u, double &s,
                       const Point3d& p, const Point3d& p1, const Point3d& p2, const Point3d& p3);

  /**
   * Line-triangle intersection
   *
   * \param[out] u Intersection point if any
   * \param[out] s Position of the intersection on the line. Between 0 and 1, 
   * the intersection is on the segment. Greater than 0, the position is on the 
   * ray starting on \p u1 and going through \p u2. If the line doesn't
   * intersect the plane, s is valued to NaN.
   * \param[in] p1 Starting point of the segment
   * \param[in] p2 Ending point of the segment
   * \param[in] tr1 First triangle point
   * \param[in] tr2 Second triangle point
   * \param[in] tr3 Third triangle point
   */
  bool lineTriangleIntersection(Point3d& u, double& s,
                                const Point3d& p1, const Point3d& p2,
                                const Point3d& tr1, const Point3d& tr2, const Point3d& tr3);

  /**
   * Find if the point p is inside the polygon using the winding number.
   */
  bool pointInPolygon(const Point2d& p, const std::vector<Point2d>& polygon);

  template <typename PointContainer>
  bool pointInPolygon(const Point2d& p, const PointContainer& polygon)
  {
    std::vector<Point2d> pol(polygon.begin(), polygon.end());
    return pointInPolygon(p, pol);
  }

  /**
   * Find if the segment [p1,p2] intersect the line going through p of normal n.
   * \param[out] u Intersection point if the function returns true, untouched otherwise
   * \param[in] p1 First point of the segment
   * \param[in] p2 Second point of the segment
   * \param[in] p Point on the line
   * \param[in] n Normal vector to the line
   */
  bool lineSegmentIntersection(Point2d& u,
                               const Point2d& p1, const Point2d& p2,
                               const Point2d& p, const Point2d& n);

  /**
   * Find if the segment [p1,p2] intersect the line going through p of normal n.
   * \param[out] u Intersection point if the function returns true, untouched otherwise
   * \param[in] p1 Point on the first line
   * \param[in] u1 Vector along the first line
   * \param[in] p2 Point on the second line
   * \param[in] u2 Vector along the second line
   */
  bool lineLineIntersection(Point2d& u,
                            const Point2d& p1, const Point2d& u1,
                            const Point2d& p2, const Point2d& u2);

  /**
   * Find if two 2D segments intersect.
   *
   * Based on "Faster line segment intersection." Graphics Gems III.
   *
   * \param u The actual intersection point, if any.
   *
   * \warning To be rewritten for numerical precision + check algo
   */
  bool segmentSegmentIntersection(Point2d& u,
                                  const Point2d& p1, const Point2d& p2,
                                  const Point2d& q1, const Point2d& q2);

};

#endif // VVELIB_GEOMETRY_INTERSECTION_H
