#ifndef VVELIB_GEOMETRY_PROJECTION_H
#define VVELIB_GEOMETRY_PROJECTION_H

/**
 * \file projection.h
 *
 * Algorithms to project on lines/planes/...
 */

#include <config.h>
#include <geometry/geometry.h>

namespace geometry
{
  /**
   * Project a point on a line
   *
   * \note Works whatever the dimension! Not restricted to 2d or 3d.
   *
   * \param pt Point to project
   * \param u1 Start of the line
   * \param u2 End of the line
   * \param strict If true, only the segment is considered, otherwise the whole 
   * line is
   *
   * \returns The coordinates of the projected point.
   */
  template <size_t N>
  util::Vector<N,double> projectPointOnLine(const util::Vector<N,double>& pt, const util::Vector<N,double>& u1, const util::Vector<N,double>& u2, bool strict=false)
  {
    util::Vector<N,double> u = u2-u1;
    util::Vector<N,double> pv = pt-u1;
    double l = util::norm(u);
    u /= l;
    double proj_val = pv*u;
    if(strict)
    {
      if(proj_val < 0)
        proj_val = 0;
      else if(proj_val > l)
        proj_val = l;
    }
    return u1 + u*proj_val;
  }

  /**
   * Project a point into a triangle.
   *
   * \param pt Point to project
   * \param p1 First point of the triangle
   * \param p2 Second point of the triangle
   * \param p3 Third point of the triangle
   * \param s Square of the distance from the triangle to the point
   * \param in_triangle If not null, set to true if the orthogonal projection is in the triangle, false otherwise.
   *
   * \returns The coordinate of the point of the triangle closest to \p pt.
   */
  Point3d projectPointOnTriangle(const Point3d& pt, const Point3d& p1, const Point3d& p2, const Point3d& p3, double &s, bool *in_triangle = 0);

  /**
   * Project a point into a plane defined by a point and a normal to the plane
   *
   * \param pos Point to project
   * \param p Point of the plane
   * \param n Normal to the plane
   */
  inline Point3d projectPointOnPlane(const Point3d& pos, const Point3d& p, const Point3d& n)
  {
    Point3d res = pos - ((pos-p)*n)*n;
    return res;
  }
};

#endif // VVELIB_GEOMETRY_PROJECTION_H

