#ifndef __GRAPH_VERTEX_H
#define __GRAPH_VERTEX_H

/**
 * \file vertex.h
 *
 * This file contain the definition of the graph::Vertex class.
 * For now, the graph::Vertex class can only be used with the graph::VVGraph 
 * class.
 */

#include <QtCore/QHash>
#include <config.h>
#include <map>
#include <utility>
#include <util/unorderedmap.h>
#include <util/mangling.h>
#include <iostream>
#include <typeinfo>
#ifndef _MSC_VER
#  include <stdint.h>
#endif

#include <util/assert.h>

#include <storage/fwd.h>

#ifdef USE_ALLOCATOR
#  include <memory>
#endif // defined USE_ALLOCATOR

#ifndef NO_NUMBER_VERTICES
#  ifdef VERTEX_ORDER_BY_NUM
#    undef VERTEX_ORDER_BY_NUM
#  endif
#endif

#define VVVERTEX_NO_CACHE
#ifndef VVGRAPH_NO_CACHE
#  undef VVVERTEX_NO_CACHE
#elif !defined(VVBIGRAPH_NO_CACHE)
#  undef VVVERTEX_NO_CACHE
#endif

#ifdef USE_ALLOCATOR
#  define TEMPLATE_VERTEX typename VertexContent, typename Alloc
#  define VERTEX_ARGS VertexContent,Alloc
#else // USE_ALLOCATOR
#  define TEMPLATE_VERTEX typename VertexContent
#  define VERTEX_ARGS VertexContent
#endif // USE_ALLOCATOR

namespace graph
{
  /**
   * Type of the identifier of a vertex
   */
  typedef uintptr_t vertex_identity_t;

#ifndef NO_NUMBER_VERTICES
  /**
   * Number used to enumerate the vertices of all types.
   *
   * This number exists only if NO_NUMBER_VERTICES is not defined at compile 
   * time.
   */
  extern size_t vertex_counter;
#endif

  template <TEMPLATE_VERTEX> class WeakVertex;

    /**
     * Type of the reference counted content
     */
  template <typename Content>
    struct CountedContent : public Content
    {
      CountedContent()
        : Content()
        , count(1)
#ifndef NO_NUMBER_VERTICES
        , num(++vertex_counter)
#endif
        {}

      /**
       * Reference count on the vertex
       */
      unsigned int count;
#ifndef NO_NUMBER_VERTICES
      size_t num;
#endif
    };

  /**
   * \class Vertex vertex.h <graph/vertex.h>
   * Vertex of a vv graph.
   *
   * The vertexes handle their associated data using a reference counting scheme. 
   * As such, they can be used as smart pointers. They are also comparable 
   * (<,>,==,!=), which allow for use in any sorted structure and hashable for 
   * use in any hash table-based structure.
   *
   * They also all have a unique identifier. This identifier can be used to 
   * retrieve a weak reference on the data.
   *
   * \warning A weak reference won't keep the data alive! So be careful when you 
   * use those. You can test if a vertex hold a weak reference using the 
   * Vertex<VertexContent>::isWeakRef() method.
   */
  template <typename VertexContent
#ifdef USE_ALLOCATOR
    , typename Alloc = DEFAULT_ALLOC(VertexContent)
#endif
           >
  class Vertex
  {
    friend class WeakVertex<VERTEX_ARGS>;

    public:
    /**
     * Type of the reference-counted content.
     *
     * Useful to share with other smart-pointers in VVE
     */
    typedef CountedContent<VertexContent> counted_content_t;

    protected:

#ifdef USE_ALLOCATOR
    //typedef std::allocator<VERTEX_ARGS> Alloc;
    typedef typename Alloc::template rebind<counted_content_t>::other RealAlloc;

    static RealAlloc alloc;
#endif

    public:

#ifdef USE_ALLOCATOR
    /**
     * Pointer to the reference-counted content.
     *
     * Useful to share with other smart-pointers in VVE
     */
    typedef typename RealAlloc::pointer real_pointer;
#else
    /**
     * Pointer to the reference-counted content.
     *
     * Useful to share with other smart-pointers in VVE
     */
    typedef counted_content_t* real_pointer;
#endif // defined USE_ALLOCATOR

    /**
     * Type of the identifier of the vertex
     */
    typedef vertex_identity_t identity_t;

    /**
     * Type of the content of the vertex
     */
    typedef VertexContent content_t;

    typedef WeakVertex<VERTEX_ARGS> weak_ref_t;

    /**
     * Type of the equivalent pointer
     */
    typedef content_t* pointer;
    typedef content_t& reference;

    /**
     * Creates a new vertex with a new content.
     *
     * Example:
     * \code
     *   struct VertexContent { int a; }
     *   // ...
     *   Vertex<VertexContent> v;
     *   v->a = 10;
     * \endcode
     */
    Vertex();

    /**
     * Create an empty vertex (argument really should be 0)
     */
    explicit Vertex( int i );

    /**
     * Creates a reference on the vertex of identifier \c id. If \c id 
     * is 0, creates a null vertex.
     *
     * \param[in] id Label of the vertex to retrieve.
     *
     * \warning This function is very unsafe if used with anything but 0 as 
     * an identifier.
     *
     * Example:
     * \code
     *   typedef Vertex<VertexContent> vertex;
     *   vertex v;
     *   vertex::identity_t i = v.id();
     *   vertex v1(i);
     *   assert(v1 == v);
     * \endcode
     */
    explicit Vertex( identity_t id );

    /**
     * Copy a vertex.
     *
     * The data is not copied. The quality of the copy (i.e. weak/strong 
     * reference) is the same as the copied.
     */
    Vertex( const Vertex& copy );

    /**
     * Create a vertex using a counted content
     *
     * The vertex add to the ownership of the content
     */
    Vertex(real_pointer content);

#ifdef USE_CXX0X
    /**
     * Move constructor
     *
     * The data is not copied, the increment not done, the other vertex made 
     * invalid
     */
    Vertex( Vertex&& copy );
#endif

    /**
     * Construct a strong reference from a weak one
     */
    explicit Vertex( const weak_ref_t& w );

    /**
     * Desctructor
     */
    ~Vertex();

    /**
     * Access to the data.
     *
     * \warning Do not try to access the data of the null vertex.
     */
    pointer operator->() const { return _content; }
    /**
     * Access to the data.
     *
     * \warning Do not try to access the data of the null vertex.
     */
    reference operator*() const { return *_content; }

    /**
     * Return the reference-counted object
     */
    real_pointer content() const { return _content; }

    /**
     * Return the content, and reset the value in the vertex
     */
    real_pointer acquire() { real_pointer p = _content; _content = 0; return p; }

    /**
     * Access to the data via pointer to member
     */
    template <typename R>
      R& operator->*(R VertexContent::*ptr)
      {
        return _content->*ptr;
      }
    /**
     * Constant access to the data via pointer to member
     */
    template <typename R>
      const R& operator->*(R VertexContent::*ptr) const
      {
        return _content->*ptr;
      }

    /**
     * Change the vertex held by the current object.
     *
     * The data is never modified by this operation. If you wish to copy the data 
     * of a vertex v1 into a vertex v2 use:
     * \code
     * *v2 = *v1;
     * \endcode
     */
    Vertex& operator=(const Vertex& other);

#ifdef USE_CXX0X
    Vertex& operator=(Vertex&& other);
#endif

    Vertex& operator=(const identity_t& id);

    Vertex& operator=(const weak_ref_t& other);

    //Vertex& operator=(const VertexContent* value);

    /**
     * Comparison operators.
     *
     * \note the comparisons work on the identifiers, not the contents!
     */
    bool operator==(const Vertex& other) const { return id() == other.id(); }

    /**
     * Two vertices of different types cannot be equal
     */
    template <typename T1>
      bool operator==(const Vertex<T1>& /*other*/) const { return false; }

    /**
     * Comparison operators.
     *
     * \note the comparisons work on the identifiers, not the contents!
     */
    bool operator!=(const Vertex& other) const { return id() != other.id(); }

    /**
     * Two vertices of different types cannot be equal
     */
    template <typename T1>
      bool operator!=(const Vertex<T1>& other) const { return true; }

    /**
     * Comparison operators.
     *
     * \note the comparisons work on the identifiers, not the contents!
     */
    bool operator>(const Vertex& other) const
    {
#ifdef VERTEX_ORDER_BY_NUM
      return num() > other.num();
#else
      return id() > other.id();
#endif
    }
    /**
     * Comparison operators.
     *
     * \note the comparisons work on the identifiers, not the contents!
     */
    bool operator<(const Vertex& other) const
    {
#ifdef VERTEX_ORDER_BY_NUM
      return num() < other.num();
#else
      return id() < other.id();
#endif
    }

    /**
     * Comparison operators.
     *
     * \note the comparisons work on the identifiers, not the contents!
     */
    bool operator>=(const Vertex& other) const
    {
#ifdef VERTEX_ORDER_BY_NUM
      return num() >= other.num();
#else
      return id() >= other.id();
#endif
    }
    /**
     * Comparison operators.
     *
     * \note the comparisons work on the identifiers, not the contents!
     */
    bool operator<=(const Vertex& other) const
    {
#ifdef VERTEX_ORDER_BY_NUM
      return num() <= other.num();
#else
      return id() <= other.id();
#endif
    }

    /**
     * Test if a vertex is a null vertex.
     */
    bool isNull() const { return _content == 0; }

    /**
     * Return the identifier of a vertex.
     */
    identity_t id() const { return (identity_t)_content; }


#ifndef NO_NUMBER_VERTICES
    /**
     * Return a number unique to each vertex, globally.
     *
     * This method can be removed by defining NO_NUMBER_VERTICES at compile 
     * time
     */
    size_t num() const { return _content->num; }
#endif

    /**
     * Convert a vertex to \c true if it is not null.
     */
    operator bool() const { return _content; }

    /**
     * Return \c true if the current object hold a weak reference on a vertex.
     */
    bool isWeakRef() const { return false; }

    /**
     * Construct a weak reference on the current vertex
     */
    weak_ref_t weakRef() const;

    /**
     * Null vertex.
     *
     * Useful to return a constant reference on a vertex all the time.
     *
     * Example:
     * \code
     * const vertex& fct(bool cond)
     * {
     *   static vertex a;
     *   if(cond)
     *     return a;
     *    return vertex::null;
     * }
     * \endcode
     */
    static Vertex null;

    /**
     * Serialization method
     *
     * \warning You need to include <storage/graph.h> to use this 
     * serialization method
     */
    bool serialize(storage::VVEStorage&);

    protected:
    /**
     * Content of the vertex
     *
     * This member is mutable to allow for modification of constant 
     * references. This is useful as no operation on the vertex depend on 
     * this.
     */
    mutable real_pointer _content;

    /**
     * Release the current pointer
     */
    void release();

#ifndef VVVERTEX_NO_CACHE
    public:
    /**
     * Label of the reference vertex for the cache information.
     * \warning You should never, on any condition, even try to read this 
     * variable, it may have no meaning whatsoever. If you modify it, expect 
     * a crash!
     */
    mutable identity_t cache_source;

    /**
     * Cache information for iteration optimization.
     * \warning If you dare to modify this variable, I will haunt your dreams 
     * until you give up and leave that member alone.
     */
    mutable void *cache;
#endif
  };

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS> Vertex<VERTEX_ARGS>::null(0);

#ifdef USE_ALLOCATOR
  template <TEMPLATE_VERTEX>
    typename Vertex<VERTEX_ARGS>::RealAlloc Vertex<VERTEX_ARGS>::alloc;
#endif // defined USE_ALLOCATOR

  /**
   * \class WeakVertex vertex.h <graph/vertex.h>
   * Weak pointer on a vertex.
   *
   * Be careful using this class as there is no garanty the vertex still 
   * exists.
   */
  template <typename VertexContent
#ifdef USE_ALLOCATOR
    , typename Alloc = DEFAULT_ALLOC(VertexContent)
#endif
    >
  class WeakVertex : public Vertex<VERTEX_ARGS>
  {
    typedef typename Vertex<VERTEX_ARGS>::counted_content_t counted_content_t;
    typedef typename Vertex<VERTEX_ARGS>::real_pointer real_pointer;
  public:
    /**
     * Type of the identifier of the vertex
     */
    typedef vertex_identity_t identity_t;

    /**
     * Type of the content of the vertex
     */
    typedef VertexContent content_t;

    /**
     * Strong reference corresponding to the weak one
     */
    typedef Vertex<VERTEX_ARGS> strong_ref;

    /**
     * Type of the equivalent pointer
     */
    typedef VertexContent* pointer;

    /**
     * Construct an empty weak vertex.
     *
     * At the difference of normal vertices, constructing a weak vertex creates 
     * a null one.
     */
    WeakVertex()
      : Vertex<VERTEX_ARGS>(0)
    {}

    /**
     * Construct a weak reference from a strong one
     */
    WeakVertex(const strong_ref& v)
      : Vertex<VERTEX_ARGS>(0)
    {
      this->_content = v._content;
    }

    /**
     * Copy constructor
     */
    WeakVertex(const WeakVertex& copy)
      : Vertex<VERTEX_ARGS>(0)
    {
      this->_content = copy._content;
    }

#ifdef USE_CXX0X
    /**
     * Move constructor
     */
    WeakVertex(WeakVertex&& ) = default;
#endif

    /**
     * Construct a weak reference from an id
     */
    explicit WeakVertex(const identity_t& id)
      : Vertex<VERTEX_ARGS>(0)
    {
      this->_content = reinterpret_cast<real_pointer>(id);
      if(this->_content and this->_content->count == 0)
        this->_content = 0;
    }

    /**
     * Set the content of the weak reference
     */
    WeakVertex& operator=(const WeakVertex& other)
    {
      this->_content = other._content;
      return *this;
    }

#ifdef USE_CXX0X
    WeakVertex& operator=(WeakVertex&& other)
    {
      this->_content = std::move(other._content);
      return *this;
    }
#endif

    WeakVertex& operator=(const identity_t& id)
    {
      this->_content = reinterpret_cast<real_pointer>(id);
      return *this;
    }

    WeakVertex& operator=(const strong_ref& other)
    {
      this->_content = other._content;
      return *this;
    }

    ~WeakVertex()
    {
      this->_content = 0;
    }

    bool isNull() const
    {
      return this->_content == 0 or this->_content->count == 0;
    }

  protected:
  };

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::~Vertex()
    {
      this->release();
    }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex()
    : _content(0)
#ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#endif
  {
#ifdef USE_ALLOCATOR
    _content = alloc.allocate(1);
    ::new(_content) counted_content_t();
#elif defined(DEBUG_ALLOC) // defined USE_ALLOCATOR
    try
    {
      _content = new counted_content_t();
    }
    catch(std::bad_alloc& )
    {
      _content = 0;
    }
#else
    _content = new counted_content_t();
#endif // defined USE_ALLOCATOR
  }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex( int id )
    : _content(0)
#ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#endif
    {
      vvassert(id == 0);
    }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex( identity_t id )
    : _content(reinterpret_cast<real_pointer>(id))
#ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#endif
  {
    if(_content)
    {
     if(_content->count == 0)
      _content = 0;
     else
       ++(_content->count);
    }
  }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex( real_pointer content )
    : _content(content)
#ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#endif
  {
    if(_content)
      ++(_content->count);
  }

//  template <TEMPLATE_VERTEX>
//    Vertex<VERTEX_ARGS>::Vertex( VertexContent* data )
//    : _content(0)
//    , cache_source(0)
//    , cache(0)
//  {
//    counted_content_t *cdata = dynamic_cast<real_pointer>(data);
//    if(cdata != 0)
//    {
//      _content = cdata;
//      ++(_content->count);
//    }
//    else
//    {
//      _content = new counted_content_t();
//      *(VertexContent*)_content = *data;
//    }
//  }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex( const typename Vertex<VERTEX_ARGS>::weak_ref_t& copy )
    : _content(copy._content)
#ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#endif
  {
    if(_content)
    {
      if(_content->count == 0)
        _content = 0;
      else
        ++(_content->count);
    }
  }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex( const Vertex& copy )
    : _content(copy._content)
#ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#endif
  {
    if(_content)
    {
      ++(_content->count);
    }
  }

#ifdef USE_CXX0X
  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>::Vertex( Vertex&& copy )
    : _content(std::move(copy._content))
#  ifndef VVVERTEX_NO_CACHE
    , cache_source(0)
    , cache(0)
#  endif
  {
    copy._content = 0;
  }
#endif

  template <TEMPLATE_VERTEX>
  Vertex<VERTEX_ARGS>& Vertex<VERTEX_ARGS>::operator=(const identity_t& id)
  {
    if((identity_t)_content == id)
      return *this;
    this->release();
    _content = reinterpret_cast<real_pointer>(id);
    if(_content)
    {
      if(_content->count == 0)
        _content = 0;
      else
        ++(_content->count);
    }
    return *this;
  }

//  template <TEMPLATE_VERTEX>
//  Vertex<VERTEX_ARGS>& Vertex<VERTEX_ARGS>::operator=(const VertexContent* id)
//  {
//    const counted_content_t* cid = dynamic_cast<const counted_content_t*>(&id);
//    if(cid != 0)
//      return *this = (const identity_t&)cid;
//    else
//    {
//      return *this = (const identity_t&)(new counted_content_t(id));
//    }
//  }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>& Vertex<VERTEX_ARGS>::operator=(const weak_ref_t& copy)
  {
    if(_content == copy._content)
      return *this;
    this->release();
    _content = copy._content;
    if(_content)
    {
      if(_content->count == 0)
        _content = 0;
      else
        ++(_content->count);
    }
    return *this;
  }

  template <TEMPLATE_VERTEX>
    void Vertex<VERTEX_ARGS>::release()
  {
    if( _content )
    {
      --(_content->count);
      if(_content->count == 0)
      {
#ifdef USE_ALLOCATOR
        _content->~counted_content_t();
        alloc.deallocate(_content, 1);
#else // defined USE_ALLOCATOR
        delete _content;
#endif
      }
    }
  }

  template <TEMPLATE_VERTEX>
    typename Vertex<VERTEX_ARGS>::weak_ref_t Vertex<VERTEX_ARGS>::weakRef() const
  {
    return weak_ref_t(*this);
  }

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>& Vertex<VERTEX_ARGS>::operator=(const Vertex& other)
  {
    if(_content == other._content)
    {
      return *this;
    }
    else
      this->release();
    _content = other._content;
    if(_content)
    {
      ++(_content->count);
    }
    return *this;
  }

#ifdef USE_CXX0X

  template <TEMPLATE_VERTEX>
    Vertex<VERTEX_ARGS>& Vertex<VERTEX_ARGS>::operator=(Vertex&& other)
  {
    if(_content == other._content)
    {
      return *this;
    }
    else
      this->release();
    _content = other._content;
    other._content = 0;
    return *this;
  }
#endif

  template <TEMPLATE_VERTEX, typename charT>
  std::basic_ostream<charT>& operator<<(std::basic_ostream<charT>& ss, const Vertex<VERTEX_ARGS>& v)
  {
    ss << "Vertex<" << util::demangle(typeid(VertexContent).name()) << ">(" << v.id() << ")";
    return ss;
  }


}

#ifdef USE_HASH
namespace std
{
#ifdef CXX_NEEDS_TR1
  namespace tr1
  {
#endif
    template <TEMPLATE_VERTEX>
      struct hash<graph::Vertex<VERTEX_ARGS> >
      {
        size_t operator()(const graph::Vertex<VERTEX_ARGS>& v) const
        {
          return ((size_t)v.id())>>4; // Shift to consider memory alignment on 64bit!
        }
      };

    template <TEMPLATE_VERTEX>
      struct hash<graph::WeakVertex<VERTEX_ARGS> >
      {
        size_t operator()(const graph::WeakVertex<VERTEX_ARGS>& v) const
        {
          return ((size_t)v.id())>>4; // Shift to consider memory alignment on 64bit!
        }
      };
#ifdef CXX_NEEDS_TR1
  }
#endif
}
#endif

template <TEMPLATE_VERTEX>
uint qHash(const graph::Vertex<VERTEX_ARGS>& v)
{
  return qHash(uint(v.id() >> 4));
}

#undef TEMPLATE_VERTEX
#undef VERTEX

#endif // __GRAPH_VERTEX_H

