#ifndef GRAPH_VVBIGRAPH_H
#define GRAPH_VVBIGRAPH_H

/**
 * \file vvbigraph.h
 *
 * Contain the definition of the VVBiGraph template class for bipartite graphs.
 */

#include <config.h>
#include <graph/vertex.h>
#include <graph/edge.h>
#include <graph/vvgraph.h>

#include <util/tie.h>
#include <util/member_iterator.h>
#include <util/circ_iterator.h>
#include <util/set_vector.h>

#include <storage/fwd.h>

#include <utility>
#include <memory>
#include <algorithm>
#include <iterator>

#include <list>
#include <set>
#include <map>

#include <iostream>

#ifdef VVBIGRAPH_NO_CACHE
#  define VVBIGRAPH_NO_EDGE_CACHE
#  define VVBIGRAPH_NO_VERTEX_CACHE
#endif

#ifdef VVBIGRAPH_NO_VERTEX_CACHE
#  define VVBIGRAPH_COMPACT_VERTEX true
#else
#  define VVBIGRAPH_COMPACT_VERTEX compact
#endif

#ifdef VVBIGRAPH_NO_EDGE_CACHE
#  define VVBIGRAPH_COMPACT_EDGE true
#else
#  define VVBIGRAPH_COMPACT_EDGE compact
#endif

#ifdef USE_ALLOCATOR
#  define BIGRAPH_TEMPLATE typename Vertex1Content, typename Vertex2Content, typename Edge1Content, typename Edge2Content, bool compact, typename Alloc
#  define BIGRAPH_ARGS Vertex1Content,Vertex2Content,Edge1Content,Edge2Content,compact,Alloc
#else // USE_ALLOCATOR
#  define BIGRAPH_TEMPLATE typename Vertex1Content, typename Vertex2Content, typename Edge1Content, typename Edge2Content, bool compact
#  define BIGRAPH_ARGS Vertex1Content,Vertex2Content,Edge1Content,Edge2Content,compact
#endif // USE_ALLOCATOR

namespace graph
{

  template <bool compact, typename T>
    struct __vvbigraph_set
    { };

  template <typename T>
    struct __vvbigraph_set<true,T>
    {
      typedef util::set_vector<T> type;
    };

  template <typename T>
    struct __vvbigraph_set<false,T>
    {
      typedef std::set<T> type;
    };

  template <bool compact, typename EdgeContent, typename VVBiGraph, typename single_neighborhood_t, typename edge_list_t, typename vertex_tgt_t>
    struct __vvbigraph_EdgeCache
    {
    };

  template <typename EdgeContent, typename VVBiGraph, typename single_neighborhood_t, typename edge_list_t, typename vertex_tgt_t>
    struct __vvbigraph_EdgeCache<false, EdgeContent, VVBiGraph, single_neighborhood_t, edge_list_t, vertex_tgt_t>
    : public EdgeContent
    {
      /**
       * Construct an edge cache
       */
      __vvbigraph_EdgeCache(const EdgeContent& copy,
                            single_neighborhood_t *n,
                            intptr_t _gid)
        : EdgeContent(copy)
        , _neighborhood(n)
        , gid(_gid)
        { }

      __vvbigraph_EdgeCache(const __vvbigraph_EdgeCache& copy)
        : EdgeContent(copy)
        , _neighborhood(copy._neighborhood)
        , gid(copy.gid)
        {}

#ifdef USE_CXX0X
      __vvbigraph_EdgeCache(__vvbigraph_EdgeCache&& ) = default;
#endif

      __vvbigraph_EdgeCache& operator=(const __vvbigraph_EdgeCache& copy)
      {
        _neighborhood = copy._neighborhood;
        gid = copy.gid;
        static_cast<EdgeContent&>(*this) = static_cast<const EdgeContent&>(copy);
        return *this;
      }

      void setIt(const typename edge_list_t::iterator& i)
      {
        _it = i;
      }

      bool eraseEdge()
      {
        if(_neighborhood->flagged && *_neighborhood->flagged == _it->target) _neighborhood->flagged = 0;
        _neighborhood->edges.erase(_it);
        return true;
      }

      void set(single_neighborhood_t* n, intptr_t id)
      {
        _neighborhood = n;
        gid = id;
      }

      single_neighborhood_t* neighborhood() { return _neighborhood; }

      typename edge_list_t::iterator it() { return _it; }

      /**
       * Edge list containing the neighbor
       */
      single_neighborhood_t* _neighborhood;
      /**
       * Iterator pointing to the neighbor
       */
      typename edge_list_t::iterator _it;

      /**
       * Id of the current graph
       */
      intptr_t gid;

      intptr_t graph_id() { return gid; }
    };

  template <typename EdgeContent, typename VVBiGraph, typename single_neighborhood_t, typename edge_list_t, typename vertex_tgt_t>
    struct __vvbigraph_EdgeCache<true, EdgeContent, VVBiGraph, single_neighborhood_t, edge_list_t, vertex_tgt_t>
    : public EdgeContent
    {
      /**
       * Construct an edge cache
       */
      __vvbigraph_EdgeCache(const EdgeContent& copy,
                            single_neighborhood_t *,
                            intptr_t)
        : EdgeContent(copy)
        { }

      /**
       * Construct an edge cache
       */
      __vvbigraph_EdgeCache(const EdgeContent& copy,
                            single_neighborhood_t *,
                            const vertex_tgt_t&,
                            intptr_t)
        : EdgeContent(copy)
        { }

      __vvbigraph_EdgeCache(const __vvbigraph_EdgeCache& copy)
        : EdgeContent(copy)
        {}

#ifdef USE_CXX0X
      __vvbigraph_EdgeCache(__vvbigraph_EdgeCache&& ) = default;
#endif

      void setIt(const typename edge_list_t::iterator&)
      {
      }

      bool eraseEdge()
      {
        return false;
      }

      intptr_t graph_id() { return 0; }

      single_neighborhood_t* neighborhood() { return 0; }

      typename edge_list_t::iterator it() { return typename edge_list_t::iterator(); }

      void set(single_neighborhood_t*, intptr_t ) {}

      __vvbigraph_EdgeCache& operator=(const __vvbigraph_EdgeCache& copy)
      {
        static_cast<EdgeContent&>(*this) = static_cast<const EdgeContent&>(copy);
        return *this;
      }
    };

  template <bool compact, typename VVBiGraph, typename single_neighborhood_t, typename vertex_src_t>
    struct __vvbigraph_VertexCache
    {
      typedef typename std::list<std::pair<vertex_src_t,single_neighborhood_t> >::iterator iterator;
      intptr_t graph_id() { return 0; }
      void update(intptr_t, const vertex_src_t& , iterator ) {}
      iterator neighborhood() { return iterator(); }
    };

  template <typename VVBiGraph, typename single_neighborhood_t, typename vertex_src_t>
    struct __vvbigraph_VertexCache<false,VVBiGraph,single_neighborhood_t,vertex_src_t>
    {
      typedef typename std::list<std::pair<vertex_src_t,single_neighborhood_t> >::iterator iterator;

      __vvbigraph_VertexCache() : gid(0) {}
      __vvbigraph_VertexCache(const __vvbigraph_VertexCache& copy ) : gid(copy.gid), _neighborhood(copy._neighborhood) {}

#ifdef USE_CXX0X
      __vvbigraph_VertexCache(__vvbigraph_VertexCache&& ) = default;
#endif

      intptr_t graph_id() { return gid; }
      iterator neighborhood() { return _neighborhood; }

      void update(intptr_t id, const vertex_src_t& v, iterator it)
      {
        gid = id;
        _neighborhood = it;
        v.cache = reinterpret_cast<void*>(this);
        v.cache_source = 0;
      }

      /**
       * Id of the current graph
       *
       * Used to checked for cache validity (you don't want to use the 
       * cache if the vertex is used in a different graph)
       */
      intptr_t gid;

      /**
       * Iterator positionned on the neighborhood of the vertex
       *
       * This is the cache itself
       */
      iterator _neighborhood;
    };

  template <typename Base>
    struct _EmptyDeriv : public Base
    {
      _EmptyDeriv() : Base() {}
      _EmptyDeriv(const _EmptyDeriv& copy) : Base(copy) {}
      _EmptyDeriv(const Base& copy) : Base(copy) {}

#ifdef USE_CXX0X
      _EmptyDeriv(_EmptyDeriv&& ) = default;
#endif

      bool operator==(const _EmptyDeriv& other) const { return *static_cast<const Base*>(this) == other; }
    };

  struct _EmptyBiEdgeContent
  {
    _EmptyBiEdgeContent& operator=(const _EmptyBiEdgeContent& /*other*/) { return *this; }
    bool serialize(storage::VVEStorage&) { return true; }
  };

  /**
   * \class VVBiGraph vvbigraph.h <graph/vvbigraph.h>
   * Class representing a VV graph.
   *
   * A VV graph is an oriented rotational graph. That is, the edges of a given 
   * vertex have a circular order. That is, if an edge has three edges e1, e2, 
   * e3, there is no first edge, but we can say e1 < e2 < e3 < e1. In that 
   * case, '<' denote the succession.
   *
   * \c VertexContent is the type used to hold the vertex data.
   *
   * \c EdgeContent is the type used to hold the edge data. If no type is 
   * specified for \c EdgeContent, then an empty structure is used.
   *
   * \section Perfs Performance notes
   * The graph includes two caching mecanisms to improve performances when 
   * iterating over the graph and retrieve edge information:
   *   -# while iterating over the graph, the vertex cache the current iterator. 
   *   So that accessing the neighborhood of the vertex is constant time (i.e. 
   *   it's dereferencing the cached iterator).
   *   -# while iterating over the neighborhood of a vertex, the target vertex 
   *   (i.e. the one returned by the iteration) cache the iterator on the 
   *   neighborhood. So that any operation using that specific edge (i.e. 
   *   getting edge data, splicing, replacing or erasing) is alos constant 
   *   time.
   *
   * To make this caching mecanism correct, the cached data are never copied. 
   * This means, to make use of the caching mecanism, you have to iterate using 
   * constant references (references will end up with compilation errors if 
   * used on the graph):
   * \code
   * forall_named(const vertex1_t &v, S, vertex1)
   * {
   *   // Here the vertex v has the vertex cache
   *   forall(const vertex2_t &n, S.neighbors(v)) // Makes use of the cache
   *   {
   *     S.edge(v,n)->pos = Point3d(1,2,3); // Uses the edge cache
   *     S.edge(n,v)->pos = Point3d(3,2,1); // Do not use any cache
   *   }
   *   v->pos = Point3d(0,0,0); // Valid
   * }
   * \endcode
   *
   * \warning Do not iterate using constant references if you ever remove the 
   * vertex you iterate on ! (i.e. remove from the graph if you iterate on the 
   * graph or remove from the neighborhood if you iterate on the neighborhood)
   *
   * Even though this code manipulate constant references, modifying the data 
   * pointed by the vertex is allowed. The vertices are coded to allow that. 
   * That means if you are using vertices in a dictionary or a set, using their 
   * content to compare them, you must not modify them ! The compiler won't 
   * guard you against it.
   */
  template <typename Vertex1Content, typename Vertex2Content,
           typename Edge1Content = _EmptyBiEdgeContent, typename Edge2Content_ = Edge1Content, bool compact = false
#ifdef USE_ALLOCATOR
             , typename Alloc = DEFAULT_ALLOC(Vertex1Content)
#endif
             >
             class VVBiGraph
             {
               /**
                * Private type used to allow Edge1Content and Edge2Content_ to be the 
                * same
                */
               typedef _EmptyDeriv<Edge2Content_> Edge2Content;

               public:
#ifdef USE_ALLOCATOR
               typedef Alloc Vertex1Alloc;
               typedef typename Alloc::template rebind<Vertex2Content>::other Vertex2Alloc;
#endif

               ///\name Smart pointer types
               //@{
               /**
                * Smart pointer on a vertex of type 1
                */
#ifdef USE_ALLOCATOR
               typedef Vertex<Vertex1Content, Vertex1Alloc> vertex1_t;
#else // USE_ALLOCATOR
               typedef Vertex<Vertex1Content> vertex1_t;
#endif // USE_ALLOCATOR
               /**
                * Smart pointer on a vertex of type 2
                */
#ifdef USE_ALLOCATOR
               typedef Vertex<Vertex2Content, Vertex2Alloc> vertex2_t;
#else // USE_ALLOCATOR
               typedef Vertex<Vertex2Content> vertex2_t;
#endif // USE_ALLOCATOR

               /**
                * Weak pointer on an edge from a vertex of type 1 to a vertex of type 2
                */
               typedef Edge<Edge1Content> edge1_t;
               /**
                * Weak pointer on an edge from a vertex of type 2 to a vertex of type 1
                */
               typedef Edge<Edge2Content> edge2_t;

               /**
                * Weak pointer on a constant edge from a vertex of type 1 to a vertex of 
                * type 2
                */
               typedef Edge<const Edge1Content> const_edge1_t;
               /**
                * Weak pointer on a constant edge from a vertex of type 2 to a vertex of 
                * type 1
                */
               typedef Edge<const Edge2Content> const_edge2_t;
               //@}

               /**
                * Default constructor
                *
                * Creates an empty VV graph
                */
               VVBiGraph() : graph_id(graph::getNextGraphId()) {}

               /**
                * Copy constructor.
                *
                * Maintains the caches in a valid state
                */
               VVBiGraph(const VVBiGraph& copy);

#ifdef USE_CXX0X
               /**
                * Move constrctor
                */
               VVBiGraph(VVBiGraph&& copy);
#endif

               protected:

               template <typename VertexSrcContent, typename VertexTgtContent, typename EdgeSrcContent>
                 struct single_neighborhood_t;

               template <typename VertexSrcContent, typename VertexTgtContent, typename EdgeSrcContent>
                 struct neighbor_t : public __vvbigraph_EdgeCache<VVBIGRAPH_COMPACT_EDGE,
                                   EdgeSrcContent,
                                   VVBiGraph,
                                   single_neighborhood_t<VertexSrcContent,VertexTgtContent,EdgeSrcContent>,
                                   std::list<neighbor_t<VertexSrcContent,VertexTgtContent,EdgeSrcContent> >,
                                   graph::Vertex<VertexTgtContent> >
               {
                 typedef single_neighborhood_t<VertexSrcContent,VertexTgtContent,EdgeSrcContent> _single_neighborhood_t;

                 typedef graph::Vertex<VertexTgtContent> vertex_tgt_t;
                 /**
                  * Structure maintaining the data for a single neighbor
                  */
                 typedef std::list<neighbor_t> edge_list_t;

                 typedef __vvbigraph_EdgeCache<VVBIGRAPH_COMPACT_EDGE,EdgeSrcContent,VVBiGraph,_single_neighborhood_t,edge_list_t,vertex_tgt_t> base;

#ifdef USE_ALLOCATOR
                 typedef typename Alloc::template rebind<EdgeSrcContent>::other EdgeAlloc;
                 static EdgeAlloc edge_alloc;
#endif

                 /**
                  * Constructor
                  */
                 neighbor_t(const vertex_tgt_t& tgt, const EdgeSrcContent &e,
                            _single_neighborhood_t& n,
                            intptr_t gid)
                   : base(e, &n, gid)
                   , target(tgt)
                   {
                   }

                 neighbor_t(const neighbor_t& copy)
                   : base(copy)
                   , target(copy.target)
                   {
                   }

                 ~neighbor_t()
                 {
                 }

                 void clear_edge()
                 {
                   static_cast<EdgeSrcContent&>(*this) = EdgeSrcContent();
                 }

                 neighbor_t& operator=(const neighbor_t& copy)
                 {
                   target = copy.target;
                   static_cast<base&>(*this) = static_cast<const base&>(copy);
                   return *this;
                 }

                 /**
                  * Target of the edge
                  *
                  * i.e. neighbor itself
                  */
                 vertex_tgt_t target;

                 /**
                  * Equality of two neighbors/edge
                  *
                  * A single neighbor is equal to another if the vertex and the content are 
                  * the same
                  */
                 bool operator==(const neighbor_t& other) const
                 {
                   return (target == other.target) &&
                     (static_cast<const EdgeSrcContent&>(*this) == static_cast<EdgeSrcContent>(other));
                 }

               };

               /**
                * Type of a neighbor of a vertex of type 1 (i.e. it is of type 2)
                */
               typedef neighbor_t<Vertex1Content, Vertex2Content, Edge1Content> neighbor1_t;

               typedef typename neighbor1_t::base EdgeCache1;

               /**
                * Type of a neighbor of a vertex of type 1 (i.e. it is of type 2)
                */
               typedef neighbor_t<Vertex2Content, Vertex1Content, Edge2Content> neighbor2_t;

               typedef typename neighbor2_t::base EdgeCache2;

               /**
                * Type of the ordered list of the outgoing edges of a vertex of type 1
                */
               typedef typename neighbor1_t::edge_list_t edge_list1_t;
               /**
                * Type of the ordered list of the outgoing edges of a vertex of type 2
                */
               typedef typename neighbor2_t::edge_list_t edge_list2_t;

               /**
                * Type of the list of sources of the in-edges of a vertex of type 1
                */
               typedef typename __vvbigraph_set<compact,vertex2_t>::type in_edges1_t;
               /**
                * Type of the list of sources of the in-edges of a vertex of type 2
                */
               typedef typename __vvbigraph_set<compact,vertex1_t>::type in_edges2_t;

               /**
                * Type of the neighborhood of a vertex
                */
               template <typename VertexSrcContent, typename VertexTgtContent, typename EdgeSrcContent>
                 struct single_neighborhood_t : public __vvbigraph_VertexCache<VVBIGRAPH_COMPACT_VERTEX,
                                              VVBiGraph,
                                              single_neighborhood_t<VertexSrcContent,VertexTgtContent,EdgeSrcContent>,
                                              graph::Vertex<VertexSrcContent> >
                                              {
                                                typedef __vvbigraph_VertexCache<VVBIGRAPH_COMPACT_VERTEX,VVBiGraph,single_neighborhood_t,graph::Vertex<VertexSrcContent> > base;

                                                typedef neighbor_t<VertexSrcContent, VertexTgtContent,EdgeSrcContent> _neighbor_t;

                                                /**
                                                 * List of the outgoing edges
                                                 */
                                                typename _neighbor_t::edge_list_t edges;

                                                typedef graph::Vertex<VertexSrcContent> vertex_src_t;
                                                typedef graph::Vertex<VertexTgtContent> vertex_tgt_t;

                                                single_neighborhood_t()
                                                  : base()
                                                  , flagged(0)
                                                  {}

                                                single_neighborhood_t(const single_neighborhood_t& copy)
                                                  : base(copy)
                                                  , edges(copy.edges)
                                                  , in_edges(copy.in_edges)
                                                  , flagged(copy.flagged)
                                                  {
                                                    if(flagged)
                                                    {
                                                      for(typename _neighbor_t::edge_list_t::iterator it = edges.begin() ; it != edges.end() ; ++it)
                                                      {
                                                        if(it->target == *flagged)
                                                        {
                                                          flagged = &it->target;
                                                          return;
                                                        }
                                                      }
                                                    }
                                                  }

                                                /**
                                                 * Set of the sources of the incoming edges
                                                 */
                                                typename __vvbigraph_set<compact,vertex_tgt_t>::type in_edges;

                                                /**
                                                 * Flagged neighbor
                                                 */
                                                const vertex_tgt_t* flagged;

                                                /**
                                                 * Equality of two neighborhood
                                                 *
                                                 * Only the outgoing edges are compared!
                                                 */
                                                bool operator==(const single_neighborhood_t& other) const
                                                {
                                                  return edges == other.edges;
                                                }
                                              };

               /**
                * Type of the neighborhood of a vertex of type 1
                */
               typedef single_neighborhood_t<Vertex1Content, Vertex2Content, Edge1Content> single_neighborhood1_t;
               /**
                * Type of the neighborhood of a vertex of type 2
                */
               typedef single_neighborhood_t<Vertex2Content, Vertex1Content, Edge2Content> single_neighborhood2_t;

               typedef typename single_neighborhood1_t::base VertexCache1;
               typedef typename single_neighborhood2_t::base VertexCache2;

               public:

               /**
                * Type of the list of vertices of type 1, together with their 
                * neighborhood
                */
               typedef std::list<std::pair<vertex1_t, single_neighborhood1_t> > neighborhood1_t;
               /**
                * Type of the list of vertices of type 2, together with their 
                * neighborhood
                */
               typedef std::list<std::pair<vertex2_t, single_neighborhood2_t> > neighborhood2_t;

               /**
                * Type of the lookup hash map for vertices of type 1
                */
               typedef std::unordered_map<vertex1_t, typename neighborhood1_t::iterator> lookup1_t;

               /**
                * Type of the lookup hash map for vertices of type 2
                */
               typedef std::unordered_map<vertex2_t, typename neighborhood2_t::iterator> lookup2_t;

               /**
                * Shortcut for the value_type of the neighborhood for a vertex of type 1
                */
               typedef typename neighborhood1_t::value_type neighborhood1_value_type;
               /**
                * Shortcut for the value_type of the neighborhood for a vertex of type 2
                */
               typedef typename neighborhood2_t::value_type neighborhood2_value_type;

               /**
                * Type of a size
                */
               typedef typename neighborhood1_t::size_type size_type;

               ///\name Iterators
               //@{
               /**
                * Iterator on the vertices of type 1
                */
               typedef util::SelectMemberIterator<typename neighborhood1_t::iterator, const vertex1_t, 
                       &neighborhood1_value_type::first> iterator1;
               /**
                * Iterator on the vertices of type 2
                */
               typedef util::SelectMemberIterator<typename neighborhood2_t::iterator, const vertex2_t, 
                       &neighborhood2_value_type::first> iterator2;
               /**
                * Constant iterator on the vertices of type 1
                */
               typedef util::SelectMemberIterator<typename neighborhood1_t::const_iterator, const vertex1_t, 
                       &neighborhood1_value_type::first> const_iterator1;
               /**
                * Constant iterator on the vertices of type 1
                */
               typedef util::SelectMemberIterator<typename neighborhood2_t::const_iterator, const vertex2_t, 
                       &neighborhood2_value_type::first> const_iterator2;

               /**
                * Iterator on the neighbors of a vertex of type 1
                */
               typedef util::SelectMemberIterator<typename edge_list1_t::iterator, vertex2_t, 
                       &neighbor1_t::target> neighbor_iterator1;
               /**
                * Iterator on the neighbors of a vertex of type 2
                */
               typedef util::SelectMemberIterator<typename edge_list2_t::iterator, vertex1_t, 
                       &neighbor2_t::target> neighbor_iterator2;
               /**
                * Range of the neighbors of a vertex of type 1. The \c first element of 
                * the pair is an iterator on the beginning of the range, the \c second 
                * element is the past-the-end iterator.
                */
               typedef util::range<neighbor_iterator1> neighbor_iterator1_range;
               /**
                * Range of the neighbors of a vertex of type 2. The \c first element of 
                * the pair is an iterator on the beginning of the range, the \c second 
                * element is the past-the-end iterator.
                */
               typedef util::range<neighbor_iterator2> neighbor_iterator2_range;

               /**
                * Iterator used to iterate over the neighbors of a vertex of type 1, but 
                * specifying the starting point
                */
               typedef util::CircIterator<neighbor_iterator1> circ_neighbor_iterator1;
               /**
                * Iterator used to iterate over the neighbors of a vertex of type 2, but 
                * specifying the starting point
                */
               typedef util::CircIterator<neighbor_iterator2> circ_neighbor_iterator2;

               /**
                * Range of circular iterators for the neighbors of a vertex of type 1
                */
               typedef util::range<circ_neighbor_iterator1> circ_neighbor_iterator1_range;
               /**
                * Range of circular iterators for the neighbors of a vertex of type 2
                */
               typedef util::range<circ_neighbor_iterator2> circ_neighbor_iterator2_range;

               /**
                * Iterator on the incoming neighbors of a vertex of type 1.
                *
                * An incoming neighbor is a source vertex of an edge whose target is the 
                * current vertex.
                */
               typedef typename in_edges1_t::const_iterator ineighbor_iterator1;
               /**
                * Iterator on the incoming neighbors of a vertex of type 2.
                *
                * An incoming neighbor is a source vertex of an edge whose target is the 
                * current vertex.
                */
               typedef typename in_edges2_t::const_iterator ineighbor_iterator2;

               /**
                * Range of the incoming neighbors of a vertex of type 1.
                *
                * The \c first element of the pair is an iterator on the beginning of 
                * the range, the \c second element is the past-the-end iterator.
                *
                * \see ineighbor_iterator
                */
               typedef util::range<ineighbor_iterator1> ineighbor_iterator1_range;
               /**
                * Range of the incoming neighbors of a vertex of type 2.
                *
                * The \c first element of the pair is an iterator on the beginning of 
                * the range, the \c second element is the past-the-end iterator.
                *
                * \see ineighbor_iterator
                */
               typedef util::range<ineighbor_iterator2> ineighbor_iterator2_range;

               /**
                * Constant iterator on the neighbors of a vertex of type 1
                */
               typedef util::SelectMemberIterator<typename edge_list1_t::const_iterator, 
                       const vertex2_t, &neighbor1_t::target> const_neighbor_iterator1;
               /**
                * Constant iterator on the neighbors of a vertex of type 2
                */
               typedef util::SelectMemberIterator<typename edge_list2_t::const_iterator, 
                       const vertex1_t, &neighbor2_t::target> const_neighbor_iterator2;

               /**
                * Constant range of the neighbors of a vertex of type 1.
                *
                * The \c first element of the pair is a constant iterator on the 
                * beginning of the range, the \c second element is the past-the-end 
                * constant iterator.
                */
               typedef util::range<const_neighbor_iterator1> const_neighbor_iterator1_range;
               /**
                * Constant range of the neighbors of a vertex of type 2.
                *
                * The \c first element of the pair is a constant iterator on the 
                * beginning of the range, the \c second element is the past-the-end 
                * constant iterator.
                */
               typedef util::range<const_neighbor_iterator2> const_neighbor_iterator2_range;

               /**
                * Iterator used to iterate over the neighbors of a vertex of type 1, but 
                * specifying the starting point
                */
               typedef util::CircIterator<const_neighbor_iterator1> const_circ_neighbor_iterator1;
               /**
                * Iterator used to iterate over the neighbors of a vertex of type 2, but 
                * specifying the starting point
                */
               typedef util::CircIterator<const_neighbor_iterator2> const_circ_neighbor_iterator2;

               /**
                * Range of circular iterators for the neighbors of a vertex of type 1
                */
               typedef util::range<const_circ_neighbor_iterator1> const_circ_neighbor_iterator1_range;
               /**
                * Range of circular iterators for the neighbors of a vertex of type 2
                */
               typedef util::range<const_circ_neighbor_iterator2> const_circ_neighbor_iterator2_range;

               /**
                * Range of vertex of type 1
                */
               typedef util::range<const_iterator1> const_range_vertex1;
               /**
                * Range of vertex of type 2
                */
               typedef util::range<const_iterator2> const_range_vertex2;

               /**
                * Range of vertex of type 1
                */
               typedef util::range<iterator1> range_vertex1;
               /**
                * Range of vertex of type 2
                */
               typedef util::range<iterator2> range_vertex2;

               //@}

               ///\name Vertex set edition methods
               //@{

               /**
                * Remove a vertex from the graph.
                *
                * \param v vertex to erase
                *
                * \returns 1 if the vertex was erased, 0 else.
                */
               size_type erase( const vertex1_t& v );
               /**
                * Remove a vertex from the graph.
                *
                * \param v vertex to erase
                *
                * \returns 1 if the vertex was erased, 0 else.
                */
               size_type erase( const vertex2_t& v );

               /**
                * Insert a new vertex in the graph.
                *
                * \returns An iterator on the inserted vertex.
                */
               iterator1 insert( const vertex1_t& v );
               /**
                * Insert a new vertex in the graph.
                *
                * \returns An iterator on the inserted vertex.
                */
               iterator2 insert( const vertex2_t& v );
               //@}

               ///\name Vertex set lookup methods
               //@{

               /**
                * Return a vertex of type 1 from the graph.
                *
                * \returns A vertex of the graph, or a null vertex if the graph is empty.
                */
               const vertex1_t& any_vertex1() const;

               /**
                * Return a vertex of type 2 from the graph.
                *
                * \returns A vertex of the graph, or a null vertex if the graph is empty.
                */
               const vertex2_t& any_vertex2() const;

               /**
                * Return the vertex of type 1 of index \c idx.
                *
                * \note This is a convenience function whose complexity if O(idx)
                *
                * \returns The element \c idx position after the first one (using 
                * iterators)
                */
               const vertex1_t& get_vertex1(size_type idx) const;
               /**
                * Return the vertex of type 2 of index \c idx.
                *
                * \note This is a convenience function whose complexity if O(idx)
                *
                * \returns The element \c idx position after the first one (using 
                * iterators)
                */
               const vertex2_t& get_vertex2(size_type idx) const;

               /**
                * Test if \c v is in the graph.
                *
                * \returns \c true is \c v is in the graph.
                */
               bool contains( const vertex1_t& v ) const;
               /**
                * Test if \c v is in the graph.
                *
                * \returns \c true is \c v is in the graph.
                */
               bool contains( const vertex2_t& v ) const;

               /**
                * Get a reference on the vertex in the graph
                *
                * If the vertex do not exist in the graph, the result is totally 
                * undefined and the program might even crash.
                */
               const vertex1_t& reference( vertex1_t v ) const
               {
                 typename neighborhood1_t::const_iterator found = this->findVertex(v);
                 if(found != neighborhood1.end())
                   return found->first;
                 return vertex1_t::null;//vertex_t(0);
               }
               /**
                * Get a reference on the vertex in the graph
                *
                * If the vertex do not exist in the graph, the result is totally 
                * undefined and the program might even crash.
                */
               const vertex2_t& reference( vertex2_t v ) const
               {
                 typename neighborhood2_t::const_iterator found = this->findVertex(v);
                 if(found != neighborhood2.end())
                   return found->first;
                 return vertex2_t::null;//vertex_t(0);
               }

               /**
                * Return the number of vertices on the graph
                */
               size_type nb_vertices1() const { return neighborhood1.size(); }
               /**
                * Return the number of vertices on the graph
                */
               size_type nb_vertices2() const { return neighborhood2.size(); }
               /**
                * Return the number of vertices on the graph
                */
               size_type size() const { return this->nb_vertices1() + this->nb_vertices2(); }

               /**
                * Test if there is any vertex in the graph
                */
               bool no_vertex1() const { return neighborhood1.empty(); }
               /**
                * Test if there is any vertex in the graph
                */
               bool no_vertex2() const { return neighborhood2.empty(); }
               /**
                * Test if there is any vertex in the graph
                */
               bool empty() const { return this->no_vertex1() and this->no_vertex2(); }

               //@}

               ///\name Neighborhood lookup methods
               //@{

               /**
                * Return a vertex in the neighborhood of \c v.
                *
                * \returns A vertex in the neighborhood of \c v, or a null vertex if \c v is 
                * not in the graph or \c v has no neighborhood.
                */
               const vertex2_t& anyIn( const vertex1_t& v ) const;
               /**
                * Return a vertex in the neighborhood of \c v.
                *
                * \returns A vertex in the neighborhood of \c v, or a null vertex if \c v is 
                * not in the graph or \c v has no neighborhood.
                */
               const vertex1_t& anyIn( const vertex2_t& v ) const;

               /**
                * Returns the number of neighbors of \c v.
                *
                * If \c v is not in the graph, the behavior is undefined.
                */
               size_type valence( const vertex1_t& v ) const;
               /**
                * Returns the number of neighbors of \c v.
                *
                * If \c v is not in the graph, the behavior is undefined.
                */
               size_type valence( const vertex2_t& v ) const;

               /**
                * Test if a vertex has a neighbor.
                *
                * Returns false if it does. If a vertex is not in the graph, it is 
                * considered as having no neighbor. In the same way, the null vertex has 
                * no neighbors.
                */
               bool empty( const vertex1_t& v ) const;
               /**
                * Test if a vertex has a neighbor.
                *
                * Returns false if it does. If a vertex is not in the graph, it is 
                * considered as having no neighbor. In the same way, the null vertex has 
                * no neighbors.
                */
               bool empty( const vertex2_t& v ) const;

               /**
                * Returns any incoming neighbor of \c v.
                *
                * \returns A vertex in the incoming neighborhood of \c v, or a null 
                * vertex if \c v is not in the graph or \c v has no incoming neighbor.
                */
               const vertex2_t& iAnyIn(const vertex1_t& v) const;
               /**
                * Returns any incoming neighbor of \c v.
                *
                * \returns A vertex in the incoming neighborhood of \c v, or a null 
                * vertex if \c v is not in the graph or \c v has no incoming neighbor.
                */
               const vertex1_t& iAnyIn(const vertex2_t& v) const;

               /**
                * Returns the number of incoming neighbors of \c v
                *
                * The incoming neighbors are the set of vertices with on edge ending on 
                * \c v.
                */
               size_type iValence( const vertex1_t& v ) const;
               /**
                * Returns the number of incoming neighbors of \c v
                *
                * The incoming neighbors are the set of vertices with on edge ending on 
                * \c v.
                */
               size_type iValence( const vertex2_t& v ) const;

               /**
                * Test if a vertex has an incoming neighbor.
                *
                * Returns false if it does. If a vertex is not in the graph, it is 
                * considered as having no incoming neighbor. In the same way, the null 
                * vertex has no incoming neighbors.
                */
               bool iEmpty( const vertex1_t& v ) const;
               /**
                * Test if a vertex has an incoming neighbor.
                *
                * Returns false if it does. If a vertex is not in the graph, it is 
                * considered as having no incoming neighbor. In the same way, the null 
                * vertex has no incoming neighbors.
                */
               bool iEmpty( const vertex2_t& v ) const;

               /**
                * Returns the nth vertex after \c neighbor in the neighborhood of \c v.
                *
                * \returns The vertex found or a null vertex if there is any problem.
                */
               const vertex2_t& nextTo( const vertex1_t& v, const vertex2_t& neighbor, unsigned int n = 1 ) const;
               /**
                * Returns the nth vertex after \c neighbor in the neighborhood of \c v.
                *
                * \returns The vertex found or a null vertex if there is any problem.
                */
               const vertex1_t& nextTo( const vertex2_t& v, const vertex1_t& neighbor, unsigned int n = 1 ) const;

               /**
                * Returns the nth vertex before ref in the neighborhood of v.
                *
                * If ref is not in the neighborhood of v, return a null vertex.
                */
               const vertex2_t& prevTo( const vertex1_t& v, const vertex2_t& ref, unsigned int n = 1 ) const;
               /**
                * Returns the nth vertex before ref in the neighborhood of v.
                *
                * If ref is not in the neighborhood of v, return a null vertex.
                */
               const vertex1_t& prevTo( const vertex2_t& v, const vertex1_t& ref, unsigned int n = 1 ) const;

               /**
                * Returns the edge from \c src to \c tgt.
                *
                * If the edge does not exists, returns a null edge.
                */
               edge1_t edge( const vertex1_t& src, const vertex2_t& tgt );
               /**
                * Returns the edge from \c src to \c tgt.
                *
                * If the edge does not exists, returns a null edge.
                */
               edge2_t edge( const vertex2_t& src, const vertex1_t& tgt );

               /**
                * Returns the edge from \c src to \c tgt.
                *
                * If the edge does not exists, returns a null edge.
                */
               const_edge1_t edge( const vertex1_t& src, const vertex2_t& tgt ) const;
               /**
                * Returns the edge from \c src to \c tgt.
                *
                * If the edge does not exists, returns a null edge.
                */
               const_edge2_t edge( const vertex2_t& src, const vertex1_t& tgt ) const;

               /**
                * Flag a neighbor of \p src for quick retrieval
                */
               bool flag(const vertex1_t& src, const vertex2_t& neighbor);
               /**
                * Flag a neighbor of \p src for quick retrieval
                */
               bool flag(const vertex2_t& src, const vertex1_t& neighbor);

               /**
                * Return the flagged neighbor of \p src
                */
               const vertex2_t& flagged(const vertex1_t& src) const;
               /**
                * Return the flagged neighbor of \p src
                */
               const vertex1_t& flagged(const vertex2_t& src) const;

               /**
                * Return the constant range of neighbors of \c v
                */
               const_neighbor_iterator1_range neighbors(const vertex1_t& v) const;
               /**
                * Return the constant range of neighbors of \c v
                */
               const_neighbor_iterator2_range neighbors(const vertex2_t& v) const;

               /**
                * Return the range of neighbors of \c v
                */
               neighbor_iterator1_range neighbors(const vertex1_t& v);
               /**
                * Return the range of neighbors of \c v
                */
               neighbor_iterator2_range neighbors(const vertex2_t& v);


               /**
                * Return the constant range of neighbors of \c v starting from \c n
                */
               const_circ_neighbor_iterator1_range neighbors(const vertex1_t& v, const vertex2_t& n) const;
               /**
                * Return the constant range of neighbors of \c v starting from \c n
                */
               const_circ_neighbor_iterator2_range neighbors(const vertex2_t& v, const vertex1_t& n) const;

               /**
                * Return the range of neighbors of \c v starting from \c n
                */
               circ_neighbor_iterator1_range neighbors(const vertex1_t& v, const vertex2_t& n);
               /**
                * Return the range of neighbors of \c v starting from \c n
                */
               circ_neighbor_iterator2_range neighbors(const vertex2_t& v, const vertex1_t& n);

               /**
                * Return the range of incoming neighbors of \c v.
                *
                * Iterating over iNeighbors will go through all the vertices having an edge 
                * toward \c v.
                */
               ineighbor_iterator1_range iNeighbors(const vertex1_t& v) const;
               /**
                * Return the range of incoming neighbors of \c v.
                *
                * Iterating over iNeighbors will go through all the vertices having an edge 
                * toward \c v.
                */
               ineighbor_iterator2_range iNeighbors(const vertex2_t& v) const;

               /**
                * Return the source vertex of the edge
                */
               const vertex1_t& source( const edge1_t& edge ) const
               {
                 typename neighborhood1_t::const_iterator found = this->findVertex(vertex1_t(edge.source()));
                 if(found != neighborhood1.end())
                   return found->first;
                 return vertex1_t::null;//vertex_t(0);
               }

               /**
                * Return the source vertex of the edge
                */
               const vertex2_t& source( const edge2_t& edge ) const
               {
                 typename neighborhood2_t::const_iterator found = this->findVertex(vertex2_t(edge.source()));
                 if(found != neighborhood2.end())
                   return found->first;
                 return vertex2_t::null;//vertex_t(0);
               }

               /**
                * Return the target vertex of the edge
                */
               const vertex2_t& target( const edge1_t& edge ) const
               {
                 typename neighborhood2_t::const_iterator found = this->findVertex(vertex2_t(edge.target()));
                 if(found != neighborhood2.end())
                   return found->first;
                 return vertex2_t::null;//vertex_t(0);
               }

               /**
                * Return the target vertex of the edge
                */
               const vertex1_t& target( const edge2_t& edge ) const
               {
                 typename neighborhood1_t::const_iterator found = this->findVertex(vertex1_t(edge.target()));
                 if(found != neighborhood1.end())
                   return found->first;
                 return vertex1_t::null;//vertex_t(0);
               }

               //@}

               ///\name Neighborhood edition methods
               //@{
               /**
                * Erase the edge from \c src to \c tgt.
                *
                * \returns True if successful.
                */
               size_type eraseEdge( const vertex1_t& src, const vertex2_t& tgt);
               /**
                * Erase the edge from \c src to \c tgt.
                *
                * \returns True if successful.
                */
               size_type eraseEdge( const vertex2_t& src, const vertex1_t& tgt);

               /**
                * Replace a vertex by another in a neighborhood.
                *
                * \param[in] v Vertex whose neighborhood is changed.
                * \param[in] neighbor Vertex to replace
                * \param[in] new_neighbor Vertex replacing \c neighbor
                *
                * If \c new_neighbor is already in the neighborhood of \c v, then the 
                * operation fails and nothing is changed.
                *
                * \returns The edge between \c v and \c new_neighbor or a null edge if 
                * anything goes wrong.
                */
               edge1_t replace( const vertex1_t& v, const vertex2_t& neighbor, const vertex2_t& new_neighbor );
               /**
                * Replace a vertex by another in a neighborhood.
                *
                * \param[in] v Vertex whose neighborhood is changed.
                * \param[in] neighbor Vertex to replace
                * \param[in] new_neighbor Vertex replacing \c neighbor
                *
                * If \c new_neighbor is already in the neighborhood of \c v, then the 
                * operation fails and nothing is changed.
                *
                * \returns The edge between \c v and \c new_neighbor or a null edge if 
                * anything goes wrong.
                */
               edge2_t replace( const vertex2_t& v, const vertex1_t& neighbor, const vertex1_t& new_neighbor );

               /**
                * Insert neighbor in the neighborhood of v after reference.
                *
                * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
                * fails.
                *
                * \returns The just created edge if everything succeed, or a null edge.
                */
               edge1_t spliceAfter( const vertex1_t& v, const vertex2_t& neighbor, const vertex2_t& new_neighbor );
               /**
                * Insert neighbor in the neighborhood of v after reference.
                *
                * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
                * fails.
                *
                * \returns The just created edge if everything succeed, or a null edge.
                */
               edge2_t spliceAfter( const vertex2_t& v, const vertex1_t& neighbor, const vertex1_t& new_neighbor );

               /**
                * Insert neighbor in the neighborhood of v before reference.
                *
                * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
                * fails.
                *
                * \returns The just created edge if everything succeed, or a null edge.
                */
               edge1_t spliceBefore( const vertex1_t& v, const vertex2_t& neighbor, const vertex2_t& new_neighbor );
               /**
                * Insert neighbor in the neighborhood of v before reference.
                *
                * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
                * fails.
                *
                * \returns The just created edge if everything succeed, or a null edge.
                */
               edge2_t spliceBefore( const vertex2_t& v, const vertex1_t& neighbor, const vertex1_t& new_neighbor );

               /**
                * Insert a new edge in the graph, without ordering.
                *
                * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
                * fails.
                *
                * \returns The just created edge if everything succeed, or a null edge.
                */
               edge1_t insertEdge( const vertex1_t& src, const vertex2_t& tgt );
               /**
                * Insert a new edge in the graph, without ordering.
                *
                * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
                * fails.
                *
                * \returns The just created edge if everything succeed, or a null edge.
                */
               edge2_t insertEdge( const vertex2_t& src, const vertex1_t& tgt );

               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges, to and from \c v will be erased.
                */
               bool eraseAllEdges( const vertex1_t& v );
               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges, to and from \c v will be erased.
                */
               bool eraseAllEdges( const vertex2_t& v );

               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges from \c v will be erased.
                */
               bool clear( const vertex1_t& v );
               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges from \c v will be erased.
                */
               bool clear( const vertex2_t& v );

               //@}

               ///\name Whole structure methods
               //@{
               /**
                * Copy the graph into another graph
                */
               VVBiGraph& operator=( const VVBiGraph& other );

               /**
                * Swap the content of two graphs
                */
               void swap(VVBiGraph& other);

               /**
                * Test equality for the graphs.
                *
                * Two graphs are equal if they shared the sames vertices and their 
                * neighborhood are equivalent.
                *
                * \note Current implementation do not consider invariance of 
                * neighborhood by rotation.
                */
               bool operator==(const VVBiGraph& other) const;

               /**
                * Clear the graph.
                */
               void clear() { neighborhood1.clear(); neighborhood2.clear(); lookup1.clear(); lookup2.clear(); }

               /**
                * Clear all the edges in the graph
                *
                * Example:
                * \code
                * vvgraph S;
                * // Initialize S
                * vvgraph::size_type s1 = S.size(); // Save the number of vertices
                * S.eraseAllEdges();
                * assert(s1 == S.size()); // The number of vertices didn't change
                * forall_named(const vertex1_t& v, S, vertex1)
                * {
                *   assert(S.empty(v)); // No neighbor
                * }
                * forall_named(const vertex2_t& v, S, vertex2)
                * {
                *   assert(S.empty(v)); // No neighbor
                * }
                * \endcode
                */
               void eraseAllEdges();

               /*
                * Extract the subgraph containing the \p vertices
                *
                * The subgraph is the set of vertices and the edges whose source and target 
                * are in the extracted vertices.
                */
               template <typename Vertex1Container, typename Vertex2Container>
                 VVBiGraph subgraph( const Vertex1Container& vertices1, const Vertex2Container& vertices2 ) const;
               //@}

               ///\name STL compatibility methods
               //@{

               /**
                * Erase element at position \p pos
                *
                * Returns the iterator on the next element (as required for 
                * sequence containers)
                */
               iterator1 erase(iterator1 pos);
               /**
                * Erase element at position \p pos
                *
                * Returns the iterator on the next element (as required for 
                * sequence containers)
                */
               iterator2 erase(iterator2 pos);

               /**
                * Erase the edge pointed to by the iterator.
                */
               neighbor_iterator1 eraseEdge(const vertex1_t& v, neighbor_iterator1 pos);

               /**
                * Erase the edge pointed to by the iterator.
                */
               circ_neighbor_iterator1 eraseEdge(const vertex1_t& v, circ_neighbor_iterator1 pos);

               /**
                * Erase the edge pointed to by the iterator.
                */
               neighbor_iterator2 eraseEdge(const vertex2_t& v, neighbor_iterator2 pos);

               /**
                * Erase the edge pointed to by the iterator.
                */
               circ_neighbor_iterator2 eraseEdge(const vertex2_t& v, circ_neighbor_iterator2 pos);

               /**
                * Insert a new vertex in the graph
                *
                * The iterator is ignored ...
                */
               iterator1 insert( iterator1 pos, const vertex1_t& v );
               /**
                * Insert a new vertex in the graph
                *
                * The iterator is ignored ...
                */
               iterator2 insert( iterator2 pos, const vertex2_t& v );

               /**
                * Insert all the vertices from first to last-1 in the graph.
                */
               void insert( iterator1 first, iterator1 last );
               /**
                * Insert all the vertices from first to last-1 in the graph.
                */
               void insert( iterator2 first, iterator2 last );

               /**
                * Returns an iterator on the beginning of the set of vertices of the graph.
                */
               iterator1 begin_vertex1() { return neighborhood1.begin(); }
               /**
                * Returns an iterator on the beginning of the set of vertices of the graph.
                */
               iterator2 begin_vertex2() { return neighborhood2.begin(); }

               /**
                * Returns an iterator on the end of the set of vertices of the graph.
                */
               iterator1 end_vertex1() { return neighborhood1.end(); }
               /**
                * Returns an iterator on the end of the set of vertices of the graph.
                */
               iterator2 end_vertex2() { return neighborhood2.end(); }

               /**
                * Returns a range of vertices
                */
               const_range_vertex1 vertices1() const { return util::make_range(begin_vertex1(), end_vertex1()); }
               /**
                * Returns a range of vertices
                */
               const_range_vertex2 vertices2() const { return util::make_range(begin_vertex2(), end_vertex2()); }

               /**
                * Returns a range of vertices
                */
               range_vertex1 vertices1() { return util::make_range(begin_vertex1(), end_vertex1()); }
               /**
                * Returns a range of vertices
                */
               range_vertex2 vertices2() { return util::make_range(begin_vertex2(), end_vertex2()); }

               /**
                * Returns an iterator on the beginning of the set of vertices of the graph.
                */
               const_iterator1 begin_vertex1() const { return neighborhood1.begin(); }
               /**
                * Returns an iterator on the beginning of the set of vertices of the graph.
                */
               const_iterator2 begin_vertex2() const { return neighborhood2.begin(); }

               /**
                * Returns an iterator on the end of the set of vertices of the graph.
                */
               const_iterator1 end_vertex1() const { return neighborhood1.end(); }
               /**
                * Returns an iterator on the end of the set of vertices of the graph.
                */
               const_iterator2 end_vertex2() const { return neighborhood2.end(); }

               /**
                * Find the vertex \c v in the graph.
                *
                * \returns An iterator on \c v if found, or end()
                */
               iterator1 find(const vertex1_t& v) { return this->findVertex(v); }
               /**
                * Find the vertex \c v in the graph.
                *
                * \returns An iterator on \c v if found, or end()
                */
               iterator2 find(const vertex2_t& v) { return this->findVertex(v); }

               /**
                * Find the vertex \c v in the graph.
                *
                * \returns An iterator on \c v if found, or end()
                */
               const_iterator1 find(const vertex1_t& v) const { return this->findVertex(v); }
               /**
                * Find the vertex \c v in the graph.
                *
                * \returns An iterator on \c v if found, or end()
                */
               const_iterator2 find(const vertex2_t& v) const { return this->findVertex(v); }

               /**
                * Count the number of vertices \p v in the graph (can be 0 or 1 only)
                */
               size_type count( const vertex1_t& v ) const;
               /**
                * Count the number of vertices \p v in the graph (can be 0 or 1 only)
                */
               size_type count( const vertex2_t& v ) const;

               /**
                * Find the vertex \c n in the neighborhood of \c v.
                *
                * \returns
                *  - An iterator on \c n in the neighborhood of \c v, if found
                *  - An iterator on the end of the neighborhood if \c n is not found in \c 
                *  v
                *  - Otherwise, the result is undefined
                */
               neighbor_iterator1 findIn(const vertex1_t& v, const vertex2_t& n);
               /**
                * Find the vertex \c n in the neighborhood of \c v.
                *
                * \returns
                *  - An iterator on \c n in the neighborhood of \c v, if found
                *  - An iterator on the end of the neighborhood if \c n is not found in \c 
                *  v
                *  - Otherwise, the result is undefined
                */
               neighbor_iterator2 findIn(const vertex2_t& v, const vertex1_t& n);

               /**
                * Find the vertex \c n in the neighborhood of \c v.
                *
                * \returns
                *  - An iterator on \c n in the neighborhood of \c v, if found
                *  - An iterator on the end of the neighborhood if \c n is not found in \c 
                *  v
                *  - Otherwise, the result is undefined
                */
               const_neighbor_iterator1 findIn(const vertex1_t& v, const vertex2_t& n) const;
               /**
                * Find the vertex \c n in the neighborhood of \c v.
                *
                * \returns
                *  - An iterator on \c n in the neighborhood of \c v, if found
                *  - An iterator on the end of the neighborhood if \c n is not found in \c 
                *  v
                *  - Otherwise, the result is undefined
                */
               const_neighbor_iterator2 findIn(const vertex2_t& v, const vertex1_t& n) const;

               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges, to and from \c *it will be erased.
                */
               void eraseAllEdges( iterator1 it );
               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges, to and from \c *it will be erased.
                */
               void eraseAllEdges( iterator2 it );

               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges from \c *it will be erased.
                */
               void clear( iterator1 it );
               /**
                * Clear the neighborhood of a vertex.
                *
                * All edges from \c *it will be erased.
                */
               void clear( iterator2 it );

               //@}

               protected:
               /**
                * Type of the result of the search for a vertex in a neighborhood
                *
                * This data structure is made such that all edit or lookup operations can 
                * be done from that data structure.
                */
               template <class Neighborhood, class Iterator>
                 struct search_result_t
                 {
                   /**
                    * Default constructor
                    *
                    * By default, the search is unsuccessful
                    */
                   search_result_t()
                     : found(false)
                     , neighborhood(0) {}

                   /**
                    * Successful constructor
                    */
                   search_result_t( Iterator i,
                                    Neighborhood* n,
                                    bool ok = true)
                     : found(ok)
                     , it(i)
                     , neighborhood(n)
                     {}

                   /**
                    * Copy constructor
                    */
                   search_result_t(const search_result_t& copy)
                     : found(copy.found)
                     , it(copy.it)
                     , neighborhood(copy.neighborhood)
                     {}

                   /**
                    * Convert the object to true if the search is successful
                    */
                   operator bool() { return found; }

                   /**
                    * True if the search was completely successful
                    */
                   bool found;
                   /**
                    * Iterator pointing in container
                    */
                   Iterator it;
                   /**
                    * Container where the element is found
                    */
                   Neighborhood* neighborhood;
                 };

               /**
                * Iterator on the outgoing edges
                */
               typedef typename edge_list1_t::iterator int_neighbor_iterator1;
               /**
                * Iterator on the outgoing edges
                */
               typedef typename edge_list2_t::iterator int_neighbor_iterator2;

               /**
                * Constant iterator on the outgoing edges
                */
               typedef typename edge_list1_t::const_iterator int_const_neighbor_iterator1;
               /**
                * Constant iterator on the outgoing edges
                */
               typedef typename edge_list2_t::const_iterator int_const_neighbor_iterator2;

               /**
                * Result of a search in the neighborhood
                */
               typedef search_result_t<single_neighborhood1_t, int_neighbor_iterator1> neighbor1_found_t;
               /**
                * Result of a search in the neighborhood
                */
               typedef search_result_t<single_neighborhood2_t, int_neighbor_iterator2> neighbor2_found_t;

               /**
                * Constant result of a search in the neighborhood
                */
               typedef search_result_t<const single_neighborhood1_t, int_const_neighbor_iterator1> const_neighbor1_found_t;
               /**
                * Constant result of a search in the neighborhood
                */
               typedef search_result_t<const single_neighborhood2_t, int_const_neighbor_iterator2> const_neighbor2_found_t;

               /**
                * Find a vertex in the graph and returns the iterator on it.
                *
                * Goal is to introduce a new optimization. When iterating over the 
                * vertices of the graph, they will cache their own iterator. So accessing 
                * their neighborhood would be constant time.
                */
               typename neighborhood1_t::iterator findVertex(const vertex1_t& v);
               /**
                * Find a vertex in the graph and returns the iterator on it.
                *
                * Goal is to introduce a new optimization. When iterating over the 
                * vertices of the graph, they will cache their own iterator. So accessing 
                * their neighborhood would be constant time.
                */
               typename neighborhood2_t::iterator findVertex(const vertex2_t& v);

               /**
                * Constant version of findVertex(const vertex1_t&)
                */
               typename neighborhood1_t::const_iterator findVertex(const vertex1_t& v) const;
               /**
                * Constant version of findVertex(const vertex2_t&)
                */
               typename neighborhood2_t::const_iterator findVertex(const vertex2_t& v) const;

               /**
                * Find a vertex neighbor in the neighborhood of v.
                *
                * If 'v' is in the graph and 'neighbor' is in its neighborhood, the 
                * result is convertible to true, 'neighborhood' points toward the 
                * neighborhood of \p v and 'it' points toward the neighbor itself.
                *
                * If 'v' is in the graph, but 'neighbor' is not in the neighborhood, the 
                * result is convertible to false, 'neighborhood' points toward the 
                * neighborhood of \p v and 'it' is neighborhood->edges.end().
                *
                * If v is not in the graph, the result is convertible to false and the 
                * address of the stored neighborhood is 0.
                */
               neighbor1_found_t findInVertex( const vertex1_t& v, const vertex2_t& neighbor);
               /**
                * Find a vertex neighbor in the neighborhood of v.
                *
                * If 'v' is in the graph and 'neighbor' is in its neighborhood, the 
                * result is convertible to true, 'neighborhood' points toward the 
                * neighborhood of \p v and 'it' points toward the neighbor itself.
                *
                * If 'v' is in the graph, but 'neighbor' is not in the neighborhood, the 
                * result is convertible to false, 'neighborhood' points toward the 
                * neighborhood of \p v and 'it' is neighborhood->edges.end().
                *
                * If v is not in the graph, the result is convertible to false and the 
                * address of the stored neighborhood is 0.
                */
               neighbor2_found_t findInVertex( const vertex2_t& v, const vertex1_t& neighbor);

               /**
                * Constant version of findInVertex(const vertex_t&, const vertex_t&)
                */
               const_neighbor1_found_t findInVertex( const vertex1_t& v, const vertex2_t& neighbor) const;
               /**
                * Constant version of findInVertex(const vertex_t&, const vertex_t&)
                */
               const_neighbor2_found_t findInVertex( const vertex2_t& v, const vertex1_t& neighbor) const;

               /**
                * Insert v in the in_edges of new_neighbor and return true if the insertion 
                * was actually done, or false if v was already there.
                */
               std::pair<ineighbor_iterator2,bool> insertInEdge(const vertex1_t& v, const vertex2_t& new_neighbor);
               /**
                * Insert v in the in_edges of new_neighbor and return true if the insertion 
                * was actually done, or false if v was already there.
                */
               std::pair<ineighbor_iterator1,bool> insertInEdge(const vertex2_t& v, const vertex1_t& new_neighbor);

               /**
                * Undo the insertion of an in edge as a roll-back mechanism if the whole 
                * edge insertion fails.
                */
               void undoInEdge(const vertex1_t& new_neighbor, ineighbor_iterator1& it);

               /**
                * Undo the insertion of an in edge as a roll-back mechanism if the whole 
                * edge insertion fails.
                */
               void undoInEdge(const vertex2_t& new_neighbor, ineighbor_iterator2& it);

               /**
                * Update the edge cache information of vertex v in its neighborhood and 
                * the in_edge too.
                */
               void updateEdgeCache(const vertex1_t& v,
                                    typename edge_list1_t::iterator new_edge_it,
                                    const vertex1_t& in_edge);
               /**
                * Update the edge cache information of vertex v in its neighborhood and 
                * the in_edge too.
                */
               void updateEdgeCache(const vertex2_t& v,
                                    typename edge_list2_t::iterator new_edge_it,
                                    const vertex2_t& in_edge);

               /**
                * Update the vertex cache for vertex v
                */
               void updateVertexCache(const vertex1_t& v,
                                      typename neighborhood1_t::iterator it);
               /**
                * Update the vertex cache for vertex v
                */
               void updateVertexCache(const vertex2_t& v,
                                      typename neighborhood2_t::iterator it);

               /**
                * Main data structure containing the neighborhood of the vertex1's
                */
               neighborhood1_t neighborhood1;
               /**
                * Main data structure containing the neighborhood of the vertex2's
                */
               neighborhood2_t neighborhood2;

               /**
                * Lookup hash map for vertex1's
                */
               lookup1_t lookup1;
               /**
                * Lookup hash map for vertex2's
                */
               lookup2_t lookup2;

               /**
                * Unique id for this graph.
                */
               intptr_t graph_id;

             };


#ifdef USE_ALLOCATOR
  template <BIGRAPH_TEMPLATE> template <typename V1, typename V2, typename E>
    typename VVBiGraph<BIGRAPH_ARGS>::template neighbor_t<V1,V2,E>::EdgeAlloc
    VVBiGraph<BIGRAPH_ARGS>::neighbor_t<V1,V2,E>::edge_alloc;
#endif

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::iterator1
    VVBiGraph<BIGRAPH_ARGS>::erase( iterator1 it )
    {
      eraseAllEdges(it);
      lookup1.erase(*it);
      return neighborhood1.erase(it.base());
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::iterator2
    VVBiGraph<BIGRAPH_ARGS>::erase( iterator2 it )
    {
      eraseAllEdges(it);
      lookup2.erase(*it);
      return neighborhood2.erase(it.base());
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::erase( const vertex1_t& v )
    {
      if(eraseAllEdges(v))
      {
        typename lookup1_t::iterator it_found = lookup1.find(v);
        if(it_found == lookup1.end()) return 0;
        typename neighborhood1_t::iterator it = it_found->second;
        neighborhood1.erase(it);
        lookup1.erase(v);
        return 1;
      }
      return 0;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::erase( const vertex2_t& v )
    {
      if(eraseAllEdges(v))
      {
        typename lookup2_t::iterator it_found = lookup2.find(v);
        if(it_found == lookup2.end()) return 0;
        typename neighborhood2_t::iterator it = it_found->second;
        neighborhood2.erase(it);
        lookup2.erase(v);
        return 1;
      }
      return 0;
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::eraseAllEdges( const vertex1_t& v )
    {
      typename neighborhood1_t::iterator it_found = this->findVertex(v);
      if(it_found != neighborhood1.end() )
      {
        eraseAllEdges(it_found);
        return true;
      }
      return false;
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::eraseAllEdges( const vertex2_t& v )
    {
      typename neighborhood2_t::iterator it_found = this->findVertex(v);
      if(it_found != neighborhood2.end() )
      {
        eraseAllEdges(it_found);
        return true;
      }
      return false;
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::eraseAllEdges( iterator1 it_found )
    {
      const vertex1_t& v = *it_found;
      // Remove all the edges and their incoming counterpart
      for(typename edge_list1_t::iterator it_el = it_found.base()->second.edges.begin() ;
          it_el != it_found.base()->second.edges.end() ; ++it_el)
      {
        typename lookup2_t::iterator it_found = lookup2.find(it_el->target);
        if(it_found != lookup2.end())
          it_found->second->second.in_edges.erase(v);
      }
      it_found.base()->second.edges.clear();
      it_found.base()->second.flagged = 0;
      // Remove the incoming edges
      in_edges1_t &in_edges = it_found.base()->second.in_edges;
      for( typename in_edges1_t::iterator it = in_edges.begin() ;
           it != in_edges.end() ; ++it )
      {
#ifndef VVBIGRAPH_NO_EDGE_CACHE
        EdgeCache2 *cache = reinterpret_cast<EdgeCache2*>(it->cache);
        if(!cache or !cache->eraseEdge())
        {
#endif
          neighbor2_found_t found = findInVertex(*it, v);
          found.neighborhood->edges.erase(found.it);
          if(found.neighborhood->flagged && *found.neighborhood->flagged == v) found.neighborhood->flagged = 0;
#ifndef VVBIGRAPH_NO_EDGE_CACHE
        }
#endif
      }
      in_edges.clear();
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::eraseAllEdges( iterator2 it_found )
    {
      const vertex2_t& v = *it_found;
      for(typename edge_list2_t::iterator it_el = it_found.base()->second.edges.begin() ;
          it_el != it_found.base()->second.edges.end() ; ++it_el)
      {
        typename lookup1_t::iterator it_found = lookup1.find(it_el->target);
        if(it_found != lookup1.end())
          it_found->second->second.in_edges.erase(v);
      }
      it_found.base()->second.edges.clear();
      // Remove the incoming edges
      in_edges2_t &in_edges = it_found.base()->second.in_edges;
      for( typename in_edges2_t::iterator it = in_edges.begin() ;
           it != in_edges.end() ; ++it )
      {
#ifndef VVBIGRAPH_NO_EDGE_CACHE
        EdgeCache1 *cache = reinterpret_cast<EdgeCache1*>(it->cache);
        if(!cache or !cache->eraseEdge())
        {
#endif
          neighbor1_found_t found = findInVertex(*it, v);
          found.neighborhood->edges.erase(found.it);
          if(found.neighborhood->flagged && *found.neighborhood->flagged == v) found.neighborhood->flagged = 0;
#ifndef VVBIGRAPH_NO_EDGE_CACHE
        }
#endif
      }
      in_edges.clear();
    }


  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::clear( const vertex1_t& v )
    {
      typename neighborhood1_t::iterator it_found = this->findVertex(v);
      if(it_found != neighborhood1.end() )
      {
        clear(it_found);
        return true;
      }
      return false;
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::clear( const vertex2_t& v )
    {
      typename neighborhood2_t::iterator it_found = this->findVertex(v);
      if(it_found != neighborhood2.end() )
      {
        clear(it_found);
        return true;
      }
      return false;
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::clear( iterator1 it_found )
    {
      const vertex1_t& v = *it_found;
      for(typename edge_list1_t::iterator it_el = it_found.base()->second.edges.begin() ;
          it_el != it_found.base()->second.edges.end() ; ++it_el)
      {
        typename neighborhood2_t::iterator it_found = this->findVertex(it_el->target);
        it_found->second.in_edges.erase(v);
      }
      it_found.base()->second.edges.clear();
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::clear( iterator2 it_found )
    {
      const vertex2_t& v = *it_found;
      for(typename edge_list2_t::iterator it_el = it_found.base()->second.edges.begin() ;
          it_el != it_found.base()->second.edges.end() ; ++it_el)
      {
        typename neighborhood1_t::iterator it_found = this->findVertex(it_el->target);
        it_found->second.in_edges.erase(v);
      }
      it_found.base()->second.edges.clear();
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::eraseAllEdges()
    {
      for(typename neighborhood1_t::iterator it = neighborhood1.begin() ;
          it != neighborhood1.end() ; ++it)
      {
        it->second.edges.clear();
        it->second.in_edges.clear();
        it->second.flagged = 0;
      }
      for(typename neighborhood2_t::iterator it = neighborhood2.begin() ;
          it != neighborhood2.end() ; ++it)
      {
        it->second.edges.clear();
        it->second.in_edges.clear();
        it->second.flagged = 0;
      }
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::count( const vertex1_t& v ) const
    {
      return lookup1.count(v);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::count( const vertex2_t& v ) const
    {
      return lookup2.count(v);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::iterator1
    VVBiGraph<BIGRAPH_ARGS>::insert( const vertex1_t& v )
    {
      typename lookup1_t::iterator it_found;
      bool inserted;
      util::tie(it_found, inserted) = lookup1.insert(std::make_pair(v, typename neighborhood1_t::iterator()));
      if(inserted)
      {
        neighborhood1.push_front(std::make_pair(v,single_neighborhood1_t()));
        typename neighborhood1_t::iterator it = neighborhood1.begin();
        it_found->second = it;
        updateVertexCache(it->first, it);
        return it;
      }
      return it_found->second;

      //neighborhood1.push_front(std::make_pair(v,single_neighborhood1_t()));
      //typename neighborhood1_t::iterator it = neighborhood1.begin();
      //lookup1[v] = it;
      //updateVertexCache(it->first, it);
      //return it;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::iterator2
    VVBiGraph<BIGRAPH_ARGS>::insert( const vertex2_t& v )
    {
      typename lookup2_t::iterator it_found;
      bool inserted;
      util::tie(it_found, inserted) = lookup2.insert(std::make_pair(v, typename neighborhood2_t::iterator()));
      if(inserted)
      {
        neighborhood2.push_front(std::make_pair(v,single_neighborhood2_t()));
        typename neighborhood2_t::iterator it = neighborhood2.begin();
        it_found->second = it;
        updateVertexCache(it->first, it);
        return it;
      }
      return it_found->second;

      //neighborhood2.push_front(std::make_pair(v,single_neighborhood2_t()));
      //typename neighborhood2_t::iterator it = neighborhood2.begin();
      //lookup2[v] = it;
      //updateVertexCache(it->first, it);
      //return it;
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::updateVertexCache(const vertex1_t& v,
                                                    typename neighborhood1_t::iterator it)
    {
#ifndef VVBIGRAPH_NO_VERTEX_CACHE
      it->second.update(graph_id, v, it);
#endif
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::updateVertexCache(const vertex2_t& v,
                                                    typename neighborhood2_t::iterator it)
    {
#ifndef VVBIGRAPH_NO_VERTEX_CACHE
      it->second.update(graph_id, v, it);
#endif
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::iterator1
    VVBiGraph<BIGRAPH_ARGS>::insert( iterator1 , const vertex1_t& v )
    {
      return this->insert(v);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::iterator2
    VVBiGraph<BIGRAPH_ARGS>::insert( iterator2 , const vertex2_t& v )
    {
      return this->insert(v);
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::insert( iterator1 first, iterator1 last )
    {
      for(iterator1 it = first ; it != last ; ++it)
      {
        insert(*it);
      }
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::insert( iterator2 first, iterator2 last )
    {
      for(iterator2 it = first ; it != last ; ++it)
      {
        insert(*it);
      }
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::any_vertex1() const
    {
      if(neighborhood1.empty())
        return vertex1_t::null;
      return *begin_vertex1();
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::any_vertex2() const
    {
      if(neighborhood2.empty())
        return vertex2_t::null;
      return *begin_vertex2();
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::get_vertex1(size_type idx) const
    {
      typename neighborhood1_t::const_iterator it = neighborhood1.begin();
      std::advance(it, idx);
      return it->first;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::get_vertex2(size_type idx) const
    {
      typename neighborhood2_t::const_iterator it = neighborhood2.begin();
      std::advance(it, idx);
      return it->first;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::anyIn( const vertex1_t& v ) const
    {
      if(v.isNull())
        return vertex2_t::null;//vertex_t(0);
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end() || it_found->second.edges.empty())
        return vertex2_t::null;//vertex_t(0);
      const edge_list1_t& lst = it_found->second.edges;
      return lst.front().target;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::anyIn( const vertex2_t& v ) const
    {
      if(v.isNull())
        return vertex1_t::null;//vertex_t(0);
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end() || it_found->second.edges.empty())
        return vertex1_t::null;//vertex_t(0);
      const edge_list2_t& lst = it_found->second.edges;
      return lst.front().target;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::valence( const vertex1_t& v ) const
    {
      if(v.isNull())
        return 0;
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return 0;
      return it_found->second.edges.size();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::valence( const vertex2_t& v ) const
    {
      if(v.isNull())
        return 0;
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return 0;
      return it_found->second.edges.size();
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::empty( const vertex1_t& v ) const
    {
      if(v.isNull())
        return true;
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return true;
      return it_found->second.edges.empty();
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::empty( const vertex2_t& v ) const
    {
      if(v.isNull())
        return true;
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return true;
      return it_found->second.edges.empty();
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::iAnyIn( const vertex1_t& v ) const
    {
      if(v.isNull())
        return vertex2_t::null;//vertex_t(0);
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end() || it_found->second.in_edges.empty())
        return vertex2_t::null;//vertex_t(0);
      const in_edges1_t& lst = it_found->second.in_edges;
      return *lst.begin();
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::iAnyIn( const vertex2_t& v ) const
    {
      if(v.isNull())
        return vertex1_t::null;//vertex_t(0);
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end() || it_found->second.in_edges.empty())
        return vertex1_t::null;//vertex_t(0);
      const in_edges2_t& lst = it_found->second.in_edges;
      return *lst.begin();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::iValence( const vertex1_t& v ) const
    {
      if(v.isNull())
        return 0;
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return 0;
      return it_found->second.in_edges.size();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::iValence( const vertex2_t& v ) const
    {
      if(v.isNull())
        return 0;
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return 0;
      return it_found->second.in_edges.size();
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::iEmpty( const vertex1_t& v ) const
    {
      if(v.isNull())
        return true;
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return true;
      return it_found->second.in_edges.empty();
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::iEmpty( const vertex2_t& v ) const
    {
      if(v.isNull())
        return true;
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return true;
      return it_found->second.in_edges.empty();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighborhood1_t::iterator
    VVBiGraph<BIGRAPH_ARGS>::findVertex( const vertex1_t& v )
    {
#ifndef VVBIGRAPH_NO_VERTEX_CACHE
      if(v.cache)
      {
        if(v.cache_source == 0)
        {
          VertexCache1* cache = reinterpret_cast<VertexCache1*>(v.cache);
          if(cache->graph_id() == graph_id)
          {
            return cache->neighborhood();
          }
        }
      }
#endif // VVBIGRAPH_NO_VERTEX_CACHE
      typename lookup1_t::const_iterator it_found = lookup1.find(v);
      if(it_found != lookup1.end())
        return it_found->second;
      return neighborhood1.end();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighborhood2_t::iterator
    VVBiGraph<BIGRAPH_ARGS>::findVertex( const vertex2_t& v )
    {
#ifndef VVBIGRAPH_NO_VERTEX_CACHE
      if(v.cache)
      {
        if(v.cache_source == 0)
        {
          VertexCache2* cache = reinterpret_cast<VertexCache2*>(v.cache);
          if(cache->graph_id() == graph_id)
          {
            return cache->neighborhood();
          }
        }
      }
#endif // VVBIGRAPH_NO_VERTEX_CACHE
      typename lookup2_t::const_iterator it_found = lookup2.find(v);
      if(it_found != lookup2.end())
        return it_found->second;
      return neighborhood2.end();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighborhood1_t::const_iterator
    VVBiGraph<BIGRAPH_ARGS>::findVertex( const vertex1_t& v ) const
    {
#ifndef VVBIGRAPH_NO_VERTEX_CACHE
      if(v.cache)
      {
        if(v.cache_source == 0)
        {
          VertexCache1* cache = reinterpret_cast<VertexCache1*>(v.cache);
          if(cache->graph_id() == graph_id)
          {
            return cache->neighborhood();
          }
        }
      }
#endif // VVBIGRAPH_NO_VERTEX_CACHE
      typename lookup1_t::const_iterator it_found = lookup1.find(v);
      if(it_found != lookup1.end())
        return it_found->second;
      return neighborhood1.end();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighborhood2_t::const_iterator
    VVBiGraph<BIGRAPH_ARGS>::findVertex( const vertex2_t& v ) const
    {
#ifndef VVBIGRAPH_NO_VERTEX_CACHE
      if(v.cache)
      {
        if(v.cache_source == 0)
        {
          VertexCache2* cache = reinterpret_cast<VertexCache2*>(v.cache);
          if(cache->graph_id() == graph_id)
          {
            return cache->neighborhood();
          }
        }
      }
#endif // VVBIGRAPH_NO_VERTEX_CACHE
      typename lookup2_t::const_iterator it_found = lookup2.find(v);
      if(it_found != lookup2.end())
        return it_found->second;
      return neighborhood2.end();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor1_found_t
    VVBiGraph<BIGRAPH_ARGS>::findInVertex( const vertex1_t& v, const vertex2_t& n)
    {
      if(v.isNull() || n.isNull())
        return neighbor1_found_t();
#ifndef VVBIGRAPH_NO_EDGE_CACHE
      if(n.cache_source == v.id())
      {
        EdgeCache1 *cache = reinterpret_cast<EdgeCache1*>(n.cache);
        if(cache->graph_id() == graph_id)
        {
          return neighbor1_found_t(cache->it(), cache->neighborhood());
        }
      }
#endif // VVBIGRAPH_NO_EDGE_CACHE
      typename neighborhood1_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return neighbor1_found_t();
      edge_list1_t &lst = it_found->second.edges;
      for(typename edge_list1_t::iterator it = lst.begin() ;
          it != lst.end() ; ++it)
      {
        if(it->target == n)
        {
          return neighbor1_found_t(it, &it_found->second);
        }
      }
      return neighbor1_found_t(lst.end(), &it_found->second, false);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor2_found_t
    VVBiGraph<BIGRAPH_ARGS>::findInVertex( const vertex2_t& v, const vertex1_t& n)
    {
      if(v.isNull() || n.isNull())
        return neighbor2_found_t();
#ifndef VVBIGRAPH_NO_EDGE_CACHE
      if(n.cache_source == v.id())
      {
        EdgeCache2 *cache = reinterpret_cast<EdgeCache2*>(n.cache);
        if(cache->graph_id() == graph_id)
        {
          return neighbor2_found_t(cache->it(), cache->neighborhood());
        }
      }
#endif // VVBIGRAPH_NO_EDGE_CACHE
      typename neighborhood2_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return neighbor2_found_t();
      edge_list2_t &lst = it_found->second.edges;
      for(typename edge_list2_t::iterator it = lst.begin() ;
          it != lst.end() ; ++it)
      {
        if(it->target == n)
        {
          return neighbor2_found_t(it, &it_found->second);
        }
      }
      return neighbor2_found_t(lst.end(), &it_found->second, false);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_neighbor1_found_t
    VVBiGraph<BIGRAPH_ARGS>::findInVertex( const vertex1_t& v, const vertex2_t& n) const
    {
      if(v.isNull() || n.isNull())
        return const_neighbor1_found_t();
#ifndef VVBIGRAPH_NO_EDGE_CACHE
      if(n.cache_source == v.id())
      {
        EdgeCache1 *cache = reinterpret_cast<EdgeCache1*>(n.cache);
        if(cache->graph_id() == graph_id)
        {
          return const_neighbor1_found_t(cache->it(), cache->neighborhood());
        }
      }
#endif // VVBIGRAPH_NO_EDGE_CACHE
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return const_neighbor1_found_t();
      const edge_list1_t &lst = it_found->second.edges;
      for(typename edge_list1_t::const_iterator it = lst.begin() ;
          it != lst.end() ; ++it)
      {
        if(it->target == n)
        {
          return const_neighbor1_found_t(it, &it_found->second);
        }
      }
      return const_neighbor1_found_t(lst.end(), &it_found->second, false);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_neighbor2_found_t
    VVBiGraph<BIGRAPH_ARGS>::findInVertex( const vertex2_t& v, const vertex1_t& n) const
    {
      if(v.isNull() || n.isNull())
        return const_neighbor2_found_t();
#ifndef VVBIGRAPH_NO_EDGE_CACHE
      if(n.cache_source == v.id())
      {
        EdgeCache2 *cache = reinterpret_cast<EdgeCache2*>(n.cache);
        if(cache->graph_id() == graph_id)
        {
          return const_neighbor2_found_t(cache->it(), cache->neighborhood());
        }
      }
#endif // VVBIGRAPH_NO_EDGE_CACHE
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return const_neighbor2_found_t();
      const edge_list2_t &lst = it_found->second.edges;
      for(typename edge_list2_t::const_iterator it = lst.begin() ;
          it != lst.end() ; ++it)
      {
        if(it->target == n)
        {
          return const_neighbor2_found_t(it, &it_found->second);
        }
      }
      return const_neighbor2_found_t(lst.end(), &it_found->second, false);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor_iterator1
    VVBiGraph<BIGRAPH_ARGS>::findIn(const vertex1_t& v, const vertex2_t& n)
    {
      neighbor1_found_t found = findInVertex(v, n);
      if(found.neighborhood)
        return neighbor_iterator1(found.it);
      return neighbor_iterator1();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor_iterator2
    VVBiGraph<BIGRAPH_ARGS>::findIn(const vertex2_t& v, const vertex1_t& n)
    {
      neighbor2_found_t found = findInVertex(v, n);
      if(found.neighborhood)
        return neighbor_iterator2(found.it);
      return neighbor_iterator2();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_neighbor_iterator1
    VVBiGraph<BIGRAPH_ARGS>::findIn(const vertex1_t& v, const vertex2_t& n) const
    {
      const_neighbor1_found_t found = findInVertex(v, n);
      if(found.neighborhood)
        return const_neighbor_iterator1(found.it);
      return const_neighbor_iterator1(found.it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_neighbor_iterator2
    VVBiGraph<BIGRAPH_ARGS>::findIn(const vertex2_t& v, const vertex1_t& n) const
    {
      const_neighbor2_found_t found = findInVertex(v, n);
      if(found.neighborhood)
        return const_neighbor_iterator2(found.it);
      return const_neighbor_iterator2(found.it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor_iterator1
    VVBiGraph<BIGRAPH_ARGS>::eraseEdge(const vertex1_t& v, neighbor_iterator1 pos)
    {
      typename neighborhood1_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return neighbor_iterator1();
      const vertex2_t& n = *pos;
      typename neighborhood2_t::iterator it_found2 = this->findVertex(n);
      if(it_found2 == neighborhood2.end())
        return neighbor_iterator1();
      size_type result = it_found2->second.in_edges.erase(v);
      if(result)
      {
        typename edge_list1_t::iterator it = pos.base();
        return it_found->second.edges.erase(it);
      }
      return neighbor_iterator1();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor_iterator2
    VVBiGraph<BIGRAPH_ARGS>::eraseEdge(const vertex2_t& v, neighbor_iterator2 pos)
    {
      typename neighborhood2_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return neighbor_iterator2();
      const vertex1_t& n = *pos;
      typename neighborhood1_t::iterator it_found2 = this->findVertex(n);
      if(it_found2 == neighborhood1.end())
        return neighbor_iterator2();
      size_type result = it_found2->second.in_edges.erase(v);
      if(result)
      {
        typename edge_list2_t::iterator it = pos.base();
        return it_found->second.edges.erase(it);
      }
      return neighbor_iterator2();
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::circ_neighbor_iterator1
    VVBiGraph<BIGRAPH_ARGS>::eraseEdge(const vertex1_t& v, circ_neighbor_iterator1 pos)
    {
      neighbor_iterator1 nit =  eraseEdge(v, pos.base());
      if(pos.isInitIterator(nit))
        return pos.end();
      return circ_neighbor_iterator1(pos, nit);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::circ_neighbor_iterator2
    VVBiGraph<BIGRAPH_ARGS>::eraseEdge(const vertex2_t& v, circ_neighbor_iterator2 pos)
    {
      neighbor_iterator2 nit =  eraseEdge(v, pos.base());
      if(pos.isInitIterator(nit))
        return pos.end();
      return circ_neighbor_iterator2(pos, nit);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::eraseEdge( const vertex1_t& v, const vertex2_t& n )
    {
      neighbor1_found_t found = this->findInVertex(v, n);
      if(!found)
        return false;
      if(found.neighborhood->flagged && *found.neighborhood->flagged == n) found.neighborhood->flagged = 0;
      found.neighborhood->edges.erase(found.it);
      typename lookup2_t::iterator it_found = lookup2.find(n);
      if(it_found != lookup2.end())
        it_found->second->second.in_edges.erase(v);
      return true;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::size_type
    VVBiGraph<BIGRAPH_ARGS>::eraseEdge( const vertex2_t& v, const vertex1_t& n )
    {
      neighbor2_found_t found = findInVertex(v, n);
      if(!found)
        return false;
      if(found.neighborhood->flagged && *found.neighborhood->flagged == n) found.neighborhood->flagged = 0;
      found.neighborhood->edges.erase(found.it);
      typename lookup1_t::iterator it_found = lookup1.find(n);
      if(it_found != lookup1.end())
        it_found->second->second.in_edges.erase(v);
      return true;
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::contains( const vertex1_t& v ) const
    {
      return this->findVertex(v) != neighborhood1.end();
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::contains( const vertex2_t& v ) const
    {
      return this->findVertex(v) != neighborhood2.end();
    }

  template <BIGRAPH_TEMPLATE>
    std::pair<typename VVBiGraph<BIGRAPH_ARGS>::ineighbor_iterator2,bool>
    VVBiGraph<BIGRAPH_ARGS>::insertInEdge(const vertex1_t& v, const vertex2_t& new_neighbor)
    {
      typename lookup2_t::iterator found = lookup2.find(new_neighbor);
      if(found == lookup2.end())
        return std::make_pair(ineighbor_iterator2(), false);
      return found->second->second.in_edges.insert(v);
    }

  template <BIGRAPH_TEMPLATE>
    std::pair<typename VVBiGraph<BIGRAPH_ARGS>::ineighbor_iterator1,bool>
    VVBiGraph<BIGRAPH_ARGS>::insertInEdge(const vertex2_t& v, const vertex1_t& new_neighbor)
    {
      typename lookup1_t::iterator found = lookup1.find(new_neighbor);
      if(found == lookup1.end())
        return std::make_pair(ineighbor_iterator1(), false);
      return found->second->second.in_edges.insert(v);
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::undoInEdge(const vertex2_t& new_neighbor,
                                             ineighbor_iterator2& it)
    {
      typename lookup2_t::iterator found = lookup2.find(new_neighbor);
      if(found == lookup2.end())
        return;
      found->second->second.in_edges.erase(it);
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::undoInEdge(const vertex1_t& new_neighbor,
                                             ineighbor_iterator1& it)
    {
      typename lookup1_t::iterator found = lookup1.find(new_neighbor);
      if(found == lookup1.end())
        return;
      found->second->second.in_edges.erase(it);
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::updateEdgeCache(const vertex1_t& v,
                                                  typename edge_list1_t::iterator new_edge_it,
                                                  const vertex1_t& in_edge)
    {
#ifndef VVBIGRAPH_NO_EDGE_CACHE
      // Caching in the neighbor
      neighbor1_t& neighbor = *new_edge_it;
      neighbor.setIt(new_edge_it);
      neighbor.target.cache_source = v.id();
      neighbor.target.cache = reinterpret_cast<void*>(&neighbor);

      // Caching in the incoming edge source
      in_edge.cache_source = v.id();
      in_edge.cache = reinterpret_cast<void*>(&neighbor);
#endif // VVBIGRAPH_NO_EDGE_CACHE
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::updateEdgeCache(const vertex2_t& v,
                                                  typename edge_list2_t::iterator new_edge_it,
                                                  const vertex2_t& in_edge)
    {
#ifndef VVBIGRAPH_NO_EDGE_CACHE
      // Caching in the neighbor
      neighbor2_t& neighbor = *new_edge_it;
      neighbor.setIt(new_edge_it);
      neighbor.target.cache_source = v.id();
      neighbor.target.cache = reinterpret_cast<void*>(&neighbor);

      // Caching in the incoming edge source
      in_edge.cache_source = v.id();
      in_edge.cache = reinterpret_cast<void*>(&neighbor);
#endif // VVBIGRAPH_NO_EDGE_CACHE
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge1_t
    VVBiGraph<BIGRAPH_ARGS>::replace( const vertex1_t& v,
                                      const vertex2_t& neighbor,
                                      const vertex2_t& new_neighbor )
    {
      if(new_neighbor.isNull() || (neighbor == new_neighbor))
        return edge1_t();
      neighbor1_found_t found = findInVertex(v, neighbor);
      if(!found)
        return edge1_t();
      typename neighborhood2_t::iterator n_found = this->findVertex(neighbor);
      ineighbor_iterator2 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge1_t();
      n_found->second.in_edges.erase(v);
      found.it->target = new_neighbor;
      found.it->clear_edge();
      updateEdgeCache(v, found.it, *it_in);
      if(n_found->second.flagged and *n_found->second.flagged == neighbor) n_found->second.flagged = 0;
      return edge1_t(v.id(), new_neighbor.id(), &*found.it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge2_t
    VVBiGraph<BIGRAPH_ARGS>::replace( const vertex2_t& v,
                                      const vertex1_t& neighbor,
                                      const vertex1_t& new_neighbor )
    {
      if(new_neighbor.isNull() || (neighbor == new_neighbor))
        return edge2_t();
      neighbor2_found_t found = findInVertex(v, neighbor);
      if(!found)
        return edge2_t();
      typename neighborhood1_t::iterator n_found = this->findVertex(neighbor);
      ineighbor_iterator1 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge2_t();
      n_found->second.in_edges.erase(v);
      found.it->target = new_neighbor;
      found.it->clear_edge();
      updateEdgeCache(v, found.it, *it_in);
      if(n_found->second.flagged and *n_found->second.flagged == neighbor) n_found->second.flagged = 0;
      return edge2_t(v.id(), new_neighbor.id(), &*found.it);
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::nextTo( const vertex1_t& v,
                                     const vertex2_t& ref,
                                     unsigned int n ) const
    {
      const_neighbor1_found_t found = findInVertex(v, ref);
      if(!found)
        return vertex2_t::null;//vertex_t(0);
      typename edge_list1_t::const_iterator it = found.it;
      const edge_list1_t& lst = found.neighborhood->edges;
      for(unsigned int i = 0 ; i < n ; ++i)
      {
        ++it;
        if(it == lst.end())
          it = lst.begin();
      }
      return it->target;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::nextTo( const vertex2_t& v,
                                     const vertex1_t& ref,
                                     unsigned int n ) const
    {
      const_neighbor2_found_t found = findInVertex(v, ref);
      if(!found)
        return vertex1_t::null;//vertex_t(0);
      typename edge_list2_t::const_iterator it = found.it;
      const edge_list2_t& lst = found.neighborhood->edges;
      for(unsigned int i = 0 ; i < n ; ++i)
      {
        ++it;
        if(it == lst.end())
          it = lst.begin();
      }
      return it->target;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::prevTo( const vertex1_t& v,
                                     const vertex2_t& ref,
                                     unsigned int n ) const
    {
      const_neighbor1_found_t found = findInVertex(v, ref);
      if(!found)
        return vertex2_t::null;//vertex_t(0);
      typename edge_list1_t::const_iterator it = found.it;
      const edge_list1_t& lst = found.neighborhood->edges;
      for(unsigned int i = 0 ; i < n ; ++i)
      {
        if(it == lst.begin())
          it = lst.end();
        --it;
      }
      return it->target;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::prevTo( const vertex2_t& v,
                                     const vertex1_t& ref,
                                     unsigned int n ) const
    {
      const_neighbor2_found_t found = findInVertex(v, ref);
      if(!found)
        return vertex1_t::null;//vertex_t(0);
      typename edge_list2_t::const_iterator it = found.it;
      const edge_list2_t& lst = found.neighborhood->edges;
      for(unsigned int i = 0 ; i < n ; ++i)
      {
        if(it == lst.begin())
          it = lst.end();
        --it;
      }
      return it->target;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge1_t
    VVBiGraph<BIGRAPH_ARGS>::edge( const vertex1_t& src,
                                   const vertex2_t& tgt )
    {
      neighbor1_found_t found = findInVertex(src, tgt);
      if(!found)
        return edge1_t();
      return edge1_t(src.id(), tgt.id(), &*found.it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge2_t
    VVBiGraph<BIGRAPH_ARGS>::edge( const vertex2_t& src,
                                   const vertex1_t& tgt )
    {
      neighbor2_found_t found = findInVertex(src, tgt);
      if(!found)
        return edge2_t();
      return edge2_t(src.id(), tgt.id(), &*found.it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_edge1_t
    VVBiGraph<BIGRAPH_ARGS>::edge( const vertex1_t& src,
                                   const vertex2_t& tgt ) const
    {
      const_neighbor1_found_t found = findInVertex(src, tgt);
      if(!found)
        return const_edge1_t();
      return const_edge1_t(src.id(), tgt.id(), &*found.it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_edge2_t
    VVBiGraph<BIGRAPH_ARGS>::edge( const vertex2_t& src,
                                   const vertex1_t& tgt ) const
    {
      const_neighbor2_found_t found = findInVertex(src, tgt);
      if(!found)
        return const_edge2_t();
      return const_edge2_t(src.id(), tgt.id(), &*found.it);
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::flag(const vertex1_t& src, const vertex2_t& neighbor)
    {
      neighbor1_found_t found = findInVertex(src, neighbor);
      if(!found)
        return false;
      found.neighborhood->flagged = &(found.it->target);
      return true;
    }

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::flag(const vertex2_t& src, const vertex1_t& neighbor)
    {
      neighbor2_found_t found = findInVertex(src, neighbor);
      if(!found)
        return false;
      found.neighborhood->flagged = &(found.it->target);
      return true;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex2_t&
    VVBiGraph<BIGRAPH_ARGS>::flagged(const vertex1_t& src) const
    {
      typename neighborhood1_t::const_iterator it_found = this->findVertex(src);
      if(it_found == neighborhood1.end() or it_found->second.flagged == 0)
        return vertex2_t::null;
      return *it_found->second.flagged;
    }

  template <BIGRAPH_TEMPLATE>
    const typename VVBiGraph<BIGRAPH_ARGS>::vertex1_t&
    VVBiGraph<BIGRAPH_ARGS>::flagged(const vertex2_t& src) const
    {
      typename neighborhood2_t::const_iterator it_found = this->findVertex(src);
      if(it_found == neighborhood2.end() or it_found->second.flagged == 0)
        return vertex1_t::null;
      return *it_found->second.flagged;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge1_t
    VVBiGraph<BIGRAPH_ARGS>::spliceAfter( const vertex1_t& v,
                                          const vertex2_t& neighbor,
                                          const vertex2_t& new_neighbor )
    {
      if(new_neighbor.isNull())
        return edge1_t();
      neighbor1_found_t found = findInVertex(v, neighbor);
      if(!found)
        return edge1_t();
      ineighbor_iterator2 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge1_t();
      ++found.it;
      typename edge_list1_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor1_t(new_neighbor, Edge1Content(), *found.neighborhood, graph_id));
      updateEdgeCache(v, new_edge_it, *it_in);
      return edge1_t(v.id(), new_neighbor.id(), &*new_edge_it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge2_t
    VVBiGraph<BIGRAPH_ARGS>::spliceAfter( const vertex2_t& v,
                                          const vertex1_t& neighbor,
                                          const vertex1_t& new_neighbor )
    {
      if(new_neighbor.isNull())
        return edge2_t();
      neighbor2_found_t found = findInVertex(v, neighbor);
      if(!found)
        return edge2_t();
      ineighbor_iterator1 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge2_t();
      ++found.it;
      typename edge_list2_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor2_t(new_neighbor, Edge2Content(), *found.neighborhood, graph_id));
      updateEdgeCache(v, new_edge_it, *it_in);
      return edge2_t(v.id(), new_neighbor.id(), &*new_edge_it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge1_t
    VVBiGraph<BIGRAPH_ARGS>::spliceBefore( const vertex1_t& v,
                                           const vertex2_t& neighbor,
                                           const vertex2_t& new_neighbor )
    {
      if(new_neighbor.isNull())
        return edge1_t();
      neighbor1_found_t found = findInVertex(v, neighbor);
      if(!found)
        return edge1_t();
      ineighbor_iterator2 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge1_t();
      typename edge_list1_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor1_t(new_neighbor, Edge1Content(), *found.neighborhood, graph_id));
      updateEdgeCache(v, new_edge_it, *it_in);
      return edge1_t(v.id(), new_neighbor.id(), &*new_edge_it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge2_t
    VVBiGraph<BIGRAPH_ARGS>::spliceBefore( const vertex2_t& v,
                                           const vertex1_t& neighbor,
                                           const vertex1_t& new_neighbor )
    {
      if(new_neighbor.isNull())
        return edge2_t();
      neighbor2_found_t found = findInVertex(v, neighbor);
      if(!found)
        return edge2_t();
      ineighbor_iterator1 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge2_t();
      typename edge_list2_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor2_t(new_neighbor, Edge2Content(), *found.neighborhood, graph_id));
      updateEdgeCache(v, new_edge_it, *it_in);
      return edge2_t(v.id(), new_neighbor.id(), &*new_edge_it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge1_t
    VVBiGraph<BIGRAPH_ARGS>::insertEdge( const vertex1_t& v,
                                         const vertex2_t& new_neighbor )
    {
      if(v.isNull() || new_neighbor.isNull())
        return edge1_t();
      typename neighborhood1_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return edge1_t();
      ineighbor_iterator2 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge1_t();
      edge_list1_t &lst = it_found->second.edges;
      lst.push_front(neighbor1_t(new_neighbor, Edge1Content(), it_found->second, graph_id));
      typename edge_list1_t::iterator it = lst.begin();
      updateEdgeCache(v, it, *it_in);
      return edge1_t(v.id(), new_neighbor.id(), &*it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::edge2_t
    VVBiGraph<BIGRAPH_ARGS>::insertEdge( const vertex2_t& v,
                                         const vertex1_t& new_neighbor )
    {
      if(v.isNull() || new_neighbor.isNull())
        return edge2_t();
      typename neighborhood2_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return edge2_t();
      ineighbor_iterator1 it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge2_t();
      edge_list2_t &lst = it_found->second.edges;
      lst.push_front(neighbor2_t(new_neighbor, Edge2Content(), it_found->second, graph_id));
      typename edge_list2_t::iterator it = lst.begin();
      updateEdgeCache(v, it, *it_in);
      return edge2_t(v.id(), new_neighbor.id(), &*it);
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_neighbor_iterator1_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex1_t& v) const
    {
      const_neighbor_iterator1_range result;
      if(v.isNull())
        return result;
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return result;
      const edge_list1_t &lst = it_found->second.edges;
      result = util::make_range(const_neighbor_iterator1(lst.begin()),
                                const_neighbor_iterator1(lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_neighbor_iterator2_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex2_t& v) const
    {
      const_neighbor_iterator2_range result;
      if(v.isNull())
        return result;
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return result;
      const edge_list2_t &lst = it_found->second.edges;
      result = util::make_range(const_neighbor_iterator2(lst.begin()), const_neighbor_iterator2(lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor_iterator1_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex1_t& v)
    {
      neighbor_iterator1_range result;
      if(v.isNull())
        return result;
      typename neighborhood1_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return result;
      edge_list1_t &lst = it_found->second.edges;
      result = util::make_range(neighbor_iterator1(lst.begin()), neighbor_iterator1(lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::neighbor_iterator2_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex2_t& v)
    {
      neighbor_iterator2_range result;
      if(v.isNull())
        return result;
      typename neighborhood2_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return result;
      edge_list2_t &lst = it_found->second.edges;
      result = util::make_range(neighbor_iterator2(lst.begin()), neighbor_iterator2(lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_circ_neighbor_iterator1_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex1_t& v, const vertex2_t& n) const
    {
      const_circ_neighbor_iterator1_range result;
      if(v.isNull())
        return result;
      const_neighbor1_found_t found = findInVertex(v, n);
      if(!found)
        return result;
      const edge_list1_t &lst = found.neighborhood->edges;
      result = util::make_range(const_circ_neighbor_iterator1(lst.begin(), lst.end(), found.it),
                                const_circ_neighbor_iterator1(lst.begin(), lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::const_circ_neighbor_iterator2_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex2_t& v, const vertex1_t& n) const
    {
      const_circ_neighbor_iterator2_range result;
      if(v.isNull())
        return result;
      const_neighbor2_found_t found = findInVertex(v, n);
      if(!found)
        return result;
      const edge_list2_t &lst = found.neighborhood->edges;
      result = util::make_range(const_circ_neighbor_iterator2(lst.begin(), lst.end(), found.it),
                                const_circ_neighbor_iterator2(lst.begin(), lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::circ_neighbor_iterator1_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex1_t& v, const vertex2_t& n)
    {
      circ_neighbor_iterator1_range result;
      if(v.isNull())
        return result;
      neighbor1_found_t found = findInVertex(v, n);
      if(!found)
        return result;
      edge_list1_t &lst = found.neighborhood->edges;
      result = util::make_range(circ_neighbor_iterator1(lst.begin(), lst.end(), found.it),
                                circ_neighbor_iterator1(lst.begin(), lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::circ_neighbor_iterator2_range
    VVBiGraph<BIGRAPH_ARGS>::neighbors(const vertex2_t& v, const vertex1_t& n)
    {
      circ_neighbor_iterator2_range result;
      if(v.isNull())
        return result;
      neighbor2_found_t found = findInVertex(v, n);
      if(!found)
        return result;
      edge_list2_t &lst = found.neighborhood->edges;
      result = util::make_range(circ_neighbor_iterator2(lst.begin(), lst.end(), found.it),
                                circ_neighbor_iterator2(lst.begin(), lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::ineighbor_iterator1_range
    VVBiGraph<BIGRAPH_ARGS>::iNeighbors(const vertex1_t& v) const
    {
      ineighbor_iterator1_range result;
      if(v.isNull())
        return result;
      typename neighborhood1_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood1.end())
        return result;
      const in_edges1_t &lst = it_found->second.in_edges;
      result = util::make_range(ineighbor_iterator1(lst.begin()),
                                ineighbor_iterator1(lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    typename VVBiGraph<BIGRAPH_ARGS>::ineighbor_iterator2_range
    VVBiGraph<BIGRAPH_ARGS>::iNeighbors(const vertex2_t& v) const
    {
      ineighbor_iterator2_range result;
      if(v.isNull())
        return result;
      typename neighborhood2_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood2.end())
        return result;
      const in_edges2_t &lst = it_found->second.in_edges;
      result = util::make_range(ineighbor_iterator2(lst.begin()), ineighbor_iterator2(lst.end()));
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    template <typename Vertex1Container, typename Vertex2Container>
    VVBiGraph<BIGRAPH_ARGS> VVBiGraph<BIGRAPH_ARGS>::subgraph( const Vertex1Container& verts1, const Vertex2Container& verts2 ) const
    {
      typename Vertex1Container::const_iterator it1;
      typename Vertex2Container::const_iterator it2;
      VVBiGraph result;
      // First, insert the vertices in the result graph
      for(it1 = verts1.begin() ; it1 != verts1.end() ; ++it1)
      {
        result.insert(*it1);
      }
      for(it2 = verts2.begin() ; it2 != verts2.end() ; ++it2)
      {
        result.insert(*it2);
      }

      for(typename neighborhood1_t::iterator it_n = result.neighborhood1.begin() ;
          it_n != result.neighborhood1.end() ; ++it_n)
      {
        const vertex1_t& v = it_n->first;
        result.updateVertexCache(v, it_n);
        // Find ou the out edges and add them
        typename neighborhood1_t::const_iterator it_orign = this->findVertex(v);
        const edge_list1_t& orig_lst = it_orign->second.edges;
        edge_list1_t& lst = it_n->second.edges;
        // Going backward because it is easier to retrieve the iterator ...
        for(typename edge_list1_t::const_reverse_iterator it_sn = orig_lst.rbegin() ;
            it_sn != orig_lst.rend() ; ++it_sn)
        {
          // For each existing edge from the vertex, keep only the ones whose 
          // target are on the subgraph
          if(result.contains(it_sn->target))
          {
            // Insert incoming edge
            ineighbor_iterator2 it_in = result.insertInEdge(v, it_sn->target).first;
            // Insert the outgoing edge
            lst.push_front(neighbor1_t(it_sn->target, *it_sn, it_n->second, result.graph_id));
            // Update the cache
            result.updateEdgeCache(v, lst.begin(), *it_in);
          }
        }
        // At last, update the flag, if any
        if(it_orign->second.flagged and result.contains(*it_orign->second.flagged))
        {
          const_neighbor1_found_t found = this->findInVertex(v, *it_orign->second.flagged);
          it_n->second.flagged = &(found.it->target);
        }
      }

      for(typename neighborhood2_t::iterator it_n = result.neighborhood2.begin() ;
          it_n != result.neighborhood2.end() ; ++it_n)
      {
        const vertex2_t& v = it_n->first;
        result.updateVertexCache(v, it_n);
        // Find ou the out edges and add them
        typename neighborhood2_t::const_iterator it_orign = this->findVertex(v);
        const edge_list2_t& orig_lst = it_orign->second.edges;
        edge_list2_t& lst = it_n->second.edges;
        // Going backward because it is easier to retrieve the iterator ...
        for(typename edge_list2_t::const_reverse_iterator it_sn = orig_lst.rbegin() ;
            it_sn != orig_lst.rend() ; ++it_sn)
        {
          // For each existing edge from the vertex, keep only the ones whose 
          // target are on the subgraph
          if(result.contains(it_sn->target))
          {
            // Insert incoming edge
            ineighbor_iterator1 it_in = result.insertInEdge(v, it_sn->target).first;
            // Insert the outgoing edge
            lst.push_front(neighbor2_t(it_sn->target, *it_sn, it_n->second, result.graph_id));
            // Update the cache
            result.updateEdgeCache(v, lst.begin(), *it_in);
          }
        }
        // At last, update the flag, if any
        if(it_orign->second.flagged and result.contains(*it_orign->second.flagged))
        {
          const_neighbor2_found_t found = this->findInVertex(v, *it_orign->second.flagged);
          it_n->second.flagged = &(found.it->target);
        }
      }
      return result;
    }

  template <BIGRAPH_TEMPLATE>
    VVBiGraph<BIGRAPH_ARGS>::VVBiGraph(const VVBiGraph& copy)
    : graph_id(graph::getNextGraphId())
    {
      *this = copy;
    }

#ifdef USE_CXX0X

  template <BIGRAPH_TEMPLATE>
    VVBiGraph<BIGRAPH_ARGS>::VVBiGraph(VVBiGraph&& copy)
    : neighborhood1(std::move(copy.neighborhood1))
    , neighborhood2(std::move(copy.neighborhood2))
    , lookup1(std::move(copy.lookup1))
    , lookup2(std::move(copy.lookup2))
    , graph_id(copy.graph_id)
    {
      copy.graph_id = graph::getNextGraphId();
      copy.neighborhood1.clear();
      copy.neighborhood2.clear();
      copy.lookup1.clear();
      copy.lookup2.clear();
    }
#endif

  template <BIGRAPH_TEMPLATE>
    bool VVBiGraph<BIGRAPH_ARGS>::operator==(const VVBiGraph& other) const
    {
      if(neighborhood1.size() != other.neighborhood1.size() or
         neighborhood2.size() != other.neighborhood2.size())
        return false;
      for(typename neighborhood1_t::const_iterator it1 = neighborhood1.begin() ;
          it1 != neighborhood1.end() ; ++it1)
      {
        const vertex1_t& v1 = it1->first;
        typename lookup1_t::const_iterator it_found = other.lookup1.find(v1);
        if(it_found == other.lookup1.end())
          return false;
        if(it1->second.edges.empty())
        {
          if(!it_found->second->second.edges.empty())
            return false;
        }
        else
        {
          const edge_list1_t& lst = it1->second.edges;
          const edge_list1_t& olst = it_found->second->second.edges;
          if(lst.size() != olst.size()) return false;
          const vertex2_t& v2 = lst.begin()->target;
          bool found = false;
          for(typename edge_list1_t::const_iterator it_olst = olst.begin() ;
              it_olst != olst.end() ; ++it_olst)
          {
            if(it_olst->target == v2)
            {
              found = true;
              for(typename edge_list1_t::const_iterator it_lst = lst.begin() ;
                  it_lst != lst.end() ; ++it_lst)
              {
                if(it_lst->target != it_olst->target) return false;
                ++it_olst;
                if(it_olst == olst.end()) it_olst = olst.begin();
              }
              break;
            }
          }
          if(!found) return false;
        }
      }
      for(typename neighborhood2_t::const_iterator it2 = neighborhood2.begin() ;
          it2 != neighborhood2.end() ; ++it2)
      {
        const vertex2_t& v2 = it2->first;
        typename lookup2_t::const_iterator it_found = other.lookup2.find(v2);
        if(it_found == other.lookup2.end())
          return false;
        if(it2->second.edges.empty())
        {
          if(!it_found->second->second.edges.empty())
            return false;
        }
        else
        {
          const edge_list2_t& lst = it2->second.edges;
          const edge_list2_t& olst = it_found->second->second.edges;
          if(lst.size() != olst.size()) return false;
          const vertex1_t& v1 = lst.begin()->target;
          bool found = false;
          for(typename edge_list2_t::const_iterator it_olst = olst.begin() ;
              it_olst != olst.end() ; ++it_olst)
          {
            if(it_olst->target == v1)
            {
              found = true;
              for(typename edge_list2_t::const_iterator it_lst = lst.begin() ;
                  it_lst != lst.end() ; ++it_lst)
              {
                if(it_lst->target != it_olst->target) return false;
                ++it_olst;
                if(it_olst == olst.end()) it_olst = olst.begin();
              }
              break;
            }
          }
          if(!found) return false;
        }
      }
      return true;
    }

  template <BIGRAPH_TEMPLATE>
    VVBiGraph<BIGRAPH_ARGS>&
    VVBiGraph<BIGRAPH_ARGS>::operator=( const VVBiGraph& other )
    {
      // Copy the structure
      neighborhood1 = other.neighborhood1;
      neighborhood2 = other.neighborhood2;
      lookup1.clear();
      lookup2.clear();
      for(typename neighborhood1_t::iterator it1 = neighborhood1.begin() ;
          it1 != neighborhood1.end() ; ++it1)
        lookup1[it1->first] = it1;
      for(typename neighborhood2_t::iterator it2 = neighborhood2.begin() ;
          it2 != neighborhood2.end() ; ++it2)
        lookup2[it2->first] = it2;
      // And reconstruct the cache for vertex type 1
      for(typename neighborhood1_t::iterator it1 = neighborhood1.begin() ;
          it1 != neighborhood1.end() ; ++it1)
      {
        const vertex1_t& src = it1->first;
        updateVertexCache(src, it1);
#ifndef VVBIGRAPH_NO_EDGE_CACHE
        edge_list1_t& lst = it1->second.edges;
        for(typename edge_list1_t::iterator it1_e = lst.begin() ;
            it1_e != lst.end() ; ++it1_e)
        {
          // Find the incoming vertex corresponding to this edge
          typename lookup2_t::iterator it_found = lookup2.find(it1_e->target);
          const vertex1_t& iv = *(it_found->second->second.in_edges.find(src));
          updateEdgeCache(src, it1_e, iv);
          it1_e->set(&it1->second, graph_id);
        }
#endif
      }
      // And reconstruct the cache for vertex type 2
      for(typename neighborhood2_t::iterator it2 = neighborhood2.begin() ;
          it2 != neighborhood2.end() ; ++it2)
      {
        const vertex2_t& src = it2->first;
        updateVertexCache(src, it2);
#ifndef VVBIGRAPH_NO_EDGE_CACHE
        edge_list2_t& lst = it2->second.edges;
        for(typename edge_list2_t::iterator it2_e = lst.begin() ;
            it2_e != lst.end() ; ++it2_e)
        {
          // Find the incoming vertex corresponding to this edge
          typename lookup1_t::iterator it_found = lookup1.find(it2_e->target);
          const vertex2_t& iv = *(it_found->second->second.in_edges.find(src));
          updateEdgeCache(src, it2_e, iv);
          it2_e->set(&it2->second, graph_id);
        }
#endif
      }
      return *this;
    }

  template <BIGRAPH_TEMPLATE>
    void VVBiGraph<BIGRAPH_ARGS>::swap( VVBiGraph& other )
    {
      neighborhood1.swap(other.neighborhood1);
      neighborhood2.swap(other.neighborhood2);
      lookup1.swap(other.lookup1);
      lookup2.swap(other.lookup2);
      std::swap(graph_id, other.graph_id);
    }

  template <typename BiGraph>
    void create_bigraph_methods(BiGraph& G)
    {
      if(G.size())
      {
        if(G.valence(G.any_vertex1()) > G.size())
        {
          // Call all fcts
          typename BiGraph::vertex1_t v1 = G.any_vertex1();
          typename BiGraph::vertex2_t v2 = G.any_vertex2();
          G.erase(v1);
          G.erase(v2);
          G.insert(v1);
          G.insert(v2);
          G.get_vertex1(10);
          G.get_vertex2(10);
          G.reference(v1);
          G.reference(v2);
          G.empty();
          G.no_vertex1();
          G.no_vertex2();
          G.anyIn(v1);
          G.anyIn(v2);
          G.valence(v1);
          G.valence(v2);
          G.empty(v1);
          G.empty(v2);
          G.iAnyIn(v1);
          G.iAnyIn(v2);
          G.iValence(v1);
          G.iValence(v2);
          G.iEmpty(v1);
          G.iEmpty(v2);
          G.nextTo(v1, v2);
          G.nextTo(v2, v1);
          G.prevTo(v1, v2);
          G.prevTo(v2, v1);
          G.edge(v1, v2);
          G.edge(v2, v1);
          G.flag(v1,v2);
          G.flag(v2,v1);
          G.flagged(v1);
          G.flagged(v2);
          G.neighbors(v1);
          G.neighbors(v2);
          G.iNeighbors(v1);
          G.iNeighbors(v2);
          G.source(typename BiGraph::edge1_t());
          G.source(typename BiGraph::edge2_t());
          G.target(typename BiGraph::edge1_t());
          G.target(typename BiGraph::edge2_t());
          G.eraseEdge(v1, v2);
          G.eraseEdge(v2, v1);
          G.replace(v1, v2, v2);
          G.replace(v2, v1, v1);
          G.spliceAfter(v1, v2, v2);
          G.spliceAfter(v2, v1, v1);
          G.spliceBefore(v1, v2, v2);
          G.spliceBefore(v2, v1, v1);
          G.insertEdge(v1,v2);
          G.insertEdge(v2,v1);
          G.eraseAllEdges(v1);
          G.eraseAllEdges(v2);
          G.begin_vertex1();
          G.begin_vertex2();
          G.end_vertex1();
          G.end_vertex1();
          G.find(v1);
          G.find(v2);
          G.count(v1);
          G.count(v2);
          G.findIn(v1, v2);
          G.findIn(v2, v1);
        }
      }
    }

}

namespace std
{
  template <BIGRAPH_TEMPLATE>
    void swap(graph::VVBiGraph<BIGRAPH_ARGS>& g1, graph::VVBiGraph<BIGRAPH_ARGS>& g2)
    {
      g1.swap(g2);
    }
}

#endif // GRAPH_VVBIGRAPH_H

