#ifndef GRAPH_VVGRAPH_H
#define GRAPH_VVGRAPH_H

/**
 * \file vvgraph.h
 *
 * Contain the definition of the VVGraph template class.
 */

#include <config.h>
#include <graph/vertex.h>
#include <graph/edge.h>

#include <util/tie.h>
#include <util/member_iterator.h>
#include <util/circ_iterator.h>
#include <util/set_vector.h>
#include <util/range.h>

#include <storage/fwd.h>

#include <utility>
#include <memory>
#include <algorithm>
#include <iterator>

#include <list>
#include <vector>
#include <set>
#include <map>

#ifdef USE_ALLOCATOR
#  include <memory>
#endif // defined USE_ALLOCATOR

#include <iostream>

#ifdef USE_CXX0X
#  include <initializer_list>
#endif

#ifdef VVGRAPH_NO_CACHE
#  define VVGRAPH_NO_EDGE_CACHE
#  define VVGRAPH_NO_VERTEX_CACHE
#endif

#ifdef VVGRAPH_NO_VERTEX_CACHE
#  define VVGRAPH_COMPACT_VERTEX true
#else
#  define VVGRAPH_COMPACT_VERTEX compact
#endif

#ifdef VVGRAPH_NO_EDGE_CACHE
#  define VVGRAPH_COMPACT_EDGE true
#else
#  define VVGRAPH_COMPACT_EDGE compact
#endif

#ifdef USE_ALLOCATOR
#  define GRAPH_TEMPLATE typename VertexContent, typename EdgeContent, bool compact, typename Alloc
#  define GRAPH_ARGS VertexContent,EdgeContent,compact,Alloc
#else // USE_ALLOCATOR
#  define GRAPH_TEMPLATE typename VertexContent, typename EdgeContent, bool compact
#  define GRAPH_ARGS VertexContent,EdgeContent,compact
#endif // USE_ALLOCATOR

/**
 * Contains all the classes related to the graphs.
 *
 * For now, the only graph available is the VVGraph.
 */
namespace graph
{

  template <bool compact, typename T>
    struct __vvgraph_set
    { };

  template <typename T>
    struct __vvgraph_set<true,T>
    {
      typedef util::set_vector<T> type;
    };

  template <typename T>
    struct __vvgraph_set<false,T>
    {
      typedef std::set<T> type;
    };

  template <bool compact, typename EdgeContent, typename VVGraph, typename single_neighborhood_t, typename edge_list_t>
    struct __vvgraph_EdgeCache
    {
    };

  template <typename VVGraph, typename EdgeContent, typename single_neighborhood_t, typename edge_list_t>
    struct __vvgraph_EdgeCache<false, EdgeContent, VVGraph, single_neighborhood_t, edge_list_t>
    : public EdgeContent
    {
      typedef typename VVGraph::vertex_t vertex_t;

      /**
       * Construct an edge cache
       */
      __vvgraph_EdgeCache(const EdgeContent& copy,
                          single_neighborhood_t *n,
                          intptr_t id)
        : EdgeContent(copy)
        , _neighborhood(n)
        , gid(id)
        { }

      /**
       * Construct an edge cache
       */
      __vvgraph_EdgeCache(const __vvgraph_EdgeCache& copy)
        : EdgeContent(copy)
        , _neighborhood(copy._neighborhood)
        , gid(copy.gid)
        { }

#ifdef USE_CXX0X
      __vvgraph_EdgeCache(__vvgraph_EdgeCache&& copy)
        : EdgeContent(std::move(copy))
        , _neighborhood(std::move(copy._neighborhood))
        , gid(std::move(copy.gid))
      {
        copy.gid = 0;
      }
#endif

      __vvgraph_EdgeCache& operator=(const __vvgraph_EdgeCache& copy)
      {
        _neighborhood = copy._neighborhood;
        gid = copy.gid;
        static_cast<EdgeContent&>(*this) = static_cast<const EdgeContent&>(copy);
        return *this;
      }

      void setIt(const typename edge_list_t::iterator& i)
      {
        _it = i;
      }

      bool eraseEdge()
      {
        if(_neighborhood->flagged && *_neighborhood->flagged == _it->target) _neighborhood->flagged = 0;
        _neighborhood->edges.erase(_it);
        return true;
      }

      void set(single_neighborhood_t* n, intptr_t id)
      {
        _neighborhood = n;
        gid = id;
      }

      single_neighborhood_t* neighborhood() { return _neighborhood; }

      typename edge_list_t::iterator it() { return _it; }

      /**
       * Edge list containing the neighbor
       */
      single_neighborhood_t* _neighborhood;
      /**
       * Iterator pointing to the neighbor
       */
      typename edge_list_t::iterator _it;

      /**
       * Id of the current graph
       */
      intptr_t gid;

      intptr_t graph_id() { return gid; }
    };

  template <typename EdgeContent, typename VVGraph, typename single_neighborhood_t, typename edge_list_t>
    struct __vvgraph_EdgeCache<true, EdgeContent, VVGraph, single_neighborhood_t, edge_list_t>
    : public EdgeContent
    {
      typedef typename VVGraph::vertex_t vertex_t;

      /**
       * Construct an edge cache
       */
      __vvgraph_EdgeCache(const EdgeContent& copy,
                          single_neighborhood_t *,
                          intptr_t)
        : EdgeContent(copy)
        { }

      /**
       * Construct an edge cache
       */
      __vvgraph_EdgeCache(const __vvgraph_EdgeCache& copy)
        : EdgeContent(copy)
        { }

#ifdef USE_CXX0X
      __vvgraph_EdgeCache(__vvgraph_EdgeCache&& ) = default;
#endif

      __vvgraph_EdgeCache& operator=(const __vvgraph_EdgeCache& copy)
      {
        static_cast<EdgeContent&>(*this) = static_cast<const EdgeContent&>(copy);
        return *this;
      }

      void setIt(const typename edge_list_t::iterator&)
      {
      }

      bool eraseEdge()
      {
        return false;
      }

      intptr_t graph_id() { return 0; }

      single_neighborhood_t* neighborhood() { return 0; }

      typename edge_list_t::iterator it() { return typename edge_list_t::iterator(); }

      void set(single_neighborhood_t*, intptr_t ) {}
    };

  template <bool compact, typename VVGraph, typename single_neighborhood_t>
    struct __vvgraph_VertexCache
    {
      __vvgraph_VertexCache() {}
      __vvgraph_VertexCache(const __vvgraph_VertexCache& ) {}

#ifdef USE_CXX0X
      __vvgraph_VertexCache(__vvgraph_VertexCache&& ) = default;
#endif

      typedef typename VVGraph::vertex_t vertex_t;
      typedef typename std::list<std::pair<vertex_t,single_neighborhood_t> >::iterator iterator;
      intptr_t graph_id() { return 0; }
      void update(intptr_t, const vertex_t& , iterator ) {}
      iterator neighborhood() { return iterator(); }
    };

  template <typename VVGraph, typename single_neighborhood_t>
    struct __vvgraph_VertexCache<false,VVGraph,single_neighborhood_t>
    {
      typedef typename VVGraph::vertex_t vertex_t;
      typedef typename std::list<std::pair<vertex_t,single_neighborhood_t> >::iterator iterator;

      __vvgraph_VertexCache() : gid(0) {}
      __vvgraph_VertexCache(const __vvgraph_VertexCache& copy ) : gid(copy.gid), _neighborhood(copy._neighborhood) {}

#ifdef USE_CXX0X
      __vvgraph_VertexCache(__vvgraph_VertexCache&& copy)
        : gid(std::move(copy.gid))
        , _neighborhood(std::move(copy._neighborhood))
        {
          copy.gid = 0;
        }
#endif

      intptr_t graph_id() { return gid; }
      iterator neighborhood() { return _neighborhood; }

      void update(intptr_t graph_id, const vertex_t& v, iterator it)
      {
        gid = graph_id;
        _neighborhood = it;
        v.cache = reinterpret_cast<void*>(this);
        v.cache_source = 0;
      }

      /**
       * Id of the current graph
       *
       * Used to checked for cache validity (you don't want to use the 
       * cache if the vertex is used in a different graph)
       */
      intptr_t gid;

      /**
       * Iterator positionned on the neighborhood of the vertex
       *
       * This is the cache itself
       */
      iterator _neighborhood;
    };

  /**
   * \class _EmptyEdgeContent vvgraph.h <graph/vvgraph.h>
   *
   * Empty class used as default for edge content
   */
  struct _EmptyEdgeContent
  {
    _EmptyEdgeContent& operator=(const _EmptyEdgeContent& /*other*/) { return *this; }
    bool serialize(storage::VVEStorage&) { return true; }
  };

  intptr_t getNextGraphId();

  /**
   * \class VVGraph vvgraph.h <graph/vvgraph.h>
   * Class representing a VV graph.
   *
   * A VV graph is an oriented rotational graph. That is, the edges of a given 
   * vertex have a circular order. That is, if an edge has three edges e1, e2, 
   * e3, there is no first edge, but we can say e1 < e2 < e3 < e1. In that 
   * case, '<' denote the succession.
   *
   * \c VertexContent is the type used to hold the vertex data.
   *
   * \c EdgeContent is the type used to hold the edge data. If no type is 
   * specified for \c EdgeContent, then an empty structure is used.
   *
   * \section Perfs Performance notes
   * The graph includes two caching mecanisms to improve performances when 
   * iterating over the graph and retrieve edge information:
   *   -# while iterating over the graph, the vertex cache the current iterator. 
   *   So that accessing the neighborhood of the vertex is constant time (i.e. 
   *   it's dereferencing the cached iterator).
   *   -# while iterating over the neighborhood of a vertex, the target vertex 
   *   (i.e. the one returned by the iteration) cache the iterator on the 
   *   neighborhood. So that any operation using that specific edge (i.e. 
   *   getting edge data, splicing, replacing or erasing) is alos constant 
   *   time.
   *
   * To make this caching mecanism correct, the cached data are never copied. 
   * This means, to make use of the caching mecanism, you have to iterate using 
   * constant references (references will end up with compilation errors if 
   * used on the graph):
   * \code
   * forall(const vertex &v, S)
   * {
   *   // Here the vertex v has the vertex cache
   *   forall(const vertex &n, S.neighbors(S)) // Makes use of the cache
   *   {
   *     S.edge(v,n)->pos = Point3d(1,2,3); // Uses the edge cache
   *     S.edge(n,v)->pos = Point3d(3,2,1); // Do not use any cache
   *   }
   *   v->pos = Point3d(0,0,0); // Valid
   * }
   * \endcode
   *
   * \warning Do not iterate using constant references if you ever remove the 
   * vertex you iterate on ! (i.e. remove from the graph if you iterate on the 
   * graph or remove from the neighborhood if you iterate on the neighborhood)
   *
   * Even though this code manipulate constant references, modifying the data 
   * pointed by the vertex is allowed. The vertexes are coded to allow that. 
   * That means if you are using vertexes in a dictionary or a set, using their 
   * content to compare them, you must not modify them ! The compiler won't 
   * guard you against it.
   */
  template <typename VertexContent, typename EdgeContent = _EmptyEdgeContent, bool compact = false
#ifdef USE_ALLOCATOR
    , typename Alloc = DEFAULT_ALLOC(VertexContent)
#endif // USE_ALLOCATOR
    >
    class VVGraph
    {
      public:
#ifdef USE_ALLOCATOR
        typedef typename Alloc::template rebind<EdgeContent>::other EdgeAlloc;
        static EdgeAlloc edge_alloc;
#endif

        ///\name Smart pointer types
        //@{
        /**
         * Smart pointer on a vertex
         */
#ifdef USE_ALLOCATOR
        typedef Vertex<VertexContent, Alloc> vertex_t;
#else // USE_ALLOCATOR
        typedef Vertex<VertexContent> vertex_t;
#endif // USE_ALLOCATOR
        /**
         * Weak pointer on an edge
         */
        typedef Edge<EdgeContent> edge_t;
        /**
         * Weak pointer on a constant edge
         */
        typedef Edge<const EdgeContent> const_edge_t;
        /**
         * Weak pointer on an arc.
         *
         * When an object of this type is destroyed, the content of the primary 
         * edge (i.e. the edge source -> target) is copied into the other.
         */
        typedef Arc<EdgeContent> arc_t;
        //@}

        /**
         * Type of the value of the graph as standard container.
         */
        typedef vertex_t value_type;

        /**
         * Default constructor
         *
         * Creates an empty VV graph
         */
        VVGraph()
          : graph_id(graph::getNextGraphId())
          {}

        /**
         * Copy constructor.
         *
         * Maintains the caches in a valid state
         */
        VVGraph(const VVGraph& copy);

#ifdef USE_CXX0X
        /**
         * Move constructor
         */
        VVGraph(VVGraph&& copy);
#endif

      protected:
        class single_neighborhood_t;

        /**
         * Structure maintaining the data for a single neighbor
         */
        struct neighbor_t : public __vvgraph_EdgeCache<VVGRAPH_COMPACT_EDGE,EdgeContent,VVGraph,single_neighborhood_t,std::list<neighbor_t> >
        {
          typedef __vvgraph_EdgeCache<VVGRAPH_COMPACT_EDGE,EdgeContent,VVGraph,single_neighborhood_t,std::list<neighbor_t> > base;
          /**
           * Type of the list of outgoing neighbors
           */
          typedef std::list<neighbor_t> edge_list_t;

          /**
           * Constructor
           */
          neighbor_t(const vertex_t& tgt, const EdgeContent& e,
                     single_neighborhood_t& n,
                     intptr_t gid)
            : base(e, &n, gid)
            , target(tgt)
            {
            }

          neighbor_t(const neighbor_t& copy)
            : base(copy)
            , target(copy.target)
            {
            }

          void clear_edge()
          {
            static_cast<EdgeContent&>(*this) = EdgeContent();
          }

          ~neighbor_t()
          {
          }

          neighbor_t& operator=(const neighbor_t& copy)
          {
            target = copy.target;
            static_cast<base&>(*this) = static_cast<const base&>(copy);
            return *this;
          }

          /**
           * Target of the edge
           *
           * i.e. neighbor itself
           */
          vertex_t target;

          /**
           * Equality of two neighbors/edge
           *
           * A single neighbor is equal to another if the vertex and the content are 
           * the same
           */
          bool operator==(const neighbor_t& other) const
          {
            return (target == other.target) &&
              (static_cast<const EdgeContent&>(*this) == static_cast<const EdgeContent&>(other));
          }

        };

        typedef typename neighbor_t::base EdgeCache;

        /**
         * Type of the list of outgoing neighbors
         */
        typedef typename neighbor_t::edge_list_t edge_list_t;

        /**
         * Type of the list of sources of the in-edges
         */
        typedef typename __vvgraph_set<compact,vertex_t>::type in_edges_t;

        /**
         * Type of the neighborhood of a vertex
         */
        struct single_neighborhood_t : public __vvgraph_VertexCache<VVGRAPH_COMPACT_VERTEX,VVGraph,single_neighborhood_t>
        {
          typedef __vvgraph_VertexCache<VVGRAPH_COMPACT_VERTEX,VVGraph,single_neighborhood_t> base;

          single_neighborhood_t()
            : base()
            , flagged(0)
            {}

          single_neighborhood_t(const single_neighborhood_t& copy)
            : base(copy)
            , edges(copy.edges)
            , in_edges(copy.in_edges)
            , flagged(copy.flagged)
            {
              if(flagged)
              {
                for(typename edge_list_t::iterator it = edges.begin() ; it != edges.end() ; ++it)
                {
                  if(it->target == *flagged)
                  {
                    flagged = &it->target;
                    return;
                  }
                }
              }
            }

          /**
           * List of the outgoing edges
           */
          edge_list_t edges;

          /**
           * Set of the sources of the incoming edges
           */
          in_edges_t in_edges;

          /**
           * Iterator toward the flagged neighbor
           */
          const vertex_t* flagged;

          /**
           * Equality of two neighborhood
           *
           * Only the outgoing edges are compared!
           */
          bool operator==(const single_neighborhood_t& other) const
          {
            return edges == other.edges;
          }
        };

        typedef typename single_neighborhood_t::base VertexCache;

      public:

        //typedef std::map<vertex_t, single_neighborhood_t> neighborhood_t;
        /**
         * Type of the list of vertexes, together with the neighborhood
         */
        typedef std::list<std::pair<vertex_t, single_neighborhood_t> > neighborhood_t;

        /**
         * Type of the fast-search map
         */
        typedef std::unordered_map<vertex_t, typename neighborhood_t::iterator> lookup_t;
        //typedef std::map<vertex_t, typename neighborhood_t::iterator> lookup_t;

        /**
         * Shortcut for the value_type of the neighborhood
         */
        typedef typename neighborhood_t::value_type neighborhood_value_type;

        /**
         * Type of a size
         */
        typedef typename neighborhood_t::size_type size_type;

        ///\name Iterators
        //@{
        /**
         * Iterator on the vertexes
         */
        typedef util::SelectMemberIterator<typename neighborhood_t::iterator, const vertex_t, 
                &neighborhood_value_type::first> iterator;
        /**
         * Constant iterator on the vertexes
         */
        typedef util::SelectMemberIterator<typename neighborhood_t::const_iterator, const vertex_t, 
                &neighborhood_value_type::first> const_iterator;

        /**
         * Iterator on the neighbors of a vertex
         */
        typedef util::SelectMemberIterator<typename edge_list_t::iterator, vertex_t, 
                &neighbor_t::target> neighbor_iterator;
        /**
         * Range of the neighbors of a vertex.
         */
        typedef util::range<neighbor_iterator> neighbor_iterator_range;

        /**
         * Iterator used to iterate over the neighbors, but specifying the 
         * starting point
         */
        typedef util::CircIterator<neighbor_iterator> circ_neighbor_iterator;

        /**
         * Range of circular iterators
         */
        typedef util::range<circ_neighbor_iterator> circ_neighbor_iterator_range;

        /**
         * Iterator on the incoming neighbors of a vertex.
         *
         * An incoming neighbor is a source vertex of an edge whose target is the 
         * current vertex.
         */
        typedef typename in_edges_t::const_iterator ineighbor_iterator;

        /**
         * Range of the incoming neighbors of a vertex.
         *
         * \see ineighbor_iterator
         */
        typedef util::range<ineighbor_iterator> ineighbor_iterator_range;

        /**
         * Constant iterator on the neighbors of a vertex
         */
        typedef util::SelectMemberIterator<typename edge_list_t::const_iterator, 
                const vertex_t, &neighbor_t::target> const_neighbor_iterator;
        /**
         * Constant range of the neighbors of a vertex. The \c first element of the 
         * pair is a constant iterator on the beginning of the range, the \c second 
         * element is the past-the-end constant iterator.
         */
        typedef util::range<const_neighbor_iterator> const_neighbor_iterator_range;

        /**
         * Iterator used to iterate over the neighbors, but spcifying the 
         * starting point
         */
        typedef util::CircIterator<const_neighbor_iterator> const_circ_neighbor_iterator;

        /**
         * Range of circular iterators
         */
        typedef util::range<const_circ_neighbor_iterator> const_circ_neighbor_iterator_range;

        //@}

        ///\name Vertex set edition methods
        //@{

        /**
         * Remove a vertex from the graph.
         *
         * \param v vertex to erase
         *
         * \returns 1 if the vertex was erased, 0 else.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Add vertices to S
         * vertex v = S.any();
         * S.erase(v); // Remove a vertex of S
         * \endcode
         */
        size_type erase( const vertex_t& v );

        /**
         * Insert a new vertex in the graph.
         *
         * \returns An iterator on the inserted vertex.
         *
         * Example:
         * \code
         * vvgraph S;
         * vertex v; // Create a vertex
         * S.insert(v); // Insert it in the graph
         * \endcode
         */
        iterator insert( const vertex_t& v );
        //@}

        ///\name Vertex set lookup methods
        //@{

        /**
         * Return a vertex from the graph.
         *
         * \returns A vertex of the graph, or a null vertex if the graph is empty.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Add vertices in S
         * const vertex& v = S.any(); // Get a vertex in S
         * \endcode
         */
        const vertex_t& any() const;

        /**
         * Return the element of index \c idx.
         *
         * \note This is a convenience function whose complexity if O(idx)
         *
         * \returns The element \c idx position after the first one (using 
         * iterators)
         *
         * Example:
         * \code
         * vvgraph S;
         * // Fill in S
         * size_t i = 0, selected = 0;
         * forall(const vertex& v, S)
         * {
         *   if(condition(v))
         *   {
         *     selected = i;
         *     break;
         *   }
         *   ++i;
         * }
         * // ...
         * const vertex& v = S[selected]; // Work with the vertex previously selected
         * \endcode
         * This construct is useful is case you need to refer to your vertices by 
         * 32bits numbers. It is true if you use selection with OpenGL for 
         * example.
         */
        const vertex_t& operator[](size_type idx) const;

        /**
         * Test if \c v is in the graph.
         *
         * \returns \c true is \c v is in the graph.
         *
         * Example:
         * \code
         * vvgraph S;
         * vertex v;
         * S.insert(v);
         * assert(S.contains(v));
         * \endcode
         */
        bool contains( const vertex_t& v ) const;

        /**
         * Get a reference on the vertex in the graph
         *
         * If the vertex do not exist in the graph, the function returns a null 
         * vertex.
         *
         * \note This function is useful if you obtained the vertex not from this 
         * graph but will use it intensively.
         *
         * Example:
         * \code
         * vvgraph S1, S2;
         * // Initialize S1 and S2 with the same vertices but different neighborhood
         * forall(const vertex& v1, S1)
         * {
         *   const vertex& v2 = S2.reference(v1);
         *   // Use v2 in S2 and v1 in v1 for speed optimizatioN
         * }
         * \endcode
         */
        const vertex_t& reference( vertex_t v ) const
        {
          typename neighborhood_t::const_iterator found = this->findVertex(v);
          if(found != neighborhood.end())
            return found->first;
          return vertex_t::null;//vertex_t(0);
        }

        /**
         * Return the number of vertexes on the graph
         */
        size_type size() const { return neighborhood.size(); }

        /**
         * Test if there is any vertex in the graph
         */
        bool empty() const { return neighborhood.empty(); }

        //@}

        ///\name Neighborhood lookup methods
        //@{

        /**
         * Return a vertex in the neighborhood of \c v.
         *
         * \returns A vertex in the neighborhood of \c v, or a null vertex if \c v is 
         * not in the graph or \c v has no neighborhood.
         *
         * Example:
         * \code
         * vvgraph S;
         * forall(const vertex& v, S)
         * {
         *   const vertex& n = S.anyIn(v);
         *   if(n)
         *   {
         *     edge e = S.edge(v,n); // e exist !
         *   }
         * }
         * \endcode
         */
        const vertex_t& anyIn( const vertex_t& v ) const;

        /**
         * Returns the number of neighbors of \c v.
         *
         * If \c v is not in the graph, the behavior is undefined.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   size_t nb_neighbors = S.valence(v); // Number of neighbors of v in S
         *   // ...
         * }
         * \endcode
         */
        size_type valence( const vertex_t& v ) const;

        /**
         * Test if a vertex has a neighbor.
         *
         * Returns false if it does. If a vertex is not in the graph, it is 
         * considered as having no neighbor. In the same way, the null vertex has 
         * no neighbors.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(cnst vertex& v, S)
         * {
         *   if(!S.empty(v))
         *   {
         *      const vertex& n = S.anyIn(v);
         *      // Working with n a neighbor of v
         *   }
         * }
         * \endcode
         */
        bool empty( const vertex_t& v ) const;

        /**
         * Returns any incoming neighbor of \c v.
         *
         * \returns A vertex in the incoming neighborhood of \c v, or a null 
         * vertex if \c v is not in the graph or \c v has no incoming neighbor.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& n, S)
         * {
         *   const vertex& v = S.iAnyIn(v);
         *   if(v)
         *   {
         *     edge e = S.edge(v, n);
         *     // Work with e the edge from v to n
         *   }
         * }
         * \endcode
         */
        const vertex_t& iAnyIn(const vertex_t& v) const;

        /**
         * Returns the number of incoming neighbors of \c v
         *
         * The incoming neighbors are the set of vertexes with on edge ending on 
         * \c v.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& n, S)
         * {
         *    size_t nb_ineighbors = S.iValence(n);
         *    // nb_ineighbors is the number of vertices source of an edge toward n
         * }
         * \endcode
         */
        size_type iValence( const vertex_t& v ) const;

        /**
         * Test if a vertex has an incoming neighbor.
         *
         * Returns false if it does. If a vertex is not in the graph, it is 
         * considered as having no incoming neighbor. In the same way, the null 
         * vertex has no incoming neighbors.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& n, S)
         * {
         *   if(!S.iEmpty(n))
         *   {
         *     const vertex& v = S.iAnyIn(n);
         *     edge e = S.edge(v,n); // It exists !
         *   }
         * }
         * \endcode
         */
        bool iEmpty( const vertex_t& v ) const;

        /**
         * Returns the nth vertex after \c neighbor in the neighborhood of \c v.
         *
         * \returns The vertex found or a null vertex if there is any problem.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     const vertex& m, S.nextTo(v);
         *     // m is the vertex after n in v
         *   }
         * }
         * \endcode
         */
        const vertex_t& nextTo( const vertex_t& v, const vertex_t& neighbor, unsigned int n = 1 ) const;

        /**
         * Returns the nth vertex before ref in the neighborhood of v.
         *
         * If ref is not in the neighborhood of v, return a null vertex.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     const vertex& m, S.prevTo(v);
         *     // m is the vertex before n in v
         *   }
         * }
         * \endcode
         */
        const vertex_t& prevTo( const vertex_t& v, const vertex_t& ref, unsigned int n = 1 ) const;

        /**
         * Returns the edge from \c src to \c tgt.
         *
         * If the edge does not exists, returns a null edge.
         *
         * Example:
         * \code
         * struct VertexContent {};
         * struct EdgeContent { int a; }
         * typedef graph::VVGraph<VertexContent,EdgeContent> vvgraph;
         * typedef vvgraph::vertex_t vertex;
         * typedef vvgraph::edge_t edge;
         * // ...
         *
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     edge e = S.edge(v,n);
         *     e->a = 10;
         *   }
         * }
         * \endcode
         */
        edge_t edge( const vertex_t& src, const vertex_t& tgt );

        /**
         * Returns the arc (i.e. undirected edge) between \c src and \c tgt.
         *
         * When the arc is destroyed, the values of both half-edge are 
         * synchronized.
         *
         * Example:
         * \code
         * struct VertexContent();
         * struct EdgeContent {int a; }
         * typedef graph::VVGraph<VertexContent,EdgeContent> vvgraph;
         * typedef vvgraph::vertex_t vertex;
         * typedef vvgraph::arc_t arc;
         * typedef vvgraph::edge_t edge;
         * // ...
         *
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *     if(v<n)
         *       S.arc(v,n)->a = 5
         *  }
         * \endcode
         * At the end, all edges have a value of 5 for \c a, even though we 
         * applied the function to only half the edges.
         */
        arc_t arc(const vertex_t& src, const vertex_t& tgt);

        /**
         * Returns the edge from \c src to \c tgt.
         *
         * If the edge does not exists, returns a null edge.
         *
         * Example:
         * \code
         * struct VertexContent {};
         * struct EdgeContent { int a; }
         * typedef graph::VVGraph<VertexContent,EdgeContent> vvgraph;
         * typedef vvgraph::vertex_t vertex;
         * typedef vvgraph::edge_t edge;
         * typedef vvgraph::const_edge_t const_edge;
         * // ...
         *
         * void fct(const vvgraph& S)
         * {
         *   forall(const vertex& v, S)
         *   {
         *     forall(const vertex& n, S.neighbors(v))
         *     {
         *       const_edge e = S.edge(v,n);
         *       cout << e->a << endl; // Here e->a cannot be modified
         *     }
         *   }
         * }
         * \endcode
         */
        const_edge_t edge( const vertex_t& src, const vertex_t& tgt ) const;

        /**
         * Flag a neighbor for quick retrieval
         *
         * \return true if the vertex has been correctly flagged
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     if(condition(v,n))
         *     {
         *       S.flag(v,n);
         *       break;
         *     }
         *   }
         * }
         * // ...
         * forall(const vertex& v, S)
         * {
         *   const vertex& n = S.flagged(v);
         *   if(n) // If a neighbor has been flagged
         *   {
         *     // Work with the flagged neighbor
         *   }
         * }
         * \endcode
         */
        bool flag(const vertex_t& src, const vertex_t& neighbor);

        /**
         * Return the flagged neighbor or a null vertex is no neighbor have been flagged.
         *
         * \see VVGraph::flag(const vertex_t&, const vertex_t&)
         */
        const vertex_t& flagged(const vertex_t& src) const;

        /**
         * Return the constant range of neighbors of \c v
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     // n is a neighbor of v
         *   }
         * }
         * \endcode
         */
        const_neighbor_iterator_range neighbors(const vertex_t& v) const;
        /**
         * Return the range of neighbors of \c v
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     // n is a neighbor of v
         *   }
         * }
         * \endcode
         */
        neighbor_iterator_range neighbors(const vertex_t& v);

        /**
         * Return the range of neighbors of \c v, starting at \c n
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& m, S.neighbors(v,n))
         * {
         *   // m is a neighbor of v
         *   // on the first loop, m == n
         * }
         * \endcode
         */
        circ_neighbor_iterator_range neighbors(const vertex_t& v, const vertex_t& n);

        /**
         * Return the range of neighbors of \c v, starting at \c n
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& m, S.neighbors(v,n))
         * {
         *   // m is a neighbor of v
         *   // on the first loop, m == n
         * }
         * \endcode
         */
        const_circ_neighbor_iterator_range neighbors(const vertex_t& v, const vertex_t& n) const;

        /**
         * Return the range of incoming neighbors of \c v.
         *
         * Iterating over iNeighbors will go through all the vertexes having an edge 
         * toward \c v.
         *
         * Example:
         * \code
         * vvgraph S;
         * // initialize S
         * forall(const vertex& n, S)
         * {
         *   forall(const vertex& v, S.iNeighbors(v))
         *   {
         *     // n is a neighbor of v
         *   }
         * }
         * \endcode
         */
        ineighbor_iterator_range iNeighbors(const vertex_t& v) const;

        /**
         * Return the source vertex of the edge
         *
         * Example:
         * \code
         * void fct(edge e, const vvgraph& S)
         * {
         *   const vertex& src = S.source(e);
         *   const vertex& tgt = S.target(e);
         *   // Work with the vertices
         * }
         * \endcode
         */
        const vertex_t& source( const edge_t& edge ) const
        {
          typename neighborhood_t::const_iterator found = this->findVertex(vertex_t(edge.source()));
          if(found != neighborhood.end())
            return found->first;
          return vertex_t::null;//vertex_t(0);
        }

        const vertex_t& source( const arc_t& arc) const
        {
          typename neighborhood_t::const_iterator found = this->findVertex(vertex_t(arc.source()));
          if(found != neighborhood.end())
            return found->first;
          return vertex_t::null;//vertex_t(0);
        }

        /**
         * Return the target vertex of the edge
         *
         * Example:
         * \code
         * void fct(edge e, const vvgraph& S)
         * {
         *   const vertex& src = S.source(e);
         *   const vertex& tgt = S.target(e);
         *   // Work with the vertices
         * }
         * \endcode
         */
        const vertex_t& target( const edge_t& edge ) const
        {
          typename neighborhood_t::const_iterator found = this->findVertex(vertex_t(edge.target()));
          if(found != neighborhood.end())
            return found->first;
          return vertex_t::null;//vertex_t(0);
        }

        const vertex_t& target( const arc_t& arc) const
        {
          typename neighborhood_t::const_iterator found = this->findVertex(vertex_t(arc.target()));
          if(found != neighborhood.end())
            return found->first;
          return vertex_t::null;//vertex_t(0);
        }

        //@}

        ///\name Neighborhood edition methods
        //@{
        /**
         * Erase the edge from \c src to \c tgt.
         *
         * \returns True if successful.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     if(condition)
         *     {
         *       S.eraseEdge(v,n);
         *       break; // The iteration must not continue after the neighborhood has been altered
         *     }
         *   }
         * }
         * \endcode
         */
        size_type eraseEdge( const vertex_t& src, const vertex_t& tgt);

        /**
         * Replace a vertex by another in a neighborhood.
         *
         * \param[in] v Vertex whose neighborhood is changed.
         * \param[in] neighbor Vertex to replace
         * \param[in] new_neighbor Vertex replacing \c neighbor
         *
         * If \c new_neighbor is already in the neighborhood of \c v, then the 
         * operation fails and nothing is changed.
         *
         * \returns The edge between \c v and \c new_neighbor or a null edge if 
         * anything goes wrong.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     if(condition)
         *     {
         *       vertex n1;
         *       S.replace(v,n,n1); // Now n1 is where n was in v
         *       break; // The iteration must not continue after the neighborhood has been altered
         *     }
         *   }
         * }
         * \endcode
         */
        edge_t replace( const vertex_t& v, const vertex_t& neighbor, const vertex_t& new_neighbor );

        /**
         * Insert neighbor in the neighborhood of v after reference.
         *
         * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
         * fails.
         *
         * \returns The just created edge if everything succeed, or a null edge.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     if(condition)
         *     {
         *       vertex n1;
         *       S.spliceAfter(v,n,n1); // Now n1 is after n in v
         *       break; // The iteration must not continue after the neighborhood has been altered
         *     }
         *   }
         * }
         * \endcode
         */
        edge_t spliceAfter( const vertex_t& v, const vertex_t& neighbor, const vertex_t& new_neighbor );

        /**
         * Splice a list of vertices after a neighbor.
         *
         * The order of the vertices is kept.
         */
        template <typename VertexContainer>
          std::vector<edge_t> spliceAfter( const vertex_t& v, const vertex_t& neighbor, const VertexContainer& new_neighbor );

        /**
         * Insert neighbor in the neighborhood of v before reference.
         *
         * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
         * fails.
         *
         * \returns The just created edge if everything succeed, or a null edge.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   forall(const vertex& n, S.neighbors(v))
         *   {
         *     if(condition)
         *     {
         *       vertex n1;
         *       S.spliceBefore(v,n,n1); // Now n1 is before n in v
         *       break; // The iteration must not continue after the neighborhood has been altered
         *     }
         *   }
         * }
         * \endcode
         */
        edge_t spliceBefore( const vertex_t& v, const vertex_t& neighbor, const vertex_t& new_neighbor );

        /**
         * Splice a list of vertices before a neighbor.
         *
         * The order of the vertices is kept.
         */
        template <typename VertexContainer>
          std::vector<edge_t> spliceBefore( const vertex_t& v, const vertex_t& neighbor, const VertexContainer& new_neighbor );

        /**
         * Insert a new edge in the graph, without ordering.
         *
         * If \c new_neighbor is already in the neighborhood of \c v, the insertion 
         * fails.
         *
         * \returns The just created edge if everything succeed, or a null edge.
         *
         * Example:
         * \code
         * vvgraph S;
         * vertex v1, v2;
         * S.insert(v1);
         * S.insert(v2);
         * S.insertEdge(v1, v2); // Insert the edge between v1 and v2
         * S.insertEdge(v2, v1); // Insert the edge between v2 and v1
         * \endcode
         */
        edge_t insertEdge( const vertex_t& src, const vertex_t& tgt );

        /**
         * Insert a set of edges, keeping the relative order of the list of edges
         */
        template <typename VertexContainer>
          std::vector<edge_t> insertEdges(const vertex_t& src, const VertexContainer& targets);

        /**
         * Clear the neighborhood of a vertex.
         *
         * All edges, to and from \c v will be erased.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   if(condition)
         *   {
         *     S.eraseAllEdges(v);
         *     assert(S.empty(v)); // No outgoing edges
         *     assert(S.iEmpty(v)); // No incoming edges
         *   }
         * }
         * \endcode
         */
        bool eraseAllEdges( const vertex_t& v );

        /**
         * Clear the outgoing edges of a vertex
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * forall(const vertex& v, S)
         * {
         *   if(condition)
         *   {
         *     S.clear(v);
         *     assert(S.empty(v));
         *   }
         * }
         * \endcode
         */
        bool clear( const vertex_t& v );

        //@}

        ///\name Whole structure methods
        //@{
        /**
         * Copy the graph into another graph
         */
        VVGraph& operator=( const VVGraph& other );

        /**
         * Swap the content of two graphs
         */
        void swap(VVGraph& other);

        /**
         * Test equality for the graphs.
         *
         * Two graphs are equal if they shared the sames vertices and their 
         * neighborhood are equivalent.
         *
         * \note Current implementation do not consider invariance of 
         * neighborhood by rotation.
         */
        bool operator==(const VVGraph& other) const;

        /**
         * Clear the graph.
         */
        void clear() { neighborhood.clear(); lookup.clear();  }

        /**
         * Clear all the edges in the graph
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * vvgraph::size_type s1 = S.size(); // Save the number of vertices
         * S.eraseAllEdges();
         * assert(s1 == S.size()); // The number of vertices didn't change
         * forall(const vertex& v, S)
         * {
         *   assert(S.empty(v)); // No neighbor
         * }
         * \endcode
         */
        void eraseAllEdges();

        /**
         * Extract the subgraph containing the \p vertexes
         *
         * The subgraph is the set of vertexes and the edges whose source and target 
         * are in the extracted vertexes.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * std::vector<vertex> to_keep;
         * // Fill in to_keep with a subset of the vertices of S
         * vvgraph S1 = S.subgraph(to_keep);
         * \endcode
         */
        template <typename VertexContainer>
          VVGraph subgraph( const VertexContainer& vertexes ) const;
        //@}

        ///\name STL compatibility methods
        //@{

        /**
         * Erase element at position \p pos
         *
         * And return the iterator on the next element (as required for sequence 
         * containers in C++)
         */
        iterator erase(iterator pos);

        /**
         * Erase the edge pointed to by the iterator.
         */
        neighbor_iterator eraseEdge(const vertex_t& v, neighbor_iterator pos);

        /**
         * Erase the edge pointed to by the iterator.
         */
        circ_neighbor_iterator eraseEdge(const vertex_t& v, circ_neighbor_iterator pos);

        /**
         * Insert a new vertex in the graph
         *
         * The iterator is ignored ...
         */
        iterator insert( iterator pos, const vertex_t& v );
        /**
         * Insert all the vertexes from first to last-1 in the graph.
         */
        template <typename Iterator>
        void insert( Iterator first, Iterator last );

        /**
         * Returns an iterator on the beginning of the set of vertexes of the graph.
         */
        iterator begin() { return neighborhood.begin(); }
        /**
         * Returns an iterator on the end of the set of vertexes of the graph.
         */
        iterator end() { return neighborhood.end(); }

        /**
         * Returns an iterator on the beginning of the set of vertexes of the graph.
         */
        const_iterator begin() const { return neighborhood.begin(); }
        /**
         * Returns an iterator on the end of the set of vertexes of the graph.
         */
        const_iterator end() const { return neighborhood.end(); }

        /**
         * Store the vertices in a STL container. By default, the cache will be 
         * setup for quick vertex access.
         *
         * Example:
         * \code
         * vvgraph S;
         * // Initialize S
         * std::vector<vertex> vertices;
         * S.store_vertices(vertices);
         * forall(const vertex& v, vertices)
         * {
         *   // The cache information is used to speed up lookup
         *   if(condition)
         *   {
         *     S.erase(v); // Valid, but do not use v in S after that
         *   }
         * }
         * \endcode
         */
        void store_vertices(std::vector<vertex_t>& vertices, bool cached = true) const
        {
          vertices.resize(this->size(), vertex_t::null);
          size_t k = 0;
          for(typename neighborhood_t::const_iterator it = neighborhood.begin() ;
              it != neighborhood.end() ; ++it)
          {
            const vertex_t& v = it->first;
            vertices[k] = v;
#ifndef VVGRAPH_NO_CACHE
            if(cached)
            {
              vertices[k].cache = v.cache;
              vertices[k].cache_source = v.cache_source;
            }
#endif
            ++k;
          }
        }

        /**
         * Find the vertex \c v in the graph.
         *
         * \returns An iterator on \c v if found, or end()
         */
        iterator find(const vertex_t& v) { return this->findVertex(v); }
        /**
         * Find the vertex \c v in the graph.
         *
         * \returns An iterator on \c v if found, or end()
         */
        const_iterator find(const vertex_t& v) const { return this->findVertex(v); }

        /**
         * Count the number of vertexes \p v in the graph (can be 0 or 1 only)
         */
        size_type count( const vertex_t& v ) const;

        /**
         * Find the vertex \c n in the neighborhood of \c v.
         *
         * \returns
         *  - An iterator on \c n in the neighborhood of \c v, if found
         *  - An iterator on the end of the neighborhood if \c n is not found in \c 
         *  v
         *  - Otherwise, the result is undefined
         */
        neighbor_iterator findIn(const vertex_t& v, const vertex_t& n);
        /**
         * Find the vertex \c n in the neighborhood of \c v.
         *
         * \returns
         *  - An iterator on \c n in the neighborhood of \c v, if found
         *  - An iterator on the end of the neighborhood if \c n is not found in \c 
         *  v
         *  - Otherwise, the result is undefined
         */
        const_neighbor_iterator findIn(const vertex_t& v, const vertex_t& n) const;

        /**
         * Clear the neighborhood of a vertex.
         *
         * All edges, to and from \c *it will be erased.
         */
        void eraseAllEdges( iterator it );

        /**
         * Clear the outgoing edges of a vertex.
         */
        void clear( iterator it );

        //@}

      protected:
        /**
         * Type of the result of the search for a vertex in a neighborhood
         *
         * This data structure is made such that all edit or lookup operations can 
         * be done from that data structure.
         */
        template <class Neighborhood, class Iterator>
          struct search_result_t
          {
            /**
             * Default constructor
             *
             * By default, the search is unsuccessful
             */
            search_result_t()
              : found(false)
              , neighborhood(0) {}

            /**
             * Successful constructor
             */
            search_result_t( Iterator i,
                             Neighborhood* n,
                             bool ok = true)
              : found(ok)
              , it(i)
              , neighborhood(n)
              {}

            /**
             * Copy constructor
             */
            search_result_t(const search_result_t& copy)
              : found(copy.found)
              , it(copy.it)
              , neighborhood(copy.neighborhood)
              {}

            /**
             * Convert the object to true if the search is successful
             */
            operator bool() { return found; }

            /**
             * True if the search was completely successful
             */
            bool found;
            /**
             * Iterator pointing in the edge list
             */
            Iterator it;
            /**
             * Neighborhood containing the element
             */
            Neighborhood* neighborhood;
          };

        /**
         * Iterator on the outgoing edges
         */
        typedef typename edge_list_t::iterator int_neighbor_iterator;
        /**
         * Constant iterator on the outgoing edges
         */
        typedef typename edge_list_t::const_iterator int_const_neighbor_iterator;

        /**
         * Result of a search in the neighborhood
         */
        typedef search_result_t<single_neighborhood_t, int_neighbor_iterator> neighbor_found_t;
        /**
         * Constant result of a search in the neighborhood
         */
        typedef search_result_t<const single_neighborhood_t, int_const_neighbor_iterator> const_neighbor_found_t;

        /**
         * Find a vertex in the graph and returns the iterator on it.
         *
         * Goal is to introduce a new optimization. When iterating over the 
         * vertices of the graph, they will cache their own iterator. So accessing 
         * their neighborhood would be constant time.
         */
        typename neighborhood_t::iterator findVertex(const vertex_t& v);

        /**
         * Constant version of findVertex(const vertex_t&)
         */
        typename neighborhood_t::const_iterator findVertex(const vertex_t& v) const;

        /**
         * Find a vertex neighbor in the neighborhood of v.
         *
         * If 'v' is in the graph and 'neighbor' is in its neighborhood, the 
         * result is convertible to true, 'neighborhood' points toward the 
         * neighborhood structure and 'it' points toward the neighbor.
         *
         * If 'v' is in the graph, but 'neighbor' is not in the neighborhood, the 
         * result is convertible to false, 'neighborhood' points toward the 
         * neighborhood structure and 'it' is neighborhood->edges.end().
         *
         * If v is not in the graph, the result is convertible to false and the 
         * address of the stored neighborhood is 0.
         */
        neighbor_found_t findInVertex( const vertex_t& v, const vertex_t& neighbor);

        /**
         * Constant version of findInVertex(const vertex_t&, const vertex_t&)
         */
        const_neighbor_found_t findInVertex( const vertex_t& v, const vertex_t& neighbor) const;

        /**
         * Insert v in the in_edges of new_neighbor and return true if the insertion 
         * was actually done, or false if v was already there.
         */
        std::pair<ineighbor_iterator,bool> insertInEdge(const vertex_t& v, const vertex_t& new_neighbor);

        /**
         * Undo the insertion of an in edge as a roll-back mechanism if the whole 
         * edge insertion fails.
         */
        void undoInEdge(const vertex_t& new_neighbor, ineighbor_iterator& it);

        /**
         * Update the edge cache information of vertex v in its neighborhood and 
         * the in_edge too.
         */
        void updateEdgeCache(const vertex_t& v,
                             typename edge_list_t::iterator new_edge_it,
                             const vertex_t& in_edge);

        /**
         * Update the vertex cache for vertex v
         */
        void updateVertexCache(const vertex_t& v,
                               typename neighborhood_t::iterator it);

        /**
         * Main data structure containing everything
         */
        neighborhood_t neighborhood;

        /**
         * Hash map to optimise look-up
         */
        lookup_t lookup;

        /**
         * Unique graph identifier
         */
        intptr_t graph_id;
    };

#ifdef USE_ALLOCATOR
  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::EdgeAlloc VVGraph<GRAPH_ARGS>::edge_alloc;
#endif

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::iterator
    VVGraph<GRAPH_ARGS>::erase( iterator it )
    {
      eraseAllEdges(it);
      lookup.erase(*it);
      return neighborhood.erase(it.base());
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::size_type
    VVGraph<GRAPH_ARGS>::erase( const vertex_t& v )
    {
      if(eraseAllEdges(v))
      {
        typename lookup_t::iterator it_found = lookup.find(v);
        if(it_found == lookup.end()) return 0;
        typename neighborhood_t::iterator it = it_found->second;
        neighborhood.erase(it);
        lookup.erase(v);
        return 1;
      }
      return 0;
    }
  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::eraseAllEdges()
    {
      for(typename neighborhood_t::iterator it = neighborhood.begin();
          it != neighborhood.end() ; ++it)
      {
        it->second.edges.clear();
        it->second.in_edges.clear();
        it->second.flagged = 0;
      }
    }

  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::eraseAllEdges( const vertex_t& v )
    {
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found != neighborhood.end() )
      {
        eraseAllEdges(it_found);
        return true;
      }
      return false;
    }

  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::eraseAllEdges( iterator it_found )
    {
      const vertex_t& v = *it_found;
      for(typename edge_list_t::iterator it_el = it_found.base()->second.edges.begin() ;
          it_el != it_found.base()->second.edges.end() ; ++it_el)
      {
        typename lookup_t::iterator it_found = lookup.find(it_el->target);
        if(it_found != lookup.end())
          it_found->second->second.in_edges.erase(v);
      }
      it_found.base()->second.edges.clear();
      it_found.base()->second.flagged = 0;
      // Remove the incoming edges
      in_edges_t &in_edges = it_found.base()->second.in_edges;
      for( typename in_edges_t::iterator it = in_edges.begin() ;
           it != in_edges.end() ; ++it )
      {
#ifndef VVGRAPH_NO_EDGE_CACHE
        EdgeCache *cache = reinterpret_cast<EdgeCache*>(it->cache);
        if(!cache or !cache->eraseEdge())
        {
#endif
          neighbor_found_t found = findInVertex(*it, v);
          found.neighborhood->edges.erase(found.it);
          if(found.neighborhood->flagged && *found.neighborhood->flagged == v) found.neighborhood->flagged = 0;
#ifndef VVGRAPH_NO_EDGE_CACHE
        }
#endif
      }
      in_edges.clear();
    }


  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::clear( const vertex_t& v )
    {
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found != neighborhood.end() )
      {
        clear(it_found);
        return true;
      }
      return false;
    }

  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::clear( iterator it )
    {
      const vertex_t& v = *it;
      for(typename edge_list_t::iterator it_el = it.base()->second.edges.begin() ;
          it_el != it.base()->second.edges.end() ; ++it_el)
      {
        typename neighborhood_t::iterator it_found = this->findVertex(it_el->target);
        it_found->second.in_edges.erase(v);
      }
      it.base()->second.edges.clear();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::size_type
    VVGraph<GRAPH_ARGS>::count( const vertex_t& v ) const
    {
      return lookup.count(v);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::iterator
    VVGraph<GRAPH_ARGS>::insert( const vertex_t& v )
    {
      typename lookup_t::iterator it_found;
      bool inserted;
      util::tie(it_found, inserted) = lookup.insert(std::make_pair(v, typename neighborhood_t::iterator()));
      if(inserted)
      {
        neighborhood.push_front(std::make_pair(v,single_neighborhood_t()));
        typename neighborhood_t::iterator it = neighborhood.begin();
        it_found->second = it;
        updateVertexCache(it->first, it);
        return it;
      }
      return it_found->second;
    }

  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::updateVertexCache(const vertex_t& v,
                                                typename neighborhood_t::iterator it)
    {
#ifndef VVGRAPH_NO_VERTEX_CACHE
      it->second.update(graph_id, v, it);
#endif
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::iterator
    VVGraph<GRAPH_ARGS>::insert( iterator , const vertex_t& v )
    {
      return this->insert(v);
    }

  template <GRAPH_TEMPLATE>
    template <typename Iterator>
    void VVGraph<GRAPH_ARGS>::insert( Iterator first, Iterator last )
    {
      for(Iterator it = first ; it != last ; ++it)
      {
        insert(*it);
      }
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::any() const
    {
      if(neighborhood.empty())
        return vertex_t::null;
      return *begin();
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::operator[](size_type idx) const
    {
      typename neighborhood_t::const_iterator it = neighborhood.begin();
      std::advance(it, idx);
      return it->first;
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::anyIn( const vertex_t& v ) const
    {
      if(v.isNull())
        return vertex_t::null;//vertex_t(0);
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end() || it_found->second.edges.empty())
        return vertex_t::null;//vertex_t(0);
      const edge_list_t& lst = it_found->second.edges;
      return lst.front().target;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::size_type
    VVGraph<GRAPH_ARGS>::valence( const vertex_t& v ) const
    {
      if(v.isNull())
        return 0;
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return 0;
      return it_found->second.edges.size();
    }

  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::empty( const vertex_t& v ) const
    {
      if(v.isNull())
        return true;
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return true;
      return it_found->second.edges.empty();
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::iAnyIn( const vertex_t& v ) const
    {
      if(v.isNull())
        return vertex_t::null;//vertex_t(0);
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end() || it_found->second.in_edges.empty())
        return vertex_t::null;//vertex_t(0);
      const in_edges_t& lst = it_found->second.in_edges;
      return *lst.begin();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::size_type
    VVGraph<GRAPH_ARGS>::iValence( const vertex_t& v ) const
    {
      if(v.isNull())
        return 0;
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return 0;
      return it_found->second.in_edges.size();
    }

  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::iEmpty( const vertex_t& v ) const
    {
      if(v.isNull())
        return true;
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return true;
      return it_found->second.in_edges.empty();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::neighborhood_t::iterator
    VVGraph<GRAPH_ARGS>::findVertex( const vertex_t& v )
    {
#ifndef VVGRAPH_NO_VERTEX_CACHE
      if(v.cache)
      {
        if(v.cache_source == 0)
        {
          VertexCache* cache = reinterpret_cast<VertexCache*>(v.cache);
          if(cache->graph_id() == graph_id)
          {
            return cache->neighborhood();
          }
        }
      }
#endif // VVGRAPH_NO_VERTEX_CACHE
      typename lookup_t::const_iterator it_found = lookup.find(v);
      if(it_found != lookup.end())
        return it_found->second;
      return neighborhood.end();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::neighborhood_t::const_iterator
    VVGraph<GRAPH_ARGS>::findVertex( const vertex_t& v ) const
    {
#ifndef VVGRAPH_NO_VERTEX_CACHE
      if(v.cache)
      {
        if(v.cache_source == 0)
        {
          VertexCache* cache = reinterpret_cast<VertexCache*>(v.cache);
          if(cache->graph_id() == graph_id)
            return cache->neighborhood();
        }
      }
#endif // VVGRAPH_NO_VERTEX_CACHE
      typename lookup_t::const_iterator it_found = lookup.find(v);
      if(it_found != lookup.end())
        return it_found->second;
      return neighborhood.end();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::neighbor_found_t
    VVGraph<GRAPH_ARGS>::findInVertex( const vertex_t& v, const vertex_t& n)
    {
      if(v.isNull() || n.isNull() || v == n)
        return neighbor_found_t();
#ifndef VVGRAPH_NO_EDGE_CACHE
      if(n.cache_source == v.id())
      {
        EdgeCache* cache = reinterpret_cast<EdgeCache*>(n.cache);
        if(cache->graph_id() == graph_id)
        {
          return neighbor_found_t(cache->it(), cache->neighborhood());
        }
      }
      else if(v.cache_source == v.id())
      {
        EdgeCache* cache = reinterpret_cast<EdgeCache*>(v.cache);
        if(cache->graph_id() == graph_id and cache->it()->target == n)
        {
          return neighbor_found_t(cache->it(), cache->neighborhood());
        }
      }
#endif // VVGRAPH_NO_EDGE_CACHE
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return neighbor_found_t();
      edge_list_t &lst = it_found->second.edges;
      single_neighborhood_t &neighborhood = it_found->second;
      for(typename edge_list_t::iterator it = lst.begin() ;
          it != lst.end() ; ++it)
      {
        if(it->target == n)
        {
          return neighbor_found_t(it, &neighborhood);
        }
      }
      return neighbor_found_t(lst.end(), &neighborhood, false);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::const_neighbor_found_t
    VVGraph<GRAPH_ARGS>::findInVertex( const vertex_t& v, const vertex_t& n) const
    {
      if(v.isNull() || n.isNull() || v == n)
        return const_neighbor_found_t();
#ifndef VVGRAPH_NO_EDGE_CACHE
      if(n.cache_source == v.id())
      {
        EdgeCache  *cache = reinterpret_cast<EdgeCache*>(n.cache);
        if(cache->graph_id() == graph_id)
        {
          return const_neighbor_found_t(cache->it(), cache->neighborhood());
        }
      }
      else if(v.cache_source == v.id())
      {
        EdgeCache *cache = reinterpret_cast<EdgeCache*>(v.cache);
        if(cache->graph_id() == graph_id and cache->it()->target == n)
        {
          return const_neighbor_found_t(cache->it(), cache->neighborhood());
        }
      }
#endif // VVGRAPH_NO_EDGE_CACHE
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return const_neighbor_found_t();
      const edge_list_t &lst = it_found->second.edges;
      const single_neighborhood_t &neighborhood = it_found->second;
      for(typename edge_list_t::const_iterator it = lst.begin() ;
          it != lst.end() ; ++it)
      {
        if(it->target == n)
        {
          return const_neighbor_found_t(it, &neighborhood);
        }
      }
      return const_neighbor_found_t(lst.end(), &neighborhood, false);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::neighbor_iterator
    VVGraph<GRAPH_ARGS>::findIn(const vertex_t& v, const vertex_t& n)
    {
      neighbor_found_t found = this->findInVertex(v, n);
      if(found.neighborhood)
        return neighbor_iterator(found.it);
      return neighbor_iterator();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::const_neighbor_iterator
    VVGraph<GRAPH_ARGS>::findIn(const vertex_t& v, const vertex_t& n) const
    {
      const_neighbor_found_t found = this->findInVertex(v, n);
      if(found.neighborhood)
        return const_neighbor_iterator(found.it);
      return const_neighbor_iterator(found.it);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::neighbor_iterator
    VVGraph<GRAPH_ARGS>::eraseEdge(const vertex_t& v, neighbor_iterator pos)
    {
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return neighbor_iterator();
      const vertex_t& n = *pos;
      typename neighborhood_t::iterator it_found2 = this->findVertex(n);
      if(it_found2 == neighborhood.end())
        return neighbor_iterator();
      size_type result = it_found2->second.in_edges.erase(v);
      if(result)
      {
        typename edge_list_t::iterator it = pos.base();
        return it_found->second.edges.erase(it);
      }
      return neighbor_iterator();
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::circ_neighbor_iterator
    VVGraph<GRAPH_ARGS>::eraseEdge(const vertex_t& v, circ_neighbor_iterator pos)
    {
      neighbor_iterator nit = eraseEdge(v, pos.base());
      if(pos.isInitIterator(nit))
        return pos.end();
      return circ_neighbor_iterator(pos, nit);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::size_type
    VVGraph<GRAPH_ARGS>::eraseEdge( const vertex_t& v, const vertex_t& n )
    {
      neighbor_found_t found = this->findInVertex(v, n);
      if(!found)
        return 0;
      if(found.neighborhood->flagged && *found.neighborhood->flagged == n) found.neighborhood->flagged = 0;
      found.neighborhood->edges.erase(found.it);
      typename lookup_t::iterator it_found = lookup.find(n);
      if(it_found != lookup.end())
        it_found->second->second.in_edges.erase(v);
      return 1;
    }

  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::contains( const vertex_t& v ) const
    {
      return this->findVertex(v) != neighborhood.end();
    }

  template <GRAPH_TEMPLATE>
    std::pair<typename VVGraph<GRAPH_ARGS>::ineighbor_iterator,bool>
    VVGraph<GRAPH_ARGS>::insertInEdge(const vertex_t& v, const vertex_t& new_neighbor)
    {
      typename lookup_t::iterator found = lookup.find(new_neighbor);
      if(found == lookup.end())
        return std::make_pair(ineighbor_iterator(), false);
      //typename neighborhood_t::iterator found = neighborhood.find(new_neighbor);
      //if(found == neighborhood.end())
      //return std::make_pair(ineighbor_iterator(), false);
      return found->second->second.in_edges.insert(v);
    }

  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::undoInEdge(const vertex_t& new_neighbor,
                                         ineighbor_iterator& it)
    {
      typename lookup_t::iterator found = lookup.find(new_neighbor);
      //typename neighborhood_t::iterator found = neighborhood.find(new_neighbor);
      //if(found == neighborhood.end())
      //return;
      if(found == lookup.end())
        return;
      found->second->second.in_edges.erase(it);
    }

  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::updateEdgeCache(const vertex_t& v,
                                              typename edge_list_t::iterator new_edge_it,
                                              const vertex_t& in_edge)
    {
#ifndef VVGRAPH_NO_EDGE_CACHE
      if(not compact)
      {
        // Caching in the neighbor
        neighbor_t& neighbor = *new_edge_it;
        neighbor.setIt(new_edge_it);
        neighbor.target.cache_source = v.id();
        neighbor.target.cache = reinterpret_cast<void*>(&neighbor);

        // Caching in the incoming edge source
        in_edge.cache_source = v.id();
        in_edge.cache = reinterpret_cast<void*>(&neighbor);
      }
#endif // VVGRAPH_NO_EDGE_CACHE
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::edge_t
    VVGraph<GRAPH_ARGS>::replace( const vertex_t& v,
                                  const vertex_t& neighbor,
                                  const vertex_t& new_neighbor )
    {
      if(new_neighbor.isNull() or (v == new_neighbor) or (neighbor == new_neighbor))
        return edge_t();
      neighbor_found_t found = this->findInVertex(v, neighbor);
      if(!found)
        return edge_t();
      typename neighborhood_t::iterator n_found = this->findVertex(neighbor);
      ineighbor_iterator it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge_t();
      n_found->second.in_edges.erase(v);
      found.it->target = new_neighbor;
      found.it->clear_edge();
      updateEdgeCache(v, found.it, *it_in);
      if(n_found->second.flagged and *n_found->second.flagged == neighbor) n_found->second.flagged = 0;
      return edge_t(v.id(), new_neighbor.id(), &*found.it);
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::nextTo( const vertex_t& v,
                                 const vertex_t& ref,
                                 unsigned int n ) const
    {
      const_neighbor_found_t found = this->findInVertex(v, ref);
      if(!found)
        return vertex_t::null;//vertex_t(0);
      typename edge_list_t::const_iterator it = found.it;
      const edge_list_t& lst = found.neighborhood->edges;
      for(unsigned int i = 0 ; i < n ; ++i)
      {
        ++it;
        if(it == lst.end())
          it = lst.begin();
      }
      return it->target;
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::prevTo( const vertex_t& v,
                                 const vertex_t& ref,
                                 unsigned int n ) const
    {
      const_neighbor_found_t found = this->findInVertex(v, ref);
      if(!found)
        return vertex_t::null;//vertex_t(0);
      typename edge_list_t::const_iterator it = found.it;
      const edge_list_t& lst = found.neighborhood->edges;
      for(unsigned int i = 0 ; i < n ; ++i)
      {
        if(it == lst.begin())
          it = lst.end();
        --it;
      }
      return it->target;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::edge_t
    VVGraph<GRAPH_ARGS>::edge( const vertex_t& src,
                               const vertex_t& tgt )
    {
      neighbor_found_t found = this->findInVertex(src, tgt);
      if(!found)
        return edge_t();
      return edge_t(src.id(), tgt.id(), &*found.it);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::arc_t
    VVGraph<GRAPH_ARGS>::arc( const vertex_t& src,
                              const vertex_t& tgt )
    {
      neighbor_found_t found = this->findInVertex(src, tgt);
      if(!found)
        return arc_t();
      neighbor_found_t other = this->findInVertex(tgt, src);
      if(!other)
        return arc_t();
      return arc_t(src.id(), tgt.id(), &*found.it, &*other.it);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::const_edge_t
    VVGraph<GRAPH_ARGS>::edge( const vertex_t& src,
                               const vertex_t& tgt ) const
    {
      const_neighbor_found_t found = this->findInVertex(src, tgt);
      if(!found)
        return const_edge_t();
      return const_edge_t(src.id(), tgt.id(), &*found.it);
    }

  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::flag( const vertex_t& src,
                                    const vertex_t& neighbor )
    {
      neighbor_found_t found = this->findInVertex(src, neighbor);
      if(!found)
        return false;
      found.neighborhood->flagged = &(found.it->target);
      return true;
    }

  template <GRAPH_TEMPLATE>
    const typename VVGraph<GRAPH_ARGS>::vertex_t&
    VVGraph<GRAPH_ARGS>::flagged( const vertex_t& src ) const
    {
      typename neighborhood_t::const_iterator found = this->findVertex(src);
      if(found == neighborhood.end() or found->second.flagged == 0)
        return vertex_t::null;
      return *found->second.flagged;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::edge_t
    VVGraph<GRAPH_ARGS>::spliceAfter( const vertex_t& v,
                                      const vertex_t& neighbor,
                                      const vertex_t& new_neighbor )
    {
      if(new_neighbor.isNull() || v == new_neighbor)
        return edge_t();
      neighbor_found_t found = this->findInVertex(v, neighbor);
      if(!found)
        return edge_t();
      ineighbor_iterator it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge_t();
      ++found.it;
      typename edge_list_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor_t(new_neighbor, EdgeContent(), *found.neighborhood, graph_id));
      updateEdgeCache(v, new_edge_it, *it_in);
      return edge_t(v.id(), new_neighbor.id(), &*new_edge_it);
    }

  template <GRAPH_TEMPLATE>
    template <typename VertexContainer>
    std::vector<typename VVGraph<GRAPH_ARGS>::edge_t>
    VVGraph<GRAPH_ARGS>::spliceAfter( const vertex_t& v,
                                      const vertex_t& neighbor,
                                      const VertexContainer& new_neighbors )
    {
      std::vector<edge_t> result;
      neighbor_found_t found = this->findInVertex(v, neighbor);
      if(!found) return result;
      ineighbor_iterator it_in;
      bool inserted;
      ++found.it;
      result.reserve(new_neighbors.size());
      for(typename VertexContainer::const_iterator it = new_neighbors.begin() ;
          it != new_neighbors.end() ; ++it) {
        const vertex_t& new_neighbor = *it;
        util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
        if(!inserted) return result;
        typename edge_list_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor_t(new_neighbor, EdgeContent(), *found.neighborhood, graph_id));
        updateEdgeCache(v, new_edge_it, *it_in);
        result.push_back(edge_t(v.id(), new_neighbor.id(), &*new_edge_it));
      }
      return result;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::edge_t
    VVGraph<GRAPH_ARGS>::spliceBefore( const vertex_t& v,
                                       const vertex_t& neighbor,
                                       const vertex_t& new_neighbor )
    {
      if(new_neighbor.isNull() || v == new_neighbor)
        return edge_t();
      neighbor_found_t found = this->findInVertex(v, neighbor);
      if(!found)
        return edge_t();
      ineighbor_iterator it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge_t();
      typename edge_list_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor_t(new_neighbor, EdgeContent(), *found.neighborhood, graph_id));
      updateEdgeCache(v, new_edge_it, *it_in);
      return edge_t(v.id(), new_neighbor.id(), &*new_edge_it);
    }

  template <GRAPH_TEMPLATE>
    template <typename VertexContainer>
    std::vector<typename VVGraph<GRAPH_ARGS>::edge_t>
    VVGraph<GRAPH_ARGS>::spliceBefore( const vertex_t& v,
                                       const vertex_t& neighbor,
                                       const VertexContainer& new_neighbors )
    {
      std::vector<edge_t> result;
      neighbor_found_t found = this->findInVertex(v, neighbor);
      if(!found) return result;
      ineighbor_iterator it_in;
      bool inserted;
      result.reserve(new_neighbors.size());
      for(typename VertexContainer::const_iterator it = new_neighbors.begin() ;
          it != new_neighbors.end() ; ++it) {
        const vertex_t& new_neighbor = *it;
        util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
        if(!inserted) return result;
        typename edge_list_t::iterator new_edge_it = found.neighborhood->edges.insert(found.it, neighbor_t(new_neighbor, EdgeContent(), *found.neighborhood, graph_id));
        updateEdgeCache(v, new_edge_it, *it_in);
        result.push_back(edge_t(v.id(), new_neighbor.id(), &*new_edge_it));
      }
      return result;
    }


  template <GRAPH_TEMPLATE>
    template <typename VertexContainer>
    std::vector<typename VVGraph<GRAPH_ARGS>::edge_t>
    VVGraph<GRAPH_ARGS>::insertEdges( const vertex_t& v, const VertexContainer& new_neighbors )
    {
      std::vector<edge_t> result;
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end()) return result;
      ineighbor_iterator it_in;
      bool inserted;
      edge_list_t& lst = it_found->second.edges;
      typename edge_list_t::iterator lst_it = lst.begin();
      result.reserve(new_neighbors.size());
      for(typename VertexContainer::const_iterator it = new_neighbors.begin() ;
          it != new_neighbors.end() ; ++it) {
        const vertex_t& new_neighbor = *it;
        if(new_neighbor.isNull() or new_neighbor == v) return result;
        util::tie(it_in, inserted) = this->insertInEdge(v, new_neighbor);
        if(!inserted) return result;
        typename edge_list_t::iterator new_lst_it = lst.insert(lst_it, neighbor_t(new_neighbor, EdgeContent(), it_found->second, graph_id));
        updateEdgeCache(v, new_lst_it, *it_in);
        result.push_back(edge_t(v.id(), new_neighbor.id(), &*new_lst_it));
      }
      return result;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::edge_t
    VVGraph<GRAPH_ARGS>::insertEdge( const vertex_t& v,
                                     const vertex_t& new_neighbor )
    {
      if(v.isNull() || new_neighbor.isNull() || v == new_neighbor )
        return edge_t();
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return edge_t();
      ineighbor_iterator it_in;
      bool inserted;
      util::tie(it_in, inserted) = insertInEdge(v, new_neighbor);
      if(!inserted)
        return edge_t();
      edge_list_t &lst = it_found->second.edges;
      lst.push_front(neighbor_t(new_neighbor, EdgeContent(), it_found->second, graph_id));
      typename edge_list_t::iterator it = lst.begin();
      updateEdgeCache(v, it, *it_in);
      return edge_t(v.id(), new_neighbor.id(), &*it);
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::const_neighbor_iterator_range
    VVGraph<GRAPH_ARGS>::neighbors(const vertex_t& v) const
    {
      const_neighbor_iterator_range result;
      if(v.isNull())
        return result;
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return result;
      const edge_list_t &lst = it_found->second.edges;
      result = util::make_range(const_neighbor_iterator(lst.begin()),
                                const_neighbor_iterator(lst.end()));
      return result;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::neighbor_iterator_range
    VVGraph<GRAPH_ARGS>::neighbors(const vertex_t& v)
    {
      neighbor_iterator_range result;
      if(v.isNull())
        return result;
      typename neighborhood_t::iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return result;
      edge_list_t &lst = it_found->second.edges;
      result = util::make_range(neighbor_iterator(lst.begin()),
                                neighbor_iterator(lst.end()));
      return result;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::const_circ_neighbor_iterator_range
    VVGraph<GRAPH_ARGS>::neighbors(const vertex_t& v, const vertex_t& n) const
    {
      const_circ_neighbor_iterator_range result;
      if(v.isNull())
        return result;
      const_neighbor_found_t found = this->findInVertex(v, n);
      if(!found)
        return result;
      const edge_list_t &lst = found.neighborhood->edges;
      result = util::make_range(const_circ_neighbor_iterator(lst.begin(), lst.end(), found.it),
                                const_circ_neighbor_iterator(lst.begin(), lst.end()));
      return result;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::circ_neighbor_iterator_range
    VVGraph<GRAPH_ARGS>::neighbors(const vertex_t& v, const vertex_t& n)
    {
      circ_neighbor_iterator_range result;
      if(v.isNull())
        return result;
      neighbor_found_t found = this->findInVertex(v, n);
      if(!found)
        return result;
      edge_list_t &lst = found.neighborhood->edges;
      result = util::make_range(circ_neighbor_iterator(lst.begin(), lst.end(), found.it),
                                circ_neighbor_iterator(lst.begin(), lst.end()));
      return result;
    }

  template <GRAPH_TEMPLATE>
    typename VVGraph<GRAPH_ARGS>::ineighbor_iterator_range
    VVGraph<GRAPH_ARGS>::iNeighbors(const vertex_t& v) const
    {
      ineighbor_iterator_range result;
      if(v.isNull())
        return result;
      typename neighborhood_t::const_iterator it_found = this->findVertex(v);
      if(it_found == neighborhood.end())
        return result;
      const in_edges_t &lst = it_found->second.in_edges;
      result = util::make_range(ineighbor_iterator(lst.begin()),
                                ineighbor_iterator(lst.end()));
      return result;
    }

  template <GRAPH_TEMPLATE>
    template <typename VertexContainer>
    VVGraph<GRAPH_ARGS> VVGraph<GRAPH_ARGS>::subgraph( const VertexContainer& verts ) const
    {
      typename VertexContainer::const_iterator it;
      VVGraph result;
      // First, insert the vertexes in the result graph
      for(it = verts.begin() ; it != verts.end() ; ++it)
      {
        result.insert(*it);
      }
      typename neighborhood_t::const_iterator it_orign;
      typename neighborhood_t::iterator it_n;
      typename edge_list_t::const_reverse_iterator it_sn;
      // Then, creates the in and out edges
      for(it_n = result.neighborhood.begin() ; it_n != result.neighborhood.end() ; ++it_n)
      {
        const vertex_t& v = it_n->first;
        result.updateVertexCache(v, it_n);
        // For each vertex, find out which edges to add
        it_orign = this->findVertex(v);
        const edge_list_t& orig_lst = it_orign->second.edges;
        edge_list_t& lst = it_n->second.edges;
        // Going backward because it is easier to retrieve the iterator ...
        for( it_sn = orig_lst.rbegin() ; it_sn != orig_lst.rend() ; ++it_sn )
        {
          // For each existing edge from the vertex, keep only the ones whose 
          // target are on the subgraph
          if(result.contains(it_sn->target))
          {
            // Insert the incoming edge
            ineighbor_iterator it_in = result.insertInEdge(v, it_sn->target).first;
            // Insert the outgoing edge
            lst.push_front(neighbor_t(it_sn->target, *it_sn, it_n->second, result.graph_id));
            // Update the cache
            result.updateEdgeCache(v, lst.begin(), *it_in);
          }
        }
        // At last, update the flag, if any
        if(it_orign->second.flagged and result.contains(*it_orign->second.flagged))
        {
          const_neighbor_found_t found = this->findInVertex(v, *it_orign->second.flagged);
          it_n->second.flagged = &(found.it->target);
        }
      }
      return result;
    }

  template <GRAPH_TEMPLATE>
    VVGraph<GRAPH_ARGS>::VVGraph(const VVGraph& copy)
    : graph_id(graph::getNextGraphId())
    {
      *this = copy;
    }

#ifdef USE_CXX0X
  template <GRAPH_TEMPLATE>
    VVGraph<GRAPH_ARGS>::VVGraph(VVGraph&& copy)
      : neighborhood(std::move(copy.neighborhood))
      , lookup(std::move(copy.lookup))
      , graph_id(copy.graph_id)
    {
      copy.neighborhood.clear();
      copy.lookup.clear();
      copy.graph_id = graph::getNextGraphId();
    }
#endif

  template <GRAPH_TEMPLATE>
    bool VVGraph<GRAPH_ARGS>::operator==(const VVGraph& other) const
    {
      if(neighborhood.size() != other.neighborhood.size())
        return false;
      for(typename neighborhood_t::const_iterator it = neighborhood.begin() ;
          it != neighborhood.end() ; ++it)
      {
        const vertex_t& v1 = it->first;
        typename lookup_t::const_iterator it_found = other.lookup.find(v1);
        if(it_found == other.lookup.end())
          return false;
        if(it->second.edges.empty())
        {
          if(!it_found->second->second.edges.empty())
            return false;
        }
        else
        {
          const edge_list_t& lst = it->second.edges;
          const edge_list_t& olst = it_found->second->second.edges;
          if(lst.size() != olst.size()) return false;
          const vertex_t& v2 = lst.begin()->target;
          bool found = false;
          for(typename edge_list_t::const_iterator it_olst = olst.begin() ;
              it_olst != olst.end() ; ++it_olst)
          {
            if(it_olst->target == v2)
            {
              found = true;
              for(typename edge_list_t::const_iterator it_lst = lst.begin() ;
                  it_lst != lst.end() ; ++it_lst)
              {
                if(it_lst->target != it_olst->target) return false;
                ++it_olst;
                if(it_olst == olst.end()) it_olst = olst.begin();
              }
              break;
            }
          }
          if(!found) return false;
        }
      }
      return true;
    }

  template <GRAPH_TEMPLATE>
    VVGraph<GRAPH_ARGS>&
    VVGraph<GRAPH_ARGS>::operator=( const VVGraph& other )
    {
      //    for(const_iterator it = other.begin() ; it != other.end() ; ++it)
      //    {
      //      this->insert(*it);
      //    }

      //    for(const_iterator it = other.begin() ; it != other.end() ; ++it)
      //    {
      //      vertex_t prev(0);
      //      const vertex_t& v = *it;
      //      const_neighbor_iterator_range ns = other.neighbors(v);
      //      for(const_neighbor_iterator itn = ns.first ; itn != ns.second ; ++itn)
      //      {
      //        const vertex_t& n = *itn;
      //        if(prev)
      //          this->spliceAfter(v, prev, n);
      //        else
      //          this->insertEdge(v,n);
      //        prev = n;
      //      }
      //    }

      // Copy the structure
      neighborhood = other.neighborhood;
      lookup.clear();
      typename neighborhood_t::iterator it;
      for(it = neighborhood.begin() ; it != neighborhood.end() ; ++it)
      {
        lookup[it->first] = it;
      }
      // And reconstruct the cache
      typename edge_list_t::iterator it_e;
      for(it = neighborhood.begin() ; it != neighborhood.end() ; ++it)
      {
        const vertex_t& src = it->first;
        updateVertexCache(src, it);
#ifndef VVGRAPH_NO_EDGE_CACHE
        if(not compact)
        {
          edge_list_t& lst = it->second.edges;
          for(it_e = lst.begin() ; it_e != lst.end() ; ++it_e)
          {
            // Find the incoming vertex corresponding the this edge
            typename lookup_t::iterator it_found = lookup.find(it_e->target);
            const vertex_t& iv = *(it_found->second->second.in_edges.find(src));
            //const vertex_t& iv = *(neighborhood[it_e->target].in_edges.find(src));
            updateEdgeCache(src, it_e, iv);
            it_e->set(&it->second, graph_id);
          }
        }
#endif
      }
      return *this;
    }

  template <GRAPH_TEMPLATE>
    void VVGraph<GRAPH_ARGS>::swap(VVGraph& other)
    {
      neighborhood.swap(other.neighborhood);
      lookup.swap(lookup);
      std::swap(graph_id, other.graph_id);
    }

  template <typename Graph>
    void create_graph_methods(Graph& G)
    {
      if(G.size())
      {
        if(G.valence(G.any()) > G.size())
        {
          // Call all fcts
          typename Graph::vertex_t v = G.any();
          G.erase(v);
          G.insert(v);
          G[10];
          G.reference(v);
          G.empty();
          G.anyIn(v);
          G.valence(v);
          G.empty(v);
          G.iAnyIn(v);
          G.iValence(v);
          G.iEmpty(v);
          G.nextTo(v, v);
          G.prevTo(v, v);
          G.edge(v, v);
          G.flag(v,v);
          G.flagged(v);
          G.neighbors(v);
          G.iNeighbors(v);
          G.source(typename Graph::edge_t());
          G.target(typename Graph::edge_t());
          G.eraseEdge(v, v);
          G.replace(v, v, v);
          G.spliceAfter(v, v, v);
          G.spliceBefore(v, v, v);
          G.insertEdge(v,v);
          G.eraseAllEdges(v);
          G.begin();
          G.end();
          G.find(v);
          G.count(v);
          G.findIn(v, v);
          G.swap(G);
          G = G;
          {
            const Graph& cG = G;
            cG[10];
            cG.reference(v);
            cG.empty();
            cG.anyIn(v);
            cG.valence(v);
            cG.empty(v);
            cG.iAnyIn(v);
            cG.iValence(v);
            cG.iEmpty(v);
            cG.nextTo(v, v);
            cG.prevTo(v, v);
            cG.edge(v, v);
            cG.flagged(v);
            cG.neighbors(v);
            cG.iNeighbors(v);
            cG.source(typename Graph::edge_t());
            cG.target(typename Graph::edge_t());
            cG.begin();
            cG.end();
            cG.find(v);
            cG.count(v);
            cG.findIn(v, v);
          }
        }
      }
    }


}

namespace std
{
  /**
   * Specialize the swap algorithm for two graphs
   */
  template <GRAPH_TEMPLATE>
  void swap(graph::VVGraph<GRAPH_ARGS>& g1, graph::VVGraph<GRAPH_ARGS>& g2)
  {
    g1.swap(g2);
  }
}

#undef GRAPH_TEMPLATE
#undef GRAPH_ARGS

#endif // GRAPH_VVGRAPH_H

