#ifndef VVMODEL_H
#define VVMODEL_H

/**
 * \file model.h
 *
 * Defines the Model class.
 */

#include <config.h>
#include <string>
#include <set>
#include <QString>
#include <QStringList>
#include <QObject>
#include <QList>
#include <QMetaMethod>

#include <util/watchdog.h>

class Viewer;
class QPoint;
class QAction;

namespace storage
{
  class VVEStorage;
}

/**
 * \brief Simulation class
 *
 * This class has to be the base class of the main class of the project.
 *
 * It handles all steps specific to the simulation and the drawing of the 
 * result. Initialization of the model should occur in the constructor of the 
 * class. And any resource deallocation should occur in the destructor. An 
 * object of this class will be created at the beginning of each simulation, 
 * and destroyed at the end.
 *
 * The main Model object has to be registered using the DEFINE_MODEL(classname) 
 * macro command defined in <vvall.h>.
 */
class Model : public util::FileObject
{
  Q_OBJECT
public:

  //@{
  /// \name Constructor/destructor
  /**
   * Constructor of the object.
   *
   * \param parent is the object handling the life of this object. It is 
   * required to handle file registration.
   */
  Model(QObject* parent);

  /**
   * Virtual destructor ensure the destruction occurring is the one of the leaf 
   * class.
   */
  virtual ~Model() {}
  //@}

  //@{
  /// \name Methods to redefine to change the behavior of the model

  /**
   * A step correspond to all the computations between two drawings.
   *
   * It is the only strictly required method.
   */
  virtual void step() = 0;

  /**
   * \fn virtual void modifiedFiles( const std::set<std::string>& files )
   * This function is called anytime one or more registered files are modified.
   *
   * \param files Registered files modified since last call to this function.
   */
  virtual void modifiedFiles( const std::set<std::string>& /*files*/ )
    { this->modifiedFiles(); }
  /**
   * Convenience function to be used if the list of files is ignored.
   */
  virtual void modifiedFiles() {}

  /**
   * Reimplement the reread method to allow using the model without a parameter 
   * file.
   */
  virtual void reread() {}

  /**
   * This function is used to setup the help in the help box.
   *
   * It should typically return a description of the model. Default 
   * implementation will read the description.txt file if available.
   */
  virtual QString helpString() const;

  //@}

  //@{
  /// \name Snapshot methods

  /*
   * \returns True if serialization is supported, False otherwise.
   */
//  virtual bool serialize() { return false; }

  /**
   * Reimplement this method to load/save the current state of the model.
   *
   * \returns true if serialization was successful, false otherwise.
   *
   * \param store Object to store the state in the model in.
   */
  virtual bool serialize(storage::VVEStorage& /*store*/) { return false; }

  /**
   * \returns an integer representing the version string for the model.
   *
   * The default implementation always return 0.
   *
   * The number has no meaning outside the model.
   *
   */
  virtual int versionNumber(const QString& /*version*/) { return 0; }

  /**
   * \returns the current version of the class.
   *
   * The default implementation returns an empty string
   */
  virtual QString version() const { return ""; }

  //@}

  //@{
  ///\name Drawing and GUI interaction methods

  /**
   * \fn virtual void preDraw( Viewer* viewer )
   * This function should be redefined to setup the OpenGL context.
   *
   * Called at the beginning of every drawing, right after the initialization 
   * of the OpenGL context.
   *
   * \param viewer Viewer object. Useful to use the capabilities of QGLViewer 
   * object.
   */
  virtual void preDraw( Viewer* /*viewer*/ ) { this->preDraw(); }
  /**
   * Convenience function used if the viewer is not necessary.
   */
  virtual void preDraw() {}

  /**
   * \fn virtual void draw(Viewer *viewer)
   * This function should be redefined to draw the representation.
   *
   * \param viewer Viewer object. Useful to use the capabilities of QGLViewer 
   * object.
   */
  virtual void draw( Viewer* /*viewer*/) { this->draw(); }
  /**
   * Convenience function used if the viewer is not necessary.
   */
  virtual void draw() {}

  /**
   * \fn virtual void postDraw(Viewer* viewer)
   * This function should be redefined to restore the original OpenGL context.
   *
   * It is called right after the draw(Viewer*) method and should restore the 
   * OpenGL context to allow for correct drawing of the decorators (grid, axes, 
   * ...)
   *
   * \param viewer Viewer object. Useful to use the capabilities of QGLViewer 
   * object.
   */
  virtual void postDraw( Viewer* /*viewer*/ ) { this->postDraw(); }
  /**
   * Convenience function used if the viewer is not necessary.
   */
  virtual void postDraw() {}

  /**
   * \fn virtual void initDraw( Viewer* viewer )
   * This function is called everytime a new OpenGL environment is created, 
   * before the first preDraw(Viewer*).
   *
   * \warning It may be called multiple times if the OpenGL environment is 
   * destroyed and recreated.
   *
   * Useful to initialize camera and lights or to gather data about the 
   * rendering capabilities. Also used for any initialization requiring a valid 
   * OpenGL context.
   *
   * \param viewer Viewer object. Useful to use the capabilities of QGLViewer 
   * object.
   */
  virtual void initDraw( Viewer* /*viewer*/) { this->initDraw(); }
  /**
   * Convenience function used if the viewer is not necessary.
   */
  virtual void initDraw() {}

  /**
   * \fn virtual void finalizeDraw( Viewer* viewer )
   * This function is called if the OpenGL environment is destroyed or if the 
   * model is detroyed.
   *
   * \param viewer Viewer object. Useful to use the capabilities of QGLViewer 
   * object.
   */
  virtual void finalizeDraw( Viewer* /*viewer*/) { this->finalizeDraw(); }
  /**
   * Convenience function used if the viewer is not necessary.
   */
  virtual void finalizeDraw() {}

  /**
   * \fn virtual void drawWithNames(Viewer* viewer)
   *
   * Draw with object naming. Used for object selection. The name is to be 
   * setup using glPushName(int) / glPopName()
   *
   * \param viewer Viewer object. Useful to use the capabilities of QGLViewer 
   * object.
   */
  virtual void drawWithNames( Viewer * /*viewer*/ ) { this->drawWithNames(); }
  /**
   * Convenience function used if the viewer is not necessary.
   */
  virtual void drawWithNames() {}

  /**
   * This function is used to resolve a selection event.
   *
   * \param pos 2D position of the user click event
   * \param viewer Viewer object on which the selection occured
   *
   * It should implement the action associated with a selection.
   * The user can call \c viewer->selectedName() to get the id of the selected 
   * item. The number returned depend on the number allocated within the 
   * drawWithNames() method.
   */
  virtual void postSelection(const QPoint& pos, Viewer* viewer);

  /**
   * Returns the current animation period.
   */
  int animationPeriod();

  /**
   * Get the current status message.
   */
  QString statusMessage() const;

  /**
   * Change the status message.
   *
   * On the gui, display the message on the status bar. On batch mode, prints the message to the console.
   */
  void setStatusMessage(const QString& msg);

  /**
   * Get the list of actions used to create the context menu on the viewer
   */
  QList<QAction*> actions() const;

  /**
   * Returns the action representing the menu item \c text
   *
   * \returns 0 if no such item exist
   */
  QAction* menuItem(const QString& text) const;

  /**
   * Add a menu item to attach to the given slot
   *
   * \returns the pointer to the action
   */
  QAction* addMenuItem(const QString& text, QObject* receiver, const char* slot);

  /**
   * Add a menu item, the receiver being the model
   */
  QAction* addMenuItem(const QString& text, const char* slot);

  /**
   * Add a menu item, without connecting any signal
   */
  QAction* addMenuItem(const QString& text);

  /**
   * Remove a menu item
   *
   * \returns true is the menu item is found
   */
  bool removeMenuItem(const QString& text);

  /**
   * Remove the slot from a menu item, if there are more connection, the item
   * is not deleted
   *
   * \returns true if the menu item and the connection are found
   */
  bool removeMenuItem(const QString& text, QObject* receiver, const char* slot);

  /**
   * Disconnect a menu item, the receiver is the current model
   */
  bool removeMenuItem(const QString& text, const char* slot);

  /**
   * Add an action to the context menu
   */
  void addAction(QAction *action);

  /**
   * Add a list of actions to the context menu
   */
  void addActions(QList<QAction*> actions);

  /**
   * Insert an action in the context menu
   */
  void insertAction(QAction* before, QAction* action);

  /**
   * Insert a list of actions in the context menu
   */
  void insertActions(QAction* before, QList<QAction*> action);

  /**
   * Remove an action from the context menu
   */
  void removeAction(QAction* action);
  //@}

  //@{
  /// \name Textual output methods

  /**
   * This method is called once before the model is ran in batch mode.
   *
   * It is intended to initialize any variable or to print an initial message. 
   * Any output should be directed on the standard output of the application.
   */
  virtual void initPrint() {}

  /**
   * This function is called after each step of a model launched in batch 
   * mode.
   *
   * The output is expected to be on the standard output (i.e. stdout or 
   * cout).
   */
  virtual void print() {}

  /**
   * This method is called once when the model is about to be destroyed in 
   * batch mode.
   *
   * It is intended to initialize any variable or to print an initial message. 
   * Any output should be directed on the standard output of the application.
   */
  virtual void finalizePrint() {}

  //@}

  //@{
  ///\name Model execution interaction

  /**
   * This function requests the model to run
   */
  virtual void run();

  /**
   * This function requests the end of the simulation
   */
  virtual void stop();

  /**
   * This function requests a restart of the simulation
   */
  virtual void restart();
  //@}

  /**
   * Retrieve the arguments left for the model on the command line (i.e. 
   * arguments after the '--')
   */
  const QStringList& arguments() const { return _arguments; }


  QString errorString() const { return error_string; }

  //@{
  ///\name Methods for internal usage only
  void setError(bool error, QString err = QString()) { _error = error; error_string = err; }
  //@}

public slots:
  /**
   * This function registers a file to the file alteration monitor.
   */
  void registerFile( std::string filename );

  /**
   * This function unregisters a file from the file alteration monitor.
   */
  void unregisterFile( std::string filename );

  /**
   * Set the exit code. Useful for testing purposes.
   */
  void setExitCode(int code);

  /**
   * Set the animation period, i.e. the period in between two calls to the 
   * step() method, in ms.
   */
  void setAnimationPeriod(int ms);

  //@{
  ///\name Snapshots methods
  /**
   * Save the current model as a snapshot
   */
  bool saveSnapshot(const QString& filename);

  /**
   * Save the current model as a snapshot on the current recording.
   *
   * If the model is not being recorded, this method has no effect
   */
  bool saveNextSnapshot();

  /**
   * Load a snapshot specified by the user of the model.
   *
   * \Note be careful with the state of the model after this call
   */
  bool loadSnapshot();

  /**
   * Load a snapshot.
   *
   * \Note be careful with the state of the model after this call
   */
  bool loadSnapshot(const QString& filename);
  //@}

  //@{
  /// \name Screenshot methods
  /**
   * Save a screenshot
   *
   * \see QGLViewer::saveSnapshot(const QString&, bool)
   */
  void screenshot(const QString &fileName, bool overwrite=false);
  //@}

signals:
  //@{
  /// \name Signals for internal usage only
  /**
   * Signal emitted anytime the user is registering a file to the file checker.
   */
  void fileRegister( std::string );
  /**
   * Signal emitted anytime the user is unregistering a file ...
   */
  void fileUnregister( std::string );
  /**
   * Signal emitted when the model request the end of computations
   */
  void runModel();
  /**
   * Signal emitted when the model request the end of computations
   */
  void stopModel();
  /**
   * Signal emitted when the model request the end of computations
   */
  void restartModel();
  /**
   * Signal emitted when the user changes the exit code
   */
  void changedExitCode(int);

  /**
   * Signal emitted when an action is inserted
   */
  void menuActionInserted(QAction *before, QAction* act);

  /**
   * Signal emitted when an action is removed
   */
  void menuActionRemoved(QAction* act);

  /**
   * Signal emitted when the user wants to save a snapshot.
   *
   * \param filename Name of the snapshot. The extension will be used to 
   * determine the format.
   */
  void savingSnapshot(const QString& filename, Model*);

  /**
   * Signal emitted when the user want to save the next snapshot. Useful to 
   * save intermediate state ...
   */
  void savingNextSnapshot(bool, Model*);

  /**
   * Signal emitted when the user wants to load a snapshot.
   */
  void loadingSnapshot(bool auto_naming, Model*);

  /**
   * Signal emitted when the user wants to load a snapshot.
   *
   * \param filename Name of the snapshot. The extension will be used to 
   * determine the format.
   */
  void loadingSnapshot(const QString& filename, Model*);

  /**
   * Signal emitted when the user request the saving of a snapshot
   */
  void savingScreenshot(const QString& filename, bool overwrite);

  //@}

private:
  QStringList _arguments;
  QList<QAction*> _actions;
  bool _error;
  QString error_string;
};

#endif // VVMODEL_H

