#ifndef SEASHELL_MODEL_H
#define SEASHELL_MODEL_H

#include <vve.h>

#include <util/parms.h>
#include <util/palette.h>

#include <algorithms/solver.h>
#include <util/vector.h>
#include <util/random.h>
#include <cmath>
#include <string.h>
#include <time.h>
#include <vector>

#include <iostream>

#ifdef WIN32
#ifndef _MSC_VER
# include <GL/glext.h>
#endif
#endif

#ifndef VertexAttributes
#  define VertexAttributes
#endif

#ifndef EdgeAttributes
#  define EdgeAttributes
#endif

#ifndef ModelInit
#  define ModelInit
#endif

using util::ran;
typedef util::Palette::Color Color;

namespace seashell_model
{
  using std::cout;
  using std::cerr;
  using std::endl;

  void reportGLError( const char* id )
  {
    GLenum error = glGetError();
    if(error)
    {
      cerr << "OpenGL error on " << id << ":" << endl
        << gluErrorString(error) << endl;
    }
  }

  typedef util::Vector<2,double> Point2d;
  typedef util::Vector<3,double> Point3d;

  typedef util::Vector<NB_CHEMICALS,double> Chemicals;

  typedef solver::Solver<NB_CHEMICALS> RDSolver;

  typedef RDSolver::tag_t rd_tag_t;

  struct VertexContent
  {
    Chemicals chems, derivs;
    RDSolver::VertexInternals interns;
    int pos;       // Position of the vertex
    VertexAttributes
  };

  struct EdgeContent
  {
    RDSolver::EdgeInternals interns;
    EdgeAttributes
  };

  // Class defining your model
  // If you change the name, don't forget to change the last line of the file too
  class SeashellModel : public Model
  {
  public:
    typedef seashell_model::Chemicals Chemicals;
    typedef seashell_model::RDSolver RDSolver;
    typedef seashell_model::rd_tag_t rd_tag_t;

    // Type of the graph
    typedef VVGraph<VertexContent,EdgeContent> vvgraph;
    // Type of a vertex
    typedef vvgraph::vertex_t vertex;
    // Type of an edge
    typedef vvgraph::edge_t edge;

    util::Palette palette;
    vvgraph LiveCells;
    RDSolver solve;
    int nbRows;
    int currentTextureHeight;
    int currentTextureWidth;
    int textureChunk;
    int textureMaxSize;

    std::vector<GLuint> texId;

    GLubyte *texture;

    int nbCells;
    int nbTextures;
    std::vector<int> posTextures;
    int cellsPerTextures;
    double cellWidth;
    double cellHeight;
    double time;
    double timeDeath;
    double timePrint;
    double timeDraw;
    double printInterval;
    double drawInterval;

    double cellDeath;

    int colorLow;             // Color for low concentrations of activator
    int colorHigh;            // Color for high concentrations of activator
    double maxView;

    double background;

    // Read parameters
    int readNbCells;
    double readCellDeath;
    double readMaxView;
    int readTextureChunk;
    int randomSeed;

    bool stop;
    bool circle;

    void readParms()
    {
      util::Parms parms("view.v");

      parms("Main", "PrintInterval", printInterval);
      parms("Main", "DrawInterval", drawInterval);
      parms("Main", "RandomSeed", randomSeed);

      solve.readParms(parms, "Solver");

      // Read seashell parameters
      parms("Seashell", "NbCells", readNbCells);
      parms("Seashell", "CellWidth", cellWidth);
      parms("Seashell", "CellHeight", cellHeight);
      parms("Seashell", "CellDeath", readCellDeath);
      parms("Seashell", "Circle", circle);

      // View parameters
      parms("View", "ColorLow", colorLow);
      parms("View", "ColorHigh", colorHigh);
      parms("View", "MaxView", readMaxView);
      parms("View", "TextureChunk", readTextureChunk);
      parms("View", "Background", background);

      readParam( parms );

      if(cellDeath > 0)
        solve.MaxDt = std::min(std::min(solve.MaxDt, cellDeath), drawInterval);
      else
        solve.MaxDt = std::min(std::min(solve.MaxDt, readCellDeath), drawInterval);
    }

    // Here, reread the files when they are modified
    void modifiedFiles( const std::set<std::string>& filenames )
    {
      forall(const std::string& fn, filenames)
      {
        if(fn == "pal.map")
          palette.reread();
        else if(fn == "view.v")
          readParms();
      }
    }

    ~SeashellModel()
    {
      if(!texId.empty())
        glDeleteTextures(nbTextures, &texId[0]);
      if(texture)
        delete texture;
    }

    virtual void initCell(vertex v) = 0;

    SeashellModel(QObject *parent)
      : Model(parent)
      , palette("pal.map")
      , nbRows(0)
      , currentTextureHeight(0)
      , currentTextureWidth(0)
      , textureChunk(0)
      , textureMaxSize(0)
      , texture(0)
      , nbCells(0)
      , cellDeath(0)
      , stop(false)
      { }

    void init()
    {
      readParms();
      // Registering the configuration files
      registerFile("pal.map");
      registerFile("view.v");

      if(randomSeed == 0)
      {
        util::sran(::time(NULL));
      }
      else
      {
        util::sran(randomSeed);
      }

      nbCells = readNbCells;
      cellDeath = readCellDeath;
      maxView = readMaxView;
      textureChunk = readTextureChunk;

      solve.initialize();

      // Initialize your structure here
      time = 0;
      timeDeath = 0;
      timePrint = printInterval;
      timeDraw = 0;

      vertex pv(0);
      vertex fv(0);
      for(int i = 0 ; i < nbCells ; ++i)
      {
        vertex v;
        v->pos = i;
        LiveCells.insert(v);
        if(!pv.isNull())
        {
          LiveCells.insertEdge(v, pv);
          LiveCells.insertEdge(pv, v);
        }
        else
        {
          fv = v;
        }
        initCell(v);
        pv = v;
      }
      if(circle)
      {
        LiveCells.insertEdge(pv, fv);
        LiveCells.insertEdge(fv, pv);
      }
    }

    void createTexture()
    {
      // Setting up the new texture buffer
      int new_height = currentTextureHeight + textureChunk;
      // First, test I can create the texture
      glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGBA, currentTextureWidth, new_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
      reportGLError("glTexImage2D(GL_PROXY_TEXTURE_2D)");
      GLint success;
      glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &success);
      if(!success)
      {
        cerr << "Texture too big, stopping model" << endl;
        stop = true;
        return;
      }
      GLubyte *new_texture = new GLubyte[nbCells*new_height*4];
      memset(new_texture, 0, nbCells*new_height*4);
      if(texture)
      {
        memcpy(new_texture, texture, nbCells*currentTextureHeight*4);
        delete [] texture;
      }
      texture = new_texture;
      currentTextureHeight = new_height;

      // Setting up the new texture
      if(texId.empty())
      {
        texId.resize(nbTextures);
        glGenTextures(nbTextures, &texId[0]);
        reportGLError("glGenTextures");
      }
      for(int i = 0 ; i < nbTextures ; ++i)
      {
        glBindTexture(GL_TEXTURE_2D, texId[i]);
        reportGLError("glBindTexture(texId)");
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        reportGLError("glPixelStorei(GL_UNPACK_ALIGNMENT)");
        glPixelStorei(GL_UNPACK_ROW_LENGTH, nbCells);
        reportGLError("glPixelStorei(GL_UNPACK_ROW_LENGTH)");
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        reportGLError("glTexParameteri(GL_TEXTURE_MAG_FILTER)");
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        reportGLError("glTexParameteri(GL_TEXTURE_MIN_FILTER)");
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        reportGLError("glTexParameteri(GL_TEXTURE_WRAP_S)");
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        reportGLError("glTexParameteri(GL_TEXTURE_WRAP_T)");
        //    cout << "Allocating texture: " << texId << " " << currentTextureWidth << "x" << currentTextureHeight << ": " << (int*)texture << endl;
        //    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, currentTextureWidth, currentTextureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
        //    reportGLError("glTexImage2D");
      }
      glBindTexture(GL_TEXTURE_2D, 0);
      reportGLError("glBindTexture(0)");
    }

    virtual Color getColor(const Chemicals& chems)
    {
      return palette.getColor(std::max(0, int(255*chems[0] / maxView)));
    }

    void drawCellsToTexture(int pos)
    {
      if(pos >= currentTextureHeight)
      {
        createTexture();
        if(stop)
          return;
      }
      int offset = nbCells*pos*4;
      GLubyte *texture_line = texture + offset;
      forall(const vertex& v, LiveCells)
      {
        //      Color c = palette.getColor(colorLow, colorHigh, v->chems[0] / maxView);
        Color c = getColor(v->chems);
        memcpy(texture_line+4*v->pos, c.c_data(), 4);
      }
      for(int i = 0 ; i < nbTextures ; ++i)
      {
        glBindTexture(GL_TEXTURE_2D, texId[i]);
        offset = posTextures[i]*4;
        reportGLError("glBindTexture(texId)");
        int width = currentTextureWidth;
        if(i == nbTextures-1)
        {
          width = nbCells - posTextures[i];
        }
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, currentTextureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture+offset);
        reportGLError("glTexImage2D(GL_TEXTURE_2D)");
      }
      glBindTexture(GL_TEXTURE_2D, 0);
      reportGLError("glBindTexture(0)");
    }

    void step()
    {
      while(timeDraw < drawInterval)
      {
        if(stop)
          return;
        solve(LiveCells, *this);
        time += solve.dt;
        timeDeath += solve.dt;
        timePrint += solve.dt;
        timeDraw += solve.dt;
        if(timeDeath > cellDeath)
        {
          if(nbRows > textureMaxSize)
          {
            stop = true;
            return;
          }
          timeDeath -= cellDeath;
          drawCellsToTexture(nbRows);
          if(stop)
            return;
          ++nbRows;
        }
        if(timePrint >= printInterval)
        {
          timePrint -= printInterval;
          Chemicals minC, maxC;
          for(int i = 0 ; i < NB_CHEMICALS ; ++i)
          {
            minC[i] = HUGE_VAL;
            maxC[i] = -HUGE_VAL;
          }
          forall(const vertex& c, LiveCells)
          {
            for(int i = 0 ; i < NB_CHEMICALS ; ++i)
            {
              if(minC[i] > c->chems[i])
                minC[i] = c->chems[i];
              if(maxC[i] < c->chems[i])
                maxC[i] = c->chems[i];
            }
          }
          cout << "Time: " << time << " - Current dt: " << solve.dt << endl;
          for(int i = 0 ; i < NB_CHEMICALS ; ++i)
          {
            cout << "   Chemical " << i << ": " << minC[i] << " - " << maxC[i] << endl;
          }
        }
      }
      timeDraw -= drawInterval;
    }

    void updateTextureSizes()
    {
      if(nbCells > textureMaxSize)
      {
        //      cerr << "Error, too many cells to render the result as a texture. Setting the number of cells to " << textureMaxSize << endl;
        nbTextures = nbCells / (textureMaxSize-2) + 1;
        cellsPerTextures = nbCells / nbTextures + 2;
        posTextures.resize(nbTextures);
        posTextures[0] = 0;
//        cout << "We have " << nbTextures << " textures for " << nbCells << " cells." << endl;
//        cout << "Size of textures: " << cellsPerTextures << endl;
        for(int i = 1 ; i < nbTextures ; ++i)
        {
          posTextures[i] = posTextures[i-1] + cellsPerTextures - 2;
//          cout << "Position of texture " << i << ": " << posTextures[i] << endl;
        }
      }
      else
      {
        nbTextures = 1;
        cellsPerTextures = nbCells;
        posTextures.resize(1);
        posTextures[0] = 0;
      }
      currentTextureWidth = cellsPerTextures;
    }

    void initDraw( Viewer* viewer )
    {
      glEnable(GL_TEXTURE_2D);
      reportGLError("glEnable(GL_TEXTURE_2D)");
      glGetIntegerv(GL_MAX_TEXTURE_SIZE, &textureMaxSize);
      reportGLError("glGetIntegerv(GL_MAX_TEXTURE_SIZE)");
      updateTextureSizes();
      viewer->camera()->setType(Camera::ORTHOGRAPHIC);
      glClearColor(background, background, background, 0.0);
      if(!texId.empty())
      {
        glDeleteTextures(nbTextures, &texId[0]);
        texId.clear();
      }
      if(texture)
      {
        delete [] texture;
        texture = 0;
        currentTextureHeight = 0;
      }
    }

    void preDraw()
    {
      glClearColor(background, background, background, 0.0);
    }

    void postDraw()
    {
    }

    void draw(Viewer* viewer)
    {
      double width = nbCells*cellWidth;
      Vec vmin(0,0,-width/10), vmax(width, (nbRows+1)*cellHeight, width/10);
      viewer->setSceneBoundingBox(vmin, vmax);
      //if(time == 0)
      //  return;
      if(!stop)
        drawCellsToTexture(nbRows);
      for(int i = 0 ; i < nbTextures ; ++i)
      {
        double texWidth = currentTextureWidth*cellWidth;
        if(i == nbTextures-1)
        {
          texWidth = (nbCells-posTextures[i])*cellWidth;
        }
        double left = posTextures[i]*cellWidth;
        double right = left+texWidth;
        double left_tex = 1/texWidth;
        double right_tex = 1-1/texWidth;
        if(i == 0)
        {
          left_tex = 0;
        }
        else
        {
          left += cellWidth;
        }
        if(i == nbTextures-1)
        {
          right_tex=1;
        }
        else
        {
          right -= cellWidth;
        }
        glBindTexture(GL_TEXTURE_2D, texId[i]);
        reportGLError("glBindTexture(texId)");
        glBegin(GL_QUADS);
        glTexCoord2f(left_tex,0);
        glVertex3f(left,0,0);
        glTexCoord2f(right_tex,0);
        glVertex3f(right, 0,0);
        glTexCoord2f(right_tex,1);
        glVertex3f(right, currentTextureHeight*cellHeight,0);
        glTexCoord2f(left_tex,1);
        glVertex3f(left,currentTextureHeight*cellHeight,0);
        glEnd();
      }
      glBindTexture(GL_TEXTURE_2D, 0);
      reportGLError("glBindTexture(0)");
      if(stop)
        viewer->stopAnimation();
    }

    // Methods for the solver
    Chemicals& values(const vertex& v, const rd_tag_t&)
    {
      return v->chems;
    }

    Chemicals& derivatives(const vertex& v, const rd_tag_t&)
    {
      return v->derivs;
    }

    RDSolver::VertexInternals& vertexInternals(const vertex& v, const rd_tag_t&)
    {
      return v->interns;
    }

    RDSolver::EdgeInternals& edgeInternals(const vertex& src, const vertex& tgt, vvgraph& S, const rd_tag_t&)
    {
      return S.edge(src, tgt)->interns;
    }

    virtual void updateDerivatives(const vertex& v, const rd_tag_t&) = 0;

    virtual void initialize() {}

    virtual void readParam( util::Parms& ) {}
  };

}

#include <QContextMenuEvent>
#include <QAction>
#include <QFileDialog>
#include <QMenu>
#include <QImageWriter>

namespace seashell_model
{
  class SeashellViewer : public Viewer
  {
    Q_OBJECT
  public:
      SeashellViewer(QWidget* parent)
        : Viewer(parent)
        {
          saveTextureAction = new QAction("Save texture", this);
          connect(saveTextureAction, SIGNAL(activated()), SLOT(saveTexture()));
        }
      public slots:
        void saveTexture()
        {
          SeashellModel *m = dynamic_cast<SeashellModel*>(model());
          QList<QByteArray> sf = QImageWriter::supportedImageFormats();
          QStringList supportedFormats;
          QStringList formats;
          forall(const QByteArray& f, sf)
          {
            QString qf = QString::fromLocal8Bit(f);
            supportedFormats += qf;
          }
          formats += "All images ( *." + supportedFormats.join(" *.") + " )";
          forall(const QString& f, supportedFormats)
          {
            formats += f.toUpper() + " images ( *." + f + ")";
          }
          formats += "All files (*.*)";
          QString filters = formats.join(";;");
          QString filename = QFileDialog::getSaveFileName(this, "Saving texture image", QString(), filters);
          if(!filename.isEmpty())
          {
            QImage img(m->texture, m->nbCells, m->nbRows+1, QImage::Format_RGB32);
            img = img.rgbSwapped();
            QImageWriter writer(filename);
            writer.write(img);
          }
        }

  protected:

      void contextMenuEvent( QContextMenuEvent* event )
      {
        QMenu *popup = new QMenu(this);
        popup->addAction(saveTextureAction);
        popup->popup(event->globalPos());
      }

      QAction *saveTextureAction;
  };

}

#include <seashell_model.moc>

DEFINE_VIEWER(seashell_model::SeashellViewer);

#define StartModel \
  struct ModelClass : public seashell_model::SeashellModel \
  { \
    typedef seashell_model::SeashellModel ParentClass; \
    ModelClass(QObject *parent) \
      : ParentClass(parent) \
      ModelInit \
    { \
      init(); \
      initialize(); \
    }

#define EndModel }; DEFINE_MODEL(ModelClass);

#endif // SEASHELL_MODEL_H
