#ifndef VVELIB_STORAGE_GRAPH_STORAGE_H
#define VVELIB_STORAGE_GRAPH_STORAGE_H

/**
 * \file storage/graph.h
 */

#include <config.h>
#include <storage/storage.h>
#include <graph/vvgraph.h>
#include <graph/vvbigraph.h>
#include <iostream>

namespace graph
{
  template <typename T>
  Vertex<T> create_object(Vertex<T> const& v)
  {
    return v;
  }

  template <typename T>
  Edge<T> create_object(Edge<T> const& e)
  {
    return e;
  }

  template <typename T>
  bool serialization(storage::VVEStorage& store, Vertex<T>& v)
  {
    return store.reference("", "Vertex", v);
  }

  template <typename T>
  bool serialization(storage::VVEStorage& store, Edge<T>& e)
  {
    return store.reference("", "Edge", e);
  }

  template <typename V, typename E, bool compact>
  bool serialization(storage::VVEStorage& store, VVGraph<V,E,compact>& G)
  {
    typedef typename VVGraph<V,E,compact>::vertex_t vertex_t;
    typedef typename VVGraph<V,E,compact>::edge_t edge_t;
    if(store.writer())
    {
      forall(const vertex_t& v, G)
      {
        vertex_t v1 = v;
        if(!store.field("Vertex", v1))
        {
          return false;
        }
      }
      forall(const vertex_t& v, G)
      {
        if(!store.startCompound("Neighborhood"))
          return false;
        vertex_t v1 = v;
        if(!store.field("Source", v1))
        {
          return false;
        }
        forall(const vertex_t& n, G.neighbors(v))
        {
          vertex_t n1 = n;
          if(!store.field("Neighbor", n1))
          {
            return false;
          }
          edge_t e = G.edge(v,n);
          if(!store.field("Edge", e))
          {
            return false;
          }
        }
        store.endCompound();
      }
    }
    else if(store.reader())
    {
      G.clear();
      while(store.checkNextField("Vertex"))
      {
        vertex_t v;
        if(!store.field("Vertex", v))
        {
          return false;
        }
        G.insert(v);
      }
      for(size_t i = 0 ; i < G.size() ; ++i)
      {
        if(!store.startCompound("Neighborhood"))
          return false;
        vertex_t v1;
        if(!store.field("Source", v1))
        {
          return false;
        }
        const vertex_t& v = G.reference(v1);
        vertex_t prev(0);
        while(store.checkNextField("Neighbor"))
        {
          vertex_t n;
          if(!store.field("Neighbor", n))
          {
            return false;
          }
          edge_t e;
          if(prev)
            e = G.spliceAfter(v,prev,n);
          else
            e = G.insertEdge(v,n);
          if(!e)
          {
            store.setLastError(10, "Error creating an edge");
            return false;
          }
          prev = n;
          if(!store.field("Edge", e))
          {
            return false;
          }
        }
        store.endCompound();
      }
    }
    return true;
  }

  template <typename V1, typename V2, typename E1, typename E2, bool compact>
  bool serialization(storage::VVEStorage& store, VVBiGraph<V1,V2,E1,E2,compact>& G)
  {
    typedef typename VVBiGraph<V1,V2,E1,E2,compact>::vertex1_t vertex1_t;
    typedef typename VVBiGraph<V1,V2,E1,E2,compact>::edge1_t edge1_t;
    typedef typename VVBiGraph<V1,V2,E1,E2,compact>::vertex2_t vertex2_t;
    typedef typename VVBiGraph<V1,V2,E1,E2,compact>::edge2_t edge2_t;
    if(store.writer())
    {
      forall_named(const vertex1_t& v, G, vertex1)
      {
        vertex1_t v1 = v;
        if(!store.field("Vertex1", v1))
        {
          return false;
        }
      }
      forall_named(const vertex2_t& v, G, vertex2)
      {
        vertex2_t v1 = v;
        if(!store.field("Vertex2", v1))
        {
          return false;
        }
      }
      forall_named(const vertex1_t& v, G, vertex1)
      {
        if(!store.startCompound("Neighborhood1"))
          return false;
        vertex1_t v1 = v;
        if(!store.field("Source", v1))
        {
          return false;
        }
        forall(const vertex2_t& n, G.neighbors(v))
        {
          vertex2_t n1 = n;
          if(!store.field("Neighbor", n1))
          {
            return false;
          }
          edge1_t e = G.edge(v,n);
          if(!store.field("Edge", e))
          {
            return false;
          }
        }
        store.endCompound();
      }
      forall_named(const vertex2_t& v, G, vertex2)
      {
        if(!store.startCompound("Neighborhood2"))
          return false;
        vertex2_t v1 = v;
        if(!store.field("Source", v1))
        {
          return false;
        }
        forall(const vertex1_t& n, G.neighbors(v))
        {
          vertex1_t n1 = n;
          if(!store.field("Neighbor", n1))
          {
            return false;
          }
          edge2_t e = G.edge(v,n);
          if(!store.field("Edge", e))
          {
            return false;
          }
        }
        store.endCompound();
      }
    }
    else if(store.reader())
    {
      G.clear();
      while(store.checkNextField("Vertex1"))
      {
        vertex1_t v;
        if(!store.field("Vertex1", v))
        {
          return false;
        }
        G.insert(v);
      }
      while(store.checkNextField("Vertex2"))
      {
        vertex2_t v;
        if(!store.field("Vertex2", v))
        {
          return false;
        }
        G.insert(v);
      }
      for(size_t i = 0 ; i < G.nb_vertices1() ; ++i)
      {
        if(!store.startCompound("Neighborhood1"))
          return false;
        vertex1_t v1;
        if(!store.field("Source", v1))
        {
          return false;
        }
        const vertex1_t& v = G.reference(v1);
        vertex2_t prev(0);
        while(store.checkNextField("Neighbor"))
        {
          vertex2_t n;
          if(!store.field("Neighbor", n))
          {
            return false;
          }
          edge1_t e;
          if(prev)
            e = G.spliceAfter(v,prev,n);
          else
            e = G.insertEdge(v,n);
          if(!e)
          {
            store.setLastError(10, "Error creating an edge");
            return false;
          }
          prev = n;
          if(!store.field("Edge", e))
          {
            return false;
          }
        }
        store.endCompound();
      }
      for(size_t i = 0 ; i < G.nb_vertices2() ; ++i)
      {
        if(!store.startCompound("Neighborhood2"))
          return false;
        vertex2_t v1;
        if(!store.field("Source", v1))
        {
          return false;
        }
        const vertex2_t& v = G.reference(v1);
        vertex1_t prev(0);
        while(store.checkNextField("Neighbor"))
        {
          vertex1_t n;
          if(!store.field("Neighbor", n))
          {
            return false;
          }
          edge2_t e;
          if(prev)
            e = G.spliceAfter(v,prev,n);
          else
            e = G.insertEdge(v,n);
          if(!e)
          {
            store.setLastError(10, "Error creating an edge");
            return false;
          }
          prev = n;
          if(!store.field("Edge", e))
          {
            return false;
          }
        }
        store.endCompound();
      }
    }
    return true;
  }
};

#endif // VVELIB_STORAGE_GRAPH_STORAGE_H

