#ifndef VVELIB_STORAGE_STORAGE_H
#define VVELIB_STORAGE_STORAGE_H
/**
 * \file storage.h
 *
 * This file contains the base class and functions to handle storage.
 */

#include <config.h>
#include <QString>
#include <cstddef>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <set>
#include <util/unorderedmap.h>

#ifdef USE_HASH
#include <QHash>

namespace std
{
#ifdef CXX_NEEDS_TR1
  namespace tr1
  {
#endif
    template <>
    struct hash<QString>
    {
      size_t operator()(const QString& s) const
      {
        return qHash(s);
      }
    };
#ifdef CXX_NEEDS_TR1
  }
#endif
}
#endif

class Model;

/**
 * \namespace storage
 *
 * Namespace containing all functions and classes related to VVE persistence
 *
 * \section Principle
 * The #storage works using the storage::VVEStorage objects. This abstract 
 * class defines the generic behavior of a serialized #storage for VVE. The 
 * implementations are created by the interpreter and given to the model. The 
 * #storage implement an interface that is the same for reading and writing.
 *
 * \section def_seri Defining the serialization
 * To allow for serialization the model should implement at least one method:
 * \code
 * bool serialize(storage::VVEStorage& store);
 * \endcode
 *
 * This method actually implements the serialization of the model. It 
 * should returns true if the serialization worked, false otherwise. To test for
 * the existence of this method, the interpreter will call it using a dummy
 * storage object, simulating a storage writer always succeeding. If the result
 * of this dummy call is true, then the object is recognised as being able to
 * serialize data.
 *
 * For every object, other than the model itself, you want to serialize, you can
 * define its serialization in two ways:
 * -# via a method in the object:
 * \code
 * bool serialize(storage::VVEStorage& store);
 * \endcode
 * -# via an external function:
 * \code
 * bool serialization(storage::VVEStorage& store, MyClass& object);
 * \endcode
 *
 * The first way is more intrusive but has the advantage to be inherited. The 
 * second way is non-intrusive but won't be inherited by sub-classes. In case 
 * the \c serialization method does not exist, the \c serialize method will be 
 * looked for.
 *
 * \subsection version File versionning
 * Optionnally, the model can define a version string. To do so, the class 
 * should define the method:
 * \code
 * QString version() const;
 * \endcode
 * The method will be called when saving a file.
 *
 * Once the version is obtained (either from a file or by calling the \c 
 * version method), the #storage object will call the method:
 * \code
 * int versionNumber(const QString& version);
 * \endcode
 * the version given is the model or file version prepended with the file 
 * format and a hyphen. For example, if the version of a model is \c 1.2.3 and 
 * the current file is an XML one, the version will be: \c XML-1.2.3
 *
 * The number returned by the \c versionNumber() method is stored by the 
 * #storage object and can be retrieved using the VVEStorage::version() method. 
 * As a helper, you can use the function storage::versionNumber(const QString&).
 *
 * \section using_store Using the VVEStorage object
 *
 * The VVEStorage object allows you to define your data structure as 
 * a hierarchical set of named fields. As the system is a serialization one, 
 * fields should always be read and written in the same order. The #storage 
 * object is not a random access object but a serial one.
 *
 * To help you structure your data, you can create three kind of elements:
 * -# Fields
 *  This is the basic type. If you want to store an element by value, you use 
 *  a field. All the basic C++ types and strings are handled directly by the 
 *  #storage object. Any other type must have either a \c serialization 
 *  function associated or a \c serialize method. If a field contains an 
 *  object, a compound will actually be created. A field is created by the 
 *  method:
 *  \code
 *  template <typename T>
 *  bool VVEStorage::field(const QString& name, T& value);
 *  \endcode
 *  The name must not be more than 255 characters but can be empty. An empty 
 *  field is valid only as sole member of a compound. In this case, the 
 *  #storage object can decide to simply create a field reusing the name of the 
 *  compound (this is actually done both for the XML and BIN implementation). 
 *  A typical exemple would be for a class with only one member:
 *  \code
 *  struct MyStructure
 *  {
 *    int value;
 *    bool serialize(storage::VVEStorage& store) {
 *      return store.field("", value);
 *    }
 *  };
 *  \endcode
 * -# References
 *  References are similar to fields but are used for pointers. The element 
 *  stored should be an actual pointer or a smart pointer. A reference is 
 *  created with the method:
 *  \code
 *  template <typename Ptr>
 *  bool VVEStorage::reference(const QString& name, const QString& ref_type, Ptr& p);
 *  \endcode
 *  The point is using a reference is that:
 *    -# if the same object is saved twice, it will be only once in the file
 *    -# if two pointers share the same object when writing, they will be 
 *    reconstructed as pointing to the same object.
 *    .
 *  A typical example in VVE is the vertices. If you store a vertex, the \c 
 *  serialization function defined in \c <storage/graph.h> is defined as:
 *  \code
 *  bool serialization(storage::VVEStorage& store, vertex& v) {
 *    return store.reference("", "Vertex", v);
 *  }
 *  \endcode
 *  The empty name tells the #storage object that it can reuse the compound name 
 *  (and actually not have a compound but just a reference). Then, the type is 
 *  usefull mostly for user-editable files where the reference ids are unique 
 *  per type. Then, if you use the same vertex twice, only one instance will be 
 *  stored and recreated later. And the two vertices will still share the same 
 *  data.
 *  Once a reference is identified, a namesless field is created to hold the 
 *  content of the reference.
 * -# Compounds
 *  Compounds are created using the two methods:
 *  \code
 *  bool VVEStorage::startCompound(const QString& name);
 *  bool VVEStorage::endCompound();
 *  \endcode
 *  This will group all element defined in between the two. A compound should 
 *  always be closed. As for references and fields, a compound can be nameless 
 *  and can then be optimized. In the end, a compound is rarely useful unless 
 *  you have a very complicated structure. It is used for example in the VVGraph 
 *  serialization. And of course, a compound is implicitely created for every 
 *  composed object.
 */
namespace storage
{

  /**
   * Common options for reader/writer
   */
  enum Options
  {
    /// Change the way type is checked while extracting a structure
    TypeChecking
  };

  /**
   * Enumeration describing how the type of the elements should be handled
   */
  enum TypeCheckingOption
  {
    /// The type asked for must be the same as the one read
    TCO_Exact,
    /**
     * The conversion should be loss-less.
     *
     * For example, you'll be able to convert from short to int but not the 
     * other way around
     */
    TCO_Strict,
    /**
     * The conversion can induce loss
     *
     * However, any loss of precision conversion will lead to a warning in the 
     * starndard output.
     */
    TCO_Permissive,
    /**
     * Same as Permissive, but without the warnings
     */
    TCO_PermissiveNoWarning,
    /**
     * No check is performed. An error will be raised only if the string cannot 
     * be interpreted as the type claimed.
     */
    TCO_NoCheck
  };

  /**
   * Enumeration describing the basic errors.
   */
  enum STORAGE_ERROR
  {
    /** No error occured
     *
     * The wierd name is due to the definition of NO_ERROR and NOERROR by 
     * windows already. Any complain should be sent to Microsoft.
     */
    NO__ERROR = 0,
    /**
     * Error occuring during an IO access.
     *
     * Typically while reading or writing a file.
     */
    IO_ERROR,
    /**
     * Bad content error.
     *
     * This mean the format of the file is not correct. For example, an XML 
     * file does not respect the XML format.
     */
    BAD_CONTENT,
    /**
     * Type error.
     *
     * It can occur either if the typing is strict and the type is different in 
     * the file and in the model or if the typing is permissive but the 
     * conversion does not exist.
     */
    TYPE_CHECK_ERROR,
    /**
     * Type conversion error.
     *
     * It can occur if the type conversion should be possible but the content 
     * is invalid. The typical case is the conversion from string to number if 
     * the string doesn't contain a number.
     */
    TYPE_CONVERSION_ERROR,
    /**
     * The field requested doesn't exist.
     */
    NO_FIELD_ERROR,
    /**
     * Error while using a reference.
     *
     * Either the reference does not exist or the user try to access 
     * a non-reference as reference.
     */
    REFERENCE_ERROR,
    /**
     * Error of the user.
     *
     * The succession of commands of the user are invalid.
     */
    USER_ERROR,
    /**
     * Error set if the serialize method fails without an error being set
     */
    UNKNOWN_ERROR
  };

  template <typename T>
  struct pointer_traits;

  template <typename T>
  struct pointer_traits<T*>
  {
    typedef T pointee;
  };

  template <typename Ptr>
  Ptr create_object(Ptr const& p)
  {
    if(p != 0)
      delete p;
    typedef typename pointer_traits<Ptr>::pointee pointee;
    return new pointee();
  }

  struct Reference
  {
    virtual ~Reference() {}
    Reference()
      : count(1)
    {}

    int count;
  };

  template <typename Ptr>
  struct RealReference : public Reference
  {
    RealReference(Ptr p=Ptr())
      : Reference()
      , value(p)
    { }

    Ptr value;
  };

  struct ReferencePtr
  {
    ReferencePtr(Reference* r = 0)
      : ptr(r)
    {
    }

    ReferencePtr(const ReferencePtr& copy)
      : ptr(copy.ptr)
    {
      if(ptr)
        ptr->count++;
    }

    void release()
    {
      if(ptr)
      {
        ptr->count--;
        if(ptr->count == 0)
        {
          delete ptr;
          ptr = 0;
        }
      }
    }

    ~ReferencePtr()
    {
      release();
    }

    bool operator==(const ReferencePtr& other) const
    {
      return ptr == other.ptr;
    }

    ReferencePtr& operator=(const ReferencePtr& other)
    {
      if(other.ptr == ptr)
        return *this;
      release();
      ptr = other.ptr;
      if(ptr)
        ptr->count++;
      return *this;
    }

    Reference *ptr;
  };

  template <typename Ptr>
  ReferencePtr make_reference(Ptr& p)
  {
    return ReferencePtr(new RealReference<Ptr>(p));
  }

  /**
   * \class VVEStorage storage.h <storage/storage.h>
   * Abstract base class defining the interactions with the persistence
   * module.
   */
  class VVEStorage
  {
  public:
    /// Type of a reference identifier
    typedef ptrdiff_t reference_t;

#ifdef USE_HASH
    typedef std::unordered_map<size_t, ReferencePtr> ref_map;
    typedef std::unordered_map<QString, ref_map> ref_map_type;
#else
    typedef std::map<size_t, ReferencePtr> ref_map;
    typedef std::map<QString, ref_map> ref_map_type;
#endif

    ref_map_type references;

    VVEStorage();
    /// Pure virtual destructor to allow for inheritance
    virtual ~VVEStorage() {}

    /**
     * Change an option.
     *
     * The options are implementation-dependent. If the option is not 
     * supported, the function returns 0.
     */
    virtual bool setOption(int , int ) { return false; }

    /**
     * Start a new compound. Useful to organize complicated data structure
     *
     * When writing, always returns true. When reading, returns true only if
     * the compound exists.
     */
    virtual bool startCompound(const QString& name) = 0;
    /**
     * End the current compound.
     */
    virtual bool endCompound() = 0;

    /**
     * Instruct a reader to ignore the next field if called \c name.
     *
     * This method is useful only for truly serial reader (like the BIN one)
     */
    virtual bool ignore(const QString& /*name*/ ) { return true; }

    /**
     * Start a reference.
     *
     * All operations preceding the call to stopReference should be stored in
     * a section reserved for references.
     *
     * References should be loaded only on need.
     *
     * When writing, always returns true. When reading, returns true only if
     * the reference exists. If the reference does not exist, it will create
     * it, so the reader has to set the fields.
     *
     * \returns the reference identifier or -1 if no reference is available.
     */
    virtual int startReference(const QString& name, const QString& ref_type, reference_t ref) = 0;
    /**
     * Stop the current reference
     */
    virtual bool endReference() = 0;

    /**
     * Clear the internal structures of the object
     */
    virtual void clear();

    /**
     * Load or save depending on the object.
     */
    virtual bool serialize(const QString& filename, Model* model) = 0;

    /**
     * \returns True is the object is a storage reader
     */
    virtual bool reader() = 0;
    /**
     * \returns True is the object is a storage writer
     */
    virtual bool writer() = 0;

    //@{
    /// \name Fields with standard types

    /// Create a boolean field
    virtual bool field(const QString& name, bool& value) = 0;
#ifndef _MSC_VER
    /// Create a wchar_t field
    virtual bool field(const QString& name, wchar_t& value) = 0;
#endif
    /// Create a char field
    virtual bool field(const QString& name, char& value) = 0;
    /// Create a signed char field
    virtual bool field(const QString& name, signed char& value) = 0;
    /// Create a short integer field
    virtual bool field(const QString& name, short& value) = 0;
    /// Create an integer field
    virtual bool field(const QString& name, int& value) = 0;
    /// Create a long integer field
    virtual bool field(const QString& name, long& value) = 0;
    /// Create a long long integer field
    virtual bool field(const QString& name, long long& value) = 0;
    /// Create a unsigned char field
    virtual bool field(const QString& name, unsigned char& value) = 0;
    /// Create a unsigned short integer field
    virtual bool field(const QString& name, unsigned short& value) = 0;
    /// Create an unsigned integer field
    virtual bool field(const QString& name, unsigned int& value) = 0;
    /// Create a unsigned long integer field
    virtual bool field(const QString& name, unsigned long& value) = 0;
    /// Create a unsigned long long integer field
    virtual bool field(const QString& name, unsigned long long& value) = 0;
    /// Create a float field
    virtual bool field(const QString& name, float& value) = 0;
    /// Create a double field
    virtual bool field(const QString& name, double& value) = 0;
    /// Create a long double field
    virtual bool field(const QString& name, long double& value) = 0;
    /// Create a QString field
    virtual bool field(const QString& name, QString& value) = 0;
    /// Create a std::string field
    virtual bool field(const QString& name, std::string& value) = 0;
    //@}

    /**
     * Create a new field.
     *
     * \param name Name of the field
     * \param value Variable to associate
     *
     * \returns Always true when writing, or False if reading and the field
     * does not exist
     */
    template <typename T>
    bool field(const QString& name, T& value)
    {
      bool result = false;
      if(startCompound(name))
      {
        result = serialization(*this, value);
        if(!result)
          return false;
        result &= endCompound();
      }
      return result;
    }

    /**
     * Create a new field, using the model to serialize it.
     *
     * \param name Name of the field
     * \param value Variable to associate
     * \param model Model used for serialization
     *
     * \returns Always true when writing, or False if reading and the field
     * does not exist
     */
    template <typename T, typename Model>
    bool field(const QString& name, T& value, Model* model)
    {
      bool result = false;
      bool no_name = name.isEmpty();
      if(startCompound(name))
      {
        result = model->serialize(*this, value);
        result &= endCompound();
      }
      return result;
    }

    /**
     * Create a reference to a field.
     *
     * The value is supposed to be a pointer on the variable, not the variable
     * itself.
     *
     * \note This could be typically a smart pointer.
     */
    template <typename Ptr>
    bool reference(const QString& name, const QString& type, Ptr& value)
    {
      int ref = startReference(name, type, (char*)(&*value) - (char*)0);
      if(ref < 0)
        return false;
      if(!find_reference(type, ref, value))
      {
        if(reader())
          value = create_object(value);
        if(!field("", *value))
        {
          return false;
        }
        references[type][ref] = make_reference(value);
      }
      return endReference();
    }

    /**
     * Create a reference to a field, using the model to serialize it.
     *
     * The value is supposed to be a pointer on the variable, not the variable
     * itself.
     *
     * \note This could be typically a smart pointer.
     */
    template <typename Ptr>
    bool reference(const QString& name, const QString& type, Ptr& value, Model* model)
    {
      int ref = startReference(name, type, (char*)(&*value) - (char*)0);
      if(ref < 0)
        return false;
      if(!find_reference(type, ref, value))
      {
        if(reader())
          value = create_object(value);
        if(!field("", *value, model))
        {
          return false;
        }
        references[type][ref] = make_reference(value);
      }
      return endReference();
    }

    /**
     * Test if there is a field with matching name and type.
     *
     * For writer, this should always return False.
     */
    virtual bool checkNextField(const QString& name) = 0;

    /**
     * Retrieve a string describing the last error that occured, if any.
     */
    QString lastErrorString() const { return last_error_string; }

    /**
     * Retrieve the identifier for the last error
     *
     * \see STORAGE_ERROR
     */
    int lastError() const { return last_error; }

    /**
     * Change the text of the last error.
     *
     * Any specific implementation might alter the text.
     */
    virtual void setLastError(int error, const QString& error_str) { if(last_error == NO__ERROR) { last_error = error; last_error_string = error_str; } }

    /**
     * Remove any error message
     */
    virtual void clearLastError() { last_error = NO__ERROR; last_error_string = QString(); }

    /**
     * Integer representing the version of the file for the model.
     *
     * This number is defined for a given model and will never appear in 
     * a file.
     */
    int version() const { return version_number; }

    /**
     * Version of the file.
     */
    const QString& fileVersion() const { return file_version; }

    /**
     * Name of the file being processed
     */
    virtual QString filename() const = 0;

  protected:

    template <typename Ptr>
    bool find_reference(const QString& type, int ref, Ptr& value)
    {
      ref_map_type::iterator found1 = references.find(type);
      if(found1 == references.end())
        return false;
      ref_map& typed_ref = found1->second;
      ref_map::iterator found = typed_ref.find(ref);
      if(found == typed_ref.end())
        return false;
      RealReference<Ptr> *r = dynamic_cast<RealReference<Ptr>*>(found->second.ptr);
      if(r)
        value = r->value;
      else
        return false;
      return true;
    }

    QString last_error_string;

    int last_error;

    QString file_version;

    int version_number;
  };

  /**
   * Function implementing the serialization.
   *
   * The default implementation expects a serialize method in the type:
   * \code
   *   bool T::serialize(VVEStorage&)
   * \endcode
   */
  template <typename T>
  bool serialization(VVEStorage& storage, T& value)
  {
    return value.serialize(storage);
  }

  /**
   * Helper function to parse version number.
   *
   * \returns the version number of 0 if no number found.
   *
   * This function will find any version on the form MAJOR[.MINOR[.MICRO]] and 
   * will return a unique integer on 32 bits supposing MAJOR and MINOR are less 
   * than 256 and micro is less than 65536. If the version numbers are too big, 
   * only the least significant bits are kept.
   *
   * Any string other than the version number will be ignored.
   *
   * Examples of valid versions:
   * \verbatim
File Version    Version     Number
VVE 1.2.20      1.2.20      0x01020014
XML 1.3         1.3         0x01030000
MyModel1.3.2a   1.3.2       0x01030002
OBJ 12.32.254   26.32.254   0x1a2000fe
FILE 12         12          0x0c000000
\endverbatim
   */
  int versionNumber(const QString& file_version);
}

#endif // VVELIB_STORAGE_STORAGE_H

