#ifndef VVELIB_STORAGE_STORAGE_BIN_H
#define VVELIB_STORAGE_STORAGE_BIN_H
/**
 * \file storage_xml.h
 *
 * This file contains the BIN implementation of the storage class.
 */

#include <config.h>
#include <storage/storage.h>
#include <storage/types.h>
#include <QFile>
#include <stack>

class Model;

/**
 * \namespace storage
 *
 * Namespace containing all functions and classes related to VVE persistence
 */
namespace storage
{
  /**
   * Data holding properties of a frame
   */
  struct Frame
  {
    Frame(int u = 0, bool h = false)
      : unnamed_subframes(u)
      , has_fields(h)
    {}

    Frame(const Frame& copy)
      : unnamed_subframes(copy.unnamed_subframes)
      , has_fields(copy.has_fields)
    {}

    int unnamed_subframes;
    bool has_fields;
  };

  /**
   * \class VVEStorage_BINWriter storage.h <storage/storage.h>
   * Implementation writing an BIN file
   */
  class VVEStorage_BINWriter : public VVEStorage
  {
  public:
    /// Type of a reference identifier
    typedef ptrdiff_t reference_t;

    VVEStorage_BINWriter();
    virtual ~VVEStorage_BINWriter() {}

    /**
     * Actually save the file.
     */
    virtual bool serialize(const QString& filename, Model* model);

    virtual bool startCompound(const QString& name);
    virtual bool endCompound();

    virtual int startReference(const QString& name, const QString& ref_type, reference_t ref);
    virtual bool endReference();

    virtual bool reader() { return false; }
    virtual bool writer() { return true; }

#define DEF_FIELD(T) virtual bool field(const QString& name, T& value);
    FOR_ALL_TYPES(DEF_FIELD)
#undef DEF_FIELD

    bool checkNextField(const QString& ) { return false; }

    /**
     * Name of the file being processed
     */
    virtual QString filename() const
    {
      if(file)
        return file->fileName();
      else
        return "";
    }

  protected:
    bool createElement(const QString& name, unsigned char type);
    bool checkFileStatus();
    bool resolveFrame(bool named_field = true);

    template <typename T>
    bool writeField(const QString& name, T& value);

    QFile *file;
    /// Stack of frames. A frame group fields.
    std::stack<Frame> frame_stack;
#ifdef USE_HASH
    typedef std::unordered_map<reference_t,int> ref_t;
    std::unordered_map<QString, ref_t> references;
#else
    typedef std::map<reference_t,int> ref_t;
    std::map<QString, ref_t> references;
#endif
  };

  /**
   * \class VVEStorage_BINReader storage.h <storage/storage.h>
   * Implementation reading an BIN file
   */
  class VVEStorage_BINReader : public VVEStorage
  {
  public:
    /// Type of a reference identifier
    typedef ptrdiff_t reference_t;

    VVEStorage_BINReader();
    virtual ~VVEStorage_BINReader() {}

    virtual bool setOption(int option, int value);

    /**
     * Actually read the file.
     */
    virtual bool serialize(const QString& filename, Model* model);

    virtual bool startCompound(const QString& name);
    virtual bool endCompound();
    virtual bool ignore(const QString& name);

    virtual int startReference(const QString& name, const QString& ref_type, reference_t ref);
    virtual bool endReference();

    virtual bool reader() { return true; }
    virtual bool writer() { return false; }

#define DEF_FIELD(T) virtual bool field(const QString& name, T& value);
    FOR_ALL_TYPES(DEF_FIELD)
#undef DEF_FIELD

    bool checkNextField(const QString& name);
    virtual void setLastError(int error, const QString& err);

    /**
     * Name of the file being processed
     */
    virtual QString filename() const
    {
      if(file)
        return file->fileName();
      else
        return "";
    }

  protected:
    unsigned char nextElementType();
    const QString& nextElementName();
    bool checkFileStatus();
    bool skipNext();
    bool skipToEnd();

    template <typename Read, typename Store>
    bool readData(QFile* file, Store& value);

    template <typename T>
    bool readValue(const QString& name, T& value);

    bool resolveFrame(bool named_field = true);

    QFile *file;
    int type_check;
    /// Stack of frames. A frame groups fields.
    std::stack<Frame> frame_stack;
    /**
     * Number of unnamed compounds on top of the stack.
     *
     * That can only happen if the named compound is replaced by an unnamed 
     * field.
     */
    int top_level_compounds;

    void elementUsed();

    QString next_element_name;
    bool has_next_element_name;

    TYPES type_conversion[NB_STORAGE_TYPES];
    size_t type_size[NB_STORAGE_TYPES];
    NUMBER_CLASS type_class[NB_STORAGE_TYPES];
  };
}

#endif // VVELIB_STORAGE_STORAGE_BIN_H

