#ifndef VVELIB_STORAGE_STORAGE_XML_H
#define VVELIB_STORAGE_STORAGE_XML_H
/**
 * \file storage_xml.h
 *
 * This file contains the XML implementation of the storage class.
 */

#include <config.h>
#include <storage/storage.h>
#include <QDomDocument>
#include <list>
#include <storage/types.h>

class Model;

/**
 * \namespace storage
 *
 * Namespace containing all functions and classes related to VVE persistence
 */
namespace storage
{
  /**
   * \class VVEStorage_XMLWriter storage.h <storage/storage.h>
   * Implementation writing an XML file
   */
  class VVEStorage_XMLWriter : public VVEStorage
  {
  public:
    /// Type of a reference identifier
    typedef ptrdiff_t reference_t;

    VVEStorage_XMLWriter();
    virtual ~VVEStorage_XMLWriter();

    virtual bool setOption(int option, int value);

    /**
     * Actually save the file.
     */
    virtual bool serialize(const QString& filename, Model* model);

    virtual bool startCompound(const QString& name);
    virtual bool endCompound();

    virtual int startReference(const QString& name, const QString& ref_type, reference_t ref);
    virtual bool endReference();

    virtual bool reader() { return false; }
    virtual bool writer() { return true; }

#define DEF_FIELD(T) virtual bool field(const QString& name, T& value);
    FOR_ALL_TYPES(DEF_FIELD)
#undef DEF_FIELD

    virtual bool checkNextField(const QString& ) { return false; }

    virtual QString filename() const { return _filename; }

  protected:
    QDomElement createElement(const QString& name, const QString& type = QString());

    QDomDocument doc;
    QDomElement root;
    std::list<QDomElement> compound_stack;

#ifdef USE_HASH
    typedef std::unordered_map<reference_t, uint> ref_map_t;
    typedef std::unordered_map<QString, ref_map_t> ref_map_typed_t;
#else
    typedef std::map<reference_t, uint> ref_map_t;
    typedef std::map<QString, ref_map_t> ref_map_typed_t;
#endif

    ref_map_typed_t typed_references_map;

    bool write_type;

    QString _filename;
  };

  /**
   * \class VVEStorage_XMLReader storage.h <storage/storage.h>
   * Implementation reading an XML file
   */
  class VVEStorage_XMLReader : public VVEStorage
  {
  public:
    /// Type of a reference identifier
    typedef ptrdiff_t reference_t;

    VVEStorage_XMLReader();
    virtual ~VVEStorage_XMLReader();

    virtual bool setOption(int option, int value);

    /**
     * Actually read the file.
     */
    virtual bool serialize(const QString& filename, Model* model);

    virtual bool startCompound(const QString& name);
    virtual bool endCompound();

    virtual int startReference(const QString& name, const QString& ref_type, reference_t ref);
    virtual bool endReference();

    virtual bool reader() { return true; }
    virtual bool writer() { return false; }

#define DEF_FIELD(T) virtual bool field(const QString& name, T& value);
    FOR_ALL_TYPES(DEF_FIELD)
#undef DEF_FIELD

    virtual bool checkNextField(const QString& name);

    virtual void setLastError(int error, const QString& err);

    virtual QString filename() const { return _filename; }

  protected:
    template <typename T>
    bool get_field(const QString& name, T& result);

    template <typename From, typename To>
    bool extract_value_with_error(QDomElement element, To& result);

    bool validType(QDomElement element, TYPES type);
    QDomElement extractElement(const QString& name);
    void setError_novalue(const QString& name);
    void setError_invalid_value(const QString& name, const QString& type, const QString& value);
    void setError_invalid_conversion(const QString& name, const QString& value, const QString& type_from, const QString& type_to);

    QDomDocument doc;
    QDomElement root;
    std::list<QDomElement> compound_stack;

    int type_checking;

    QString _filename;
  };
}

#endif // VVELIB_STORAGE_STORAGE_XML_H

