#ifndef TEST_TEST_H
#define TEST_TEST_H

/**
 * \file test.h
 *
 * File containing macros and functions to for main testing
 */


#include <config.h>
#include <QTextStream>
#include <util/random.h>
#include <util/vector.h>

#ifndef TEST_RESULT
#  define TEST_RESULT test::result
#endif

/**
 * \def CHECK
 *
 * check if tst is true. If not, output the line of the error and set the 
 * variable "TEST_RESULT" to false.
 *
 * \relates test
 */
#define CHECK( tst ) \
  TEST_RESULT &= test::check(tst, __LINE__, #tst)

/**
 * \def CHECK_EQUAL
 *
 * check if tst is equal to value. The operator == must exist between the two. 
 * If not, output the line of the error, show the content of both member (i.e. 
 * the output stream operator must exist) and set the variable "TEST_RESULT" to 
 * false.
 *
 * \relates test
 */
#define CHECK_EQUAL( tst, value ) \
  TEST_RESULT &= test::check_equal(tst, value, __LINE__, #tst, #value)

/**
 * \def CHECK_NOTEQUAL
 *
 * check if tst is equal to value. The operator == must exist between the two. 
 * If not, output the line of the error, show the content of both member (i.e. 
 * the output stream operator must exist) and set the variable "TEST_RESULT" to 
 * false.
 *
 * \relates test
 */
#define CHECK_NOTEQUAL( tst, value ) \
  TEST_RESULT &= test::check_notequal(tst, value, __LINE__, #tst, #value)

/**
 * \def CHECK_FLOAT
 *
 * Check if tst is equal to value with a relative error of epsilon. This 
 * function is used for floating points number. The absolute epsilon is 
 * computed as the relative epsilon multiplied by the biggest absolute values 
 * of tst and value;
 */
#define CHECK_FLOAT( tst, value, epsilon ) \
  TEST_RESULT &= test::check_float(tst, value, epsilon, __LINE__, #tst, #value)

/**
 * \def CHECK_FLOAT_DIFF
 *
 * Check if tst is different to value with a relative error of epsilon. This 
 * function is used for floating points number. The absolute epsilon is 
 * computed as the relative epsilon multiplied by the biggest absolute values 
 * of tst and value;
 */
#define CHECK_FLOAT_DIFF( tst, value, epsilon ) \
  TEST_RESULT &= test::check_float_diff(tst, value, epsilon, __LINE__, #tst, #value)

/**
 * \def CHECK_ZERO
 *
 * Check if tst is equal to zero with an error of epsilon. This function is 
 * used for floating points number.
 */
#define CHECK_ZERO( tst, epsilon ) \
  TEST_RESULT &= test::check_zero(tst, epsilon, __LINE__, #tst)

/**
 * \def TEST_FCT
 *
 * Launch the function test_tst (replacing tst with the argument), expect it to 
 * return a boolean and display a message indicating if the test was successful 
 * or not. If the test failed, the boolean variable put as second argument is 
 * set to true. the variable is not changed otherwise.
 *
 * \relates test
 */
#define TEST_FCT( tst, failed ) \
{ \
  bool _11532fd4s85fds_test_failed = !test_##tst(); \
  if(_11532fd4s85fds_test_failed) \
  { \
    failed = true; \
    test::out << #tst " test failed" << endl; \
    setExitCode(10); \
  } \
  else \
    test::out << #tst " test passed" << endl;\
}

#define FAILED_TEST(line, tst) \
  test::out << "Test line " << line << " " << tst << " failed as ";

#define FAILED_FUNCTION(line) \
  test::out << "Test line " << line << " " << __PRETTY_FUNCTION__ << "() failed as ";

/**
 * Namespace containing test facilities
 */
namespace test
{
  extern QTextStream out;
  extern bool result;

  inline void resetResult() { result = true; }

  template <size_t N, typename T>
  void shuffle(T array[])
  {
    for(size_t i = 0 ; i < N ; ++i)
    {
      long int n = util::random(N-i);
      T dropped = array[i];
      array[i] = array[n+i];
      array[n+i] = dropped;
    }
  }

  inline bool check(bool tst, size_t line_nb, const char* tst_str)
  {
    if(!tst)
    {
      test::out << "Test line " << line_nb << " " << tst_str << " failed" << endl;
    }
    return tst;
  }

  template <typename T1, typename T2>
  bool check_equal(const T1& tst, const T2& value, size_t line_nb,
                   const char* tst_str, const char* value_str)
  {
    if(not (tst == value))
    {
      FAILED_TEST(line_nb, tst_str << " == " << value_str);
      test::out << tst_str << " = " << tst << " and " << value_str << " = "
                << value << endl;
      return false;
    }
    return true;
  }

  template <typename T1, typename T2>
  bool check_notequal(const T1& tst, const T2& value, size_t line_nb,
                      const char* tst_str, const char* value_str)
  {
    if(not (tst != value))
    {
      FAILED_TEST(line_nb, tst_str << " != " << value_str);
      test::out << tst_str << " = " << tst << " and " << value_str << " = "
                << value << endl;
      return false;
    }
    return true;
  }

  template <typename T1, typename T2, typename T3>
  bool check_float(const T1& tst, const T2& value, const T3& epsilon, size_t line_nb,
                   const char* tst_str, const char* value_str)
  {
    if( util::norm(tst-value) > epsilon*std::max(util::norm(tst), util::norm(value)) )
    {
      FAILED_TEST(line_nb, tst_str << " ~= " << value_str);
      test::out << tst_str << " = " << tst << " and "
                << value_str << " = " << value << endl;
      return false;
    }
    return true;
  }

  template <typename T1, typename T2, typename T3>
  bool check_float_diff(const T1& tst, const T2& value, const T3& epsilon, size_t line_nb,
                        const char* tst_str, const char* value_str)
  {
    if( util::norm(tst-value) <= epsilon*std::max(util::norm(tst), util::norm(value)) )
    {
      FAILED_TEST(line_nb, tst_str << " != " << value_str);
      test::out << tst_str << " = " << tst << " and "
                << value_str << " = " << value << endl;
      return false;
    }
    return true;
  }

  template <typename T1, typename T2>
  bool check_zero(const T1& tst, const T2& epsilon, size_t line_nb,
                  const char* tst_str)
  {
    if( util::norm(tst) > epsilon )
    {
      FAILED_TEST(line_nb, tst_str << " ~= 0");
      test::out << tst_str << " = " << tst << endl;
      return false;
    }
    return true;
  }

}

#endif // TEST_TEST_H

