#ifndef TISSUE_MODEL_H
#define TISSUE_MODEL_H

/**
 * \file tissue_model.h
 *
 * This files include the tissue model helper. This helper define a tissue that 
 * is already drawn.
 *
 * The namespace tissue_model contains the base class for this helper.
 *
 * To use the helper, simply include <tissue_model.h>.
 * Any extra attributes for vertices or edges should be defined before 
 * including this file.
 *
 * Once included, the model starts with
 * \code
 * StartModel
 * \endcode
 * and ends with
 * \code
 * EndModel
 * \endcode
 *
 * The tissue defined in the model uses cells which contains these attributes:
 * \code
 *  double area;              // Area of the cell if the vertex is of type
 *  Point3d pos;              // Position of the vertex
 *  Point3d normal;           // Normal to the surface
 *  double value;             // Value used to draw the vertex
 * \endcode
 * and junctions which contains these attributes:
 * \code
 *  Point3d pos;              // Position of the junction
 *  Point3d normal;           // Normal to the surface
 *  Point2d uv;               // Local coordinate of the junction on the bspline
 * \endcode
 *
 * In addition, the user can add its own attributes by defining the macro \c 
 * VertexAttributes \b before including this file:
 * \code
 * #define CellAttribute \
 *   double attr1; \
 *   int attr2; \
 *   bool attr3;
 * #define JunctionAttribute \
 *   double attr1; \
 *   bool attr2; \
 *   int attr3;
 *
 * #include <tissue_model.h>
 * \endcode
 * \note As this is a macro, to span over many lines you need to end the inner lines with a backslash.
 *
 * Similarly, the user can add attributes to the cell and tissue edges by 
 * defining the macros \c CellEdgeAttributes, \c WallAttributes, \c 
 * CellJunctionEdgeAttributes and \c JunctionCellEdgeAttributes.
 *
 * Within the model class, named \c ModelClass, all the methods and variable of 
 * the class tissue_model::TissueModel are available. Reference to the parent 
 * class can be done using the \c ParentClass typedef.
 *
 * For example, if yo want to augment the drawing method, you might write:
 * \code
 * void draw(Viewer* viewer)
 * {
 *   ParentClass::draw(viewer);
 *   // my other commands
 * }
 * \endcode
 *
 * As a small example, here is a code that creates a single cell and change its 
 * color when the user click on it.
 * \code
 * #define CellAttributes \
 *   bool clicked; // Is the current cell clicked?
 *
 * #include <tissue_model.h>
 *
 * StartModel
 *
 * // Define
 * int clickedColor, unclickedColor;
 *
 * // Read the extra parameter
 * void readParam(util::Parms& parms)
 * {
 *   parms("View", "ClickedColor", clickedColor);
 *   parms("View", "UnclickedColor", unclickedColor);
 * }
 *
 * // Initialize the tissue
 * void initialize()
 * {
 *   // First, create the cell
 *   cell c;
 *   junction j1, j2, j3, j4;
 *   j1->pos = Point3d(-1,-1,0);
 *   j2->pos = Point3d(1,-1,0);
 *   j3->pos = Point3d(1,1,0);
 *   j4->pos = Point3d(-1,1,0);
 *   c->pos = Point3d(0,0,0);
 *   T.addCell(c, j1, j2, j3, j4);
 *   // Then, initialise the content of it
 *   c->clicked = false;
 * }
 *
 * // Update the status of the cell after user interaction
 * void postSelection(const QPoint&, Viewer* viewer)
 * {
 *   const cell& c = cellFromId(viewer->selectedName());
 *   // If a cell was selected
 *   if(c)
 *   {
 *     c->clicked ^= true; // Invert its state
 *   }
 * }
 *
 * // Define the color depending on the clicked state
 * Color getCellColor(const cell& c)
 * {
 *   if(c->clicked)
 *     return palette.getColor(clickedColor);
 *   else
 *     return palette.getColor(unclickedColor);
 * }
 *
 * // Define the color depending on the clicked state
 * Color getCellCenterColor(const cell& c)
 * {
 *   if(c->clicked)
 *     return palette.getColor(clickedColor);
 *   else
 *     return palette.getColor(unclickedColor);
 * }
 *
 * // All the normals are the same and upward
 * Point3d normal(const cell&)
 * {
 *   return Point3d(0,0,1);
 * }
 * Point3d normal(const junction&)
 * {
 *   return Point3d(0,0,1);
 * }
 *
 * EndModel
 *
 * \endcode
 */

#include <vve.h>
#include <util/parms.h>
#include <util/palette.h>
#include <iostream>

#include <util/watchdog.h>
#include <algorithms/tissue.h>

#include <geometry/geometry.h>

using geometry::Point3d;

typedef util::Palette::Color Color;

/**
 * Namespace containing the base class for the ::tissue model helper
 */
namespace tissue_model
{
  /**
   * Relative constant for geometrical error
   */
  const double epsilon=0.0001;

  /**
   * Base class for the tissue_model helper
   */
  template <typename RealModel, typename TissueClass>
  struct TissueModel : public Model
  {
    /**
     * Documentation for the types imported
     *
     * \typedef typename VVComplex::cell cell
     * \brief Type of a cell
     *
     * \typedef typename VVComplex::junction junction
     * \brief Type of a junction
     *
     * \typedef typename VVComplex::cell_edge cell_edge
     * \brief Type of an edge in the cell graph
     *
     * \typedef typename VVComplex::wall wall
     * \brief Type of a wall (i.e. an edge in the wall graph)
     *
     * \typedef typename VVComplex::cell_junction_edge cell_junction_edge
     * \brief Type of an edge from a cell to a junction
     *
     * \typedef typename VVComplex::junction_cell_edge junction_cell_edge
     * \brief Type of an edge from a junction to a cell
     *
     * \typedef typename VVComplex::complex_graph complex_graph;
     * \brief Type of the complex graph (i.e. linking cells to junctions and 
     * vice-versa)
     *
     * \typedef typename VVComplex::cell_graph cell_graph;
     * \brief Type of the cell graph
     *
     * \typedef typename VVComplex::wall_graph wall_graph
     * \brief Type of the wall graph
     */
    IMPORT_COMPLEX_TYPES(TissueClass);

    /**
     * Operator class to compare the size of cells
     */
    struct CompareSize : public std::binary_function<cell, cell, bool>
    {
      bool operator()(const cell& c1, const cell& c2) const
      {
        return c1->area < c2->area;
      }
    };

    /**
     * Ordered set of vertices.
     *
     * The vertices must represent cells. They will be ordered by area.
     */
    typedef std::set<cell,CompareSize> ordered_cells_t;

    /// Watchdog object to monitor file modification
    util::WatchDog rex;
    /// Color palette
    util::Palette palette;
    /// Tissue created
    TissueClass T;

    /// Color of the background
    int backColor;
    /// Draw the neighborhood as arrows ?
    bool drawNeighborhood;

    /**
     * Default constructor.
     *
     * Initialise the watchdog, palette, tissue and aliases.
     */
    TissueModel(QObject *parent)
      : Model(parent)
      , rex(this)
      , palette("pal.map")
      , T(palette, (RealModel*)this)
    {
      rex.addObject(&palette);
    }

    void modifiedFiles( const std::set<std::string>& filenames )
    {
      if(filenames.find("view.v") != filenames.end())
        readTissueParam();
      rex.watch(filenames);
    }

    /**
     * Function to be redefined to read extra parameters
     */
    virtual void readParam( util::Parms& ) {}

    /**
     * Update the area of all the cells in the tissue.
     */
    void updateCellsArea()
    {
      forall_named(const cell& c, T.S, cells)
      {
        c->area = 0.0;
        forall( const junction& n, T.S.neighbors(c) )
        {
          // Find area
          const junction& m = T.S.nextTo(c, n);
          c->area += geometry::triangleArea(c->pos, m->pos, n->pos);
        }
        if(c->area <= epsilon ) // Avoid inf
        {
          std::cout << "Bad area " << c->area << std::endl;
          c->area = 0.1;
        }
      }
    }

    void registerFiles()
    {
      registerFile("view.v"); 
    }

    /**
     * Read the parameters for the tissue_model helper (automatically called by 
     * the helper)
     */
    void readTissueParam()
    {
      util::Parms parms("view.v");
      T.readViewParms(parms, "View");
      T.readParms(parms, "Tissue");
      parms("View", "BackColor", backColor);
      parms("View", "DrawNeighborhood", drawNeighborhood);
      readParam(parms);
    }

    //@{
    /// \name Drawing methods

    void initDraw()
    {
      glEnable(GL_NICEST);
      glDisable(GL_CULL_FACE);
    }

    void preDraw()
    {
      Color bg = palette.getColor(backColor);
      glClearColor(bg.r(), bg.g(), bg.b(), 1.0);
      T.preDraw();
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    }

    void postDraw()
    {
      glDisable(GL_BLEND);
      T.postDraw();
    }

    /**
     * Draw the tissue using the ModelClass::getCellColor and 
     * ModelClass::getCellCenterColor methods.
     */
    void draw(Viewer *viewer)
    {
      Vec vmin(HUGE_VAL,HUGE_VAL,HUGE_VAL), vmax(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
      forall(const junction& v, T.W)
      {
        const Point3d& pos = v->pos;
        if(pos.x() < vmin.x)
          vmin.x = pos.x();
        if(pos.y() < vmin.y)
          vmin.y = pos.y();
        if(pos.z() < vmin.z)
          vmin.z = pos.z();
        if(pos.x() > vmax.x)
          vmax.x = pos.x();
        if(pos.y() > vmax.y)
          vmax.y = pos.y();
        if(pos.z() > vmax.z)
          vmax.z = pos.z();
      }
      viewer->setSceneBoundingBox(vmin, vmax);

      forall_named(const cell& c, T.S, cells)
      {
        Color color = getCellColor(c);
        Color center_color = getCellCenterColor(c);
        T.drawCell(c, color, center_color);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        if(drawNeighborhood)
        {
          int cnt = 3;
          double ds = 0.1 / (3+T.C.valence(c));
          glColor3f(1.0, 0, 0);
          Vec from(c->pos);
          forall(const cell& n, T.C.neighbors(c))
          {
            glColor3f(1.0, 1.0, 1.0);
            Vec to(n->pos);
            Vec u = to-from;
            u.normalize();
            to = from + cnt*ds*u;
            cnt++;
            viewer->drawArrow(from, to, -1, 6);
          }
        }
      }
    }

    /**
     * Color using the \c value attribute of the cell
     */
    virtual Color getCellColor(const cell &c)
    {
      return T.valueColor(c->value);
    }

    /**
     * Color using the \c value attribute of the cell
     */
    virtual Color getCellCenterColor(const cell &c)
    {
      return T.valueCenterColor(c->value);
    }

    //@}

    //@{
    /// \name Selection methods

    /**
     * Draw for cell selection
     */
    void drawWithNames()
    {
      bool db = T.drawBorders;
      T.drawBorders = false;
      int i = 0;
      forall_named(const cell& c, T.S, cells)
      {
        glPushName(i++);
        T.drawCell(c, c->value);
        glPopName();
      }
      T.drawBorders = db;
    }

    /**
     * Return a reference of the cell whose id is \c i.
     *
     * The id is the one used to name cells in drawWithNames.
     */
    const cell& cellFromId(int i)
    {
      if(i == -1)
        return cell::null;
      return T.S.get_cell(i);
    }
    //@}

    //@{
    /// \name Methods required by the tissue library
    Point3d position( const cell& v) const
    {
      return v->pos;
    }

    Point3d position( const junction& v) const
    {
      return v->pos;
    }

    void setPosition( const cell& v, const Point3d& pos)
    {
      v->pos = pos;
    }

    void setPosition( const junction& v, const Point3d& pos)
    {
      v->pos = pos;
    }

    void setVertexPositionHint( const junction& , const junction& , const junction& , double )
    {
    }

    Point3d normal( const cell& v) const
    {
      return v->normal;
    }

    Point3d normal( const junction& v) const
    {
      return v->normal;
    }

    void updateFromOld( const cell& , const cell& , const cell& ,
                        const typename TissueClass::division_data&, TissueClass& )
    {
    }
    //@}

    /**
     * Default step method: do nothing
     */
    void step() { }

    /**
     * Function to override to initialize the model.
     */
    void initialize() {}
  };
}

#ifndef CellAttributes
#  define CellAttributes
#endif

#ifndef JunctionAttributes
#  define JunctionAttributes
#endif

#ifndef CellEdgeAttributes
#  define CellEdgeAttributes
#endif

#ifndef WallAttributes
#  define WallAttributes
#endif

#ifndef CellJunctionEdgeAttributes
#  define CellJunctionEdgeAttributes
#endif

#ifndef JunctionCellEdgeAttributes
#  define JunctionCellEdgeAttributes
#endif

#ifndef ModelInit
#  define ModelInit
#endif

/**
 * \def Use this macro to define where the model definition starts.
 */
#define StartModel \
  struct CellType \
  { \
    double area; \
    Point3d pos; \
    Point3d normal; \
    double value; \
    CellAttributes\
  }; \
 \
 struct JunctionType \
 { \
   Point3d pos; \
   Point3d normal; \
   JunctionAttributes\
 }; \
 \
  struct CellEdgeType \
  { \
    CellEdgeAttributes\
  }; \
 \
  struct WallType \
  { \
    WallAttributes\
  }; \
 \
  struct CellJunctionEdgeType \
  { \
    CellJunctionEdgeAttributes\
  }; \
 \
  struct JunctionCellEdgeType \
  { \
    JunctionCellEdgeAttributes\
  }; \
 \
  class ModelClass; \
  typedef tissue::Tissue<ModelClass, CellType, JunctionType, WallType, CellEdgeType, CellJunctionEdgeType, JunctionCellEdgeType> Tissue; \
 \
 EXPORT_COMPLEX_TYPES(Tissue); \
 \
  struct ModelClass : public tissue_model::TissueModel<ModelClass,Tissue> \
  { \
    typedef tissue_model::TissueModel<ModelClass,Tissue> ParentClass; \
    ModelClass(QObject *parent) \
      : ParentClass(parent)\
      ModelInit \
    { \
      readTissueParam(); \
      registerFiles(); \
      initialize(); \
    }

/**
 * \def Use this macro to end the model definition
 */
#define EndModel \
  }; DEFINE_MODEL(ModelClass)

#endif // TISSUE_MODEL_H
