#ifndef __BSURFACE_H__
#define __BSURFACE_H__

#include <config.h>
#include <vector>
#include <string>

#include <util/vector.h>
#include <util/matrix.h>
#include <geometry/geometry.h>

namespace util
{
  //#define Debug false

  using std::vector;

  using geometry::Matrix3d;
  using geometry::Point3d;
  using geometry::Point2d;

  /**
   * \namespace util::Shapes
   *
   * This namespace contains various shapes.
   *
   * For now, only BSplineSurface is included.
   */
  namespace Shapes
  {

    /**
     * \class BSplineSurface bsurface.h <util/bsurface.h>
     *
     * Class describing a bspline surface.
     */
    class BSplineSurface
    {
    public:

      //@{
      /// \name Interface to LPFG
      /**
       * Load a file describing the BSplineSurface
       * \param fname File name
       * \param scale Scale modification
       * \param TextureId texture used for rendering
       * \param divU number of division in U to use in the description
       * \param divV number of division in V to use in the description
       */
      void Load(const char* fname, double scale, int TextureId, int divU, int divV);
      /// Reread the last loaded file (if modified)
      void Reread();
      /// Compute the bounding box of the surface
      void BoundingBox(double &x1,double &y1,double &z1,double &x2,double &y2,double &z2);
      /// Transform the surface using a 3D linear transformation and a translation
      void Transform(Matrix3d M, double x, double y, double z);
      /// True if a texture is attached to the surface
      bool IsTextured() const
      { return _TextureId >= 0; }

      /// Get the texture id attached to the object (0 if none)
      int TextureId() const
      { return _TextureId; }
      //@}

      /// Render the object in the current OpenGL context
      void Draw(int uDiv, int vDiv);

      /// Force recomputation of the surface
      void Recompute()
      {precomputed=false;}

      void GetControlMesh(int i, int j, double &x,double &y,double &z)
      {x=Control_points[0][i][j][0];y=Control_points[0][i][j][1];z=Control_points[0][i][j][2];}

      void SetControlMesh(int i, int j,double x, double y, double z)
      {Control_points[0][i][j][0]=x;Control_points[0][i][j][1]=y;Control_points[0][i][j][2]=z;}

      //Functions used in splineEdit and frameEdit

      //Constructors and Destructors
      BSplineSurface();
      ~BSplineSurface();
      BSplineSurface(const BSplineSurface &B);

      //Save/Load the surface in fname
      bool  LoadPatch(const char *fname);
      bool  SavePatch(const char *fname);
      void Reset(int u,int v);	

      //Resizes the patch to have num_rows rows and num_cols columns using forward and reverse subdivision
      //It is assumed that the dimensions passed in can be created simply by using forward and reverse subdivision
      void Resize(unsigned int num_rows,unsigned int num_cols);

      //Returns first direvative of change in area at u,v
      float HasContracted(BSplineSurface *b, float u, float v);
      //Provides previous b-spline surface to perform the above comparison on when rendering
      void SetPrevious(BSplineSurface *b)
      {psurf = b;};
      void VisualizeContraction(bool c)
      {contraction=c;}


      //Applys forward subdivision along the U-curves of the control points
      void DoubleU();
      //Applys reverse subdivision along the U-curves of the control points
      void HalveU();
      //Applys forward subdivision along the V-curves of the control points
      void DoubleV();
      //Applys reverse subdivision along the V-curves of the control points
      void HalveV();

      //Toggles between wireframe and shaded renderings
      void IsShaded(bool shaded){wireframe=!shaded;};
      //Sets the degree of the surface, if degree!=3 then forward and reverse subdivision will not work
      //  void setdegree(int i){degree=i;};
      //Toggles between a shaded wireframe mesh and shaded polygons
      void ToggleShadedMesh(){s_mesh=false;};
      void TogglePolygonMesh(){s_mesh=true;};
      void ToggleNormals(){disp_normal=!disp_normal;};
      void ToggleControlMesh(){disp_c_mesh=!disp_c_mesh;};
      //Changes the colors used when rendering
      void SetColors(double c1[3],double c2[3],double c3[3]);

      //Specifies which resolution of control points is being edited
      //returns false if level specified does not exist
      bool EditLevel(unsigned int i);

      //Specifies the length of u and v steps to be used when rendering
      //the surface
      void SetPrecision(double u, double v){u_prec=u;v_prec=v;};

      //Select a point for editing
      void SelectPoint(Point3d n, Matrix3d A, Point3d t);
      //Move the selected point
      void MovePoint(Point3d m, Matrix3d A, Point3d t);

      //Select a region for editing
      void SelectRegion(vector<Point3d > s_reg, Matrix3d A, Point3d t);
      //Determine whether to_check fall in the contour defined by  shape
      bool InsideContour(vector<Point3d > shape, Point3d to_check);

      int SelectedLevel()
      {return selected_level;}
      vector<int*> SelectedIndicies()
      {return selected_region;}
      vector<vector<vector<Point3d > > >* ControlPoints()
      {return &Control_points;}

      //Move the selected region
      void  MoveRegion(Point3d dis);
      //Apply the bending tool to the selected region
      //void RotateRegion(Point2d c_rot,double theta,double r_dis, Matrix3d A);

      //renders the control points as circles
      void DisplayControlPoints(Matrix3d A, Point3d t);
      //renders the surface based on options contained in the class
      void DisplaySurface();
      //draws a chircel at x,y,z with radius w
      void DrawCircle(double x, double y, double z, double w);

      //Functions working on the parameter range (1.0,n_u+degree-1.0)x(1.0,n_v+degree-1.0)
      //evaluates the b-spline surface at (u,v)
      Point3d Eval(double u, double v) const;
      //evaluates the u-curve at u
      Point3d EvalU(double u, int m) const;
      //evaluates the normal at u,v
      Point3d NormalEval(double u, double v) const;
      //evaluates F(u,v)/du
      Point3d PartialU(double u, double v) const;
      //evaluates F(u,v)/dv
      Point3d PartialV(double u, double v) const;
      //Locates the closest point to q on the surface using p0 as
      //an initial guess
      bool PointInversion(Point2d &surf_p, Point2d p0, Point3d q) const;

      //The same functionality as above, but for the parameter range (0,1)x(0,1)
      Point3d EvalN(double u, double v) const;
      Point3d NormalEvalN(double u, double v) const; 
      Point3d PartialUN(double u, double v) const;
      Point3d PartialVN(double u, double v) const;

      bool PointInversionN(Point2d &surf_p, Point2d p0, Point3d q) const;
      inline double ConvertU(double u) const;
      inline double ConvertV(double v) const;

      //Parameterized representation of the curve along the lines
      //(0,0)(0,1),(0,1)(1,1),(1,1)(1,0),(1,0)(0,0) in the order given
      //this traces out the boundary of the b-spline patch, u varies between
      //0 and 1
      Point3d ContourEval(double u) const;

      Point2d ContourInverse(double u) const;

      BSplineSurface Interpolate(const BSplineSurface &BS1, const BSplineSurface &BS2, double alpha, int num_rows, int num_cols) const;
      unsigned int LowestResDim(int i);
      BSplineSurface LowestRes();

      void UpdateFootprint();
      void RenderSupport(unsigned int i,unsigned int j,unsigned int level);
      void RenderFootprint();
      void CalcFootprint(unsigned int i,unsigned  int j,unsigned int level);
      void CalcPreImage(int i, int j, unsigned int level);

      void  RenderLine(int i1, int j1, int level1, int i2, int j2, int level2);

      BSplineSurface& operator=(const BSplineSurface& B);
      BSplineSurface operator+(const BSplineSurface& B);
      BSplineSurface operator-(const BSplineSurface& B);
      BSplineSurface operator*(const double c);

      int DimU(){return n_u;};
      int DimV(){return n_v;};

    private:
      //	UVPrecision _uvPrecision;
      inline void Set(int r, int c, Point3d v)
      { Control_points[0][r][c] = v; }
      inline Point3d Get(int r, int c) const
      { return Control_points[0][r][c]; }

      std::string _fname;
      double _scale;

      int _TextureId;

      int PtId(int s, int t) const
      {
        return s+(_divU+1)*t;
      }

      int _divU, _divV;
      vector<Point3d > _vrtx;
      vector<Point3d > _nrml;
      vector<float > _ctract; 
      float c_min,c_max;

      void Precompute(int, int);
      bool precomputed;
      std::string _f_name;

      bool s_mesh;
      bool wireframe;
      bool lines;
      bool contraction;
      bool disp_normal;
      bool disp_c_mesh;
      bool periodic;

      unsigned int selected[2];
      int selected_level;
      int n_u,n_v;

      double sel_point[3];
      double unsel_point[3];
      double poly_line[3];
      int degree;

      vector< vector< vector< Point3d > > > Control_points;
      vector< vector< vector< Point3d > > > Details;

      vector< int* > footprint;
      vector< int* > selected_region;
      int selected_bb[4];

      double u_prec,v_prec;  

      BSplineSurface *psurf;
      /*
         void Update_Mres();
         */

    };
  } // namespace Shapes

  using Shapes::BSplineSurface;
}
#endif

