#ifndef VVELIB_UTIL_CIRC_ITERATOR_H
#define VVELIB_UTIL_CIRC_ITERATOR_H

#include <config.h>
#include <iterator>

namespace util
{
  /**
   * \class CircIterator circ_iterator.h <util/circ_iterator.h>
   *
   * Creates a circular iterator from a range of forward iterators
   */
  template <typename ForwardIterator>
  class CircIterator
  {
  public:
    typedef ForwardIterator base_type;
    typedef std::forward_iterator_tag iterator_category;
    typedef typename std::iterator_traits<ForwardIterator>::value_type        value_type;
    typedef typename std::iterator_traits<ForwardIterator>::difference_type   difference_type;
    typedef typename std::iterator_traits<ForwardIterator>::pointer           pointer;
    typedef typename std::iterator_traits<ForwardIterator>::reference         reference;

    CircIterator() {}

    /**
     * Create an iterator within a range, starting at \c c.
     */
    CircIterator(const ForwardIterator& f, const ForwardIterator& l, const ForwardIterator& c)
      : first(f)
      , last(l)
      , init(c)
      , cur(c)
    {
    }

    /**
     * Creates an 'end' iterator
     */
    CircIterator(const ForwardIterator& f, const ForwardIterator& l)
      : first(f)
      , last(l)
      , init(l)
      , cur(l)
    {
    }

    /**
     * Copy the iterator, but change the current position to somewhere else.
     */
    CircIterator(const CircIterator& copy, const ForwardIterator& new_cur)
      : first(copy.first)
      , last(copy.last)
      , init(copy.init)
      , cur(new_cur)
    {}

    CircIterator(const CircIterator& copy)
      : first(copy.first)
      , last(copy.last)
      , init(copy.init)
      , cur(copy.cur)
    {}

#ifdef USE_CXX0X
    CircIterator(CircIterator&& copy) = default;
#endif

    CircIterator& operator++()
    {
      ++cur;
      if(cur == last)
        cur = first;
      if(cur == init)
        cur = last;
      return *this;
    }

    CircIterator operator++(int)
    {
      CircIterator temp(*this);
      this->operator++();
      return temp;
    }

    reference operator*() { return *cur; }
    pointer operator->() { return cur.operator->(); }

    bool operator==(const ForwardIterator& other) const
    {
      return cur == other;
    }

    bool operator==(const CircIterator& other) const
    {
      return cur == other.cur;
    }

    bool operator!=(const ForwardIterator& other) const
    {
      return cur != other;
    }

    bool operator!=(const CircIterator& other) const
    {
      return cur != other.cur;
    }

    /**
     * Check if the base iterator given as argument correspond to the position 
     * this circular iterator starts from
     */
    bool isInitIterator(const base_type& cmp) const
    {
      return cmp == init;
    }

    /**
     * Create the end iterator from a current one
     */
    CircIterator end() const
    {
      return CircIterator(first, last);
    }

    base_type base() const { return cur; }

  protected:
    ForwardIterator first, last, init, cur;
  };
}

#endif // VVELIB_UTIL_CIRC_ITERATOR_H

