#ifndef __UTIL__COLOUR_H__
#define __UTIL__COLOUR_H__
/**
 * \file color.h
 *
 * Defines the util::Color class template
 */

#include <config.h>
#include <util/vector.h>
#include <util/clamp.h>
#include <QtGui/QColor>

namespace util {
  /**
   * \class Color color.h <util/color.h>
   * \brief A utility class to encapsulate color data.
   */
  template<class T> class Color : public Vector<4, T> {
  public:

    /**
     * Constructor to convert from one color type to another
     */
    template <typename T1>
    explicit Color(const Vector<4,T1>& color, const T& scale = 1)
      : Vector<4,T>(color[0]*scale, color[1]*scale, color[2]*scale, color[3]*scale)
    { }

    template <typename T1>
    explicit Color(const Vector<4,T1>& color, const T1& scale)
      : Vector<4,T>(color[0]*scale, color[1]*scale, color[2]*scale, color[3]*scale)
    { }

    Color(const Vector<4,T>& copy)
      : Vector<4,T>(copy)
    { }

    Color(const QColor& c)
    {
      convertFromQColor(*this, c);
    }

#ifdef USE_CXX0X
    Color(Color&& copy) = default;
#endif

    /**
     * \brief Constructor.
     * \param r Value for red.
     * \param g Value for green.
     * \param b Value for blue.
     * \param a Value for alpha.
     */
    Color(const T& r = T(), const T& g = T(), const T& b = T(), const T& a = T())
      : Vector<4,T>(r,g,b,a) { }

    /**
     * \brief Return the red component.
     */
    T& r() {return this->x();}

    /**
     * \brief Return the green component.
     */
    T& g() {return this->y();}

    /**
     * \brief Return the blue component.
     */
    T& b() {return this->z();}

    /**
     * \brief Return the alpha component.
     */
    T& a() {return this->t();}


    /**
     * \brief Return the red component.
     */
    const T& r() const {return this->x();}

    /**
     * \brief Return the green component.
     */
    const T& g() const {return this->y();}

    /**
     * \brief Return the blue component.
     */
    const T& b() const {return this->z();}

    /**
     * \brief Return the alpha component.
     */
    const T& a() const {return this->t();}


    /**
     * \brief Set the red component.
     */
    void r(const T& val) {this->x(val);}

    /**
     * \brief Set the green component.
     */
    void g(const T& val) {this->y(val);}

    /**
     * \brief Set the blue component.
     */
    void b(const T& val) {this->z(val);}

    /**
     * \brief Set the alpha component.
     */
    void a(const T& val) {this->t(val);}

    Color& operator=(const Color& c);
    Color& operator=(const Vector<4,T>& c);
    Color& operator=(const T& val);


#ifdef USE_CXX0X
    Color& operator=(Color&& copy) = default;
    Color& operator=(Vector<4,T>&& c);
#endif


    operator QColor() const
    {
      return convertToQColor(*this);
    }
  };

  QColor convertToQColor(const Color<float>& c);
  QColor convertToQColor(const Color<double>& c);
  QColor convertToQColor(const Color<long double>& c);
  QColor convertToQColor(const Color<unsigned char>& c);
  QColor convertToQColor(const Color<unsigned short>& c);
  QColor convertToQColor(const Color<unsigned int>& c);
  QColor convertToQColor(const Color<unsigned long>& c);
  QColor convertToQColor(const Color<unsigned long long>& c);
  QColor convertToQColor(const Color<char>& c);
  QColor convertToQColor(const Color<short>& c);
  QColor convertToQColor(const Color<int>& c);
  QColor convertToQColor(const Color<long>& c);
  QColor convertToQColor(const Color<long long>& c);

  void convertFromQColor(Color<float>& c, const QColor& col);
  void convertFromQColor(Color<double>& c, const QColor& col);
  void convertFromQColor(Color<long double>& c, const QColor& col);
  void convertFromQColor(Color<unsigned char>& c, const QColor& col);
  void convertFromQColor(Color<unsigned short>& c, const QColor& col);
  void convertFromQColor(Color<unsigned int>& c, const QColor& col);
  void convertFromQColor(Color<unsigned long>& c, const QColor& col);
  void convertFromQColor(Color<unsigned long long>& c, const QColor& col);
  void convertFromQColor(Color<char>& c, const QColor& col);
  void convertFromQColor(Color<short>& c, const QColor& col);
  void convertFromQColor(Color<int>& c, const QColor& col);
  void convertFromQColor(Color<long>& c, const QColor& col);
  void convertFromQColor(Color<long long>& c, const QColor& col);

  /** \brief Assignment of color data. */
  template <class T>
    Color<T>& Color<T>::operator=(const Color<T>& c) {
      this->Vector<4,T>::operator=(c);
      return *this;
    }

  template <class T>
    Color<T>& Color<T>::operator=(const T& val) {
      this->Vector<4,T>::operator=(val);
      return *this;
    }

  template <class T>
    Color<T>& Color<T>::operator=(const Vector<4,T>& c) {
      this->Vector<4,T>::operator=(c);
      return *this;
    }

#ifdef USE_CXX0X
  template <class T>
    Color<T>& Color<T>::operator=(Vector<4,T>&& c) {
      this->Vector<4,T>::operator=(std::move(c));
      return *this;
    }
#endif

  /** \brief Return a color based on HSV values. */
  template <class T>
  Color<T> convertHSVtoRGB(T h, T s, T v) {
    // based on Jo's code in medit

    Color<T> rgb;
    rgb.a(1.0);

    while (h > 360.0) h -= 360.0;
    while (h < 0.0) h += 360.0;

    h /= 60.0;

    int i = int(h);

    double f = h - i;
    double p = v * (1 - s);
    double q = v * (1 - (s * f));
    double t = v * (1 - (s * (1 - f)));

    switch (i) {
    case 0: rgb.r(v); rgb.g(t); rgb.b(p); break;
    case 1: rgb.r(q); rgb.g(v); rgb.b(p); break;
    case 2: rgb.r(p); rgb.g(v); rgb.b(t); break;
    case 3: rgb.r(p); rgb.g(q); rgb.b(v); break;
    case 4: rgb.r(t); rgb.g(p); rgb.b(v); break;
    case 5: rgb.r(v); rgb.g(p); rgb.b(q); break;
    }

    return rgb;
  }

  typedef Color<float> Colorf;
  typedef Color<double> Colord;

}

#endif
