#ifndef __GRAPH_INSET_H
#define __GRAPH_INSET_H
/**
 * \file graph_inset.h
 *
 * Defines the util::GraphInset class.
 */

#include <config.h>
#include <vector>
#include <list>
#include <string>

#include <graph/vertex.h>
#include <util/vector.h>
#include <util/parms.h>

class Viewer;
class QString;

namespace util
{
  class Palette;

  /**
   * \class GraphInset graph_inset.h <util/graph_inset.h>
   *
   * Class used to draw a graph of time stamped data in the model window.
   */
  class GraphInset
  {
  public:
    /**
     * Data structure used to store the time stamped values
     */
    struct Data
    {
      Data(double t, const std::vector<double>& vs, bool c)
        : time(t)
        , values(vs)
        , changed(c)
      { }

      Data(const Data& copy)
        : time(copy.time)
        , values(copy.values)
        , changed(copy.changed)
      { }

      /**
       * Time stamp
       */
      double time;

      /**
       * Vector of values with one value per line
       */
      std::vector<double> values;

      /**
       * This value comes from a different source than the previous one
       */
      bool changed;
    };

    /**
     * Constant iterator on the stamped data
     */
    typedef std::list<Data>::const_iterator iterator;

    /**
     * Constant reverse iterator on the stamped data
     */
    typedef std::list<Data>::const_reverse_iterator reverse_iterator;

    /**
     * Type of the data
     */
    typedef Data value_type;
    /**
     * Type of a pointer on the data
     */
    typedef const Data* pointer;
    /**
     * Type of a reference on the data
     */
    typedef const Data& reference;

    /**
     * Construct a new graph inset
     *
     * \param nb_lines Number of lines to draw (i.e. nb of values given each 
     * time)
     */
    GraphInset(int nb_lines = 0);

    /**
     * Read the parameters for the graph.
     *
     * \param section Section to read the parameters from
     * \param parms util::Parms object to read the parameters from
     */
    void readParms(const QString& section, util::Parms parms);

    /**
     * Add time stamped values
     *
     * \param time Timestamp
     * \param values Values to add, must be in equal number as declared before
     * \param changed Indicate the source of the value changed (a line will be 
     * drawn at that time on the graph)
     */
    void addData( double time, const std::vector<double>& values, bool changed = false);

    /**
     * Helper to add time stamped values
     */
    void addData( double time, double value, bool changed = false)
    {
      std::vector<double> vs(1, value);
      addData(time, vs, changed);
    }

    /**
     * Helper to add time stamped values
     */
    void addData( double time, double value1, double value2, bool changed = false)
    {
      std::vector<double> vs(2, value1);
      vs[1] = value2;
      addData(time, vs, changed);
    }

    /**
     * Helper to add time stamped values
     */
    void addData( double time, double value1, double value2, double value3, bool changed = false)
    {
      std::vector<double> vs(3, value1);
      vs[1] = value2;
      vs[2] = value3;
      addData(time, vs, changed);
    }

    /**
     * Helper to add time stamped values
     */
    void addData( double time, double value1, double value2, double value3, double value4, bool changed = false)
    {
      std::vector<double> vs(4, value1);
      vs[1] = value2;
      vs[2] = value3;
      vs[3] = value4;
      addData(time, vs, changed);
    }

    /**
     * Helper to add time stamped values
     */
    void addData( double time, double value1, double value2, double value3, double value4, double value5, bool changed = false)
    {
      std::vector<double> vs(5, value1);
      vs[1] = value2;
      vs[2] = value3;
      vs[3] = value4;
      vs[4] = value5;
      addData(time, vs, changed);
    }

    /**
     * Add time stamped values
     */
    template <size_t N>
    void addData( double time, const util::Vector<N,double>& values, bool changed = false)
    {
      std::vector<double> vec_values(N);
      for(size_t i = 0 ; i < N ; ++i)
      {
        vec_values[i] = values[i];
      }
      addData(time, vec_values, changed);
    }

    /**
     * Add time stamped values
     */
    void addData( double time, const double* values, bool changed = false)
    {
      std::vector<double> vec_values(_numberOfLines);
      for(size_t i = 0 ; i < _numberOfLines ; ++i)
      {
        vec_values[i] = values[i];
      }
      addData(time, vec_values, changed);
    }

    void dataSourceChanged();

    /**
     * Remove all data that won't be drawn because they are older than the time 
     * span of the graph.
     */
    void cleanData();

    /**
     * Draw the graph in the viewer
     */
    void draw(Viewer* viewer, const util::Palette& palette);

    /**
     * Draw the graph in the viewer with a given name
     *
     * Used for selection purposes. Useful if the background is selected as 
     * transparent.
     */
    void drawWithName(size_t id, Viewer* viewer);

    /**
     * Draw the frame only in the viewer.
     *
     * Used for selection purposes.
     */
    void drawFrame(size_t id, Viewer* viewer);

    /**
     * Order the graph to close
     */
    void closeGraph();

    /**
     * Order the graph to open up
     */
    void openGraph();

    /**
     * Tell if the graph is closed
     */
    bool graphClosed() const { return closed; }

    /**
     * Kepp the graph informed of the current time in the system.
     */
    void setTime(double t) { time = t; }

    /**
     * Remove all data from the graph
     */
    void clear();

    /**
     * Returns the numbers of lines drawn
     */
    int numberOfLines() const { return (int)_numberOfLines; }

    /**
     * Change the number of lines
     *
     * If the number of lines is actually changed, then the system is reset
     */
    void setNumberOfLines(size_t nb_lines);

    /**
     * Check if the object is correctly initialized.
     *
     * This function must be called before using the object. It is 
     * automatically called within readParms.
     */
    void check();

    /**
     * Write the data in a .csv file
     */
    bool writeCSV(const QString& filename);

    ///\name STL-like methods
    //@{
    /**
     * Number of values stored
     */
    size_t size() const { return data.size(); }

    iterator begin() const { return data.begin(); }

    iterator end() const { return data.end(); }

    reverse_iterator rbegin() const { return data.rbegin(); }

    reverse_iterator rend() const { return data.rend(); }

    reference front() const { return data.front(); }

    reference back() const { return data.back(); }
    //@}

  protected:
    /**
     * Number of lines drawn by the graph
     */
    size_t _numberOfLines;

    /**
     * Check the system is correctly initialized.
     */
    bool initialized;

    /**
     * State if the graph window should be closed.
     */
    bool closed;

    /**
     * Data to be draw
     */
    std::list<Data> data;

  public:

    /**
     * Time span to keep the data
     */
    double timeSpan;

    /**
     * Color of the back of the graph
     *
     * If negative, the background will be transparent
     */
    int backColor;

    /**
     * Color of the lines
     */
    std::vector<int> linesColor;

    /**
     * Width of the lines
     */
    std::vector<double> linesWidth;

    /**
     * Extrema value shown for each line
     */
    std::vector<util::Vector<2, double> > linesMinMaxView;

    /**
     * Color of the contour
     */
    int contourColor;

    /**
     * Width of the line used for the contour
     */
    int contourWidth;

    /**
     * Position of the top-left corner of the graph window in pixels
     *
     * (0,0) is the top-left corner of the screen.
     *
     * Negative values are considered distance from the bottom-right corner.
     * Positives and negatives values can be mixed (i.e. (-100,30) = 100 pixel 
     * from the right, 30 pixels from the top)
     */
    util::Vector<2,int> windowPosition;

    /**
     * Size of the graph window in pixels
     */
    util::Vector<2,size_t> windowSize;

    /**
     * Color of the line used to indicate a data change (typically if the data 
     * correspond to another cell)
     */
    int changeSourceColor;

    /**
     * Width of the line used to indicate a data change (typically if the data 
     * correspond to another cell)
     */
    int changeSourceWidth;

    /**
     * Offset to draw the background of the graph
     */
    double backgroundOffset;

    /**
     * Offset to draw the contour of the graph
     */
    double contourOffset;

    /**
     * Offset to draw the data source change line
     */
    double changeSourceOffset;

    /**
     * Offset to draw the lines
     */
    std::vector<double> linesOffset;

    /**
     * Time of the system
     */
    double time;

    /**
     * Order the object to keep all the data
     */
    bool keepAllData;

    /**
     * Automatically open/close the graph.
     *
     * Graph is opened as soon as some data is added to it.
     *
     * Graph is closed when the data set is empty.
     */
    bool autoOpenClose;

    /**
     * If true, show all the data
     */
    bool showAll;

    /**
     * If not showing all data, contains the indices to show
     */
    std::vector<size_t> showIndices;

  };
}

#endif // GRAPH_INSET_H

