#ifndef VVLIB_UTIL_MEMBER_ITERATOR_H
#define VVLIB_UTIL_MEMBER_ITERATOR_H
/**
 * \file member_iterator.h
 *
 * Defines the util::SelectMemberIterator class template
 */

#include <config.h>
#include <iterator>

namespace util
{
  /**
   * \class SelectMemberIterator member_iterator.h <util/member_iterator.h>
   *
   * \brief Iterate over a container of structure, dereferencing only a member 
   * of it.
   *
   * This iterator is used when iterating over a container of a complexe 
   * structure. It allows for dereferencing only a member of that structure. 
   * A typical example is iterating over a map but dereferencing either the key 
   * or the value:
   * \code
   * std::map<int,double> my_map;
   * typedef std::map<int,double>::iterator map_iterator;
   * typedef std::map<int,double>::value_type map_value;
   * util::SelectMemberIterator<map_iterator,const int, &map_value::first> key_iterator;
   * util::SelectMemberIterator<map_iterator,int,&map_value::second> data_iterator;
   * util::SelectMemberIterator<map_iterator,int,&map_value::second,const int&, const int*> data_const_iterator;
   * // Fill in my_map
   * for(key_iterator it = my_map.begin() ; it != my_map.end() ; ++it)
   * {
   *   cout << *it << endl; // Output the keys !
   * }
   * for(data_iterator it = my_map.begin() ; it != my_map.end() ; ++it)
   * {
   *   *it += 5;
   *   cout << *it << endl; // Output the data !
   * }
   * for(data_const_iterator it = my_map.begin() ; it != my_map.end() ; ++it)
   * {
   *   // It is not possible to modify the value pointed
   *   cout << *it << endl; // Output the data !
   * }
   * \endcode
   */
  template <class Iterator, class T,
           T std::iterator_traits<Iterator>::value_type::* member,
           class Reference = T&, class Pointer = T*>
  struct SelectMemberIterator
  {
    /**
     * Type of the underlying iterator
     */
    typedef Iterator base_iterator;
    /**
     * Category of the iterator
     */
    typedef typename std::iterator_traits<Iterator>::iterator_category iterator_category;
    /**
     * Type of the value iterated on
     */
    typedef T value_type;
    /**
     * Type of the difference between two iterators
     */
    typedef typename std::iterator_traits<Iterator>::difference_type difference_type;
    /**
     * Type of a reference on the values
     */
    typedef Reference reference;
    /**
     * Type of a pointer on the values
     */
    typedef Pointer pointer;

    /**
     * Default constructor
     */
    SelectMemberIterator() {}

    /**
     * Conversion from the base iterator
     */
    SelectMemberIterator(const base_iterator& i)
      : it(i)
      { }

    /**
     * Copy constructor
     */
    SelectMemberIterator(const SelectMemberIterator& other)
      : it(other.it)
      {}

#ifdef USE_CXX0X
    SelectMemberIterator(SelectMemberIterator&& other) = default;
#endif

    ///\name 1 - Forward/input iterator methods
    //@{

    bool operator==(const SelectMemberIterator& other) const
    { return it == other.it; }

    bool operator!=(const SelectMemberIterator& other) const
    { return it != other.it; }

    /**
     * Assignment operator
     */
    SelectMemberIterator& operator=(const SelectMemberIterator& copy)
    { it = copy.it; return *this; }

#ifdef USE_CXX0X
    SelectMemberIterator& operator=(SelectMemberIterator&& other) = default;
#endif

    /**
     * Prefix increment operator
     */
    SelectMemberIterator& operator++() { ++it; return *this; }
    /**
     * Postfix increment operator
     */
    SelectMemberIterator& operator++(int) { SelectMemberIterator tmp(*this); ++it; return tmp; }

    /**
     * Dereference operator
     */
    reference operator*() { return (*it).*member; }
    /**
     * Constant dereference operator
     */
    const reference operator*() const { return (*it).*member; }

    /**
     * Pointer-like arrow operator
     */
    pointer operator->() { return &((*it).*member); }
    /**
     * Pointer-like constant arrow operator
     */
    const pointer operator->() const { return &((*it).*member); }
    //@}

    ///\name 2 - Bidirectional iterator methods
    //@{
    /**
     * Prefix decrement operator
     */
    SelectMemberIterator& operator--() { --it; return *this; }
    /**
     * Postfix decrement operator
     */
    SelectMemberIterator& operator--(int) { SelectMemberIterator tmp(*this); --it; return tmp; }
    //@}

    ///\name 3 - Random access methods
    //@{
    /**
     * In-place random increment operator
     */
    SelectMemberIterator& operator+=(difference_type n)
    { it += n; return *this; }

    /**
     * In-place random decrement operator
     */
    SelectMemberIterator& operator-=(difference_type n)
    { it -= n; return *this; }

    bool operator<(const SelectMemberIterator& other) const
    { return it < other.it; }

    bool operator>(const SelectMemberIterator& other) const
    { return it > other.it; }

    bool operator<=(const SelectMemberIterator& other) const
    { return it <= other.it; }

    bool operator>=(const SelectMemberIterator& other) const
    { return it >= other.it; }
    //@}

    ///\name Functions for random access iterators
    //@{
    friend SelectMemberIterator operator+(const SelectMemberIterator& it, difference_type n)
    { SelectMemberIterator tmp(it); tmp += n; return tmp; }

    friend SelectMemberIterator operator+(difference_type n, const SelectMemberIterator& it )
    { SelectMemberIterator tmp(it); tmp += n; return tmp; }

    friend SelectMemberIterator operator-(const SelectMemberIterator& it, difference_type n )
    { SelectMemberIterator tmp(it); tmp -= n; return tmp; }

    friend SelectMemberIterator operator-(difference_type n, const SelectMemberIterator& it )
    { SelectMemberIterator tmp(it); tmp -= n; return tmp; }

    /**
     * Distance between two iterators
     */
    friend difference_type operator-(const SelectMemberIterator& last, const SelectMemberIterator& first)
    { return last.it - first.it; }
    //@}

    /**
     * Direct access to the base iterator
     */
    base_iterator base() const { return it; }

  protected:
    /**
     * Underlying iterator
     */
    base_iterator it;
  };

  template <typename MapType>
  struct KeyIterator
  {
   typedef SelectMemberIterator<typename MapType::iterator,
                                const typename MapType::key_type,
                                &MapType::value_type::first> type;
  };

  template <typename MapType>
  struct MappedIterator
  {
   typedef SelectMemberIterator<typename MapType::iterator,
                                const typename MapType::mapped_type,
                                &MapType::value_type::second> type;
  };

}

#endif // VVLIB_UTIL_MEMBER_ITERATOR_H

