#ifndef VVELIB_UTIL_RANDOM_H
#define VVELIB_UTIL_RANDOM_H
/**
 * \file random.h
 *
 * Defines various functions to generate random numbers
 */

#include <config.h>
#include <util/vector.h>

namespace util
{
  /**
   * Initialize the random number with the current time of the day (in 
   * microsecond)
   *
   * \returns the seed used.
   */
  unsigned int sran_time();

  /**
   * Initialize the random number generator
   */
  void sran(unsigned int seed);

  /**
   * Generate a random number uniformely distributed between 0 and M
   */
  double ran(double M);

  /**
   * Generate a random vector uniformely distributed between Vector(0) and V
   */
  template <size_t dim>
  Vector<dim,double> ran(const Vector<dim,double>& V)
  {
    Vector<dim,double> result;
    for(size_t i = 0 ; i < dim ; ++i)
      result[i] = ran(V[i]);
    return result;
  }

  /**
   * Generate a random number with gaussian distribution
   *
   * \param mean Mean of the gaussian distribution
   * \param sigma Standard deviation of the gaussian distribution
   */
  double gaussRan(double mean, double sigma);

  /**
   * Generate a random vector with gaussian distribution
   *
   * \param mean Vector with mean of the gaussian distribution for each 
   * dimension
   * \param sigma Vector with standard deviation of the gaussian distribution 
   * for each dimension
   */
  template <size_t dim>
  Vector<dim,double> gaussRan(const Vector<dim,double>& mean, const Vector<dim,double>& sigma)
  {
    Vector<dim,double> result;
    for(size_t i = 0 ; i < dim ; ++i)
      result[i] = gaussRan(mean[i], sigma[i]);
    return result;
  }

  /**
   * Returns a random number between 0 and RAND_MAX.
   */
  long int random();

  /**
   * Returns a vector of random numbers between 0 and RAND_MAX.
   */
  template <size_t dim>
  Vector<dim,long int> random()
  {
    Vector<dim,double> result;
    for(size_t i = 0 ; i < dim ; ++i)
      result[i] = random();
    return result;
  }

  /**
   * Returns a random number between 0 and n (excluded), for n <= RAND_MAX
   */
  long int random(long int n);

  /**
   * Returns a vector of random numbers
   * \param n Vector with the maximum number for each dimension
   */
  template <size_t dim>
  Vector<dim,long int> random(const Vector<dim,long int>& n)
  {
    Vector<dim,double> result;
    for(size_t i = 0 ; i < dim ; ++i)
      result[i] = random(n[i]);
    return result;
  }
};


#endif // VVELIB_UTIL_RANDOM_H

