#ifndef __UTIL_RANGE_HPP__
#define __UTIL_RANGE_HPP__
/**
 * \file util/range.h
 *
 * Defines the range util template and various functions to deal with numerical range.
 */

#include <config.h>
#include <utility>
#include <iterator>

namespace util {

  /**
   * \class range range.h <util/range.h>
   *
   * Class representing a range of values from two iterators
   */
  template <typename Iterator>
  struct range
  {
    typedef Iterator iterator;
    typedef Iterator const_iterator;
    typedef typename std::iterator_traits<iterator>::difference_type difference_type;

    /**
     * Create an invalid range
     */
    range()
      : first()
      , last()
    {}

    /**
     * Range from two iterators
     */
    range(const iterator& fst,
          const iterator& lst)
      : first(fst)
      , last(lst)
    {}

    /**
     * Range from a pair of iterators
     */
    explicit range(const std::pair<iterator,iterator>& p)
      : first(p.first)
      , last(p.second)
    {}

    /**
     * Copy constructor
     */
    range(const range& copy)
      : first(copy.first)
      , last(copy.last)
    {}

    /**
     * Conversion constructor
     */
    template <typename It>
    explicit range(const range<It>& copy)
      : first(copy.first)
      , last(copy.last)
    { }

#ifdef USE_CXX0X
    /**
     * Move constructor
     */
    range(range&& copy)
      : first(std::move(copy.first))
      , last(std::move(copy.last))
    { }

    /**
     * Move constructor from another type
     */
    template <typename It>
    range(range<It>&& copy)
      : first(std::move(copy.first))
      , last(std::move(copy.last))
    {}
#endif

    /**
     * Begin of the range
     */
    iterator begin() const { return first; }
    /**
     * Past-the-End of the range
     */
    iterator end() const { return last; }

    /**
     * Begin of the range
     */
    const_iterator cbegin() const { return first; }
    /**
     * Past-the-End of the range
     */
    const_iterator cend() const { return last; }

    /**
     * Set the start of the range
     */
    void setBegin(const iterator& it) { first = it; }
    /**
     * Set the past-the-end iterator for the range
     */
    //void setEnd(const iterator& it) { end = it; }
    // Pascal's hack to make it compile
    void setEnd(const iterator& it) { last = it; }

    /**
     * Return the size of the range.
     *
     * Complexity depends on the iterator.
     */
    difference_type size() const { return std::distance(first, last); }

    /**
     * Assignment operator
     */
    range& operator=(const range& other)
    {
      first = other.first;
      last = other.last;
      return *this;
    }

    /**
     * Assignment operator
     */
    template <typename It>
    range& operator=(const range<It>& other)
    {
      first = other.first;
      last = other.last;
      return *this;
    }

#ifdef USE_CXX0X
    /**
     * Assignment operator
     */
    range& operator=(range&&) = default;

    /**
     * Assignment operator
     */
    template <typename It>
    range& operator=(range<It>&& other)
    {
      first = std::move(other.first);
      last = std::move(other.last);
      return *this;
    }
#endif

    iterator first, last;
  };

  /**
   * Function to create a range from two iterators
   */
  template <typename Iterator>
  range<Iterator> make_range(const Iterator& first, const Iterator& last)
  {
    return range<Iterator>(first, last);
  }

  /**
   * Function to create a range from a pair of iterators
   */
  template <typename Iterator>
  range<Iterator> make_range(const std::pair<Iterator,Iterator>& p)
  {
    return range<Iterator>(p);
  }

  /**
   * \brief Check if a value is in an closed range.
   * \param min The minimum range value.
   * \param max The maximum range value.
   * \param val The value to check.
   */
  template <typename T> bool range_closed(const T& min, const T& max, const T& val) {
    return (val >= min && val <= max);
  }

  /**
   * \brief Check if a value is in an open range.
   * \param min The minimum range value.
   * \param max The maximum range value.
   * \param val The value to check.
   */
  template <typename T> bool range_open(const T& min, const T& max, const T& val) {
    return (val > min && val < max);
  }

  /**
   * \brief Check if a value is in a C-style range (closed on lower bound and 
   * open on the upper)
   * \param min The minimum range value.
   * \param max The maximum range value.
   * \param val The value to check.
   */
  template <typename T> bool range_c(const T& min, const T& max, const T& val) {
    return (val >= min && val < max);
  }
}

#endif
