#ifndef __UTIL_SET_VECTOR_H
#define __UTIL_SET_VECTOR_H

#include <vector>
#include <functional>
#include <algorithm>

namespace util
{

  template <typename T, typename Equal>
  struct __set_vector_FindEqual
  {
    __set_vector_FindEqual(const T& t)
      : value(t)
      {}

    bool operator()(const T& other) const
    {
      return equal(value, other);
    }

    const T& value;
    static const Equal equal;
  };

  template <typename T, typename Equal>
    const Equal __set_vector_FindEqual<T,Equal>::equal = Equal();



  /**
   * Define a class using (small) vectors as unordered sets
   */
  template <typename T, typename Equal = std::equal_to<T>, typename Alloc = std::allocator<T> >
    struct set_vector : public std::vector<T,Alloc>
    {
      typedef std::vector<T,Alloc> base;
      typedef typename base::value_type key_type;
      typedef typename base::value_type value_type;
      typedef typename base::reference reference;
      typedef typename base::const_reference const_reference;
      typedef typename base::difference_type difference_type;
      typedef typename base::size_type size_type;
      typedef typename base::allocator_type allocator_type;
      typedef typename base::pointer pointer;
      typedef typename base::const_pointer const_pointer;

      typedef Equal key_equal;
      typedef typename base::iterator iterator;
      typedef typename base::const_iterator const_iterator;

      typedef typename base::reverse_iterator reverse_iterator;
      typedef typename base::const_reverse_iterator const_reverse_iterator;

      set_vector() : base() {}

      set_vector(const set_vector& copy) : base(copy) {}

      set_vector(const base& copy) : base(copy) {}

#ifdef USE_CXX0X
      set_vector(set_vector&&) = default;
      set_vector& operator=(set_vector&&) = default;

      set_vector(base&& copy)
        : base(std::move(copy))
      { }

      set_vector& operator=(base&& other)
      {
        base::operator=(std::move(other));
        return *this;
      }
#endif

      std::pair<iterator,bool> insert(const T& v, bool test_presence = true)
      {
        if(test_presence)
        {
          iterator it_found = this->find(v);
          if(it_found != this->end())
            return std::make_pair(it_found, false);
        }
        this->push_back(v);
        iterator e = this->end();
        --e;
        return std::make_pair(e, true);
      }

      int erase(const T& v)
      {
        iterator it_found = this->find(v);
        if(it_found != this->end())
        {
          this->erase(it_found);
          return true;
        }
        return false;
      }

      iterator erase(iterator it)
      {
        return base::erase(it);
      }

      int count(const T& v) const
      {
        if(this->find(v) != this->end())
          return 1;
        return 0;
      }

      iterator find(const T& v)
      {
        return std::find_if(this->begin(), this->end(), __set_vector_FindEqual<value_type,key_equal>(v));
      }

      const_iterator find(const T& v) const
      {
        return std::find_if(this->begin(), this->end(), __set_vector_FindEqual<value_type,key_equal>(v));
      }

      set_vector& operator=(const set_vector& other)
      {
        base::operator=(other);
        return *this;
      }
    };

}

#endif // __UTIL_SET_VECTOR_H

