#ifndef STATIC_ASSERT_HPP
#define STATIC_ASSERT_HPP
/**
 * \file static_assert.h
 *
 * Define the STATIC_ASSERT macro
 */

#include <config.h>
namespace util
{

#ifndef GXX_USE_GNUXX0X

  // Extracted from boost static_asset.h ...
  template <bool x> struct STATIC_ASSERTION_FAILURE;
  template <> struct STATIC_ASSERTION_FAILURE<true> { enum { value = 1 }; };
  template<int x> struct static_assert_test{};

#define JOIN( X, Y ) DO_JOIN( X, Y )
#define DO_JOIN( X, Y ) DO_JOIN2(X,Y)
#define DO_JOIN2( X, Y ) X##Y

  /**
   * \def STATIC_ASSERT
   * Assertion that works at compile time.
   *
   * If the parameters evaluate to false, the program won't compile. Useful to 
   * check that a template method is not used when it is invalid.
   *
   * A typical example can be found in the implementation of the util::Vector 
   * class to make sure the constructor with n (fixed) values are used only if 
   * the vector is of dimension n:
   * \code
   *   template <size_t N, typename T>
   *   Vector<N,T>::Vector( const T& v1, const T& v2)
   *   {
   *    STATIC_ASSERT(N == 2); // Compilation will fail if N != 2
   *    // ...
   *   }
   * \endcode
   *
   * \note This will work because, as specified by the standard, 
   * a (non-virtual) method of class template is instanciated only if used. So, 
   * as long as the user do not try to construct a vector with two values, this 
   * specific constructor won't even exist.
   */
#  define STATIC_ASSERT(B, txt) \
   typedef util::static_assert_test<\
      sizeof(util::STATIC_ASSERTION_FAILURE< (bool)( B ) >)>\
         JOIN(static_assert_typedef_, __LINE__)

#else

#  define STATIC_ASSERT(B, txt) static_assert(B, txt)

#endif

}

#endif // STATIC_ASSERT_HPP
