#ifndef VVLIB_UTIL_TIE_H
#define VVLIB_UTIL_TIE_H
/**
 * \file tie.h
 *
 * Defines the util::tie function
 */

#include <config.h>
#include <utility>
/**
 * Various utility classes for generic programming
 */
namespace util
{

  // Code taken from the Boost library !
  /**
   * \class refpair tie.h <util/tie.h>
   *
   * Class used to hold references for the util::tie() function.
   */
  template< typename T, typename U >
    struct refpair
    {
      typedef T first_type;
      typedef U second_type;

      /// Construct a pair of references to \c x and \c y.
      refpair( T& x, U& y ) : first(x), second(y) {}
      /// Construct a copy.
      refpair(refpair const& rp) : first(rp.first), second(rp.second) {}
#ifdef USE_CXX0X
      refpair(refpair&&) = default;
#endif
      /// Assign the values of \c p to the references in this pair.
      refpair& operator=( std::pair<first_type, second_type> const& p )
      {
        first = p.first;
        second = p.second;
        return *this;
      }

#ifdef USE_CXX0X
      refpair& operator=( std::pair<first_type, second_type> && p)
      {
        first = std::move(p.first);
        second = std::move(p.second);
        return *this;
      }
#endif

      /// The first member of the pair.
      T& first;
      /// The second member of the pair.
      U& second;
    };

  /**
   * Tie two variables to the values of a pair.
   *
   * Example:
   * \code
   * std::pair<int,double> p(1,2.5);
   * int a;
   * double b;
   * util::tie(a,b) = p;
   * \endcode
   *
   * At the end, \c a is \c 1 and \c b is \c 2.5
   */
  template< typename T, typename U >
    inline refpair<T, U> tie( T& x, U& y ) { return refpair<T, U>(x, y); }

};

#endif // VVLIB_UTIL_TIE_H

