#ifndef WATCHDOG_H
#define WATCHDOG_H
/**
 * \file watchdog.h
 *
 * Defines the util::WatchDog and util::FileObject classes
 */

#include <config.h>
#include <string>
#include <set>
#include <map>
#include <QObject>
#include <QString>

#include <util/forall.h>

namespace util
{

  /**
   * \class FileObject watchdog.h <util/watchdog.h>
   *
   * This class is the base class for any object that might be watched by the 
   * WatchDog.
   *
   * Anytime the FileObject::update() method is called, the 
   * FileObject::modified() signal is emuitted by the object.
   */
  class FileObject : public QObject
  {
    Q_OBJECT

  public:
    /**
     * Default constructor
     *
     * Just initialize the QObject.
     */
    FileObject(QObject *parent = 0) : QObject(parent) {}
    /**
     * Initialized constructor
     *
     * \param fn Name of the file handled by this object
     */
    FileObject(std::string fn, QObject *parent = 0) : QObject(parent), filename(fn) {}
    /**
     * Initialized constructor
     *
     * \param fn Name of the file handled by this object
     */
    FileObject(QString fn, QObject *parent = 0) : QObject(parent), filename(fn.toStdString()) {}
    /**
     * Copy constructor
     *
     * \param copy Object to copy
     */
    FileObject(const FileObject& copy, QObject *parent = 0) : QObject(parent), filename(copy.filename) {}
    /**
     * Assignment operator
     *
     * \param other Object to copy
     *
     * If the object to copy handle a different filename, then the 
     * FileObject::modified() signal is emitted.
     */
    FileObject& operator=(const FileObject& other)
    { if(other.filename != filename) { filename = other.filename; } return *this; }

#ifdef USE_CXX0X
    FileObject(FileObject&& other, QObject * parent = 0)
      : QObject(parent)
      , filename(std::move(other.filename))
    { }

    FileObject& operator=(FileObject&& other)
    {
      filename = std::move(other.filename);
      return *this;
    }
#endif

    /**
     * Method to be implemented in subblasses to reread the file.
     */
    virtual void reread() = 0;

    /**
     * Get the current filename of the object
     */
    std::string getFilename() { return filename; }

    /**
     * Change the file handled
     *
     * If the call actually change the file handled, then 
     * FileObject::modified() is emitted.
     */
    void setFilename(std::string fn);

    /**
     * Change the file handled
     *
     * If the call actually change the file handled, then 
     * FileObject::modified() is emitted.
     */
    void setFilename(QString fn) { setFilename(fn.toStdString()); }

    /**
     * Change the file handled
     *
     * If the call actually change the file handled, then 
     * FileObject::modified() is emitted.
     */
    void setFilename(const char* fn) { setFilename(std::string(fn)); }

    void update();

  signals:
    /**
     * Signal emitted to indicate a change in the object
     */
    void modified();

  protected:
    std::string filename;
  };

  /**
   * \class WatchDog watchdog.h <util/watchdog.h>
   *
   * Watch the modifications of file objects for you.
   */
  class WatchDog : public QObject
  {
    Q_OBJECT

  public:
    /**
     * Constructor
     *
     * \param m Pointer toward the model to watch files for
     */
    WatchDog(QObject* m);

    /**
     * Test if an object is being watched
     */
    bool guarded(FileObject *obj) const;
    bool guarded(FileObject& obj) const { return guarded(&obj); }

    /**
     * Watch a new object
     */
    void addObject(FileObject* obj);
    /**
     * Watch a new object
     */
    void addObject(FileObject& obj) { addObject(&obj); }
    /**
     * Change the filename of an object.
     *
     * \note When an object is handled by the WatchDog, you have to use this 
     * method to change its filename.
     */
    void changeFilename(FileObject* obj, std::string fn);
    /**
     * Do not watch this object anymore
     */
    void removeObject(FileObject *obj);
    /**
     * Do not watch this object anymore
     */
    void removeObject(FileObject &obj) { removeObject(&obj); }

    /**
     * Check if any of the file names correspond to one of the watched object
     *
     * \param filenames Set of file names that were recently modified
     *
     * If the WatchDog finds a file name it handles in the set, it emits 
     * a FileObject::modified() signal for this object.
     */
    bool watch(const std::set<std::string>& filenames);

  signals:
    void registerFile(std::string);
    void unregisterFile(std::string);

  protected:
    /**
     * Model to guard
     */
    QObject *model;
    /**
     * List of file objects to handle
     */
    std::set<FileObject*> fileObjects;
    /**
     * Map the file names to the file objects
     */
    std::map<std::string, std::set<FileObject*> > names;
  };

}
#endif // WATCHDOG_H

