#ifndef VVVIEWER_H
#define VVVIEWER_H

#include <config.h>
#include <QGLViewer/qglviewer.h>
#include <model.h>
#include <QPointer>

class QDomElement;
class QShowEvent;
class QToolBar;
class QMainWindow;
class QMenuBar;
class QMenu;
class QKeyEvent;
class QGLViewer;

/**
 * \brief ::Viewer widget
 *
 * This class handle the simulation window and the life of the simulation 
 * object.
 *
 * You might want to inherit this class to personalize the interactions. You 
 * might also want to look at the documentation of its base class.
 *
 * If you inherit from it, you have to register the derived class using the 
 * DEFINE_VIEWER(classname) macro defined in <vvall.h>.
 */
class Viewer : public QGLViewer
{
  Q_OBJECT

public:

  //@{
  /// \name Constructors/destructors
  /**
   * Creates the object within the parent.
   *
   * \param parent Enclosing widget.
   */
  Viewer( QWidget* parent, const QGLWidget *shareWidget = 0, int idx = -1  );
  /**
   * Destroy the viewer properly
   */
  ~Viewer();
  //@}

  static void initFormat();

  /**
   * Function returning a pointer onto the current model.
   */
  Model* model() { return _model; }

  //@{
  ///\name Internal methods
  /**
   * Function used to save the current state of the viewer (i.e. camera, ...)
   */
  virtual QDomElement domElement(const QString & name, QDomDocument &document) const;

  /**
   * Returns a description of the model to show on the help screen.
   *
   * \see Model::helpString()
   */
  virtual QString helpString() const
  {
    if(_model) return _model->helpString();
    else return QString(tr("Generic VV model"));
  }

  /**
   * Set the creation index to the position in the viewer list
   */
  void setIndex(int idx) { _index = idx; }

  //@}

  //@{
  /// \name Screenshot methods

  /**
   * Get the current screenshot file name
   *
   * \see QGLViewer::snapshotFileName
   */
  const QString & screenshotFileName() const
  { return QGLViewer::snapshotFileName(); }

  /**
   * Get the current screenshot format
   * \see QGLViewer::snapshotFormat
   */
  const QString & screenshotFormat() const
  { return QGLViewer::snapshotFormat (); }

  /**
   * Get the current screenshot counter
   *
   * \see QGLViewer::snapshotCounter
   */
  int screenshotCounter() const
  { return QGLViewer::snapshotCounter(); }

  /**
   * Get the current screenshot quality
   *
   * \see QGLViewer::snapshotQuality
   */
  int screenshotQuality()
  { return QGLViewer::snapshotQuality(); }

  /**
   * Exposing the drawLight method ...
   *
   * \see QGLViewer::drawLight
   */
  virtual void drawLight(GLenum light, float scale = 1.0f) { QGLViewer::drawLight(light, scale); }


  /**
   * Returns a pointer to the menu bar of the application.
   *
   * \param clear If true, remove any element from the menu bar but the help.
   */
  QMenuBar* menuBar(bool clear = true);

  /**
   * Returns a pointer to the help menu
   *
   * This is useful for two purposes:
   * 1 - Adding your own help elements
   * 2 - Placing other menus before the help
   */
  QMenu* helpMenu();

  /**
   * Returns true if this viewer has already been initialized
   */
  bool wasInitialized() const { return initialized; }

  /**
   * Return the creation index, or -1 if the viewer is not registered with the 
   * main GUI.
   *
   * This number can be used to identified what viewer is suppose to do what.
   */
  int index() const { return _index; }

  /**
   * Create a new viewer.
   *
   * The viewer created is boxed in a widget for easy replacement.
   */
  void newViewer(const QGLWidget* shareWidget = 0);

  /**
   * Create a new viewer.
   *
   * The viewer created is boxed in a widget for easy replacement.
   */
  void addViewer(QLayout *l, const QGLWidget* shareWidget = 0);

  /**
   * Create a new viewer.
   *
   * The viewer created is boxed in a widget for easy replacement.
   */
  void addViewer(QWidget *w, const QGLWidget* shareWidget = 0);

public slots:
  /**
   * Start recording a movie with the various frames
   */
  void startRecordingCameraAnimation();

  /**
   * Start recording a movie with the various frames
   */
  void stopRecordingCameraAnimation();

  /**
   * Start recording a movie with the various frames
   */
  void startRecordingModelAnimation();

  /**
   * Start recording a movie with the various frames
   */
  void stopRecordingModelAnimation();

  /**
   * Save a screenshot
   *
   * \see QGLViewer::saveSnapshot(bool,bool)
   */
  void saveScreenshot (bool automatic=true, bool overwrite=false)
  { QGLViewer::saveSnapshot(automatic, overwrite); }

  /**
   * Save a screenshot
   *
   * \see QGLViewer::saveSnapshot(const QString&, bool)
   */
  void saveScreenshot (const QString &fileName, bool overwrite=false)
  { QGLViewer::saveSnapshot(fileName, overwrite); }

  /**
   * Set the screenshot file name
   *
   * \see QGLViewer::setSnapshotFileName
   */
  void setScreenshotFileName (const QString &name)
  { QGLViewer::setSnapshotFileName(name); }

  /**
   * Set the screenshot format
   *
   * \see QGLViewer::setSnapshotFormat
   */
  void setScreenshotFormat (const QString &format)
  { QGLViewer::setSnapshotFormat(format); }

  /**
   * Set the screenshot counter
   *
   * \see QGLViewer::setSnapshotCounter
   */
  void setScreenshotCounter (int counter)
  { QGLViewer::setSnapshotCounter(counter); }

  /**
   * Set the screenshot quality
   *
   * \see QGLViewer::setSnapshotQuality
   */
  void setScreenshotQuality (int quality)
  { QGLViewer::setSnapshotQuality(quality); }

  /**
   * Open the screenshot format dialog
   *
   * \see QGLViewer::openSnapshotFormatDialog
   */
  bool openScreenshotFormatDialog ()
  { return QGLViewer::openSnapshotFormatDialog(); }

  //@}

  /**
   * This function registers the model creation function to the viewer.
   */
  virtual void setModel( Model* model );

  /**
   * Read a saved state
   */
  virtual void initFromDOMElement(const QDomElement &element );

  //@}

  //@{
  /// \name Snapshot methods

  void blockDraw(bool block=true) { if(block) block_draw++; else block_draw--; }

  //@}

  /// Open the dialog box to configure the OpenGL context
  void configOpenGL();

  /// Update first OpenGL, then the view
  void updateAll();

signals:
  //@{
  /// \name Internal signals
  /**
   * Signal emitted at the end of each simulation step (including the drawing)
   */
  void stepDrawFinished(bool);

  /**
   * Open OpenGL config dialog box
   */
  void openOpenGLConfig();

  /**
   * Signal sent to the GUI to start recording animations.
   */
  void modelRecordingStart();

  /**
   * Signal sent to the GUI to top recording animations.
   */
  void cameraRecordingStop();

  /**
   * Signal sent to the GUI to start recording animations.
   */
  void cameraRecordingStart();

  /**
   * Signal sent to the GUI to top recording animations.
   */
  void modelRecordingStop();

  /**
   * Signal sent to request a new viewer
   */
  void createViewer(const QGLWidget* shareWidget);

  /**
   * Signal sent to request a new viewer
   */
  void createViewer(QLayout *l, const QGLWidget* shareWidget);

  /**
   * Signal sent to request a new viewer
   */
  void createViewer(QWidget *w, const QGLWidget* shareWidget);

  //@}

public:

  //@{
  /// \name Method to redefine to personalize drawing or user interaction
  /**
   * Initialize the OpenGL context before drawing.
   *
   * \see Model::preDraw(Viewer*)
   */
  virtual void preDraw()
  { QGLViewer::preDraw(); if( _model and !block_draw ) _model->preDraw(this); }
  /**
   * Redraw the simulation result.
   *
   * \see Model::draw(Viewer*)
   */
  virtual void draw()
  { if( _model and !block_draw ) _model->draw(this); }
  /**
   * Restore the original state of the OpenGL context.
   *
   * \see Model::postDraw(Viewer*)
   */
  virtual void postDraw();

  /**
   * Initialize the viewer with a valid OpenGL context.
   *
   * \see Model::initDraw(Viewer*)
   */
  virtual void init();

  /**
   * Draw with object naming. Used for object selection.
   *
   * \see Model::drawWithNames(Viewer*)
   */
  virtual void drawWithNames()
  { if (_model) _model->drawWithNames(this); }

  /**
   * Handle post selection actions
   */
  virtual void postSelection(const QPoint& pos)
  { if(_model) _model->postSelection(pos, this); }

  //@}

  /**
   * Restore the state (including position) on the show event
   */
  void showEvent(QShowEvent* event);

  /**
   * Get the current status message.
   */
  QString statusMessage() const;

  /**
   * Change the status message.
   *
   * On the gui, display the message on the status bar. On batch mode, prints the message to the console.
   */
  void setStatusMessage(const QString& msg);

  /**
   * Creates a toolbar associated with the main window of the application
   */
  QToolBar* addToolBar(const QString& title);

  /**
   * Add a toolbar to the main window of the application
   */
  void addToolBar(Qt::ToolBarArea area, QToolBar* toolbar);

  /**
   * Returns the main window associated with the application
   */
  QMainWindow *mainWindow();

  /**
   * Define a new shortcut ...
   */
  void keyPressEvent(QKeyEvent *e);

private:
  QString fullSnapshotFileName();
  void initObject();
  QPointer<Model> _model;
  bool animation;
  bool initialized;
  qglviewer::Camera *lastInitCamera;

  size_t _index;
  void initCamera();
  int block_draw;
  int minFrameInterval;

  /**
   * Pointer to the QMainWindow
   */
  QMainWindow *main_window;
};

#endif // VVVIEWER_H

