// Implement surface and surface point class
#include <string>
#include <iostream>
#include <cstdio>
#include <cmath>

#include <util/point.hpp>
#include <util/function.hpp>
#include <util/contour.hpp>

#include "surface.h"

using std::cout;
using std::cerr;
using std::endl;

typedef util::Point<double> UtilPoint;

// Fix angle, make positive and < 2PI
double FixAngle(double a)
{
  if(a != a || a > PI * 10.0 || a < PI * -10.0)
    cerr << "Surface::FixAngle Error Bad angle " << a << endl;
  while(a < 0.0)
    a += PI * 2.0;
  while(a > PI * 2.0)
    a -= PI * 2.0;
  return(a);
}

// Find polar angle of a point
double PolarAngle(double x, double y)
{
  double angle = acos(x/sqrt(pow(x, 2) + pow(y, 2)));
  if(y < 0.0)
    angle = PIx2 - angle;
  return(angle);
}

// Return angle between two angles
double MidAngle(double a, double b)
{
  double s, l, d;

  if(a > b) {
    s = b; l = a;
  } else {
    s = a; l = b;
  }
  d = l - s;
  if(d > PI) {
    d = PI * 2.0 - d;
    l += d/2.0;
    if(l >= PI * 2.0)
      l -= PI * 2.0;
    return(l);
  } else
    return(s + d/2.0);
}

// Return Arclength
double SurfacePoint::ArcLength()
{
  return(arcLength);
}

// Return Polar Angle
double SurfacePoint::Angle()
{
  return(angle);
}

UtilPoint SurfacePoint::Pos()
{
  return(pos);
}

UtilPoint SurfacePoint::Normal()
{
  return(normal);
}

Surface::Surface(util::Parms &parms, string section)
{
  // Get Growth Type
  parms(section, "GrowthType", growthType);

  // Get Scaling constants
  parms(section, "SurfaceScale", surfaceScale);
  parms(section, "GrowthScale", growthScale);
  parms(section, "RootSearchMaxSteps", rootSearchMaxSteps);

  if(growthType == ARCLENGTH) {
    // For arc length 1st file is surface contour 2nd is arclength growth func
    parms(section, "SurfaceContour", surfaceContourFile);
    ArcLengthParam(surfaceContourFile, arcLV);
    
    // Integrate growth curve
    parms(section, "ArcLengthGrowth", arcLengthGrowthFile);
    IntegrateFunction(arcLengthGrowthFile, growthVA);
  
    cerr << "Surface::Surface:Info Arc Length Growth, " <<
                              "Total Arc Length " << dArc * SAMPLES << endl;
    cerr << "Surface::Surface:Info Sample size " << SAMPLES << endl;
  }
}

Surface::~Surface() 
{
}

// Zero a surface Point
void Surface::Zero(SurfacePoint &p)
{
  p.arcLength = p.angle = 0.0;
  CalcPos(p);
  CalcNormal(p);
}

// Find closest point on surface to a given surface Point (Newton)
bool Surface::SetPoint(SurfacePoint &p, SurfacePoint sp, UtilPoint cp)
{
  bool found = true;
  if(cp.length() <= 0.0)
    Zero(p);
  else {
    // Zip close point around to same angle as given point and start from there
    p = sp;
    p.angle = PolarAngle(cp.x(), cp.z());
    UtilPoint zp(cp.x(), 0.0, cp.y());
    if(zp.length() <= DX) {
      p.arcLength = zp.length();
      CalcPos(p);
    } else {
      int max = rootSearchMaxSteps;
      CalcPos(p);
      UtilPoint pp;
      do {
        // Save previous position
        pp = p.Pos();

        // Calculate normal
        SurfacePoint n, s;
        n.arcLength = p.arcLength - DX;
        if(n.arcLength < 0.0)
          n.arcLength = 0.0;
        s.arcLength = p.arcLength + DX;
        n.angle = s.angle = p.angle;
        CalcPos(n); CalcPos(s);
        UtilPoint pn = (s.Pos() - n.Pos());
        pn.normalise();

        // Get projection onto normal
        double darc = pn * (cp - p.Pos());

        // Exit is small enough
        if(fabs(darc) < DX * 10)
          break;

        // Set new arc length
        p.arcLength += darc;

        // Set to 0 if hit tip
        if(p.arcLength <= 0.0) {
          Zero(p);
          break;
        }

        // Calc new position
        CalcPos(p);
      } while(--max > 0);

      // Print error if hit max iterations
      if(max <= 0) {
        //cout << "Surface::SetPoint:Error failed to find root " << cp << endl;
        found = false;
      }
    }
    CalcNormal(p);
  }
  return(found);
}

// Grow surface point 
void Surface::Grow(SurfacePoint &p, double dt)
{
  int i;

  if(p.arcLength > 0.0) { // don't grow apex tip
    i = (int)(p.arcLength/dArc);
    if(i > SAMPLES)
      i = SAMPLES;

    p.arcLength += growthVA[i] * dt * growthScale;
  }
  CalcPos(p);
  CalcNormal(p);
}

// Return distance between u and v
double Surface::Distance(SurfacePoint &u, SurfacePoint &v)
{
  // For now use Euclidean distance
  return(u.pos.distance(v.pos));
}

// Reparamaterize contour by arclength
void Surface::ArcLengthParam(string contFile, UtilPoint *v) {
  util::Contour surfCont(contFile);
  UtilPoint curr, prev, last;
  double totArcLen = 0.0;

  // First calculate arc length of contour based on initial paramaterization
  prev = surfaceScale * surfCont((double)0.0);
  for(int i = 1; i <= SAMPLES; i++) {
    curr = surfaceScale * surfCont((double)i/(double)SAMPLES);
    totArcLen += prev.distance(curr);
    prev = curr;
  }
    
  // re-parameterize by arc length
  dArc  = totArcLen/(double)SAMPLES;
  v[0] = prev = curr = surfaceScale * surfCont((double)0.0);
    
  { // MS C++ requires this
    int i = 1, j = 1;
    for(; i <= SAMPLES; i++) {
      double dis;
      while((dis = v[i - 1].distance(curr)) < dArc && j <= SAMPLES)
        curr = surfaceScale * surfCont((double)j++/(double)SAMPLES);
      
      if(j >= SAMPLES)
        break;
    
      v[i] = (v[i - 1] * (dis - dArc) + curr * dArc)/dis;
    }
    
    last = v[i - 1];
    dPos = v[i - 1] - v[i - 2];
    j = i - 1;
    
    for(; i <= SAMPLES; i++)
      v[i] = last + dPos * (double)(i - j);
  }
    
  // Scale dPos to 1.0
  dPos /= dArc;
    
  // translate so start of curve is at 0,0,0
  UtilPoint v0 = v[0];
  for(int i = 0; i <= SAMPLES; i++)
    v[i].set(v[i].x() - v0.x(), v[i].y() - v0.y(), v[i].z());
    //v[i].set(v[i].x() - minx, v[i].y() - miny, v[i].z());
    
// Not sure how to do this now
//  SurfacePoint pt0, pt1;
//  SetPoint(pt0, 0.0, 0.0);
//  SetPoint(pt0, 0.0, dArc * (double)SAMPLES);
//  
//  // Check x value to see if contour backwards
//  if(pt0.Pos().x() > pt1.Pos().x())
//    cerr << "Surface::ParamArcLength:Warning " << contFile << 
//                                                  " curve backwards?" << endl;
}

void Surface::LoadFunction(string funcFile, double *v)
{
  util::Function func(funcFile);
  double dx = 1.0/SAMPLES, x = 0;
  int negy = 0;
  
  for(int i = 0; i <= SAMPLES; i++, x += dx) {
    v[i] = func(x);
    if(v[i] < 0.0)
      negy = 1;
  }

  if(negy)
    cerr << "Surface::LoadFunction:Warning " << funcFile << 
                                                " goes negative" << endl;
}

void Surface::IntegrateFunction(string funcFile, double *v)
{
  util::Function func(funcFile);
  double dx = 1.0/SAMPLES, x = 0, sum = 0, y;
  int negy = 0;
  
  for(int i = 0; i <= SAMPLES; i++) {
    v[i] = sum;
    x += dx;
    y = func(x);
    if(y < 0.0)
      negy = 1;
    sum += dx * y;
  }

  if(negy)
    cerr << "Surface::IntegrateFunction:Warning " << funcFile << 
                                                 " goes negative" << endl;
}

// Calculate xyz postition
void Surface::CalcPos(SurfacePoint &p)
{
  // Get pos from table based on arclength
  if(p.arcLength <= 0.0)
    p.pos = arcLV[0];
  else {
    int i = (int)(p.arcLength/dArc);
    if(i >= SAMPLES - 1)
      i = SAMPLES - 1;
    
    // If exact use table value, otherwise extrapolate, or interpolate
    double diff = p.arcLength - (double)i * dArc;
    if(i >= SAMPLES - 1) // extrapolate, past end of curve
      p.pos = arcLV[SAMPLES - 1] + dPos * diff;
    else if(diff == 0.0) // exact
      p.pos = arcLV[i];
    else // interpolate
      p.pos = (arcLV[i] * (dArc - diff) + arcLV[i + 1] * diff)/dArc;

    // Rotate based on angle
    if(p.angle > 0.0)
      p.pos.set(p.pos.x() * cos(p.angle), p.pos.y(), p.pos.x() * sin(p.angle));
  }
}

void Surface::CalcNormal(SurfacePoint &p)
{
  if(p.arcLength <= DX)
    p.normal.set(0.0, 1.0, 0.0);
  else {
    // Calc normal from surface.
    SurfacePoint n, s, e, w;
    n.arcLength = p.arcLength - DX; n.angle = p.angle;
    s.arcLength = p.arcLength + DX; s.angle = p.angle;
    e.arcLength = p.arcLength; e.angle = p.angle + DX;
    w.arcLength = p.arcLength; w.angle = p.angle - DX;
    CalcPos(n); CalcPos(s);
    CalcPos(e); CalcPos(w);
    
    p.normal = s.Pos() - n.Pos();
    p.normal = p.normal.cross(w.Pos() - e.Pos());
    p.normal.normalise();
  }
}
