// Implement surface and surface point class
#include "surface.h"
#include "bezier.h"

#include <string>
#include <iostream>

#include <cstdio>
#include <cmath>
#include <util/parms.hpp>
#include <util/point.hpp>
#include <util/texture.hpp>

using std::cout;
using std::cerr;
using std::endl;

typedef util::Point<double> UtilPoint;

// Clip to bounds
static double InBounds(double s, double min, double max) 
{
  if(s < min)
    return(min);
  else if(s > max)
    return(max);
  else
    return(s);
}

UtilPoint SurfacePoint::Pos()
{
  return(pos);
}

UtilPoint SurfacePoint::Normal()
{
  return(normal);
}

Surface::Surface(util::Parms &parms, string section)
{
  string surffile;

  time = 0;

  // Load surfaces 
  parms(section, "Surfaces", surfaces);
  parms(section, "SurfTimeScale", surfTimeScale);
  parms(section, "SurfMaxDist", surfMaxDist);
  for(int i = 0; i < surfaces; i++) {
    // Bezier surface
    ostringstream key;
    key << "Surface" << i;
    parms(section, key.str(), surffile);
    surface[i].Load(surffile);

    // Scaling constant
    key.str("");
    key << "SurfaceScale" << i;
    parms(section, key.str(), surfScale[i]);

    // Time constant
    key.str("");
    key << "SurfaceTime" << i;
    parms(section, key.str(), surfTime[i]);
  }
}

// Zero a surface Point
void Surface::Zero(SurfacePoint &p)
{
  p.u = p.v = 0.0;
  CalcPos(p);
  CalcNormal(p);
}

// Initial surface Point
void Surface::InitPoint(SurfacePoint &p, double u, double v)
{
  p.u = InBounds(u, 0.0, 1.0);
  p.v = InBounds(v, 0.0, 1.0);
 
  CalcPos(p);
  CalcNormal(p);
}

// Find closest point on surface to a given surface point
bool Surface::SetPoint(SurfacePoint &p, SurfacePoint sp, UtilPoint cp)
{
  // Initial guess for p
  p.u = sp.u;
  p.v = sp.v;
  CalcPos(p);
  double lastd = -(DX * 1000);
  double count = 0;
  double du, dv;

  while(fabs(lastd - cp.distance(p.pos)) > DX * 2 && count++ < MAXSEARCHSTEPS) {
    // Save previous distance
    lastd = cp.distance(p.pos);

    // Calc partials
    SurfacePoint u1 = p, u2 = p;
    u1.u -= DX;
    u2.u += DX;
    CalcPos(u1);
    CalcPos(u2);
    du = (cp.distance(u1.pos) - cp.distance(u2.pos))/fabs(u1.u - u2.u);

    SurfacePoint v1 = p, v2 = p;
    v1.v -= DX;
    v2.v += DX;
    CalcPos(v1);
    CalcPos(v2);
    dv = (cp.distance(v1.pos) - cp.distance(v2.pos))/fabs(v1.v - v2.v);

    // Do line minimization
    double step = .01;
    while(step > DX) {
      SurfacePoint t = p;
      t.u += step * du;
      t.v += step * dv;
      CalcPos(t);
      if(cp.distance(t.pos) < cp.distance(p.pos)) {
        step *= 1.11;
        p = t;
      } else
        step /= 2.0;
    }
  }

  CalcNormal(p);
  if(count >= MAXSEARCHSTEPS || p.pos.distance(cp) > surfMaxDist) {
    cerr << "Surface::SetPoint:Error Failed, point " << cp << 
            " closest point " << p.pos << " du " << du << " dv " << dv << 
            " count " << count << " distance " <<  p.pos.distance(cp) << endl;
  }

  return(true);
}

// Advance time
void Surface::GrowStep(double dt)
{
  time += dt * surfTimeScale;
  int surf = 1;
  double surftime = 0;
  while(time > surftime + surfTime[surf] && surf < surfaces - 1)
    surftime += surfTime[surf++ - 1];
    
  surfCurr.Interpolate(surface[surf-1], surface[surf], 
        surfScale[surf-1], surfScale[surf], (time - surftime)/surfTime[surf]);
//  surfCurr.Interpolate(surface[1], surface[2], 
//        surfScale[1], surfScale[2], time/surfTime[0]);

}

// Get position at current time 
void Surface::GetPos(SurfacePoint &p)
{
  CalcPos(p);
  CalcNormal(p);
}

// Return distance between u and v
double Surface::Distance(SurfacePoint &u, SurfacePoint &v)
{
  // For now use Euclidean distance
  return(u.pos.distance(v.pos));
}

// Calculate xyz postition
void Surface::CalcPos(SurfacePoint &p)
{
  p.u = InBounds(p.u, 0.0, 1.0);
  p.v = InBounds(p.v, 0.0, 1.0);
  p.pos = surfCurr.EvalCoord(p.u, p.v);
}

// Calculate normal
void Surface::CalcNormal(SurfacePoint &p)
{
  // Calc normal from surface.
  SurfacePoint n, s, e, w;
  n.u = p.u + DX; n.v = p.v; 
  s.u = p.u - DX; s.v = p.v;
  e.u = p.u; e.v = p.v + DX;
  w.u = p.u; w.v = p.v - DX;
  CalcPos(n); CalcPos(s);
  CalcPos(e); CalcPos(w);
    
  p.normal = s.Pos() - n.Pos();
  p.normal = p.normal.cross(w.Pos() - e.Pos());
  p.normal.normalise();
}
