#include <util/vector.h>

#define NB_CHEMICALS 2

typedef util::Vector<4,double> Point4d;

#define EXTRA_CONTENT \
  double sigma;  /* Substrate production in the cell*/\
  double rho;    /* Activator production in the cell*/\
  double mu;     /* Activator decay in the cell*/\
  double width;  /* Width of the cell*/

#include "seashell_model.cpp"

class NaticaEnzonaModel : public SeashellModel
{
public:
  double diffA;              // Activator diffusion
  double mumax, mumin;       // Activator decay
  double rhomax, rhomin;     // Activator production
  double rhorand;            // Variation on rho
  double kappa;              // Activator saturation
  double rho0;               // Activator base production
  double diffS;              // Substrate diffusion
  double nu;                 // Substrate decay
  double sigmamax, sigmamin; // Substrate production
  Point4d initCurve;         // Parameter for the initial parameter curve
  double initCurveWidth;     // Width of the initial parameter curve

  NaticaEnzonaModel(QObject *parent)
    : SeashellModel(parent)
  {
    init();
  }

  void readRDParams( util::Parms& parms )
  {
    // Read chemicals parameters
    parms("Chemicals", "DiffA", diffA);
    parms("Chemicals", "mumax", mumax);
    parms("Chemicals", "mumin", mumin);
    parms("Chemicals", "rhomax", rhomax);
    parms("Chemicals", "rhomin", rhomin);
    parms("Chemicals", "rhorand", rhorand);
    parms("Chemicals", "kappa", kappa);
    parms("Chemicals", "rho0", rho0);
    parms("Chemicals", "DiffS", diffS);
    parms("Chemicals", "nu", nu);
    parms("Chemicals", "sigmamax", sigmamax);
    parms("Chemicals", "sigmamin", sigmamin);

    parms("Seashell", "InitCurve", initCurve);
    parms("Seashell", "InitCurveWidth", initCurveWidth);
  }

  void initCell(vertex v)
  {
    int i = v->pos;
    double x = i*cellWidth/initCurveWidth;
    double radd = ran(rhomax)*rhorand;
    double newrho = (ran(1) < 0.5)? rhomax+radd : rhomax - radd;
    double t = std::fmod(x, 1)*initCurve[3];
    if(t < initCurve[0])
    {
      double tprime = t*M_PI_2/initCurve[0];
      double coef = (1+sin(tprime))/2;
      v->sigma = coef*(sigmamax - sigmamin) + sigmamin;
      v->rho = coef*(newrho-rhomin) + rhomin;
      v->mu = coef*(mumax-mumin)+mumin;
    }
    else if(t < initCurve[1])
    {
      v->sigma = sigmamax;
      v->rho = newrho;
      v->mu = mumax;
    }
    else if(t < initCurve[2])
    {
      double tprime = (t-initCurve[1])*M_PI_2/(initCurve[2]-initCurve[1]) + M_PI_2;
      double coef = (1+sin(tprime))/2;
      v->sigma = coef*(sigmamax - sigmamin) + sigmamin;
      v->rho = coef*(newrho-rhomin) + rhomin;
      v->mu = coef*(mumax-mumin)+mumin;
    }
    else
    {
      v->sigma = sigmamin;
      v->rho = rhomin;
      v->mu = mumin;
    }
  }

  void updateDerivatives(const vertex& v, const rd_tag_t&)
  {
    double A = v->chems[0];
    double S = v->chems[1];
    double AxA = A*A;
    double AProd = v->rho*S*AxA/(1+kappa*AxA) + rho0;

    double dA = AProd - v->mu*A;
    double dS = v->sigma - AProd - nu*S;

    // Diffusion
    forall(const vertex& n, LiveCells.neighbors(v))
    {
      dA += diffA*(n->chems[0] - A)/cellWidth;
      dS += diffS*(n->chems[1] - S)/cellWidth;
    }

    v->derivs[0] = dA;
    v->derivs[1] = dS;
  }

};

DEFINE_MODEL(NaticaEnzonaModel);

