#include <util/vector.h>

#define NB_CHEMICALS 2

typedef util::Vector<4,double> Point4d;

#define EXTRA_CONTENT \
  double rho;    /* coefficient of proportionality after randomizing*/\

#include "seashell_model.cpp"

class ConusMarmoreusModel : public SeashellModel
{
public:
  double diffA;              // Activator diffusion
  double mu;                 // Activator decay
  double rho;                // Activator production
  double kappa;              // Activator saturation
  double rho0;               // Activator base production
  double HAprob;             // Probability of an initial cell
                             // Having a high activator concentration

  double diffS;              // Substrate diffusion
  double nu;                 // Substrate decay
  double sigma;              // Substrate production

  double R; // ???

  ConusMarmoreusModel(QObject *parent)
    : SeashellModel(parent)
    , R(0)
  {
    init();
  }

  void readRDParams( util::Parms& parms )
  {
    // Read chemicals parameters
    parms("Chemicals", "DiffA", diffA);
    parms("Chemicals", "mu", mu);
    parms("Chemicals", "rho", rho);
    parms("Chemicals", "kappa", kappa);
    parms("Chemicals", "rho0", rho0);
    parms("Chemicals", "HAprob", HAprob);

    parms("Chemicals", "DiffS", diffS);
    parms("Chemicals", "nu", nu);
    parms("Chemicals", "sigma", sigma);
  }

  void initCell(vertex v)
  {
    int i = v->pos;
    if(i == 0 || i == nbCells-1)
    {
      v->rho = 0;
      v->chems[0] = 0;
      v->chems[1] = 0;
      return;
    }
    //double radd = ran((rho*5)/100.0);
    //double r = (ran(1) < 0.5) ? (rho+radd) : (rho-radd);
    double radd = rho*5/100.0;
    double r = rho + ran(2*radd) - radd;
    double act = (ran(1) < HAprob) ? 1 : 0;
    v->chems[0] = act;
    v->chems[1] = 0;
    v->rho = r;
  }

  void step()
  {
    double aTotal = 0;
    forall(const vertex& v, LiveCells)
    {
      aTotal += v->chems[0];
    }
    R += 0.05*((aTotal/nbCells) - R)*solve.dt;
    this->SeashellModel::step();
  }

  void updateDerivatives(const vertex& v, const rd_tag_t&)
  {
    double A = v->chems[0];
    double S = v->chems[1];
    double AxA = A*A;
    double AProd = v->rho*S*AxA/(1+kappa*AxA) + rho0;

    double dA = AProd - mu*A;
    double dS = sigma/R - AProd - nu*S;

    // Diffusion
    forall(const vertex& n, LiveCells.neighbors(v))
    {
      dA += diffA*(n->chems[0] - A)/cellWidth;
      dS += diffS*(n->chems[1] - S)/cellWidth;
    }

    v->derivs[0] = dA;
    v->derivs[1] = dS;
  }

};

DEFINE_MODEL(ConusMarmoreusModel);

