#include <util/vector.h>

#define NB_CHEMICALS 2

#include "seashell_model.cpp"

class OliviaPorphyria : public SeashellModel
{
public:
  double diffA;              // Activator diffusion
  double mu;                 // Activator decay
  double rho;                // Activator production
  double kappa;              // Activator saturation
  double rho0;               // Activator base production
  double HAprob;             // Probability of an initial cell
                             // Having a high activator concentration

  double diffH;              // Inhibitor diffusion
  double nu;                 // Inhibitor decay
  double rho1;               // Inhibitor production

  double R; // ???

  OliviaPorphyria(QObject *parent)
    : SeashellModel(parent)
    , R(0)
  {
    init();
  }

  void readRDParams( util::Parms& parms )
  {
    // Read chemicals parameters
    parms("Chemicals", "DiffA", diffA);
    parms("Chemicals", "mu", mu);
    parms("Chemicals", "rho", rho);
    parms("Chemicals", "kappa", kappa);
    parms("Chemicals", "rho0", rho0);
    parms("Chemicals", "HAprob", HAprob);

    parms("Chemicals", "DiffH", diffH);
    parms("Chemicals", "nu", nu);
    parms("Chemicals", "rho1", rho1);
  }

  void initCell(vertex v)
  {
    double act = (ran(1) < HAprob) ? 1 : 0;
    v->chems[0] = act;
  }

  void step()
  {
    double aTotal = 0;
    forall(const vertex& v, LiveCells)
    {
      aTotal += v->chems[0];
    }
    R += 0.1*((aTotal/nbCells) - R)*solve.dt;
    this->SeashellModel::step();
  }

  void updateDerivatives(const vertex& v, const rd_tag_t&)
  {
    double A = v->chems[0];
    double H = v->chems[1];
    double AxA = A*A;
    double AProd = rho*AxA/(1+kappa*AxA);

    double dA = (AProd+rho0)/(0.1+H) - mu*A;
    double dH = AProd - (nu/R)*H;

    // Diffusion
    forall(const vertex& n, LiveCells.neighbors(v))
    {
      dA += diffA*(n->chems[0] - A)/cellWidth;
      dH += diffH*(n->chems[1] - H)/cellWidth;
    }

    v->derivs[0] = dA;
    v->derivs[1] = dH;
  }

};

DEFINE_MODEL(OliviaPorphyria);

