#include <vve.h>

#include <util/parms.h>
#include <util/palette.h>

#include <algorithms/solver.h>
#include <util/vector.h>
#include <util/random.h>
#include <cmath>
#include <string.h>

#include <iostream>

using util::ran;
using std::sin;
using std::cout;
using std::endl;

typedef util::Vector<2,double> Point2d;
typedef util::Vector<3,double> Point3d;

typedef solver::Solver<NB_CHEMICALS> RDSolver;
typedef RDSolver::tag_t rd_tag_t;

#ifndef EXTRA_CONTENT
#  define EXTRA_CONTENT
#endif

typedef util::Vector<NB_CHEMICALS,double> Chemicals;

struct VertexContent
{
  Chemicals chems, derivs;
  RDSolver::VertexInternals interns;
  int pos;       // Position of the vertex
  EXTRA_CONTENT
};

struct EdgeContent
{
  RDSolver::EdgeInternals interns;
};

// Type of the graph
typedef VVGraph<VertexContent,EdgeContent> vvgraph;
// Type of a vertex
typedef vvgraph::vertex_t vertex;
// Type of an edge
typedef vvgraph::edge_t edge;

// Class defining your model
// If you change the name, don't forget to change the last line of the file too
class SeashellModel : public Model
{
public:

  util::Palette palette;
  vvgraph LiveCells;
  RDSolver solve;
  int nbRows;
  int currentTextureHeight;
  int currentTextureWidth;
  int textureChunk;
  int textureMaxSize;
  int nbTextureWidth;
  int nbTextureHeight;
  int widthLastTexture;
  bool running;

  GLuint *texIds; // array of texture ids
  GLubyte **textures; // array of textures

  int nbCells;
  double cellWidth;
  double cellHeight;
  double time;
  double timeDeath;
  double timePrint;
  double printInterval;

  double cellDeath;

  int colorLow;             // Color for low concentrations of activator
  int colorHigh;            // Color for high concentrations of activator
  double maxView;

  double background;

  void readParms()
  {
    util::Parms parms("view.v");

    parms("Main", "PrintInterval", printInterval);

    solve.readParms(parms, "Solver");

    readRDParams( parms );

    // Read seashell parameters
    parms("Seashell", "NbCells", nbCells);
    parms("Seashell", "CellWidth", cellWidth);
    parms("Seashell", "CellHeight", cellHeight);
    parms("Seashell", "CellDeath", cellDeath);

    // View parameters
    parms("View", "ColorLow", colorLow);
    parms("View", "ColorHigh", colorHigh);
    parms("View", "MaxView", maxView);
    parms("View", "TextureChunk", textureChunk);
    parms("View", "Background", background);

    solve.MaxDt = std::min(solve.MaxDt, cellDeath);

    if(textureMaxSize)
      updateNbTextureWidth();
  }

  void updateNbTextureWidth()
  {
    if(running)
      return;
    nbTextureWidth = nbCells / textureMaxSize;
    widthLastTexture = nbCells % textureMaxSize;
    if(widthLastTexture)
      nbTextureWidth++;
    else
      widthLastTexture = textureMaxSize;
    if(textureChunk > textureMaxSize)
    {
      cout << "Warning, texture shunk too big for the video cards, reduced to " << textureMaxSize << endl;
      textureChunk = textureMaxSize;
    }
  }

  // Here, reread the files when they are modified
  void modifiedFiles( const std::set<std::string>& filenames )
  {
    forall(const std::string& fn, filenames)
    {
      if(fn == "pal.map")
        palette.reread();
      else if(fn == "view.v")
        readParms();
    }
  }

  ~SeashellModel()
  {
    if(texIds)
    {
      glDeleteTextures(nbTextureHeight*nbTextureWidth, &texId);
      delete [] texIds;
    }
    if(texture)
      delete [] texture;
  }

  virtual void readRDParams( util::Parms& ) = 0;

  virtual void initCell(vertex v) = 0;

  SeashellModel(QObject *parent)
    : Model(parent)
    , palette("pal.map")
    , texIds(0)
    , textures(0)
    , textureMaxSize(0)
    , running(false)
  { }

  void init()
  {
    readParms();
    // Registering the configuration files
    registerFile("pal.map");
    registerFile("view.v");

    solve.initialize();

    // Initialize your structure here
    time = 0;
    timeDeath = 0;
    timePrint = 0;
    currentTextureHeight = 0;
    currentTextureWidth = nbCells;
    nbRows = 0;

    vertex pv(0);
    for(int i = 0 ; i < nbCells ; ++i)
    {
      vertex v;
      v->pos = i;
      LiveCells.insert(v);
      if(!pv.isNull())
      {
        LiveCells.insertEdge(v, pv);
        LiveCells.insertEdge(pv, v);
      }
      initCell(v);
      pv = v;
    }
  }

  void extendTexture()
  {
    // Setting up the new texture buffer
    int new_height = std::min(currentTextureHeight + textureChunk, textureMaxSize);
    bool new_texture = false;
    int offset = nbTextureWidth*(nbTextureHeight-1);
    if(new_height == currentTextureHeight)
    {
      offset += nbTextureWidth;
      new_height = textureChunk;
      new_texture = true;
      nbTextureHeight++;
    }
    if(new_texture)
    {
      // First, create a new array of ids
      GLuint *new_ids = new GLuint[nbTextureWidth*nbTextureHeight];
      memcpy(new_ids, texIds, offset*sizeof(GLuint));
      glGenTextures(nbTextureWidth, new_ids + offset);
      delete [] texIds;
      texIds = new_ids;
    }
    // Now, create the new array of textures
    GLubyte **new_textures;
    new_textures = new (GLubytes*)[nbTextureWidth*nbTextureHeight];
    // Copy the old (finished) textures
    memcpy(new_textures, textures, offset*sizeof(GLubyte*));
    // Create the new textures
    for(int i = 0 ; i < nbTextureWidth ; ++i)
    {
      int width = textureMaxSize;
      if(i == nbTextureWidth-1)
        width = widthLastTexture;
      new_textures[i+offset] = new GLubyte[new_height*width];
      memset(new_textures+i+offset, 0, new_height*width);
    }
    // If necessary, copy the content of the old active textures
    if(!new_texture)
    {
      for(int i = 0 ; i < nbTextureWidth ; ++i)
      {
        int width = textureMaxSize;
        if(i == nbTextureWidth-1)
          width = widthLastTexture;
        memcpy(new_textures+i+offset, textures+i+offset, width*currentTextureHeight*4);
        delete [] textures[i+offset];
      }
    }
    delete [] textures;
    texture = new_textures;
    currentTextureHeight = new_height;

    // Setting up the new texture
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    for(int i = 0 ; i < nbTextureWidth ; ++i)
    {
      int width = textureMaxSize;
      if(i == nbTextureWidth-1)
        width = widthLastTexture;
      glBindTexture(GL_TEXTURE_2D, texIds[offset+i]);
      GLubyte *texture = textures[offset+i];
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, currentTextureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glBindTexture(GL_TEXTURE_2D, 0);
    }
  }

  void drawCellsToTexture(int pos)
  {
    if(pos >= currentTextureHeight)
    {
      extendTexture();
    }
    int offset = currentTextureWidth*pos*4;
    GLubyte *texture_line = texture + offset;
    forall(const vertex& v, LiveCells)
    {
      util::Palette::Color c = palette.getColor(colorLow, colorHigh, v->chems[0] / maxView);
      memcpy(texture_line+4*v->pos, c.c_data(), 4);
    }
    glBindTexture(GL_TEXTURE_2D, texId);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, currentTextureWidth, currentTextureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
    glBindTexture(GL_TEXTURE_2D, 0);
  }

  void step()
  {
    running = true;
    solve(LiveCells, *this);
    time += solve.dt;
    timeDeath += solve.dt;
    timePrint += solve.dt;
    if(timeDeath > cellDeath)
    {
      timeDeath -= cellDeath;
      drawCellsToTexture(nbRows);
      ++nbRows;
    }
    if(timePrint > printInterval)
    {
      timePrint -= printInterval;
      double minA = HUGE_VAL, maxA = 0, minS = HUGE_VAL, maxS = 0;
      forall(const vertex& c, LiveCells)
      {
        if(c->chems.x() > maxA)
          maxA = c->chems.x();
        if(c->chems.x() < minA)
          minA = c->chems.x();
        if(c->chems.y() > maxS)
          maxS = c->chems.y();
        if(c->chems.y() < minS)
          minS = c->chems.y();
      }
      cout << "Current dt: " << solve.dt << endl
        << "Chemical 1: " << minA << " - " << maxA << " ; Chemical 2: " << minS << " - " << maxS << endl;
    }
  }

  void initDraw()
  {
    glEnable(GL_TEXTURE_2D);
    glGetIntergerv(GL_MAX_TEXTURE, &textureMaxSize);
    updateNbTextureWidth();
  }

  void preDraw()
  {
    glEnable(GL_TEXTURE_2D);
    glClearColor(background, background, background, 1.0);
  }

  void postDraw()
  {
    glDisable(GL_TEXTURE_2D);
  }

  void draw(Viewer* viewer)
  {
    double width = nbCells*cellWidth;
    drawCellsToTexture(nbRows);
    glBindTexture(GL_TEXTURE_2D, texId);
    glBegin(GL_QUADS);
    glTexCoord2f(0,0);
    glVertex3f(0,0,0);
    glTexCoord2f(1,0);
    glVertex3f(width, 0,0);
    glTexCoord2f(1,1);
    glVertex3f(width, currentTextureHeight*cellHeight,0);
    glTexCoord2f(0,1);
    glVertex3f(0,currentTextureHeight*cellHeight,0);
    glEnd();
    glBindTexture(GL_TEXTURE_2D, 0);
    Vec vmin(0,0,0), vmax(width, (nbRows+1)*cellHeight, 0);
    viewer->setSceneBoundingBox(vmin, vmax);
  }

  // Methods for the solver
  Chemicals& values(const vertex& v, const rd_tag_t&)
  {
    return v->chems;
  }

  Chemicals& derivatives(const vertex& v, const rd_tag_t&)
  {
    return v->derivs;
  }

  RDSolver::VertexInternals& vertexInternals(const vertex& v, const rd_tag_t&)
  {
    return v->interns;
  }

  RDSolver::EdgeInternals& edgeInternals(const vertex& src, const vertex& tgt, vvgraph& S, const rd_tag_t&)
  {
    return S.edge(src, tgt)->interns;
  }

  virtual void updateDerivatives(const vertex& v, const rd_tag_t&) = 0;
};

#include <QContextMenuEvent>
#include <QAction>
#include <QFileDialog>
#include <QMenu>
#include <QImageWriter>

class SeashellViewer : public Viewer
{
  Q_OBJECT
public:
  SeashellViewer(QWidget* parent)
    : Viewer(parent)
  {
    saveTextureAction = new QAction("Save texture", this);
    connect(saveTextureAction, SIGNAL(activated()), SLOT(saveTexture()));
  }
public slots:
  void saveTexture()
  {
    SeashellModel *m = dynamic_cast<SeashellModel*>(model());
    QList<QByteArray> sf = QImageWriter::supportedImageFormats();
    QStringList supportedFormats;
    QStringList formats;
    forall(const QByteArray& f, sf)
    {
      QString qf = QString::fromLocal8Bit(f);
      supportedFormats += qf;
    }
    formats += "All images ( *." + supportedFormats.join(" *.") + " )";
    forall(const QString& f, supportedFormats)
    {
      formats += f.toUpper() + " images ( *." + f + ")";
    }
    formats += "All files (*.*)";
    QString filters = formats.join(";;");
    QString filename = QFileDialog::getSaveFileName(this, "Saving texture image", QString(), filters);
    if(!filename.isEmpty())
    {
      QImage img(m->texture, m->currentTextureWidth, m->nbRows+1, QImage::Format_RGB32);
      img = img.rgbSwapped();
      QImageWriter writer(filename);
      writer.write(img);
    }
  }

protected:
  void contextMenuEvent( QContextMenuEvent* event )
  {
    QMenu *popup = new QMenu(this);
    popup->addAction(saveTextureAction);
    popup->popup(event->globalPos());
  }

  QAction *saveTextureAction;
};

#include "seashell_model.moc"

DEFINE_VIEWER(SeashellViewer);

